/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

// com.oracle.iot.client.device
/**
 * Implementation of protocol Equatable
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: AdvancedDirectlyConnectedDevice,
               rhs: AdvancedDirectlyConnectedDevice) -> Bool {
    return lhs.hashValue == rhs.hashValue
}

/**
 * A directly&ndash;connected device is able to send messages to, and receive 
 * messages from, the IoT server. A directly&ndash;connected device has a 
 * physical&ndash;device identifier that is registered with the server. When 
 * the directly&ndash;connected device is activated on the server, the server 
 * assigns a logical&ndash;endpoint identifier. This logical&ndash;endpoint 
 * identifier is required for sending messages to, and receiving messages from, 
 * the server.
 */
public class AdvancedDirectlyConnectedDevice : Hashable {
    
    var sendReceiveImpl: SendReceiveImpl? = nil
    var secureConnection: HttpSecureConnection? = nil
    var trustedAssetsManager: TrustedAssetsManager!
    // TODO: Until we have persistence
    let messagingPolicyImplQueue:DispatchQueue =
        DispatchQueue(label:"messagingPolicyImpl queue", qos:.utility)
    var messagingPolicyImpl:MessagingPolicyImpl? = nil
    
    let syncQueue:DispatchQueue = DispatchQueue.global(qos:.utility)

    // TODO: Not checked by all methods. Need much tighter synchronization
    // if no method should run if closed. Basically you don't want the
    // device to close while exeuting a method. This means every method
    // must be run in a sync queue, unless the method is async.
    // Checked by all methods. If a method is called and isClosed is true
    // call the callback with the error or throw `ClientError.state`
    // depending on the method signature.
    private var _isClosed:Bool = false
    var isClosed: Bool {
        get {
            return syncQueue.sync() {
                return _isClosed
            }
        }
        set {
            syncQueue.sync() {
                _isClosed = newValue
            }
        }
    }
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        
        //return ObjectIdentifier(self).hashValue
        var hash:Int = 37
        hash = 37 &* hash &+ self.trustedAssetsManager.getServerHost().hashValue
        hash = 37 &* hash &+ self.trustedAssetsManager.getClientId().hashValue
        return hash
    }
    
    /**
     * This constructor is used internally and is not intended for general use. 
     *
     * - parameter trustedAssetsManager: the `TrustedAssetsManager` to store,
     *                            load and handle the device trust material.
     * - throws
     *       - `ClientError.network` if policy resources cannot be registered.
     */
    init(trustedAssetsManager: TrustedAssetsManager) throws {
        self.trustedAssetsManager = trustedAssetsManager
        self.secureConnection =
            HttpSecureConnection.createHttpSecureConnection(tam: trustedAssetsManager,
                                                    useOnlySharedSecret: false)
        if trustedAssetsManager.isActivated() {
            self.sendReceiveImpl =
                SendReceiveImpl(secureConnection: secureConnection!)
            
            // TODO: Need general storage for references like these
            // This causes the DevicePolicyManager to store a reference mapped
            // to this endpointId.
            _ = DevicePolicyManager(endpointId:getEndpointId(),
                                    secureConnection:secureConnection!)
            /* TODO: No Persistence yet.
             let dpm = DevicePolicyManager(secureConnection:secureConnection!)
             let persistenceStore:PersistenceStore =
             PersistenceStoreManager.getPersistenceStore(endpointId:getEndpointId())
             persistenceStore
             .openTransaction()
             .putOpaque(key:getClassName(instance:dpm), value:dpm)
             .commit()
             */
            try registerDevicePolicyResource()
        }
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will
     * use a custom or default `TrustedAssetsManager` to store, load and handle
     * the device trust material.
     *
     * - throws
     *       - `ClientError.network` if policy resources cannot be registered.
     *       - `ClientError.state` if a TrustedAssetsManager cannot be acquired
     *       - `ClientError.security` if server policy cannot be set.
     *       - `ClientError.file` if the provision file is empty or has the
     *          wrong format version, or can't be decrypted or the server url is
     *
     */
    convenience init() throws {
        // TODO: can getTrustedAssetsManager return nil ?
        guard let tam:TrustedAssetsManager =
                try TrustedAssetsManagerFactory.getTrustedAssetsManager(
                    context: AdvancedDirectlyConnectedDevice.self) else {
             throw ClientError.state("Failed to acquire TrustedAssetsManager")
        }
        try self.init(trustedAssetsManager: tam)
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will
     * use the given filename, fileextension, subdirectory and password to 
     * provision a `TrustedAssetManager` instance
     * - throws: 
     *      - `ClientError.security` if server policy cannot be set.
     *      - `ClientError.file` if the provision file is empty or has the
     *        wrong format version, or can't be decrypted or the server url is
     *        invalid.
     *      - `ClientError.network` if policy resources cannot be registered.
     *      - `ClientError.state` if a TrustedAssetsManager cannot be acquired
     */
    convenience init(file: String, fileExtension: String? = nil,
                     subdirectory: String? = nil, password: String) throws {
        guard let tam:TrustedAssetsManager =
                try TrustedAssetsManagerFactory.createTAMWithClassName(file: file,
                    fileExtension: fileExtension, subdirectory: subdirectory,
                    password: password) else {
            throw ClientError.state("Failed to acquire TrustedAssetsManager")
        }
        try self.init(trustedAssetsManager: tam)
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will 
     * use the given path and password to provision a `TrustedAssetManager` 
     * instance
     * 
     * - parameter path: the path to the provisioning file
     *
     * - parameter password: the password to protect the provisioning file 
     *                       during the creation of the file
     *
     * - throws: 
     *       - `ClientError.file` if the provisioning file is not present
     *       - `ClientError.security` if the provisioning file cannot be
     *           decrypted or there is invalid information in the file
     *       - `ClientError.state` if a TrustedAssetsManager cannot be acquired
     */
    public convenience init(path: String, password: String) throws {
        guard let tam:TrustedAssetsManager =
                try TrustedAssetsManagerFactory.createTrustedAssetsManager(path: path,
                    password: password) else {
            throw ClientError.state("Failed to acquire TrustedAssetsManager")
        }
        try self.init(trustedAssetsManager: tam)
    }

    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the device's endpoint ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX` to `true`.
     *
     * If `contentType` is `nil`, the mime-type defaults to
     *  "application/octet&ndash;stream".
     *
     * - parameter name: the unique name to be used to reference the content
     *                    in storage
     * - parameter contentType: The mime-type of the content
     *
     * - returns: a `StorageObject`
     * - throws:
     *      - `ClientError.network` if there is an exception raised by the 
     *                             runtime, or an abnormal response from the
     *                             storage cloud
     *      - `ClientError.security` if there is an exception establishing a
     *                              secure connection to the storage cloud
     *      - `ClientError.state` if the dcd is closed
     */
    public func createStorageObject(name:String,
                                    contentType:String? = StorageObject.OCTET_STREAM)
                                    throws -> AdvancedStorageObject {
        return try createStorageObject(name:name, contentType:contentType!, mutable:true)
    }
    
    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the device's endpoint ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX` to `true`.
     *
     * If `contentType` is `nil`, the mime-type defaults to
     *  "application/octet&ndash;stream".
     *
     * - parameter name: the unique name to be used to reference the content
     *                    in storage
     * - parameter contentType: The mime-type of the content
     *
     * - parameter mutable: mutable `true` means mutable, `false` means immutable
     *        If `mutable` == `false`, then the AdvancedStorageObject will be made
     *        immutable by setting it’s *Write-Once-Read-Many* (WORM) policy when
     *        uploading it to the container to prevent the users
     *        from deleting it for a 7 days or modifying it.
     *        Number of days is 7 by default and can be overridden by setting property
     *        `Config.STORAGE_WORM_DAYS`. If the property is "0" then no WORM policy
     *        will be applied.
     *
     * - returns: a `StorageObject`
     * - throws:
     *      - `ClientError.network` if there is an exception raised by the
     *                             runtime, or an abnormal response from the
     *                             storage cloud
     *      - `ClientError.security` if there is an exception establishing a
     *                              secure connection to the storage cloud
     *      - `ClientError.state` if the dcd is closed
     */
    public func createStorageObject(name:String,
                                    contentType:String,
                                    mutable:Bool) throws -> AdvancedStorageObject {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed.",
                                                      comment:""))
        }
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection!)
            return try storageConnection.createStorageObject(clientId:getEndpointId(),
                                                             name:name,
                                                             contentType:contentType,
                                                             mutable:mutable)
    }

    /**
     * Create a new `StorageObject` from the URI for a named object in storage.
     *
     * - parameter url: the URL of the object in the storage cloud
     *
     * - returns: a `StorageObject` for the object in storage
     *
     * - throws: 
     *      - `ClientError.network` if there is an exception raised by the
     *                              runtime, or a failure reported by the
     *                              storage cloud
     *      - `ClientError.security` if there is an exception establishing a
     *                               secure connection to the storage cloud
     *      - `ClientError.state` if the dcd is closed
     */
    public func createStorageObject(uri:String,
                callback: @escaping (AdvancedStorageObject?, ClientError?)->()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed.",
                                                      comment:""))
        }
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection!)
        try storageConnection.createStorageObject(storageUrl:uri,
                                                      callback: callback)
    }

    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of 
     *                           this directly connected device
     *
     * - throws: `ClientError.state` if the device is already activated or
     *           `close` has been called
     *
     * - see: `isActivated()`
     */
    public func activate(deviceModels: String...) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment:""))
        }
        try self.activate(deviceModels: deviceModels)
    }
 
    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of 
     *                           this directly connected device
     *
     * - throws: `ClientError.state` if the device is already activated or 
     *           'close' has been called
     *
     * - see: isActivated()
     */
    public func activate(deviceModels: [String]) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment:""))
        }
        try self.activate(deviceModels: deviceModels,
                          callback: { endpointId, error in })
    }
    
    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of
     *                           this directly connected device
     *
     * - parameter callback: The function to be executed after completion of
     *                      activation. If activation is not successful then
     *                      the object will be `nil` and an error object is 
     *                      passed in the form `callback(endpoint, error)` and 
     *                      the reason can be taken from error
     *
     * - throws: `ClientError.state` if the device is already activated or
     *           `close` has been called
     *
     * - see: isActivated()
     */
    public func activate(deviceModels: [String],
            callback: @escaping (String?,ClientError?) -> ()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString(
                "Client is closed", comment:""))
        }

        if self.isActivated() {
            throw ClientError.state(NSLocalizedString("Already activated",
                                                      comment: ""))
        }
            
        var deviceModelSet: [String] = []
        if deviceModels.count > 0 {
            for deviceModel in deviceModels {
                deviceModelSet.append(deviceModel)
            }
        }
        deviceModelSet.append(ActivationManager.DIRECT_ACTIVATION_URN)
        if self is AdvancedGatewayDevice {
            deviceModelSet.append(ActivationManager.INDIRECT_ACTIVATION_URN)
        }
        // TODO: SE does not do this
        //  --- Question this has to be also set for DCD or only for Gateway
        deviceModelSet.append(ActivationManager.MESSAGE_DISPATCHER_URN)
        deviceModelSet.append(ActivationManager.DIAGNOSTICS_URN)
        //
        
        let activationManager: ActivationManager = ActivationManager.getInstance()
        try activationManager.getActivationPolicy(secureConnection: secureConnection!,
                deviceId: trustedAssetsManager!.getClientId(),
                callback: {(activationPolicyResponse: ActivationPolicyResponse?,
                    error: ClientError?) -> Void in
            if (error != nil) {
                callback(nil, error)
                return
            }
            
            if (activationPolicyResponse == nil) {
                callback(nil, ClientError.argument(NSLocalizedString(
                    "AdvancedDirectlyConnectedDevice activate() activationPolicyResponse should not be nil",
                    comment: "")))
                return
            }
                    
            do {
                try self.trustedAssetsManager!.generateKeyPair(
                    algorithm: activationPolicyResponse!.getKeyType(),
                    keySize: activationPolicyResponse!.getKeySize())
            } catch {
                callback(nil, ClientError.security(String(format:
                    NSLocalizedString("Issue while generating Key Pair: %@",
                                      comment: ""), "\(error)")))
                return
            }
            
            var directActivationRequest: DirectActivationRequest?
            do {
                directActivationRequest =
                    try activationManager.createDirectActivationRequest(
                        trustedAssetsManager: self.trustedAssetsManager!,
                        hashAlgorithm: activationPolicyResponse!.getHashAlgorithm(),
                        deviceModels: deviceModelSet)
            } catch {
                callback(nil, ClientError.security(NSLocalizedString(
                    "Issue while generating ActivationRequest", comment:"")))
                return
            }
            do {
                try activationManager.postDirectActivationRequest(
                        secureConnection: self.secureConnection!,
                        deviceId: self.trustedAssetsManager!.getClientId(),
                        directActivationRequest: directActivationRequest!,
                        callback: { (directActivationResponse: DirectActivationResponse?,
                            error: ClientError?) -> Void in
                    if error != nil {
                        callback(nil, error)
                        return
                    }
                            
                    if (directActivationResponse == nil) {
                        callback(nil, ClientError.argument(NSLocalizedString(
                            "AdvancedDirectlyConnectedDevice activate() directActivationResponse should not be nil",
                            comment: "")))
                        return
                    }
                            
                    let endpointId:String = directActivationResponse!.getEndpointId()
                    let certificate: Foundation.Data =
                        directActivationResponse!.getCertificate()
                    
                    do {
                        try self.trustedAssetsManager!.setEndPointCredentials(
                            endpointId: endpointId, certificate: certificate)
                        // request to update accessToken
                        self.secureConnection!.disconnect()
                        self.sendReceiveImpl =
                            SendReceiveImpl(secureConnection: self.secureConnection!)
                        // TODO: Need generalized storage for refeerences like this
                        // Initialize a devicePolicyManager
                        _ = DevicePolicyManager(endpointId:endpointId,
                            secureConnection:self.secureConnection!)
                        /* TODO: No persistence yet
                        let dpm = DevicePolicyManager(secureConnection:self.secureConnection!)
                        let persistenceStore:PersistenceStore =
                            PersistenceStoreManager.getPersistenceStore(key:getEndpointId())
                        persistenceStore
                            .openTransaction()
                            .putOpaque(getClassName(instance:dpm), dpm)
                            .commit()
                        */
                        
                        try self.registerDevicePolicyResource()
                    } catch let error as ClientError {
                        callback(nil, error)
                        return
                    } catch {
                        callback(nil, ClientError.security(NSLocalizedString(
                            "Issue while setting endpoint credentials",
                            comment: "")))
                        return
                    }
                    callback(endpointId, nil)
                })
            } catch {
                callback(nil, ClientError.network(NSLocalizedString(
                    "Issue while retrieving ActivationResponse", comment: "")))
                return
            }
        })
    }
    
    /**
     * Returns whether the device is activated.
     *
     * - returns: `true` if the device is activated, else `false`.
     */
    public func isActivated() -> Bool {
        if isClosed {
            return false
        }
        return self.trustedAssetsManager!.isActivated()
    }
    
    /**
     * Send messages to the server. This call will block until the server
     * responds to the message delivery, or a network timeout occurs.
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                      `ResponseMessage` and/or `ResourceMessage` objects 
     *                      authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. If send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     *                      reason can be taken from error
     * - throws: `ClientError.state` if the device is not activated or 'close'
     *           has been called
     */
    public func send(messages: Message...,
                     callback: @escaping (ClientError?) -> ()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        try send(messages: messages, callback: callback)
    }
    
    /**
     * Send messages to the server. This call will block until the server
     * responds to the message delivery, or a network timeout occurs. 
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                       `ResponseMessage` and/or `ResourceMessage` objects
     *                       authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. If send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     *                      reason can be taken from error
     * - throws: `ClientError.state` if the device is not activated or 'close'
     *           has been called
     */
    public func send(messages: [Message]?,
                     callback: @escaping (ClientError?) -> ()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        
        if self.isActivated() {
            // FINE 
            /*
            if let messages = messages {
                if messages.count > 1 {
                    log(type:.debug,
                        message:"%@ : Send : (%d messages follow)",
                        args:getEndpointId(), messages.count)
                }
                var dumpedMessages:String = ""
                var count:Int = 1
                for message in messages {
                    let dumpedMessage:String = dumpMessage(message:message)
                    log(type:.debug,
                        message: "%@ : %d : Send : %@",
                        args:getEndpointId(), count, dumpedMessage)
                }
            }
            */
            try self.sendReceiveImpl?.send(messages: messages,
                                           callback: callback)
        } else {
            throw ClientError.argument(NSLocalizedString(
                "Directly Connected Device not activated", comment: ""))
        }
    }
    
    /**
     * Offer message to be sent to the server. Whether or not the message
     * is sent depends on policies that have been configured for the
     * attributes in the message. If there are no policies for the attributes
     * in the message, then this call is equivalent to calling `send(Message...)`.
     * Depending on the policies, it is possible that them message
     * may not be sent.
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                    `ResponseMessage` and/or `ResourceMessage`
     *                     objects.
     *
     * - throws: `ClientError.network` if there is an I/O exception when trying to send
     *                               messages.
     *     `ClientError.security` when key or signature algorithm class
     *                            cannot be loaded, or the key is not in
     *                            the trusted assets store, or the
     *                            private key is invalid
     *      `ClientError.state` if the gateway device has not been
     *                          authenticated with the server.
     */
    public func offer(message:Message,
        callback: @escaping (ClientError?)->()) throws {
        try offer(messages:[message], callback:callback)
    }
    
    /**
     * Offer messages to be sent to the server. Whether or not the messages
     * are sent depends on policies that have been configured for the
     * attributes in the message. If there are no policies for the attributes
     * in the message, then this call is equivalent to calling `send(Message...)`.
     * Depending on the policies, it is possible that all, some, or none of
     * the messages will be sent.
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                    `ResponseMessage` and/or `ResourceMessage`
     *                     objects.
     *
     * - throws: `ClientError.network` if there is an I/O exception when trying to send
     *                               messages.
     *     `ClientError.security` when key or signature algorithm class
     *                            cannot be loaded, or the key is not in
     *                            the trusted assets store, or the
     *                            private key is invalid
     *      `ClientError.state` if the gateway device has not been
     *                          authenticated with the server or if the client
     *                          is closed.
     */
    public func offer(messages:[Message],
                          callback: @escaping (ClientError?)->()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed.",
                                                      comment:""))
        }
        if !isActivated() {
            throw ClientError.state("not activated")
        }
        
        let _offer:(ClientError?)->() =
        { (error) in
            // TODO: Should this continue anyway or fail ?
            if error != nil {
                log(type:OSLogType.fault,
                    message:"AdvancedDirectlyConnectedDevice.offeer %@  error: %@",
                    args: self.getEndpointId(), "\(error!)")
            }
            do {
                // TODO: No persistence yet.
                // let the dcd instance maintain a reference
                /*
                let persistenceStore:PersistenceStore =
                    PersistenceStoreManager.getPersistenceStore(key:getEndpointId())
                let messagingPolicyImpl:MessagingPolicyImpl!
                
                if let mpiObj:Any? = persistenceStore.getOpaque(
                    key:getClaassName(instance:MessagingPolicyImpl.self / *defaultValue:nil* /)) {
                    messagingPolicyImpl = (mpiObj as! MessagingPolicyImpl)
                } else {
                    messagingPolicyImpl = MessagingPolicyImpl(directlyConnectedDevice:self)
                    persistenceStore
                        .openTransaction()
                        .putOpaque(getClassName(instance:MessagingPolicyImpl.self,
                                                value:messagingPolicyImpl))
                        .commit()
                        
                    let dpm:DevicePolicyManager? =
                        DevicePolicyManager.getDevicePolicyManager(dcd:self)
                    dpm?.addChangeListener(messagingPolicyImpl)
                }
                */
            
                self.messagingPolicyImplQueue.sync() {
                    if self.messagingPolicyImpl == nil {
                        self.messagingPolicyImpl =
                            MessagingPolicyImpl(directlyConnectedDevice:self)
                    }
                }
                // TODO: What if it returns nil ?
                if let devicePolicyManager:DevicePolicyManager =
                        DevicePolicyManager.getDevicePolicyManager(dcd:self) {
                    devicePolicyManager.addChangeListener(
                        changeListener:self.messagingPolicyImpl!)
                }
                
                // Now we know here that messages list is not empty.
                // If the message list is not empty after applying policies,
                // then send the messages. If it is empty after applying the
                // policies, then there is nothing to send because messages
                // were filtered, or are aggregating values (e.g., mean policy).
                var messagesToPost:[Message] = [Message]()
                for message in messages {
                    let messagesFromPolicies:[Message] =
                        try self.messagingPolicyImpl!.applyPolicies(message:message)
                    if messagesFromPolicies.count > 0 {
                        messagesToPost.append(contentsOf:messagesFromPolicies)
                    }
                }
                
                if messagesToPost.count > 0 {
                    try self.send(messages:messagesToPost,
                                  callback: callback)
                }
            } catch {
                log(type:OSLogType.fault,
                    message:"AdvancedDirectlyConnectedDevice.offeer %@ error: %@",
                    args: self.getEndpointId(), "\(error)")
                callback(error as? ClientError)
            }
        }
        // We need to distinguish between an empty list of messages
        // that has been passed in versus an empty list of message
        // that has resulted from policies being applied.
        // So if the list we receive is empty, let send handle it.
        if messages.count == 0 {
            // TODO: can send handle nil ?
            try send(messages: messages, callback: { error in
                _offer(error)
            })
        } else {
            _offer(nil)
        }
    }
    
    
    /**
     * Receive a `RequestMessage` from the server, if any. This call may
     * block if the implementation sends a message to the server to poll
     * for requests (see `send(Message...)`). This call may return `nil`
     * if there are no requests from the server.
     *
     * - parameter timeout: maximum time, in milliseconds, to wait for a 
     *                      response from the server
     * - parameter callback: The function to be executed after receiving a
     *                      a message. If receive is not successful then
     *                      the `ReqeuestMessage` will be `nil` and an error
     *                      object is passed in the form `callback(nil, error)`
     *                      and the reason can be taken from error
     *
     * - throws: `ClientError.state` if the device is not activated or 'close' 
     *           has been called
     */
    public func receive(timeout: Int,
            callback: @escaping (RequestMessage?, ClientError?) -> ()) throws {
        if isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        if !self.isActivated() {
            throw ClientError.state(NSLocalizedString(
                "Directly Connected Device not activated", comment: ""))
        }
        
        var connectionTimeout:Int = timeout
        if timeout == 0 {
            connectionTimeout = -1 // USE_DEFAULT_TIMEOUT_VALUE
        } else if timeout < 0 {
            connectionTimeout = 0
        }
        let t0:Double = Date().timeIntervalSince1970 * 1000
        
        let _receive0 = {
            let remainingTime:Double = Double(connectionTimeout) -
                (Date().timeIntervalSince1970 * 1000) - t0
            if remainingTime > 0 {
                do {
                    try self.receive(timeout: Int(remainingTime),
                                     callback: callback)
                } catch {
                    log(type:OSLogType.fault,
                        message:"\(error)")
                    callback(nil, error as? ClientError)
                }
            }
        }
        
        var _receive:(RequestMessage?,ClientError?)->() = {_,_ in }
        _receive = {(requestMessage, error) in
            if error != nil {
                callback(nil, error)
                return
            }
            guard let requestMessage = requestMessage else {
                callback(nil, nil)
                return
            }
            let path:String = requestMessage.getURL()
            if path != DevicePolicyManager.PATH {
                callback(requestMessage,nil)
                return
            }
            do {
                let responseMessage:ResponseMessage? =
                    try DevicePolicyManager.getDevicePolicyManager(dcd: self)?
                        .policyChanged(directlyConnectedDevice:self,
                                       requestMessage: requestMessage)
                let messages:[Message] = responseMessage == nil ?
                    [Message]() : [responseMessage!]

                try self.send(messages:messages,
                               callback: { error in
                    // TODO: Should _receive0 be called even if
                    // error is not nil ? In JCL the functionality in
                    // _receive0 is called if either handleRequest or
                    // offer throws exception, and that case is also handled
                    // the same way here. The question is, if the offer callback
                    // receives an error, is that the same as if offer throws
                    // exception immediately ? For now assume it is the
                    // same case and call _receive0 which will eventually
                    // call the callback.
                    if error != nil {
                        log(type:OSLogType.fault,
                            message: "AdvancedDirectlyConnectedDevice.receive: offer failed, %@",
                            args: "\(error!)")
                    }
                    _receive0()
                })
            } catch {
                log(type:OSLogType.fault,
                    message:"\(error)")
                _receive0()
            }
        }
    
        try sendReceiveImpl?.receive(timeout: Int64(connectionTimeout),
                                     callback: { (requestMessage,error) in
            _receive(requestMessage, error)
         })

    }
    
    /**
     * Return the logical&ndash;endpoint identifier of this
     * directly&ndash;connected device. The logical&ndash;endpoint identifier 
     * is assigned by the server as part of the activation process.
     *
     * If the client is closed, the return value is undefined.
     *
     * - returns: the logical&ndash;endpoint identifier of this 
     *            directly&ndash;connected device.
     *
     * - see: `activate(String...)`
     */
    public func getEndpointId() -> String {
        // TODO verify 'isClosed' value in order to know if client has been 
        // closed. Can't do much here if closed, since the interface does not
        // throw or return nil
        return self.trustedAssetsManager?.getClientId() ?? ""
    }
    
    /**
     * Get the `DeviceModel` for the device model urn. This method may
     * return `nil` if there is no device model for the URN. `nil` may also be
     * returned if the device model is a &quot;draft&quot; and the property
     * `Config.ALLOW_DRAFT_DEVICE_MODELS` is set to
     * `false`, which is the default.
     *
     * - parameter deviceModelUrn: The URN of the device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be `nil` and an
     *                       error object is passed in the form callback(device,
     *                       error) and the reason can be taken from 
     *                       error
     *
     * On error `callback` receives a `ClientError.network` error if
     *    the object is not found (StatusCode.NOT_FOUND),
     *    the request fails (!StatusCode.OK), or
     *    the returned data is `nil` or JSON parsing fails
     */
    public func getDeviceModel(deviceModelUrn: String,
            callback: @escaping (DeviceModel?, ClientError?)->()) throws {
        
        if isClosed {
            throw ClientError.state(NSLocalizedString(
                "Client is closed", comment:""))
        }
        /*
         * The `DirectlyConnectedDevice` class has no trusted
         * assets manager and this class gives no access to the one it has,
         * so this method is here.
         */
        return try DeviceModelFactory.getDeviceModel(secureConnection: self.secureConnection!,
                                    urn: deviceModelUrn, callback: callback)
    }
    
    deinit {
        self.close()
    }
    
    /**
     * Close the AdvancedDirectlyConnectedDevice.
     */
    public func close() {
        syncQueue.sync() {
            if !self._isClosed {
                // This should be "tied" to some class
                removeDevicePolicyManager(endpointId:getEndpointId())
                self._isClosed = true
                self.secureConnection?.close()
                self.secureConnection = nil
                self.sendReceiveImpl = nil
                // This is risky because of Hashable
                //self.trustedAssetsManager = nil
                self.messagingPolicyImpl = nil
            }
        }
    }
    
    private func registerDevicePolicyResource() throws {
        if !isActivated() {
            return
        }
    
        let resourceBuilder:ResourceBuilder = ResourceBuilder()
        resourceBuilder.setEndpointName(endpointName:getEndpointId())
        resourceBuilder.addMethod(method:Method.PUT)
        resourceBuilder.setPath(path:DevicePolicyManager.PATH)
        resourceBuilder.setName(name:DevicePolicyManager.CAPABILITY_URN)
        let resource:Resource = try resourceBuilder.build()
        
        let resourceMessageBuilder:ResourceMessageBuilder = ResourceMessageBuilder()
        resourceMessageBuilder.setEndpointName(endpointName:getEndpointId())
        resourceMessageBuilder.setSource(source:getEndpointId())
        resourceMessageBuilder.register(ri:resource)
        let resourceMessage:ResourceMessage = try resourceMessageBuilder.build()
        
        try sendReceiveImpl?.send(messages:resourceMessage,
                                  callback: { error in
            if error != nil {
                log(type:OSLogType.fault,
                    message: "AdvancedDirectlyConnectedDevice register device policy resources error: %@",
                    args: "\(error!)")
            }
        })
    }

    /*
     * Private only
     */
    fileprivate class StorageConnectionMap  {
        private static
            var STORAGE_CONNECTION_MAP:Dictionary<SecureConnection,StorageConnectionImpl> =
                Dictionary<SecureConnection,StorageConnectionImpl>();

        fileprivate class func getStorageConnection(secureConnection:SecureConnection)
                -> StorageConnectionImpl {
            var storageConnection = STORAGE_CONNECTION_MAP[secureConnection]
            if storageConnection == nil {
                storageConnection = StorageConnectionImpl(secureConnection:secureConnection)
                STORAGE_CONNECTION_MAP[secureConnection] = storageConnection
            }
            return storageConnection!
        }
    }
    
    // If data or alert message, return data items formatted
    // as "<key>"="<value>"[,"<key>"="<value>"...]
    // (whether there are quotes around value depends on type)
    // e.g., "temperature"=24,"humidity"=51
    // Otherwise, return the type of message.
    fileprivate func dumpMessage(message:Message) -> String {
        var dataItems:[DataItem]!
        if message is DataMessage {
            dataItems = (message as! DataMessage).getDataItems()
        } else if message is AlertMessage {
            dataItems = (message as! AlertMessage).getDataItems()
        } else {
            dataItems = [DataItem]()
        }
        
        var dumpedMessage:String = ""
        var comma:String = ""
        for dataItem in dataItems {
            
            let key:String = dataItem.getKey()
            var val:String = ""
            
            //stringBuilder.append('"').append(dataItem.getKey()).append("\"=");
            //stringBuilder.append('"').append(val).append('"');
            
            switch dataItem.getType() {
            case .STRING:
                val = dataItem.getValue() as! String
                break
            case .DOUBLE:
                val = "\(Double(truncating: dataItem.getValue() as! NSNumber))"
                break
            case .BOOLEAN:
                val = "\(dataItem.getValue() as! Bool)"
                break
            }
            
            dumpedMessage = String(format:"%@\"%@\" = \"%@\"",
                                   comma, key, val)
            comma = ","
        }
        // add (DATA)
        if !dataItems.isEmpty {
            dumpedMessage += " "
        }
        return String(format:"%@(%@)", dumpedMessage, message.getType().rawValue)
    }
}
