/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

/**
 * A *AdvancedGatewayDevice* is capable of registering indirectly&ndash;connected devices
 * and proxies messages for indirectly&ndash;connected devices.
 */
public class AdvancedGatewayDevice : AdvancedDirectlyConnectedDevice {
    /**
     * The *manufacturer* attribute. This attribute is to be used
     * when setting the *manufacturer* value in the meta-data.
     */
    public static let MANUFACTURER:String = "manufacturer"
    
    /**
     * The *modelNumber* attribute. This attribute is to be used
     * when setting the *modelNumber* value in the meta-data.
     */
    public static let MODEL_NUMBER:String = "modelNumber"
    
    /**
     * The *serialNumber* attribute. This attribute is to be used
     * when setting the *serialNumber* value in the meta-data.
     */
    public static let SERIAL_NUMBER:String = "serialNumber"
    
    /**
     * The *deviceClass* attribute. This attribute is to be used
     * when setting the *deviceClass* value in the meta-data.
     */
    public static let DEVICE_CLASS:String = "deviceClass"
    
    /**
     * The *protocol* attribute. This attribute is to be used
     * when setting the *protocol* value in the meta-data.
     */
    public static let PROTOCOL:String = "protocol"
    
    /**
     * The *protocolDeviceClass* attribute. This attribute is to be used
     * when setting the *protocolDeviceClass* value in the meta-data.
     */
    public static let PROTOCOL_DEVICE_CLASS:String = "protocolDeviceClass"
    
    /**
     * The *protocolDeviceId* attribute. This attribute is to be used
     * when setting the *protocolDeviceId* value in the meta-data.
     */
    public static let PROTOCOL_DEVICE_ID:String = "protocolDeviceId"
    
    /**
     * This constructor is used internally and is not intended for general use.
     *
     * - parameter trustedAssetManager: the *TrustedAssetManager* to store,
     *                            load and handle the device trust material.
     */
    override init(trustedAssetsManager: TrustedAssetsManager) throws {
        try super.init(trustedAssetsManager: trustedAssetsManager)
    }
    
    /**
     * Constructs a new *AdvancedGatewayDevice* instance.
     */
    convenience init() throws {
        try self.init(trustedAssetsManager: TrustedAssetsManagerFactory.getTrustedAssetsManager(context: AdvancedDirectlyConnectedDevice.self)!)
    }
    
    /**
     * Register an indirectly-connected device with the cloud service and
     * specifiy whether the gateway device is required to have the appropriate
     * credentials for activating the indirectly-connected device.
     *
     * The `restricted` parameter controls whether or not the client
     * library is *required* to supply credentials for activating
     * the indirectly-connected device. The client library will
     * *always* supply  credentials for an indirectly-connected
     * device whose trusted assets have been provisioned to the client.
     * If, however, the trusted assets of the indirectly-connected device
     * have not been provisioned to the client, the client library can
     * create credentials that attempt to restrict the indirectly connected
     * device to this gateway device.
     *
     *
     * Pass `true` for the `restricted` parameter
     * to ensure the indirectly-connected device cannot be activated
     * by this gateway device without presenting credentials. If `restricted`
     * is `true`, the client library will provide credentials to the server.
     * The server will reject the activation request if the indirectly connected
     * device is not allowed to roam to this gateway device.
     *
     *
     * Pass `false` to allow the indirectly-connected device to be activated
     * without presenting credentials if the trusted assets of the
     * indirectly-connected device have not been provisioned to the client.
     * If `restricted` is `false`, the client library will provide
     * credentials if, and only if, the credentials have been provisioned to the
     * client. The server will reject the activation if credentials are required
     * but not supplied, or if the provisioned credentials do not allow the
     * indirectly connected device to roam to this gateway device.
     *
     *
     * The `hardwareId` is a unique identifier within the cloud service
     * instance and may not be `nil`. If one is not present for the device,
     * it should be generated based on other metadata such as: model, manufacturer,
     * serial number, etc.
     *
     *
     * The `metaData` Map should typically contain all the standard
     * metadata (the constants documented in this class) along with any other
     * vendor defined metadata.
     *
     * - parameter restricted: indicate whether or not credentials are required
     *                         for activating the indirectly connected device
     *                         This is an optional parameter, the default is
     *                         `false`.
     *

     * - parameter hardwareId: an identifier unique within the cloud service
     *                         instance
     *
     * - parameter metaData: The meta-data of the device
     *
     * - parameter deviceModels: should contain the device model type URNs
     *                           supported by the indirectly connected device
     *
     * - parameter callback: closure to be called after registration with the
     *                      endpoint id of the indirectly-connected device
     *                      If registration is not successful then
     *                      the object will be `nil` and an error object is
     *                      passed in the form `callback(endpointId, error)`
     *                      and the reason can be taken from error
     *
     * - throws:
     *      - `ClientError.state` if device is already activated or the
     *                            shared secret has not been set for the device
     *      - `ClientError.network` if there was a problem with the network
     *        call
     */
    public func registerDevice(restricted: Bool = false, hardwareId: String,
                               metaData: [String:String], deviceModels: [String],
                               callback: @escaping (String?, ClientError?) -> ()) throws {
        
        if (self.isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment: "")))
            callback(nil, ClientError.state(NSLocalizedString("Client is closed", comment: "")))
            return
        }
        
        guard !deviceModels.isEmpty else {
            let msg:String = logMsg(type:.fault,
                message:"%@: deviceModels cannot be empty, hardwareId: %@",
                args:["AdvancedGatewayDevice:registerDevice()", hardwareId])
            log(type:.fault, message:msg)
            throw ClientError.argument(msg)
        }
    
        var signature: String?
        do {
            signature = try trustedAssetsManager?.signWithSharedSecretString(
                        data: getEndpointId(), algorithm: "HmacSHA256",
                        activationId:hardwareId)
            
        } catch ClientError.state {
            // An implicitly created icd with no shared secret
            signature = nil
        } catch {
            throw error
        }
        if (restricted && signature == nil) {
            signature = try trustedAssetsManager?.signWithPrivateKeyString(
                str: getEndpointId(), algorithm: "SHA256withRSA");
        }
        
        let activationManager: ActivationManager = ActivationManager.getInstance()
        let indirectActivationRequest:IndirectActivationRequest =
            activationManager.createIndirectActivationRequest(hardwareId: hardwareId,
                                            metadata: metaData,
                                            deviceModels: deviceModels,
                                            signature: signature)!
        try activationManager.postIndirectActivationRequest(secureConnection: self.secureConnection!,
            indirectActivationRequest: indirectActivationRequest,
            endpointId: self.getEndpointId(),
            callback: { (indirectActivationResponse: IndirectActivationResponse?, error: ClientError?) -> () in
                if error != nil {
                    callback(nil, error)
                    return
                }

                if (indirectActivationResponse == nil) {
                    callback(nil, ClientError.argument(NSLocalizedString("AdvancedGatewayDevice registerDevice() indirectActivationResponse should not be nil", comment: "")))
                    return
                }
                
                let endpointId: String = indirectActivationResponse!.getEndpointId()
                callback(endpointId, nil)
        })
    }
}
