/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import QuartzCore

/**
 * Credentials is a singleton that handles client credentials
 */
class Credentials {
    
    // Only access is static interface
    private init() {
    }
    
    static func getClientAssertionCredentials(tam: TrustedAssetsManager, useOnlySharedSecret: Bool) throws -> Foundation.Data {
        return try getAssertionCredentialsPostData(tam: tam, useOnlySharedSecret: useOnlySharedSecret)
    }
    
    // JWT expiration claim adjustment value, in milliseconds. This is added
    // to the current millisecond time since January 1, 1970 in order to get
    // some time "not to far" in the future for the JWT "exp" claim.
    static var EXP_CLAIM_DELTA: Int64 = 15 * 1000 * 60 // 15 minutes
    
    static var UTF_8: String = "UTF-8"
    static var DEFAULT_MESSAGE_DIGEST_ALGORITHM: String = "HmacSHA256"
    
    
    /*package for testing*/
    static func getAssertionCredentialsPostData(tam: TrustedAssetsManager, useOnlySharedSecret: Bool) throws -> Foundation.Data {
    
        let scope: String = (tam.isActivated()) ? "" : "oracle/iot/activation"
        var postData: String = "grant_type=client_credentials"
        postData = postData + "&client_assertion_type=urn%3Aietf%3Aparams%3Aoauth%3Aclient-assertion-type%3Ajwt-bearer" // already url-encoded
        try postData = postData + "&client_assertion=" + buildClientAssertion(tam: tam, useOnlySharedSecret: useOnlySharedSecret)
        postData = postData + "&scope=" + scope
    
        return postData.data(using: .utf8)!
    }
    
    /*package for testing*/
    static func buildClientAssertion(tam: TrustedAssetsManager, useOnlySharedSecret: Bool) throws -> String {
    
        // Expiration claim is in units of seconds since January 1, 1970 UTC.
        // Note that EXP_CLAIM_DELTA is in units of milliseconds.
        let exp: Int64 = (TimeManager.currentTimeMillis() + EXP_CLAIM_DELTA) / 1000
    
        var id: String
        let tamac: Bool = tam.isActivated()
        if !useOnlySharedSecret && !tamac {
            id = tam.getClientId()
        } else {
            id = tam.getEndpointId()
        }
    
        let useSharedSecret: Bool = useOnlySharedSecret || tamac
        let alg: String = useSharedSecret ?  "HS256" : "RS256"
    
        let header: String = "{\"typ\":\"JWT\",\"alg\":\"" + alg + "\"}"
        var claims: String = "{\"iss\":\"" + id + "\""
        claims = claims + ", \"sub\":\"" + id + "\""
        claims = claims + ", \"aud\":\"oracle/iot/oauth2/token\""
        claims = claims + ", \"exp\":" + exp.description  + "}"
    
        let headerData = header.data(using: .utf8)
        let headerBase64String = headerData!.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
        let claimsData = claims.data(using: .utf8)
        let claimsBase64String = claimsData!.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
        var inputToSign: String = headerBase64String + "." + claimsBase64String
    
        let bytesToSign: [UInt8] = Array(inputToSign.utf8)
        let signByteData: Foundation.Data = Foundation.Data(bytes: UnsafePointer<UInt8>(bytesToSign), count: bytesToSign.count)
        var signature: String? = nil
        do {
            if (useSharedSecret) {
                let signatureData = try tam.signWithSharedSecret(data: inputToSign,
                    algorithm: DEFAULT_MESSAGE_DIGEST_ALGORITHM,
                    activationId: nil)
                signature = signatureData.base64EncodedString(
                    options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
            } else {
                let signedBytes: Foundation.Data? =
                    try tam.signWithPrivateKey(data: signByteData, algorithm: "SHA256withRSA")
                if signedBytes == nil {
                    print("Signing with private key returns nil")
                    throw ClientError.security(NSLocalizedString(
                        "Error while signing", comment: ""))
                }
                signature = signedBytes!.base64EncodedString(
                    options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
            }
        } catch {
            throw ClientError.security(NSLocalizedString(
                "Error while signing", comment: ""))
        }
    
        inputToSign = inputToSign + "."
        inputToSign = inputToSign + signature!
        return inputToSign
    }
}
