/*
 * Copyright (c) 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device

import SystemConfiguration
import Foundation
import CommonCrypto
import os

typealias Stack<T> = Array<T>

// TODO: Should be in shared utils
fileprivate func isRunningOnDevice() -> Bool {
    #if targetEnvironment(simulator)
        return true
    #else
        return false
    #endif
}

fileprivate var zeroAddress:sockaddr?
fileprivate var ref:SCNetworkReachability?
fileprivate var cellularOk:Bool = true

fileprivate func convertNetworkCost(networkCost:String) -> String {
    let cost:String = networkCost.lowercased()
    if cost.starts(with:"ethernet") {
        return "ethernet"
    } else if cost.starts(with:"cellular") {
        return "cellular"
    } else if cost.starts(with:"satellite") {
        return "satellite"
    } else {
        return "ethernet"
    }
}

fileprivate func getNetworkCost() throws -> NetworkCost {
    
    if zeroAddress == nil {
        zeroAddress = sockaddr()
        zeroAddress!.sa_len = UInt8(MemoryLayout<sockaddr>.size)
        zeroAddress!.sa_family = sa_family_t(AF_INET)
        ref = SCNetworkReachabilityCreateWithAddress(nil, &zeroAddress!)
    }
    
    var flags = SCNetworkReachabilityFlags()
    if SCNetworkReachabilityGetFlags(ref!, &flags) {
        
        // TODO: What if there is no network ?
        guard flags.contains(.reachable) else {
            throw ClientError.network(NSLocalizedString(
                "getNetworkCost(): network not reachable", comment:""))
        }
        
        // TODO: Probably don't need "isRunningOnDevice" since cost
        // does not discriminate between WiFi and ethernet
        // If we're reachable, but not on an iOS device
        // (i.e. simulator), we must be on WiFi.
        // For our case wifi IS ethernet
        guard isRunningOnDevice() else {
            // Could be ethernet or wifi, simulator
            // simulator does not discriminate
            return .ethernet
        }
        // lowest cost, but on device can only be wifi
        var connection:NetworkCost = .ethernet
        if !flags.contains(.connectionRequired) {
            connection = .ethernet
        }
        
        if !flags.intersection([.connectionOnTraffic,
                                .connectionOnDemand]).isEmpty {
            if !flags.contains(.interventionRequired) {
                connection = .ethernet
            }
        }
        
        if flags.contains(.isWWAN) {
            if !cellularOk {
                connection = .ethernet
            } else {
                connection = .cellular
            }
        }
        return connection
    } else {
        return .ethernet
    }
}

// NetworkCost in order of increasing cost.
fileprivate enum NetworkCost : Int {
    
    case ethernet = 0
    case cellular = 1
    case satellite = 2
    
    // Get the cost of the NetworkCost given by the string value.
    // The "property" parameter is just used for logging.
    // The defaultValue is the value returned if the value is nil or
    // not a valid NetworkCost
    static func getCost(value:String?, property:String,
                        defaultValue:NetworkCost) -> Int {
        var networkCost:NetworkCost = .ethernet
        if let value = value {
            networkCost =
                NetworkCost.valueOf(networkCostString:convertNetworkCost(networkCost:value),
                                              defaultValue:defaultValue)
        }
        return networkCost.rawValue
    }
    
    static func valueOf(networkCostString:String,
                        defaultValue:NetworkCost = .ethernet) -> NetworkCost {
        var networkCost:NetworkCost!
        let cost = networkCostString.uppercased()
        switch(cost) {
        case "ETHERNET":
            networkCost = .ethernet
            break
        case "CELLULAR":
            networkCost = .cellular
            break
        case "SATELLITE":
            networkCost = .satellite
            break
        default:
            networkCost = defaultValue
        }
        return networkCost
    }
    
    func toString(cost:NetworkCost) -> String {
        var networkCost:String!
        switch(cost) {
        case .ethernet:
            networkCost = "ethernet"
            break
        case .cellular:
            networkCost = "cellular"
            break
        case .satellite:
            networkCost = "satellite"
            break
        }
        return networkCost
    }
}

// TODO: In general, Any should be used vs. AnyObject. But there was conflict
// with the library's use of AnyObject for attribute values. At some point
// change this file to use "Any" and defer casting to AnyObject until the
// value is used as an attribute value.

//
// Policy definitions
//
fileprivate class FilterCondition : DeviceFunction {
    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {

        data["filterCondition.value"] = value

        var condition:FormulaParserNode? =
            data["filterCondition.condition"] as? FormulaParserNode
        if condition == nil {
            let str:String = configuration["condition"] as! String
            let tokens:[FormulaParserToken]!
            do {
                tokens = try FormulaParser.tokenize(formula:str)
            } catch {
                // TODO: What should be done here ?
                // Should this really throw ?
                // I think in SE the exception is a RuntimeException so 
                // bubles up without throws declaration
                print(error)
                return false
            }
            var stack:Stack<FormulaParserNode> = Stack<FormulaParserNode>()
            do {
                _ = try FormulaParser.parseConditionalOrExpression(stack:&stack,
                                                       tokens:tokens,
                                                       formula:str, index:0)
            } catch {
                // TODO: What should be done here ?
                // Should this really throw ?
                // I think in SE the exception is a RuntimeException so
                // bubbles up without throws declaration
                log(type:.fault,
                    message:"%@.%@: Caught exception parsing forumula '%@': %@",
                    args: getId(), "apply()", str, "\(error)")
            }
            condition = stack.popLast()
            data["filterCondition.condition"] = condition
        }

        let computedValue:Double = DeviceFunction.compute(node:condition!,
                                                          deviceAnalog:deviceAnalog) as! Double
        // For a filter condition, if the computation returns 0.0, meaning
        // the condition evaluated to false, then we want to return 'true'
        // because "filter" means out, not in.
        return -1.0 < computedValue && computedValue < 1.0
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        return data.removeValue(forKey:"filterCondition.value")
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let condition = configuration["condition"] ?? "NoConditionFound"
        return super.getDetails(configuration:configuration) + 
            "[condition=\"\(condition)\"]"
    }
}

fileprivate class Sample : DeviceFunction {
    override
    init(id:String) {
        super.init(id:id)
    }

    //Random rand = new Random()
    //let rand:UInt32 = arc4random()
    //let rand:UInt32 = arc4random_uniform(100) // range is 0 to 99
    // convert the UInt32 to some other  types
    //let randomTime:TimeInterval = TimeInterval(randomNum)
    //let someInt:Int = Int(randomNum)
    //let someString:String = String(randomNum) //string works too
 
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // Always put the value in the data map.
        data["sample.value"] = value
        
        var terms:Int? = data["sample.terms"] as? Int
        if terms == nil || terms! == Int.max {
            terms = 0
        }
        terms! += 1
        data["sample.terms"] = terms as AnyObject
        
        // -1 is random, 0 is all
        // TODO: Can never be nil ? And if it is nil what is implied ? Nothing ?
        // JCL does not discriminate, does not catch Cast exception
        guard let criterion = configuration["rate"] else {
            log(type:.error,
                message:"%@.%@: rate cannot be nil",
                args: getId(), "apply()")
            // Don't lose a value
            return true
        }

        var result:Bool = false
        if criterion is Int {
            let number:Int = criterion as! Int
            switch (number) {
            case -1: //random
                // TODO: Make configurable
                result = arc4random_uniform(30) == 15
                break
            case 0: // all
                // zero out terms or else it will just accumulate
                data["sample.terms"] = 0 as AnyObject
                result = true
                break
            default:
                // Sepcific number
                result = number == terms
            }
        } else {
            // zero out terms or else it will just accumulate
            data["sample.terms"] = 0 as AnyObject
            log(type:.fault,
                message:"%@.%@: unknown type for 'rate' %@",
                args: getId(), "apply()", "\(criterion)")
            // Don't lose a value
            result = true
        }
        return result
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        let sample:AnyObject? = data.removeValue(forKey:"sample.value") as AnyObject
        data.removeValue(forKey:"sample.terms")
        return sample
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let rate:Any = configuration["rate"] ?? "NoValueForRate"
        return super.getDetails(configuration:configuration) + "[rate=\"\(rate)\"]"
    }
}

fileprivate class Mean : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }
    
    private func applyWindow(deviceAnalog:DeviceAnalog,
                                attribute:String?,
                                configuration:Dictionary<String,Any>,
                                data: inout Dictionary<String,AnyObject>,
                                value:AnyObject?) -> Bool {
    
        let number:NSNumber =  value as? NSNumber ?? 0.0
        var sum:Double = data["mean.sum"] as? Double ?? 0.0
        var terms:Int = data["mean.terms"] as? Int ?? 0
    
        sum += Double(truncating: number)
        terms += 1
    
        data["mean.sum"] = sum as AnyObject
        data["mean.terms"] = terms as AnyObject
    
        return false
    }
    
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return applyWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data,
                               value:value)
        }
        // The window is sliding
        let bucket:Int = getBucket(function:"mean", configuration:configuration,
                                           data: &data, initialValue: 0.0)
        var buckets:[Bucket<Double>]! = data["mean.buckets"] as? [Bucket<Double>]
        let number:Double =  value as? Double ?? 0.0
        buckets[bucket].value! += number
        buckets[bucket].terms! += 1
        data["mean.buckets"] = buckets as AnyObject
        return false
    }
    
    private func getWindow(deviceAnalog:DeviceAnalog,
             attribute:String?,
             configuration:Dictionary<String,Any>,
             data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        let sum:Double? = data.removeValue(forKey:"mean.sum") as? Double
        let terms:Int? = data.removeValue(forKey:"mean.terms") as? Int
        
        if sum == nil || sum == ZERO || terms == nil || terms == 0 {
            return nil
        } else {
            return (sum! / Double(terms!)) as AnyObject
        }
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {

        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return getWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data)
        }
        // The window is sliding
        
        // TODO: Slide class
        guard let bucketState:(bucketZero:Int, bucketsPerWindow:Int,
                               bucketsPerSlide:Int) =
                getBucketState(function:"mean", configuration:configuration,
                           data: &data,
                           bucketType:Double.self) else {
            return nil
        }
        let bucketsPerWindow:Int = bucketState.bucketsPerWindow
        let bucketsPerSlide:Int = bucketState.bucketsPerSlide
        let bucketZero:Int = bucketState.bucketZero
        let buckets:[Bucket<Double>] = data["mean.buckets"] as! [Bucket<Double>]
        
        var sum:Double = 0.0
        var terms:Int = 0
        
        // TODO: window bucket iterator
        // Loop through the number of buckets in the
        // window and sum them up.
        for i in 0..<bucketsPerWindow {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<Double> = buckets[index]
            sum += bucket.value
            terms += bucket.terms
        }
        
        // TODO: slide bucket iterator
        // Now slide the window
        for i in 0..<bucketsPerSlide {
            let bucket:Bucket<Double> = buckets[(bucketZero + i) % buckets.count]
            bucket.value = 0.0
            bucket.terms = 0
        }
        data["mean.buckets"] = buckets as AnyObject
        
        if sum.isZero || terms == 0 {
            return nil
        } else {
            return sum / Double(terms) as AnyObject
        }
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {

        var details:String = super.getDetails(configuration:configuration)
        
        var commaOrBracket:String = "["
        if let window = configuration["window"] {
            details.append("[window=")
            details.append("\(window)")
            commaOrBracket = ","
        }
        
        if let slide = configuration["slide"] {
            details.append(commaOrBracket)
            details.append("slide=")
            details.append("\(slide)")
        }
        details.append("]")
        return details
    }
}

fileprivate class Min : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }
    
    private func applyWindow(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>,
               value:AnyObject?) -> Bool {
        
        var min = data["min.value"]
        if min == nil {
            min = value
        }
        if let value = value {
            min = DeviceFunction.min(x:value as! NSNumber, y:min as! NSNumber)
        }
        data["min.value"] = min
        return false
    }
    
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return applyWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data,
                               value:value)
        }
        // The window is sliding
        let bucket:Int = getBucket(function:"min", configuration:configuration,
                                   data: &data,
                                   initialValue: Double.greatestFiniteMagnitude)
        var buckets:[Bucket<Double>] = data["min.buckets"] as! [Bucket<Double>]
        let min:Double = buckets[bucket].value
        buckets[bucket].value =
            Double.minimum((value as? Double) ?? Double.greatestFiniteMagnitude,
                           min)
        data["min.buckets"] = buckets as AnyObject
        
        return false
    }

    private func getWindow(deviceAnalog:DeviceAnalog,
             attribute:String?,
             configuration:Dictionary<String,Any>,
             data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        let min:AnyObject? = data.removeValue(forKey:"min.value")
        return min
    }
    
    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return getWindow(deviceAnalog:deviceAnalog,
                             attribute:attribute,
                             configuration:configuration,
                             data: &data)
        }
        // The window is sliding
        guard let bucketState:(bucketZero:Int, bucketsPerWindow:Int,
                               bucketsPerSlide:Int) =
                getBucketState(function:"min", configuration:configuration,
                               data: &data,
                               bucketType:Double.self) else {
            return nil
        }
        let bucketsPerWindow:Int = bucketState.bucketsPerWindow
        let bucketsPerSlide:Int = bucketState.bucketsPerSlide
        let bucketZero:Int = bucketState.bucketZero
        let buckets:[Bucket<Double>] = data["min.buckets"] as! [Bucket<Double>]
        
        var min:Double = Double.greatestFiniteMagnitude
        
        for i in 0..<bucketsPerWindow {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<Double> = buckets[index]
            let num:Double = bucket.value
            min = Double.minimum(num, min)
        }
        
        for i in 0..<bucketsPerSlide {
            let bucket:Bucket<Double> = buckets[(bucketZero + i) % buckets.count]
            bucket.value = Double.greatestFiniteMagnitude
        }
        
        data["min.buckets"] = buckets as AnyObject
        
        return min as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        var details:String = super.getDetails(configuration:configuration)
        
        var commaOrBracket:String = "["
        if let window:Int64 = configuration["window"] as? Int64 {
            details.append("[window=")
            details.append("\(window)")
            commaOrBracket = ","
        }
        
        if let slide:Int64 = configuration["slide"] as? Int64 {
            details.append(commaOrBracket)
            details.append("slide=")
            details.append("\(slide)")
        }
        details.append("]")
        return details
    }
}

fileprivate class Max : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    private func applyWindow(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>,
               value:AnyObject?) -> Bool {
        
        var max = data["max.value"]
        if max == nil {
            max = value
        }
        if let value = value {
            max = DeviceFunction.max(x:value as! NSNumber, y:max as! NSNumber)
        }
        data["max.value"] = max
        return false
    }
    
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return applyWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data,
                               value:value)
        }
        // The window is sliding
        let bucket:Int = getBucket(function:"max", configuration:configuration,
                                   data: &data,
                                   initialValue: 0.0)
        var buckets:[Bucket<Double>]! = data["max.buckets"] as? [Bucket<Double>]
        let max:Double = buckets[bucket].value
        buckets[bucket].value =
            Double.maximum((value as? Double) ?? Double.leastNonzeroMagnitude,
                           max)
        data["max.buckets"] = buckets as AnyObject
        return false
    }
    
    private func getWindow(deviceAnalog:DeviceAnalog,
             attribute:String?,
             configuration:Dictionary<String,Any>,
             data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        let max:AnyObject? = data.removeValue(forKey: "max.value") as AnyObject
        return max
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return getWindow(deviceAnalog:deviceAnalog,
                             attribute:attribute,
                             configuration:configuration,
                             data: &data)
        }
        // The window is sliding
        guard let bucketState:(bucketZero:Int, bucketsPerWindow:Int,
                               bucketsPerSlide:Int) =
                getBucketState(function:"max", configuration:configuration,
                               data: &data,
                               bucketType:Double.self) else {
            return nil
        }
        let bucketsPerWindow:Int = bucketState.bucketsPerWindow
        let bucketsPerSlide:Int = bucketState.bucketsPerSlide
        let bucketZero:Int = bucketState.bucketZero
        let buckets:[Bucket<Double>] = data["max.buckets"] as! [Bucket<Double>]
        
        var max:Double = Double.leastNonzeroMagnitude
        
        for i in 0..<bucketsPerWindow {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<Double> = buckets[index]
            let num:Double = bucket.value
            max = Double.maximum(num, max)
        }
        
        for i in 0..<bucketsPerSlide {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<Double> = buckets[index]
            bucket.value = Double.leastNonzeroMagnitude
        }
        
        data["max.buckets"] = buckets as AnyObject
        return max as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        var details:String = super.getDetails(configuration:configuration)
        
        // TODO: if window and slide are nil then details ends up with
        // a terminating "]"
        var commaOrBracket:String = "["
        if let window:Int64 = configuration["window"] as? Int64 {
            details.append("[window=")
            details.append("\(window)")
            commaOrBracket = ","
        }
        
        if let slide:Int64 = configuration["slide"] as? Int64 {
            details.append(commaOrBracket)
            details.append("slide=")
            details.append("\(slide)")
        }
        details.append("]")
        return details
    }
}

fileprivate class StandardDeviation : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    private func applyWindow(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>,
               value:AnyObject?) -> Bool {

        guard let number:NSNumber = value as? NSNumber else {
            return false
        }
        var terms:[Double]! = data["standardDeviation.terms"] as? [Double]
        if terms == nil {
            terms = [Double]()
        }
        terms.append(Double(truncating: number))
        data["standardDeviation.terms"] = terms as AnyObject
        return false
    }
    
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return applyWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data,
                               value:value)
        }
        
        // Not in JCL slide implementation
        if value == nil {
            return false
        }

        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return applyWindow(deviceAnalog:deviceAnalog,
                               attribute:attribute,
                               configuration:configuration,
                               data: &data,
                               value:value)
        }
        // The window is sliding
        let bucket:Int = getBucket(function:"standardDeviation", configuration:configuration,
                                   data: &data,
                                   initialValue:[Double]())
        var buckets:[Bucket<[Double]>]! = data["standardDeviation.buckets"] as? [Bucket<[Double]>]
        let number:Double = value as? Double ?? 0.0
        buckets[bucket].value.append(number)
        data["standardDeviation.buckets"] = buckets as AnyObject
        
        return false
    }
    
    private func standardDeviation(terms:[Double]) -> Double {
        
        var _terms:[Double] = [Double](repeating:0.0, count:terms.count)
        var sum:Double = 0.0
        
        for n in 0..<_terms.count {
            _terms[n] = terms[n]
            let d:Double = _terms[n]
            sum += d
        }
        
        var mean:Double = sum / Double(terms.count)
        for n in 0..<_terms.count {
            let d:Double = _terms[n] - mean
            _terms[n] = pow(d, 2)
        }
        
        sum = 0.0
        for n in 0..<terms.count {
            let d:Double = _terms[n]
            sum += d
        }
        mean = sum / Double(terms.count)
        
        return sqrt(mean)
    }

    private func getWindow(deviceAnalog:DeviceAnalog,
             attribute:String?,
             configuration:Dictionary<String,Any>,
             data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        guard let terms:[Double] =
                data.removeValue(forKey:"standardDeviation.terms") as? [Double],
                !terms.isEmpty else {
            return nil
        }
        return standardDeviation(terms:terms) as AnyObject
    }
    
    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        // Special case slide, slide == 0, most common scenario
        if !haveSlide(configuration:configuration) {
            return getWindow(deviceAnalog:deviceAnalog,
                             attribute:attribute,
                             configuration:configuration,
                             data: &data)
        }
        // The window is sliding
        guard let bucketState:(bucketZero:Int, bucketsPerWindow:Int,
                bucketsPerSlide:Int) =
                    getBucketState(function:"standardDeviation", configuration:configuration,
                                   data: &data, bucketType:[Double].self) else {
            return nil
        }
        let bucketsPerWindow:Int = bucketState.bucketsPerWindow
        let bucketsPerSlide:Int = bucketState.bucketsPerSlide
        let bucketZero:Int = bucketState.bucketZero
        let buckets:[Bucket<[Double]>] =
            data["standardDeviation.buckets"] as! [Bucket<[Double]>]
        
        var terms:[Double] = [Double]()
        
        for i in 0..<bucketsPerWindow {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<[Double]> = buckets[index]
            let values:[Double] = bucket.value
            terms.append(contentsOf:values)
        }

        let stdDeviation:Double = standardDeviation(terms: terms)
        
        for i in 0..<bucketsPerSlide {
            let index:Int = (bucketZero + i) % buckets.count
            let bucket:Bucket<[Double]> = buckets[index]
            var _value:[Double] = bucket.value
            _value.removeAll()
            bucket.value = _value
        }
        data["standardDeviation.buckets"] = buckets as AnyObject
        
        return stdDeviation as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        var details:String = super.getDetails(configuration: configuration)
        
        var commaOrBracket:String = "["
        if let window:Int64 = configuration["window"] as? Int64 {
            details.append("[window=")
            details.append("\(window)")
            commaOrBracket = ","
        }
        
        if let slide:Int64 = configuration["slide"] as? Int64 {
            details.append(commaOrBracket)
            details.append("slide=")
            details.append("\(slide)")
        }
        details.append("]")
        return details
    }
}

fileprivate class BatchBySize : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        if let _ = batchByPersistence {
            putPersistedBatchedData(deviceAnalog: deviceAnalog,
                                    value: value as? Pair<Message,StorageObject?>)
        } else {
            var list:[Any]? = data["batchBySize.value"] as? [Any]
            if list == nil {
                list = [Any]()
            }
            list!.append(value as Any)
            data["batchBySize.value"] = list as AnyObject
        }
        // TODO: Should this count nil Pairs ?
        var batchCount:Int = data["batchBySize.batchCount"] as? Int ?? 0
        batchCount += 1
        data["batchBySize.batchCount"] = batchCount as AnyObject
        
        let batchSize:Int? = configuration["batchSize"] as? Int
        if batchSize == nil || batchSize! == batchCount {
            return true
        }
        return false
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        data["batchBySize.batchCount"] = 0 as AnyObject
        if let _ = batchByPersistence {
            let value:[Pair<Message,StorageObject?>] =
                getPersistedBatchedData(deviceAnalog: deviceAnalog)
            return value as AnyObject
        } else {
            return data.removeValue(forKey:"batchBySize.value")
        }
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let batchSize:Int = configuration["batchSize"] as? Int ?? 0
        return super.getDetails(configuration:configuration) + "[batchSize=\(batchSize)]"
    }
}

fileprivate class BatchByTime : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        // TODO: What if value is nil ?
        
        // TODO: Need to know if StorageObject being nil is consistent
        // everywhere
        if let _ = batchByPersistence {
            putPersistedBatchedData(deviceAnalog: deviceAnalog,
                                    value: value as? Pair<Message,StorageObject?>)
        } else  {
            // TODO: What if value is nil ?
            var list:[Any?]! = data["batchByTime.value"] as? [Any?]
            if list == nil {
                list = [Any?]()
            }
            list.append(value as Any)
            data["batchByTime.value"] = list as AnyObject
        }
        return false
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        if let _ = batchByPersistence {
            return getPersistedBatchedData(deviceAnalog: deviceAnalog) as AnyObject
        } else {
            return data.removeValue(forKey:"batchByTime.value")
        }
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let delayLimit:Any = configuration["delayLimit"] ?? "nil"
        return super.getDetails(configuration:configuration) + "[delayLimit=\(delayLimit)]"
    }

}

// Will batch data until networkCost (Satellite > Cellular > Ethernet)
// lowers to the configured value
fileprivate class BatchByCost : DeviceFunction {
    override
    init(id:String) {
        super.init(id:id)
    }
    
    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                      attribute:String?,
                      configuration:Dictionary<String,Any>,
                      data: inout Dictionary<String,AnyObject>,
                      value:AnyObject?) -> Bool {
        
        // TODO: what if value is nil ?
        if let _ = batchByPersistence {
            putPersistedBatchedData(deviceAnalog: deviceAnalog,
                                    value: value as? Pair<Message,StorageObject?>)
        } else {
            var list:[AnyObject?]! = data["batchByCost.value"] as? [AnyObject?]
            if list == nil {
                list = [AnyObject?]()
            }
            list.append(value)
            data["batchByCost.value"] = list as AnyObject
        }
        
        // Assume the configured cost is the most expensive
        let configuredCost:Int = NetworkCost.getCost(
            value:configuration["networkCost"] as? String,
            property:"networkCost",
            defaultValue:.satellite)
        
        // Assume the client cost is the least expensive
        // Get the symtem or configured default
        var networkCost:NetworkCost =
            NetworkCost.valueOf(networkCostString: Config.ORACLE_IOT_CLIENT_NETWORK_COST)
        do {
            // Get the system network cost
            networkCost = try getNetworkCost()
            // To test set the device cost to cellular and set the
            // policy configured cost to ethernet
            //networkCost = .cellular
        } catch {
            log(type:.fault,
                message:"%@.%@: could not get system network cost, using default %@: %@",
                args: getId(), "apply()",
                Config.ORACLE_IOT_CLIENT_NETWORK_COST, "\(error)")
        }
        
        // If the cost of the network the client is on (networkCost) is
        // greater than the cost of the network the policy is willing to
        // bear (configuredCost), then return false (the value is
        // filtered).
        if networkCost.rawValue > configuredCost {
            return false
        } else {
            return true
        }
    }
    
    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
                    attribute:String?,
                    configuration:Dictionary<String,Any>,
                    data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        if let _ = batchByPersistence {
            return getPersistedBatchedData(deviceAnalog: deviceAnalog) as AnyObject
        } else {
            return data.removeValue(forKey:"batchByCost.value")
        }
    }
    
    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let networkCost:Any = configuration["networkCost"] ?? "nil"
        return super.getDetails(configuration:configuration) +
            "[networkCost= \(networkCost)]"
    }
}

// TODO: What are the semantics of duplicates ?
// This implementation only considers adjacent values.
// Are all data points within a window subject to the duplicates tast ?
// i.e. data points within the window will never have a duplicate ?

// In order to see any effect the window must be greater than the data rate.

fileprivate class EliminateDuplicates : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {

        var isDuplicate:Bool = false
        let now:Double = Date().timeIntervalSince1970
        let lastValue:AnyObject? = data["eliminateDuplicates.lastValue"]
        let window:Int64 = DeviceFunction.getWindow(configuration:configuration)

        // First time through, value is not a duplicate
        guard  let lastTime:Double =
                data["eliminateDuplicates.lastTime"] as? Double else {
            data["eliminateDuplicates.lastValue"] = value
            data["eliminateDuplicates.lastTime"] = now as AnyObject
            data["eliminateDuplicates.startTime"] = now as AnyObject
            return true
        }
        // Get the window start time.
        let startTime:Double = data["eliminateDuplicates.startTime"] as! Double
        
        // Always done
        data["eliminateDuplicates.lastValue"] = value
        data["eliminateDuplicates.lastTime"] = now as AnyObject
        
        // The window has expired, this value is the first value of the
        // new window
        if (lastTime - startTime) > Double(window)  {
            /*
            log(type:.debug,
                message:"%@: data arrived outside window last:%@ current:%@",
                args:"EliminateDuplicates.apply()",
                lastValue == nil ? "nil" : "\(lastValue!)",
                value == nil ? "nil" : "\(value!)")
            */
            // Restart the window
            data["eliminateDuplicates.startTime"] = now as AnyObject
            return true
        }

        let lv:Foundation.Data = NSKeyedArchiver.archivedData(withRootObject: lastValue as Any)
        let v:Foundation.Data = NSKeyedArchiver.archivedData(withRootObject: value as Any)
        isDuplicate = (lv == v)
        
        // If the new data is not a duplicate move the window
        if !isDuplicate {
            data["eliminateDuplicates.startTime"] = now as AnyObject
        }
        // debug
        /*
        if isDuplicate {
            log(type:.debug,
                message:"%@: duplicate last:%@ current:%@",
                args:"EliminateDuplicates.apply()", "\(lastValue)",
                value == nil ? "nil" : "\(value!)")
        } else {
            log(type:.debug,
                message:"%@: not duplicate last:%@ current:%@",
                args:"EliminateDuplicates.apply()", "\(lastValue)",
                value == nil ? "nil" : "\(value!)")
        }
        */
        
        // returning false tells the caller to not call get, thereby 
        // filtering the duplicate
        
        return !isDuplicate
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        return data["eliminateDuplicates.lastValue"] as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let window:Any = configuration["window"] ?? "nil"
        return super.getDetails(configuration:configuration) + "[window=\(window)]"
    }

}

fileprivate class DetectDuplicates : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {
        
        let now:Double = Date().timeIntervalSince1970
        let lastValue:AnyObject? = data["detectDuplicates.lastValue"]
        
        // First time through, value is not a duplicate
        guard  let lastTime:Double =
            data["detectDuplicates.lastTime"] as? Double else {
                data["detectDuplicates.lastValue"] = value
                data["detectDuplicates.lastTime"] = now as AnyObject
                data["detectDuplicates.startTime"] = now as AnyObject
                return true
        }
        // Get the window start time.
        let startTime:Double = data["detectDuplicates.startTime"] as! Double
        
        // Always done
        data["detectDuplicates.lastValue"] = value
        data["detectDuplicates.lastTime"] = now as AnyObject
        let window:Int64 = DeviceFunction.getWindow(configuration:configuration)
        var isDuplicate:Bool = false
        
        // The window has expired, this value is the first value of the
        // new window
        if (lastTime - startTime) > Double(window)  {
            /*
             log(type:.debug,
             message:"%@: data arrived outside window last:%@ current:%@",
             args:"DetectDuplicates.apply()",
             lastValue == nil ? "nil" : "\(lastValue!)",
             value == nil ? "nil" : "\(value!)")
             */
            // Restart the window
            data["detectDuplicates.startTime"] = now as AnyObject
            return true
        }
        
        let lv:Foundation.Data = NSKeyedArchiver.archivedData(withRootObject: lastValue as Any)
        let v:Foundation.Data = NSKeyedArchiver.archivedData(withRootObject: value as Any)
        isDuplicate = (lv == v)

        if isDuplicate {
            do {
                let alertMessage:AlertMessage = try DeviceFunction.createAlert(
                    deviceAnalog:deviceAnalog, configuration:configuration)
                try deviceAnalog.queueMessage(message:alertMessage)
            } catch {
                log(type:.fault,
                    message:"%@.%@: could not create and queue alert: %@",
                    args: getId(), "apply()", "\(error)")
            }
        } else {
            // Restart the window
            data["detectDuplicates.startTime"] = now as AnyObject
        }

        return true
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {

        return data["detectDuplicates.lastValue"]
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let window:Any = configuration["window"] ?? "nil"
        let format:Any = configuration["format"] ?? "nil"
        return super.getDetails(configuration:configuration) +
                "[window=\(window), format=\"\(format)\"]"
    }

}

/* Untested */
fileprivate class PrivacyPolicy : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {

        data["privacyPolicy.value"] = value
        return true
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {

        
        guard let value:String = data["privacyPolicy.value"] as? String else {
            return nil
        }
        var result:AnyObject? = nil
        
        //let content:[Int8]? = [Int8](value.utf8)
        if let level:String = configuration["level"] as? String {
            switch (level) {
            case "one-way":
                //https://docs.oracle.com/javase/6/docs/technotes/guides/security/StandardNames.html#MessageDigest
                var digest = [UInt8](repeating: 0, count: Int(CC_SHA256_DIGEST_LENGTH))
                if let data = value.data(using: .utf8) {
                    CC_SHA256((data as NSData).bytes, CC_LONG(data.count), &digest)
                } else {
                    log(type:.fault,
                        message:"%@.%@: Could not get Data from String",
                        args: getId(), "get()")
                    return nil
                }
                var digestHex = ""
                for index in 0..<Int(CC_SHA256_DIGEST_LENGTH) {
                    digestHex += String(format: "%02x", digest[index])
                }
                result = digestHex.data(using:.utf8)?.base64EncodedData() as AnyObject
        
            break
            case "two-way":
                guard let hashingKey:String = configuration["hashingKey"] as? String else {
                    // WARNING
                    log(type:.fault,
                        message:"%@.%@: '%@' cannot be nil",
                        args: getId(), "get()", "hashingKey")
                    return nil
                }
                if let signature = HMAC.sign(message: value,
                                             algorithm: HMAC.Algorithm.sha256,
                                             key: hashingKey) {
                    if let sigData:Foundation.Data =
                            signature.data(using:.utf8)?.base64EncodedData() {
                        result = String(data: sigData, encoding: .utf8) as AnyObject
                    }
                }
                break
            case "random":
                // WARNING
                log(type:.fault,
                    message:"%@.%@: random privacy not supported",
                    args:getId(), "get()")
            break
            default:
                // level is "none" or not given (default)
                result = value as AnyObject
            break
            }
        }
        return result
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        // TODO: What is 'xget("level")"
        //return super.getDetails(configuration:configuration) + "[level=\"" + config[xget("level") + "\"]"
        let level:Any = configuration["level"] ?? "nil"
        return super.getDetails(configuration:configuration) + "[level=\"\(level)\"]"
    }

}

fileprivate class AlertCondition : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {

        var condition:FormulaParserNode? =
            data["alertCondition.condition"] as? FormulaParserNode
        if condition == nil {
            if let str:String = configuration["condition"] as? String {
                let tokens:[FormulaParserToken]!
                do {
                    tokens = try FormulaParser.tokenize(formula:str)
                } catch {
                    // TODO: What should be done here ?
                    // Should this really throw ?
                    // I think in SE the exception is a RuntimeException so
                    // bubles up without throws declaration
                    log(type:.fault,
                        message:"%@.%@: Caught exception parsing forumula '%@': %@",
                        args: getId(), "apply()", str, "\(error)")
                    return false
                }
                var stack:Stack<FormulaParserNode> = Stack<FormulaParserNode>()
                do {
                    let _ = try FormulaParser.parseConditionalOrExpression(stack: &stack,
                                                           tokens:tokens,
                                                           formula:str, index:0)
                } catch {
                    // TODO: What should be done here ?
                    // Should this really throw ?
                    // I think in SE the exception is a RuntimeException so
                    // bubles up without throws declaration
                    log(type:.fault,
                        message:"%@.%@: Caught exception parsing forumula '%@': %@",
                        args: getId(), "apply", str, "\(error)")
                    return false
                }
                condition = stack.popLast()
                data["alertCondition.condition"] = condition
            } else {
                // TODO: SE assumes this cannot happen.
                return false
            }
        }

        let computedValue:Double = DeviceFunction.compute(node:condition, deviceAnalog:deviceAnalog) as! Double
        if computedValue.isNaN || computedValue.isInfinite || computedValue.isZero { // zero is false
            data["alertCondition.value"] = value
            return true
        }

        do {
            let alertMessage:AlertMessage = try DeviceFunction.createAlert(
                    deviceAnalog:deviceAnalog,
                    configuration:configuration)
            try deviceAnalog.queueMessage(message:alertMessage)
            
            let valuestr = value ?? ("nil" as AnyObject)
            
            // FINE
            var eventTime:Double!
            if let evt = alertMessage.getEventTime() {
                eventTime = Double(evt/1000)
            } else {
                eventTime = Date().timeIntervalSince1970
            }
            
            log(type:.debug,
                message:"%@ : Alert : %@ \"%@\"=%@ (%@)",
                args: deviceAnalog.getEndpointId(),
                formatDate(date:Date(timeIntervalSince1970:eventTime)),
                    attribute ?? "model policy",
                    "\(valuestr)", alertMessage.getDescription())
        } catch {
            // SEVERE
            log(type:.error,
                message:"%@.%@: %@",
                args:getId(), "apply()", "\(error)")
        }
        
        guard let filter:Bool = configuration["filter"] as? Bool, !filter else {
            // if this is a filter, returning false stops the pipeline
            return false
            
        }
        data["alertCondition.value"] = value
        return true
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {

        return data.removeValue(forKey: "alertCondition.value") as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        
        let filter:Any = configuration["filter"] ?? true
        let condition:Any = configuration["condition"] ?? "nil"
        let urn:Any = configuration["urn"] ?? "nil"
        let fields:Any = configuration["fields"] ?? "nil"
        
        return String(format:"%@[condition=\"%@\", urn=\"%@\", fields=%@, filter=%@]",
                      super.getDetails(configuration:configuration),
                      "\(condition)", "\(urn)", "\(fields)", "\(filter)")
    }
}

fileprivate class ComputedMetric : DeviceFunction {

    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                  attribute:String?,
                  configuration:Dictionary<String,Any>,
                  data: inout Dictionary<String,AnyObject>,
                  value:AnyObject?) -> Bool {

        var formula:FormulaParserNode? = data["computedMetric.formula"] as? FormulaParserNode
        if formula == nil {
            if let str:String = configuration["formula"] as? String {
                let tokens:[FormulaParserToken]!
                do {
                    tokens = try FormulaParser.tokenize(formula:str)
                    formula = try FormulaParser.parseFormula(tokens:tokens, formula:str)
                } catch {
                    // TODO: What should be done here ?
                    // Should this really throw ?
                    // I think in SE the exception is a RuntimeException so
                    // bubbles up without throws declaration
                    print(error)
                    log(type:.fault,
                        message: "%@.%@: Caught exception parsing formula '%@': %@",
                        args: getId(), "apply()", str, "\(error)")
                }
                data["computedMetric.formula"] = formula
            } else {
                // SE assumes formula is never nil
                log(type:.fault,
                    message:"%@.%@: formula cannot be nil",
                    args:getId(), "apply()")
                return false
            }
        }

        let computedValue:Double = DeviceFunction.compute(node:formula, deviceAnalog:deviceAnalog) as! Double
        if computedValue.isNaN || computedValue.isInfinite {
            return false
        }

        data["computedMetric.value"] = computedValue as AnyObject
        return true
    }

    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
               attribute:String?,
               configuration:Dictionary<String,Any>,
               data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        return data.removeValue(forKey:"computedMetric.value") as AnyObject
    }

    override
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        let formula:String = (configuration["formula"] as? String) ?? "nil"
        return super.getDetails(configuration:configuration) + "[formula=\"\(formula)\"]"
    }
}

fileprivate class ActionCondition : DeviceFunction {
    override
    init(id:String) {
        super.init(id:id)
    }

    override
    fileprivate func apply(deviceAnalog:DeviceAnalog,
                      attribute:String?,
                      configuration:Dictionary<String,Any>,
                      data: inout Dictionary<String,AnyObject>,
                      value:AnyObject?) -> Bool {
        
        var condition:FormulaParserNode? =
            data["actionCondition.condition"] as? FormulaParserNode
        if condition == nil {
            let str:String = configuration["condition"] as? String ?? ""
            var tokens:[FormulaParserToken]? = nil
            do {
                tokens = try FormulaParser.tokenize(formula:str)
            } catch {
                tokens = [FormulaParserToken]()
                
            }
            var stack:Stack<FormulaParserNode> = Stack<FormulaParserNode>()
            do {
                _ = try FormulaParser.parseConditionalOrExpression(stack:&stack,
                                                       tokens:tokens!,
                                                       formula:str, index:0)
            } catch {
                
            }
            condition = stack.popLast()
            data["actionCondition.condition"] = condition
        }
        
        let computedValue:Double = DeviceFunction.compute(node:condition,
                                                          deviceAnalog:deviceAnalog) as! Double
        if computedValue.isNaN || computedValue.isInfinite
            || computedValue.isZero { // zero is false
                data["actionCondition.value"] = value
                    return true
        }

        let actionArgs:[Any?] = getActionArgs(deviceAnalog: deviceAnalog,
                                              configuration: configuration)
        let actionName:String = configuration["name"] as? String ?? ""
        
        deviceAnalog.call(actionName:actionName, args:actionArgs
                                    /*,callback: { error in }*/)
        
        // filter == nil || filter return false
        guard let filter:Bool = configuration["filter"] as? Bool, !filter else {
            // if this is a filter, returning false stops the pipeline
            return false
            
        }
        
        data["actionCondition.value"] = value
        return true
    }
    
    override
    fileprivate func get(deviceAnalog:DeviceAnalog,
                    attribute:String?,
                    configuration:Dictionary<String,Any>,
                    data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        
        return data.removeValue(forKey:"actionCondition.value")
    }
    
    override
    public func getDetails(configuration:Dictionary<String,Any>) -> String {
        
        let filter:Any = configuration["filter"] ?? true
        let arguments:Any = configuration["arguments"] ?? "nil"
        return String(format:"%@ [condition=\"%@\", action=\"%@\", arguments=\"%@\", filter=%@",
            super.getDetails(configuration:configuration),
            configuration["condition"] as? String ?? "nil",
            configuration["name"] as? String ?? "nil",
            "\(arguments)", "\(filter)")
    }
}

fileprivate func getActionArgs(deviceAnalog:DeviceAnalog,
                                 configuration:Dictionary<String,Any>) -> [Any?] {
    
    // This list comes from handling the "action" parameter in
    // DevicePolicyManager.devicePolicyFromJSON()
    guard let arguments:[Any] = configuration["arguments"] as? [Any],
        !arguments.isEmpty else {
            return [Any?]() //actionArgs
    }
    
    var actionArgs:[Any?] = [Any?]()
    for arg in arguments {
        
        let deviceModel:DeviceModelImpl = deviceAnalog.getDeviceModel() as! DeviceModelImpl
        let actionMap:[String:DeviceModelAction] = deviceModel.getDeviceModelActions()
        if actionMap.isEmpty {
            // TODO: this could get annoying
            // WARNING
            log(type:.fault,
                message:"DeviceFunction.%@: no actions in device model '%@'",
                args:"getActionArgs()", deviceModel.getURN())
            actionArgs.append(nil)
            continue
        }
        guard let actionName:String = configuration["name"] as? String,
                let deviceModelAction:DeviceModelAction = actionMap[actionName] else {
            // TODO: this could also get annoying
            // WARNING
            log(type:.fault,
                message:"DeviceFunction.%@: " +
                "no action named '%@' in device model '%@'",
                args:"getActionArgs()", (configuration["name"] as? String ?? "nil"),
                deviceModel.getURN())
            actionArgs.append(nil)
            continue
        }
        
        guard let argType:DeviceModelAttributeType = deviceModelAction.getArgType() else {
            log(type:.fault,
                message:"DeviceFunction.%@: " +
                "no argument type for '%@' in device model '%@'",
                args:"getActionArgs()", actionName, deviceModel.getURN())
            actionArgs.append(nil)
            continue
        }
        
        do {
            let convertedArg = try convertArg(deviceAnalog:deviceAnalog,
                                          argType: argType,
                                          arg: arg)
            actionArgs.append(convertedArg)
        } catch {
            // IllegalArgumentException
            // WARNING
            log(type:.fault,
                message: "DeviceFunction.%@: " +
                "bad argument to '%@' int '%@': '%@'",
                args: "getActionArgs()", actionName,
                deviceModel.getURN(), "\(error)")
            // maybe this was purposeful - let application handle
            actionArgs.append(arg)
        }
    }
    return actionArgs
}

fileprivate func convertArg(deviceAnalog:DeviceAnalog,
                            argType:DeviceModelAttributeType,
                            arg:Any?) throws -> Any? {
    
    guard let arg = arg else {
        return nil
    }
    // arg in an ACTION is a formula
    // ACTION type is represented as JSON of form
    // {
    //     "name": "name-of-the-action",
    //     "arguments": [
    //         "formula:argument-1",
    //         "formula:argument-2"
    //     ]
    // }
    // field in an ALERT is also a formula
    // ALERT type is represented as JSON of form
    // {
    //     "urn": "URN-of-the-alert",
    //     "fields": {
    //         "name-of-the-field-1": "<formula>",
    //         "name-of-the-field-2": "<formula>"
    //         ...
    //      },
    //     "severity": "severity level"
    // }
    var convertedArg:Any? = nil
    //let conversion = convertFormula(deviceAnalog: deviceAnalog, formula: arg as! String)
    var formulaArg:String = ""
    if arg is String {
        formulaArg = arg as! String
    } else if arg is NSNumber {
        formulaArg = (arg as! NSNumber).stringValue
    }
    // Not sure what "" will return, but this either has to be a String or an NSNumber or nil
    let conversion = convertFormula(deviceAnalog: deviceAnalog, formula: formulaArg)

    switch (argType) {
    case .STRING, .URI:
        convertedArg = conversion as! String
        break
    case .BOOLEAN:
        if conversion is Bool {
            convertedArg = conversion as! Bool
        } else if conversion is String {
            convertedArg = Bool(conversion as! String)
        } else if conversion is NSNumber {
            convertedArg = Double(truncating: conversion as! NSNumber) != 0.0
        } else {
            throw ClientError.argument("cannot convert \(arg) to \(argType)")
        }
        break
    case .DATETIME:
        if conversion is NSNumber {
            convertedArg = conversion as! NSNumber
        } else {
            throw ClientError.argument("cannot convert \(arg) to \(argType)")
        }
        break
    case .INTEGER:
        if conversion is NSNumber {
            convertedArg = Int32(truncating: conversion as! NSNumber)
        } else {
            throw ClientError.argument("cannot convert \(arg) to \(argType)")
        }
        break
    case .NUMBER:
        if conversion is NSNumber {
            convertedArg = conversion as! NSNumber
        } else {
            throw ClientError.argument("cannot convert \(arg) to \(argType)")
        }
        break
    }
    return convertedArg
}

fileprivate func convertFormula(deviceAnalog:DeviceAnalog,
                                formula:String) -> Any {
    
    // If arg is a String, it should be a FORMULA
    do {
        let tokens:[FormulaParserToken] =
            try FormulaParser.tokenize(formula:formula)
        let node:FormulaParserNode =
            try FormulaParser.parseFormula(tokens:tokens, formula:formula)
        return DeviceFunction.compute(node:node, deviceAnalog:deviceAnalog)
    } catch {
        // IllegalArgumentException
        // WARNING
        log(type:.fault,
            message: "DeviceFunction.%@: " +
            "field in formula not in device model: %@",
            args: "convertFormula()", formula)
    }
    return Double.nan as NSNumber
}

fileprivate func  convertFormulaToString(deviceAnalog:DeviceAnalog, formula:String) -> Any {
    
    var _formula:String = formula
    // If arg is a String, it should be a FORMULA
    do {
        let tokens:[FormulaParserToken] =
            try FormulaParser.tokenize(formula:formula)
        let node:FormulaParserNode =
            try FormulaParser.parseFormula(tokens:tokens, formula:formula)
        
        if node is FormulaParserTerminal {
            let terminal:FormulaParserTerminal = node as! FormulaParserTerminal
            let nodeValue:String = (node as! FormulaParserTerminal).getValue()
            switch (terminal.getTerminalType()) {
            case .currentAttribute:
                guard let value:AnyObject =
                    try deviceAnalog.getAttributeValue(attribute:nodeValue) else {
                        break
                }
                if value is String {
                    _formula = value as! String
                }
                break
            case .inprocessAttribute:
                if let value:Any = DeviceFunction.getInProcessValue(
                        endpointId:deviceAnalog.getEndpointId(),
                        deviceModelURN: deviceAnalog.getDeviceModel().getURN(),
                        attribute: nodeValue) {
                    if value is String {
                        _formula = value as! String
                    }
                }
                if let value = try deviceAnalog.getAttributeValue(attribute:nodeValue) {
                    if value is String {
                        _formula = value as! String
                    }
                }
                break
            case .ident:
                _formula = nodeValue
                break
            default:
                break
            }
        }
    } catch {
        // WARNING
        log(type:.fault,
            message:"DeviceFunction.%@: could not parse: %@, error: %@",
            args: "convertToFormula()", formula, "\(error)")
    }
    return _formula
}

fileprivate let FILTER_CONDITION:DeviceFunction = FilterCondition(id:"filterCondition")
fileprivate let ELIMINATE_DUPLICATES:DeviceFunction = EliminateDuplicates(id:"eliminateDuplicates")
fileprivate let DETECT_DUPLICATES:DeviceFunction = DetectDuplicates(id:"detectDuplicates")
fileprivate let SAMPLE:DeviceFunction = Sample(id:"sampleQuality")
fileprivate let MEAN:DeviceFunction = Mean(id:"mean")
fileprivate let MAX:DeviceFunction = Max(id:"max")
fileprivate let MIN:DeviceFunction = Min(id:"min")
fileprivate let STANDARD_DEVIATION:DeviceFunction = StandardDeviation(id:"standardDeviation")
fileprivate let BATCH_BY_SIZE:DeviceFunction = BatchBySize(id:"batchBySize")
fileprivate let BATCH_BY_TIME:DeviceFunction = BatchByTime(id:"batchByTime")
fileprivate let BATCH_BY_COST:DeviceFunction = BatchByCost(id:"batchByCost")
fileprivate let ALERT_CONDITION:DeviceFunction = AlertCondition(id:"alertCondition")
fileprivate let COMPUTED_METRIC:DeviceFunction = ComputedMetric(id:"computedMetric")
fileprivate let PRIVACY_POLICY:DeviceFunction = PrivacyPolicy(id:"privacyPolicy")
fileprivate let ACTION_CONDITION:DeviceFunction = ActionCondition(id:"actionCondition")


private let DEVICE_FUNCTIONS:[DeviceFunction] = [
     FILTER_CONDITION,
     ELIMINATE_DUPLICATES,
     DETECT_DUPLICATES,
     SAMPLE,
     MAX,
     MIN,
     MEAN,
     STANDARD_DEVIATION,
     BATCH_BY_SIZE,
     BATCH_BY_TIME,
     ALERT_CONDITION,
     COMPUTED_METRIC,
     PRIVACY_POLICY,
     ACTION_CONDITION
]

fileprivate let POLICY_MAP:Dictionary<String, DeviceFunction> = [
    "filterCondition" : FILTER_CONDITION,
    "eliminateDuplicates" : ELIMINATE_DUPLICATES,
    "detectDuplicates" : DETECT_DUPLICATES,
    "sampleQuality" : SAMPLE,
    "mean" : MEAN,
    "max" : MAX,
    "min" : MIN,
    "standardDeviation" : STANDARD_DEVIATION,
    "batchBySize" : BATCH_BY_SIZE,
    "batchByTime" : BATCH_BY_TIME,
    "batchByCost" : BATCH_BY_COST,
    "alertCondition" : ALERT_CONDITION,
    "computedMetric" : COMPUTED_METRIC,
    "privactPolicy" : PRIVACY_POLICY,
    "actionCondition" : ACTION_CONDITION
]


fileprivate let inProcessValuesSyncQueue:DispatchQueue =
    DispatchQueue(label:"inProcessValues sync Queue", qos:.utility)
// A map of attribute-name -> in-process attribute value.
// The in-process value of an attribute is the value that is being
// passed along the pipeline; i.e., $(attribute) in a formula or condition
fileprivate var inProcessValues:Dictionary<String,Any> = Dictionary<String,Any>()

/**
 * DeviceFunction is an abstraction of a policy device function.
 */
class DeviceFunction {
    
    // A bucket used for slide (see comments in the 'mean' function)
    // <T> Is the type of value
    class Bucket<T> {
        var value:T!
        var terms:Int!
        
        init(initialValue:T) {
            self.value = initialValue
            self.terms = 0
        }
        
        public func toString() -> String {
            return String(format:
                "{\"value\" : %@, \"terms\" : %@}", "\(value)", "\(terms)")
        }
    }

    var batchByPersistence:BatchByPersistence? {
        get {
            return BatchByPersistence.getInstance()
        }
    }
    let syncQueue:DispatchQueue!

    let ZERO:Double = 0.0
    private var id:String!

    init(id:String) {
        self.id = id
        self.syncQueue = DispatchQueue(label:"\(self.id) sync queue",
                                       qos:.utility)
    }

    /**
     * The `apply` method is where the logic for the function is coded.
     * This method returns `true` if the conditions for the function have
     * been met. Only when this function's apply method returns true
     * will the next function in the pipeline be applied.
     *
     * After this method returns `true`, use
     * `get(DeviceAnalog, String, Map, Map)` to retrieve
     * the value from the function.
     *
     * - parameter deviceAnalog: the `VirtualDevice`, never `nil`
     *
     * - parameter attribute: the `String`, which may be `nil` if the
     *                             function is being applied at the device model level
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - parameter data: a place for storing data between invocations of the function
     *
     * - parameter value: the value to which the function is being applied
     *
     * - returns: `true if the conditions for the function have been satisfied.
     */
    public func apply(deviceAnalog:DeviceAnalog, attribute:String?,
                      configuration:Dictionary<String,Any>,
                      data: inout Dictionary<String,AnyObject>,
                      value:AnyObject?) -> Bool {
        //fatalError("Subclass must implement 'apply' function")
        return syncQueue.sync(execute: {
            var _data:Dictionary<String,AnyObject> = data
            let result:Bool = self.apply(deviceAnalog:deviceAnalog, attribute:attribute,
                              configuration:configuration,
                              data: &_data,
                              value:value)
            data = _data
            return result
        })
    }
    
    /**
     * Return the value from the function. This method should only be called after
     * `apply(DeviceAnalog, String, Map, Map, Object)`
     * returns `true`, or when a window expires.
     *
     * - parameter deviceAnalog: the `VirtualDevice`, never `nil`
     *
     * - parameter attribute: the `String`, which may be `nil` if the
     *                      function is being applied at the device model level
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - parameter data: a place for storing data between invocations of the function
     *
     * - returns: the value from having applied the function
     */
    public func get(deviceAnalog:DeviceAnalog, attribute:String?,
                    configuration:Dictionary<String,Any>,
                    data: inout Dictionary<String,AnyObject>) -> AnyObject? {
        //fatalError("Subclass must implement 'get' function")
        var _data:Dictionary<String,AnyObject> = data
        let result:AnyObject? = syncQueue.sync(execute: {
            return self.get(deviceAnalog:deviceAnalog, attribute:attribute,
                            configuration:configuration,
                            data: &_data)
            
        })
        data = _data
        return result
    }

    /**
     * Utility for getting a "window" value from a configuration.
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - returns: a window value, or -1 if the configuration is not time based
     */
    static func getWindow(configuration:Dictionary<String,Any>) -> Int64 {
        for key in ["window", "delayLimit"] {
            if let criterion:NSNumber = configuration[key] as? NSNumber {
                return Int64(truncating: criterion)
            }
        }
        return -1
    }
    
    /**
     * Quick check for slide
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - returns: true if slide != 0
     */
    func haveSlide(configuration:[String:Any]) -> Bool {
        return configuration["slide"] as? NSNumber != 0
    }
    
    /**
     * Utility for getting a "slide" value from a configuration.
     *
     * - parameter configuration: the parameters for this function from the
     *                            device policy
     * - parameter window: the corresponding window for the slide
     * - returns: the configured slide value, or window if there is no slide or
     *            slide is zero
     */
    static func getSlide(configuration:[String:Any], window:Int64) -> Int64 {
        if let slideParameter:NSNumber = configuration["slide"] as? NSNumber {
            let slide:Int64 = Int64(truncating: slideParameter)
            return slide > 0 ? slide : window
        } else {
            return window
        }
    }
    
    /**
     * Utility for getting the "slide" bucket index given the function state.
     *
     * - parameter function: the map key prefix
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - parameter data: a place for storing data between invocations of the function
     *
     * - returns: the slide bucket index to update
     */
    func getBucket<T>(function:String, configuration:[String:Any],
                    data: inout [String:AnyObject],
                    initialValue:T) -> Int {
        
        // Handling slide:
        //
        // Slide is how much the window moves after the
        // window expires. If there is a window of 5 seconds
        // with a slide of 2 seconds, then at the end of
        // 5 seconds, the window slides over by two seconds.
        // That means that the next window's worth of data
        // would include 3 seconds of data from the previous
        // window, and 2 seconds of new data.
        //
        // To handle this, we divide up the window into buckets.
        // Each bucket represents a period of time, such
        // that the time period is the greatest common factor
        // between the window and the slide. For example, if
        // the window is 60 seconds and the slide is 90
        // seconds, a bucket would span 30 seconds, and
        // there would be three buckets.
        //
        // When the window expires and the get method is called,
        // the return value of the mean policy function will
        // include the value and number of terms of bucket[0]
        // through bucket[n]. Then the buckets that don't
        // contribute to the next window are emptied (so to speak)
        // and the cycle continues.
        //
        // Best case is that the slide equal to the window.
        // In this case, there is only ever one bucket.
        // The worst case is when greatest common factor between
        // slide and window is small. In this case, you end up
        // with a lot of buckets, potentially one bucket per
        // slide time unit (e.g., 90 seconds, 90 buckets).
        // But this is no worse (memory wise) than keeping
        // an array of values and timestamps.
        //
        let now:Int64 = TimeManager.currentTimeMillis()
        
        // windowStartTime is the time at which the first
        // call to apply was made for the current window
        // of time. We need to know when this window
        // started so that we can figure out what bucket
        // the data goes into.
        var windowStartTime:Int64! = data["\(function).windowStartTime"] as? Int64
        if windowStartTime == nil {
            windowStartTime = now
            data["\(function).windowStartTime"] = windowStartTime as AnyObject
        }
        
        // The greatest common factor between the
        // window and the slide represents the greatest
        // amount of time that goes evenly into
        // both window and slide.
        let window:Int64 = DeviceFunction.getWindow(configuration:configuration)
        let slide:Int64 = DeviceFunction.getSlide(configuration:configuration, window:window)
        // Each bucket spans this amount of time.
        let span:Int64 = DeviceFunction.gcd(x:window, y:slide)
        
        var buckets:[Bucket<T>]! = data["\(function).buckets"] as? [Bucket<T>]
        if buckets == nil {
            
            // The number of buckets is the window or span
            // (which ever is greater) divided
            // by the amount of time it spans. Thus, if
            // there is a 5 second window with a 2 second slide,
            // the greatest common factor is 1 second and we end
            // up with 5 buckets. But if the slide was 7 seconds,
            // you'd end up with 7 buckets. This lets us fill
            // up buckets without worrying about whether the
            // window is greater than, equal to, or less than
            // the slide.
            // Note: we add 1 so there is a bucket for when
            // a value comes in for the next window, but before
            // the window has been moved.
            let numberOfBuckets:Int =
                Int(Int64(truncating: DeviceFunction.max(x:NSNumber(value:slide),
                                                         y:NSNumber(value:window))) / span) + 1
            buckets = [Bucket]()
            buckets.reserveCapacity(numberOfBuckets)
            for _ in 0..<numberOfBuckets {
                buckets.append(Bucket<T>(initialValue:initialValue))
            }
            data["\(function).buckets"] = buckets as AnyObject
        }
        
        // bucketZero is the index of the zeroth bucket
        // in the buckets array. This allows the buckets array
        // to be treated as a circular buffer so we don't have
        // to move array elements when the window slides.
        var bucketZero:Int! = data["\(function).bucketZero"] as? Int
        if bucketZero == nil {
            bucketZero = 0
            data["\(function).bucketZero"] = bucketZero as AnyObject
        }
        
        // Which bucket are we working on is calculated
        // by the dividing the amount of time we are into
        // the window by the span of time represented by
        // one bucket. For example, say we have a 2 second
        // slide and a 10 second window giving us 5 buckets.
        // Say our window started at 20 seconds and the
        // value arrives at 25 seconds (5 seconds into the
        // window). The value, then should be added to the
        // third bucket (buckets[2]) since that bucket
        // represents the time from 4 seconds to 6 seconds
        // into the current window.
        let bucketIndex:Int = Int((now - windowStartTime) / span)
        let bucket:Int = (bucketZero + bucketIndex) % buckets.count
        
        return bucket
    }
    
    /**
     * Utility for getting the "slide" bucket state for "get".
     *
     * - parameter function: the map key prefix
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - parameter data: a place for storing data between invocations of the function
     *
     * - returns: the `bucketsPerWindow` and `bucketsPerSlide` counts
     */
    func getBucketState<T>(function:String, configuration:[String:Any],
                        data: inout [String:AnyObject],
                        bucketType:T.Type) ->
                (bucketZero:Int, bucketsPerWindow:Int,bucketsPerSlide:Int)? {
        
        guard let buckets:[Bucket<T>] =
            data["\(function).buckets"] as? [Bucket<T>] else {
                // must have called get before apply
                return nil
        }
        
        // JCL Assert
        guard let bucketZero:Int = data["\(function).bucketZero"] as? Int else {
            // if buckets is not null, but bucketZero is,
            // something is wrong with our implementation
            return nil
        }
        
        // The greatest common factor between the
        // window and the slide represents the greatest
        // amount of time that goes evenly into
        // both window and slide.
        let window:Int64 = DeviceFunction.getWindow(configuration:configuration)
        let slide:Int64 = DeviceFunction.getSlide(configuration:configuration, window:window)
        
        // Each bucket spans this amount of time.
        let span:Int64 = DeviceFunction.gcd(x:window, y:slide)
        
        // The number of buckets that make up a window
        let bucketsPerWindow:Int = Int(window / span)
        
        // The number of buckets that make up the slide
        let bucketsPerSlide:Int = Int(slide / span)
        
        // update windowStartTime for the next window.
        // The new windowStartTime is just the current window
        // start time plus the slide.
        var windowStartTime:Int64! = data["\(function).windowStartTime"] as? Int64
        if windowStartTime == nil {
            //assert windowStartTime != null;
            windowStartTime = TimeManager.currentTimeMillis()
        }
        data["\(function).windowStartTime"] = windowStartTime + span * Int64(bucketsPerSlide) as AnyObject
        
        // update bucketZero index. bucketZero is the index
        // of the zeroth bucket in the circular buckets array.
        data["\(function).bucketZero"] = (bucketZero + bucketsPerSlide) % buckets.count as AnyObject
        
        return (bucketZero:bucketZero, bucketsPerWindow:bucketsPerWindow, bucketsPerSlide:bucketsPerSlide)
    }

    /**
     * Get the id of the function. This is the unique id from the function definition.
     *
     * - returns: the policy id
     */
    func getId() -> String {
        return id
    }
    
    func toString() -> String {
        return id
    }

    /**
     * Return a string representation of this function. Useful for logging.
     *
     * - parameter configuration: the parameters for this function from the device policy
     *
     * - returns: a string representation of this function
     */
    func getDetails(configuration:Dictionary<String,Any>) -> String {
        return getId()
    }
    
    static func getDeviceFunction(functionId:String) -> DeviceFunction? {
        return POLICY_MAP[functionId]
    }
    
    private static func createInProcessMapKey(endpointId:String,
                                              deviceModelURN:String,
                                              attribute:String) -> String {
        return endpointId + "/deviceModels/" + deviceModelURN +
               ":attributes/" + attribute
    }

    public static func putInProcessValue(endpointId:String,
                                         deviceModelURN:String,
                                         attribute:String, value:AnyObject?) {
        inProcessValuesSyncQueue.sync() {
            inProcessValues[DeviceFunction.createInProcessMapKey(
                endpointId:endpointId, deviceModelURN:deviceModelURN,
                attribute:attribute)] = value
        }
    }
    
    public static func removeInProcessValue(endpointId:String,
                                            deviceModelURN:String,
                                            attribute:String) -> Any? {
        return inProcessValuesSyncQueue.sync(execute: {
            return inProcessValues.removeValue(forKey:createInProcessMapKey(
                endpointId:endpointId, deviceModelURN:deviceModelURN,
                attribute:attribute))
        })
    }
    
    public static func getInProcessValue(endpointId:String,
                                         deviceModelURN:String,
                                         attribute:String) -> Any? {
        return inProcessValuesSyncQueue.sync(execute: {
            return inProcessValues[createInProcessMapKey(endpointId:endpointId,
                deviceModelURN:deviceModelURN, attribute:attribute)]
        })
    }

    //
    //  routines called from device functions
    //
    // Called from alertCondition
    fileprivate static func createAlert(deviceAnalog:DeviceAnalog,
                configuration:Dictionary<String,Any>) throws -> AlertMessage {
        
        let deviceModel:DeviceModelImpl =
            deviceAnalog.getDeviceModel() as! DeviceModelImpl

        let deviceModelFormatMap:Dictionary<String,DeviceModelFormat> =
                deviceModel.getDeviceModelFormats()

        if deviceModelFormatMap.isEmpty {
            throw ClientError.argument(String(format:NSLocalizedString(
                "%@ does not contain alert formats", comment: ""),
                deviceModel.getURN()))
        }
        
        guard let format:String = configuration["urn"] as? String else {
            throw ClientError.argument(String(format:NSLocalizedString(
                "configuration for %@ does not contain alert format", comment: ""),
                    deviceModel.getURN()))
        }

        guard let deviceModelFormat:DeviceModelFormat =
                                        deviceModelFormatMap[format] else {
                
            throw ClientError.argument(String(format:NSLocalizedString(
                    "%@ does not contain alert format %@", comment: ""),
                                                deviceModel.getURN(), format))
        }

        var alertSeverity:AlertMessageSeverity = AlertMessageSeverity.normal
        if let severityConfig:String = configuration["severity"] as? String {
            alertSeverity =
                AlertMessageSeverity.valueOf(severity: severityConfig,
                                             defaultValue:.normal)
        }
        
        let builder:AlertMessageBuilder = AlertMessageBuilder()
        builder.setSource(source:deviceAnalog.getEndpointId())
        builder.setFormat(format:format)
        builder.setDescription(description:deviceModelFormat.getName())
        builder.setSeverity(severity:alertSeverity)
        
        let fieldsFromPolicy:Dictionary<String,AnyObject>? =
            configuration["fields"] as? Dictionary<String,AnyObject>
        let fields:[DeviceModelFormatField] = deviceModelFormat.getFields()
        for field in fields {
            
            guard let policyValue:AnyObject =
                    fieldsFromPolicy?[field.getName()] else {
                continue
            }
            do {
                let value:Any? = try convertArg(deviceAnalog: deviceAnalog,
                                            argType: field.getType(),
                                            arg: policyValue)
                try addDataItem(builder:builder, field:field,
                                value:value as AnyObject)
            } catch {
                // WARNING
                log(type:.fault,
                    message:"DeviceFunction.%@: bad value for '%@' in '%@' : %@",
                    args:"createAlert()", field.getName(),
                         deviceModel.getURN(), "\(error)")
                
            }
        }
        return try builder.build()
    }
    
    // TODO: Should this method throw ?
    fileprivate static func addDataItem(builder:AlertMessageBuilder,
                                    field:DeviceModelFormatField,
                                    value:AnyObject) throws {
        switch (field.getType()) {
        case .INTEGER, .NUMBER:
            if value is NSNumber {
                try builder.addDataItem(key:field.getName(),
                                        value:Double(truncating: value as! NSNumber))
            } else {
                throw ClientError.argument(String(format:NSLocalizedString(
                    "value of attribute '%@' is not a %@", comment: ""),
                    field.getName(), "\(field.getType())"))
            }
        break
        case .BOOLEAN:
            if value is Bool {
                try builder.addDataItem(key:field.getName(),
                                    value: Bool(value as! Bool))
            } else {
                throw ClientError.argument(String(format:NSLocalizedString(
                    "value of attribute '%@' is not a %@", comment: ""),
                    field.getName(), "\(field.getType())"))
            }
        break
        case .DATETIME:
            if value is NSNumber {
                try builder.addDataItem(key:field.getName(),
                                        value: Double(truncating: value as! NSNumber))
            } else if value is Date {
                try builder.addDataItem(key:field.getName(),
                            value:(value as! Date).timeIntervalSince1970)
            } else {
                throw ClientError.argument(String(format:NSLocalizedString(
                    "value of attribute '%@' is not a %@", comment: ""),
                    field.getName(), "\(field.getType())"))
            }
        break
        //case .STRING, .URI:
        default:
            try builder.addDataItem(key:field.getName(), value:"\(value)")
        break
        }
    }
    
    static func sqlMatches(s:String, regEx:String) -> Bool {
        var prev = ""
        var next = ""
        var nsregex = ""
        for c in regEx {
            if c == "_" && prev != "\\" {
                next = "."
            } else if c == "%" && prev != "\\" {
                nsregex.append(".")
                next = "*"
            } else {
                next = String(c)
            }
            nsregex.append(next)
            prev = next
        }
        do {
            let re = try NSRegularExpression(pattern: nsregex)
            let nmatches = re.numberOfMatches(in: s, range: NSRange(0..<s.count))
            return nmatches > 0
        } catch {
            return false
        }
    }
        
    static func compute(node:FormulaParserNode?,
                        deviceAnalog:DeviceAnalog) -> Any {

        guard let node = node else {
            return Double.nan
        }

        var result:Double = Double.nan
        
        if node is FormulaParserTerminal {
            let terminal:FormulaParserTerminal = node as! FormulaParserTerminal
            let attr:String = (node as! FormulaParserTerminal).getValue()
            switch (terminal.getTerminalType()) {
                case .currentAttribute:
                    do {
                        guard let value = try deviceAnalog.getAttributeValue(attribute:attr) else {
                            break
                        }
                        if value is NSNumber {
                            result = Double(truncating:(value as! NSNumber))
                        } else if value is Bool {
                            result = Bool(value as! Bool) ? 1.0 : 0.0
                        }
                    } catch {
                        // WARNING
                        log(type:.fault,
                            message:"DeviceFunction.%@: Cannot get value for attribute '%@':  %@",
                            args: "compute()", attr, "\(error)")
                    }
                    break
                case .inprocessAttribute:
                    do {
                        var value:Any? =
                            DeviceFunction.getInProcessValue(
                                endpointId: deviceAnalog.getEndpointId(),
                                deviceModelURN: deviceAnalog.getDeviceModel().getURN(),
                                attribute:attr)
                        if value == nil {
                            value = try deviceAnalog.getAttributeValue(
                                attribute:attr) as Any
                        }
                        if let value = value {
                            if value is NSNumber {
                                result = Double(truncating:value as! NSNumber)
                            } else if value is Bool {
                                result = Bool(value as! Bool) ? 1.0 : 0.0
                            }
                        }
                    } catch {
                        // WARNING
                        log(type:.fault,
                            message:"DeviceFunction.%@: Cannot get value for attribute '%@': %@",
                            args: "compute()", attr, "\(error)")
                    }
                    break
                case .number:
                    if let value:Double = Double(attr) {
                        result = value
                    }
                case .string, .ident:
                    return attr;
            }
            return result
        }

        if node.getNodeOperation() == .ternary {
            let cond:Double = DeviceFunction.compute(node:node.getLeftHandSide(),
                                                     deviceAnalog:deviceAnalog) as! Double
            if cond == 1.0 {
                return DeviceFunction.compute(node:node.getRightHandSide()!.getLeftHandSide(),
                                              deviceAnalog:deviceAnalog)
            } else {
                return DeviceFunction.compute(node:node.getRightHandSide()!.getRightHandSide(),
                                              deviceAnalog:deviceAnalog)
            }
        } else if node.getNodeOperation() == .group {
            return DeviceFunction.compute(node:node.getLeftHandSide(),
                                          deviceAnalog:deviceAnalog)
        }

        let lhs:Any = DeviceFunction.compute(node:node.getLeftHandSide(),
                                                deviceAnalog:deviceAnalog)
        let rhs:Any = DeviceFunction.compute(node:node.getRightHandSide(),
                                                deviceAnalog:deviceAnalog)
        
        //
        // TODO: According to many sources Java lang spec as seen here
        // https://www.securecoding.cert.org/confluence/display/java/NUM07-J.+Do+not+attempt+comparisons+with+NaN
        // excerpt:
        // NaN (not-a-number) is unordered, so the numerical comparison
        // operators <, <=, >, and >= return false if either or both operands
        // are NaN. The equality operator == returns false if either operand is
        // NaN, and the inequality operator != returns true if either operand
        // is NaN.
        //
        // I will code to these specs vs. the JCL implementation
        // Also in native iOS NaN == NaN -> false
        //
        
        func equalAny<BaseType: Equatable>(lhs: Any, rhs: Any,
                                            baseType: BaseType.Type) -> Bool {
            guard
                let lhsEquatable = lhs as? BaseType,
                let rhsEquatable = rhs as? BaseType
                else { return false }
            return lhsEquatable == rhsEquatable
        }

        let operation:FormulaParserNodeOperation = node.getNodeOperation()
        switch (operation) {
        case .function:
            let fn:String = (node.getLeftHandSide() as! FormulaParserTerminal).getValue()
            if "LOWER".caseInsensitiveCompare(fn) == .orderedSame {
                return (rhs as? String) == nil ? "" : (rhs as! String).lowercased()
            } else if "UPPER".caseInsensitiveCompare(fn) == .orderedSame {
                return (rhs as? String) == nil ? "" : (rhs as! String).uppercased()
            } else {
                // WARNING
                log(type:.fault,
                    message:"%@.%@: unknown function '%@'",
                    args: "DeviceFunction", "compute()", fn)
                return Double.nan
            }
        case .lower:
            return (lhs as! String).lowercased()
        case .upper:
            return (lhs as! String).uppercased()
        case .eq:
            if type(of:lhs) == type(of:rhs) {
                return equalAny(lhs:lhs, rhs:rhs, baseType: NSObject.self) ? 1.0 : 0.0
            } else {
                return 0.0
            }
        case .neq:
            if type(of:lhs) == type(of:rhs) {
                return equalAny(lhs:lhs, rhs:rhs, baseType: NSObject.self) ? 0.0 : 1.0
            } else {
                return 1.0;
            }
        case .plus:
            if lhs is String && rhs is String {
                let lhsString = (lhs as? String) ?? ""
                let rhsString = (rhs as? String) ?? ""
                return "\(lhsString)\(rhsString)"
            }
            break;
        case .like:
            if lhs is String && rhs is String {
                guard let lhsString = (lhs as? String),
                    let rhsString = (rhs as? String) else {
                        return 0.0
                }
                return sqlMatches(s: lhsString, regEx: rhsString) ? 1.0 : 0.0;
            } else {
                return 0.0;
            }
        default:
            break
        }
        
        guard let lhsdbl = lhs as? Double else {
            return Double.nan
        }
    
        switch (operation) {
            case .unaryMinus:
                return -lhsdbl
            case .unaryPlus:
                return +lhsdbl
            case .not:
                return lhsdbl == 0.0 ? 1.0 : 0.0
            default:
                break
        }
        
        guard let rhsdbl = rhs as? Double else {
            return Double.nan
        }
        
        switch (operation) {
            case .div:
                return lhsdbl / rhsdbl
            case .mul:
                return lhsdbl * rhsdbl
            case .plus:
                return lhsdbl + rhsdbl
            case .minus:
                return lhsdbl - rhsdbl
            case .mod:
                //return lhsdbl % rhsdbl
                return lhsdbl.truncatingRemainder(dividingBy: rhsdbl)
            case .or:
                // Let NaN or NaN be false
                if lhsdbl.isNaN && rhsdbl.isNaN {
                    return 0.0
                }
                // Anything or'd with NaN is true except 0.0
                return lhsdbl != 0.0 || rhsdbl != 0.0 ? 1.0 : 0.0
            case .and:
                // If lhsdbl or rhsdbl is NaN, return false
                if lhsdbl.isNaN || rhsdbl.isNaN {
                    return 0.0
                }
                return lhsdbl != 0.0 && rhsdbl != 0.0 ? 1.0 : 0.0
//            case .eq:
//                // NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
//                return lhsdbl == rhsdbl ? 1.0 : 0.0
//            case .neq:
//                return lhsdbl == rhsdbl ? 0.0 : 1.0
            case .gt:
                // TODO: As in the above excerpt anything relation with NaN
                // is false
                // NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
                // Let NaN > 42 return false, and 42 > NaN return true
                if lhsdbl.isNaN {
                    return 0.0
                }
                if rhsdbl.isNaN {
                    // JCL
                    return 1.0
                    //return 0.0
                }
                return lhsdbl > rhsdbl ? 1.0 : 0.0
            case .gte:
                // TODO: As in the above excerpt anything relation with NaN
                // is false
                // NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
                // Let NaN >= 42 return false, and 42 >= NaN return true
                if lhsdbl.isNaN {
                    // JCL
                    return rhsdbl.isNaN ? 1.0 : 0.0
                    //return 0.0
                }
                if rhsdbl.isNaN {
                    // JCL
                    return 1.0
                    //return 0.0
                }
                return lhsdbl >= rhsdbl ? 1.0 : 0.0
            case .lt:
                // TODO: As in the above excerpt anything relation with NaN
                // returns false
                // NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
                // Let NaN < 42 return false, and 42 < NaN return true
                if lhsdbl.isNaN {
                    return 0.0
                }
                if rhsdbl.isNaN {
                    // JCL
                    return 1.0
                    //return 0.0
                }
                return lhsdbl < rhsdbl ? 1.0 : 0.0
            case .lte:
                // TODO: As in the above excerpt anything relation with NaN
                // returns false
                // NaN.compareTo(42) == 1, 42.compareTo(NaN) == -1
                // Let NaN <= 42 return false, and 42 <= NaN return true
                if lhsdbl.isNaN {
                    // JCL
                    return rhsdbl.isNaN ? 1.0 : 0.0
                    //return 0.0
                }
                if rhsdbl.isNaN {
                    // JCL
                    return 1.0
                    //return 0.0
                }
                return lhsdbl <= rhsdbl ? 1.0 : 0.0
            case .ternary:
                break
            case .alternative:
                break
            // TODO: handled above ?
//            case .function:
//                break
            case .group:
                break
            case .terminal:
                break
            default:
                break
        }
        return Double.nan
    }
    
    // greatest common factor, e.g., gcd(90,60) = 30
    static func gcd(x:Int64, y:Int64) -> Int64 {
        return (y == 0) ? x : gcd(x:y, y:x % y)
    }

    // return x if x is >= y else y
    // x >= y -> orderedSame, orderedAscending
    // x < y -> orderedDescending
    static func max(x:NSNumber, y:NSNumber) -> NSNumber {
        var result:NSNumber = y
        switch(y.compare(x)) {
        case .orderedSame, .orderedAscending:
            result = x
            break
        case .orderedDescending:
            break
        }
        return result
    }

    // return x if x <= y else y
    // x <= y -> orderedSame, orderedDescending
    // x > y -> orderedAscending
    static func min(x:NSNumber, y:NSNumber) -> NSNumber {
        var result:NSNumber = x
        switch(y.compare(x)) {
        case .orderedSame, .orderedDescending:
            break
        case .orderedAscending:
            result = y
            break
        }
        return result
    }
    
    fileprivate func putPersistedBatchedData(deviceAnalog:DeviceAnalog,
                                             value:Pair<Message,StorageObject?>?) {
        guard let bbp:BatchByPersistence = batchByPersistence,
              let message:Message = value?.getKey() else {
            return
        }
        let saveSemaphore:DispatchSemaphore = DispatchSemaphore(value:0)
        bbp.save(messages: [message], endpointId:deviceAnalog.getEndpointId(),
                 callback: { saveSemaphore.signal() })
        saveSemaphore.wait()
    }
    
    fileprivate func getPersistedBatchedData(deviceAnalog:DeviceAnalog) ->
                                            [Pair<Message,StorageObject?>] {
        var pairs:[Pair<Message,StorageObject?>] = [Pair<Message,StorageObject?>]()
        guard let messages:[Message]  =
            batchByPersistence?.load(endpointId: deviceAnalog.getEndpointId()) else {
                return pairs
        }
        batchByPersistence?.delete(messages: messages)
                                                
        // TODO: JCK need to handle StorageObject
        /*
        let deviceModel:DeviceModelImpl = deviceAnalog.getDeviceModel() as! DeviceModelImpl
        let deviceModelAttributes:[String:DeviceModelAttribute] =
                                    deviceModel.getDeviceModelAttributes()
        */
    
        for message in messages {
            pairs.append(Pair(key: message, value: nil /*StorageObject()*/))
            /*
            //TODO: JCL need to handle storage object somehow
            if message is DataMessage {
                let dataMessage:DataMessage = message as! DataMessage
                let dataItems:[DataItem] = dataMessage.getDataItems()
                // go through data items to see if any are related to a storage object
                for dataItem in dataItems {
                    let key:String = dataItem.getKey()
                    // JCL assert not nil
                    if let deviceModelAttribute:DeviceModelAttribute =
                                deviceModelAttributes[key],
                            deviceModelAttribute.getType() ==
                                DeviceModelAttributeType.URI {
                        let uri:String = dataItem.getValue() as! String
                        // TODO: JCL ?
                        //deviceAnalog.
                    }
                }
            }
            */
        }
        return pairs
    }
}
