/*
 * Copyright (c) 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */
//com.oracle.iot.client.impl.device;

import Foundation
import os
import SwiftyJSON

public class DevicePolicyManagerChangeListener : NSObject {
    
    override
    init() {
        super.init()
    }
    
    /**
     * A callback for notification that a policy has been assigned to some
     * devices. The listener should check to see if the device is in the set of
     * assigned devices. If the device is not in the set of assigned devices,
     * the listener should take no action.
     *
     * - parameter devicePolicy: the policies that have been assigned.
     *
     * - parameter assignedDevices: the set of devices that to which the policy
     *                              has been assigned.
     */
    func policyAssigned(devicePolicy:DevicePolicy, assignedDevices:Set<String>) throws {
        fatalError("Subclass must implement 'policyAssigned'")
    }

    /**
     * A callback for notification that a policy has been unassigned from some
     * devices. The listener should check to see if the device is in the set of
     * unassigned devices. If the device is not in the set of unassigned
     * devices, the listener should take no action.
     *
     * - parameter devicePolicy: the policies that have been unassigned.
     *
     * - parameter unassignedDevices: the set of devices from which the policy
     *                                has been unassigned.
     */
    func policyUnassigned(devicePolicy:DevicePolicy, unassignedDevices:Set<String>) throws {
        fatalError("Subclass must implement 'policyUnassigned'")
    }
}

fileprivate let syncQueue:DispatchQueue =
    DispatchQueue(label:"DevicePolicyManager sync queue", qos: .utility)

fileprivate let policyMapQueue:DispatchQueue =
    DispatchQueue(label:"DevicePolicyManager policy map sync queue", qos: .utility)

fileprivate let getPolicyQueue:DispatchQueue =
    DispatchQueue(label:"DevicePolicyManager get policy sync queue", qos: .utility)

fileprivate let devicePolicyManagerMapQueue:DispatchQueue =
    DispatchQueue(label:"DevicePolicyManagerMap queue", qos:.utility)
fileprivate var devicePolicyManagerMap:Dictionary<String,DevicePolicyManager> =
    Dictionary<String,DevicePolicyManager>()

fileprivate func addDevicePolicyManager(devicePolicyManager:DevicePolicyManager,
                                        endpointId:String) {
    devicePolicyManagerMapQueue.sync() {
        devicePolicyManagerMap[endpointId] = devicePolicyManager
    }
}

func removeDevicePolicyManager(endpointId:String) {
    _ = devicePolicyManagerMapQueue.sync() {
        devicePolicyManagerMap.removeValue(forKey:endpointId)
    }
}

class DevicePolicyMap {
    
    // TODO: refactor naming of methods and prune what is not needed
    
    private let mapQueue:DispatchQueue =
        DispatchQueue(label:"policyMap queue", qos: .utility)
    
    // Map a device id to the policies that are available to it.
    // The key is the device id.
    // The value is a map of device model URN to plicy id. The policy id
    // gets us to the configuration data.
    //
    // { <device-id> : { <device-model-urn> : <policy-id> }}
    private var policiesByDeviceId:[String:[String:String?]]! //=
        //[String:[String:String?]]()
    
    // model deviceModels/<device-model-urn>/devicePolicies/<policy-id>/devices
    // policiesByDeviceModelUrn is { <device-model-urn> : { <policy-id> : [ <device-id>...] }
    private var policiesByDeviceModelUrn:[String:[String:Set<String>]]! // =
        //[String:[String:Set<String>]]()
    
    // map policy id to policies.
    // The key is a policy id.
    // The value is a map of policy attributes and values.
    // { policyId : DevicePolicy }
    //private let policiesByPolicyIdQueue:DispatchQueue =
    //    DispatchQueue(label:"policiesByPolicyId queue", qos: .utility)
    private var policiesByPolicyId:[String:DevicePolicy]!// =
        //[String:DevicePolicy]()
    
    init() {
        self.policiesByPolicyId = [String:DevicePolicy]()
        self.policiesByDeviceId = [String:[String:String?]]()
        self.policiesByDeviceModelUrn = [String:[String:Set<String>]]()
    }
    
    init(policiesByPolicyId:[String:DevicePolicy],
              policiesByDeviceId:[String:[String:String?]],
              policiesByDeviceModelUrn:[String:[String:Set<String>]]) {
        self.policiesByPolicyId = policiesByPolicyId
        self.policiesByDeviceId = policiesByDeviceId
        self.policiesByDeviceModelUrn = policiesByDeviceModelUrn
    }

    // Update all maps with a new policy - from getPolicy
    // Update all maps policy is unassigned - from unassignPolicyFromDevice
    // Update all maps policy changed - from lookupPolicyForDevice
    func update(deviceId:String, deviceModelUrn:String, devicePolicyId:String?,
                removeUrn:Bool = false) {
        mapQueue.sync() {
            var urn2policy:[String:String?]! = policiesByDeviceId[deviceId]
            if urn2policy == nil {
                urn2policy = [String:String?]()
            }
            if removeUrn {
                urn2policy.removeValue(forKey:deviceModelUrn)
            } else {
                urn2policy[deviceModelUrn] = devicePolicyId
            }
            policiesByDeviceId[deviceId] = urn2policy
            
            var policy2devices:[String:Set<String>]! =
                policiesByDeviceModelUrn[deviceModelUrn]
            if policy2devices == nil {
                policy2devices = [String:Set<String>]()
                policiesByDeviceModelUrn[deviceModelUrn] = policy2devices
            }
            if let devicePolicyId = devicePolicyId {
                var devices:Set<String>! = policy2devices[devicePolicyId]
                if devices == nil {
                    devices = Set<String>()
                }
                devices.insert(deviceId)
                policy2devices[devicePolicyId] = devices
                policiesByDeviceModelUrn[deviceModelUrn] = policy2devices
            }
        }
    }
    // Return true if the map is updated.
    // Update the policiesByPolicyId map if force is true or if the
    // devicePolicy is not already mapped. - from lookupPolicyForDevice
    func update(devicePolicy:DevicePolicy, force:Bool) -> Bool {
        return mapQueue.sync(execute: {
            let policyId:String = devicePolicy.getId()
            if !force && policiesByPolicyId[policyId] != nil {
                return false
            } else {
                policiesByPolicyId[policyId] = devicePolicy
                return true
            }
        })
    }
    
    /**
     * - returns: the `DevicePolicy` for `devicePolicyId`
     */
    func getDevicePolicy(devicePolicyId:String) -> DevicePolicy? {
        return mapQueue.sync(execute: {
            return policiesByPolicyId[devicePolicyId]
        })
    }
    
    /**
     * - return: the mapped `DevicePolicy`
     */
    func getDevicePolicy(deviceId:String, deviceModelUrn:String) -> DevicePolicy? {
        return mapQueue.sync(execute: {
            if let m2P:[String:String?] = self.policiesByDeviceId[deviceId],
                    let policyId:String = m2P[deviceModelUrn]! {
                return policiesByPolicyId[policyId]
            } else {
                return nil
            }
        })
    }
    
    /**
     * Just add the policy to policiesByPolicyId, the assumption is that
     * other maps are up to date, called from "processPipelineChanged"
     */
    func putDevicePolicy(devicePolicy:DevicePolicy) {
        mapQueue.sync() {
            policiesByPolicyId[devicePolicy.getId()] = devicePolicy
        }
    }

    /**
     * Full population of maps with this policy.
     */
    func putDevicePolicy(devicePolicy:DevicePolicy,
                         deviceId:String,
                         deviceModelUrn:String) {
        mapQueue.sync() {
            let policyId:String = devicePolicy.getId()
            policiesByPolicyId[policyId] = devicePolicy
            // assignedDevices { <policyId> : { <deviceId>... }
            var assignedDevices:[String:Set<String>]! =
                policiesByDeviceModelUrn[deviceModelUrn]
            if assignedDevices == nil {
                assignedDevices = [String:Set<String>]()
            }
            // devices { <deviceId>... }
            var devices:Set<String>! = assignedDevices[policyId]
            if devices == nil {
                devices = Set<String>()
            }
            devices.insert(deviceId)
            assignedDevices[policyId] = devices
            policiesByDeviceModelUrn[deviceModelUrn] = assignedDevices
            
            var urn2policy:[String:String?]? = policiesByDeviceId[deviceId]
            if urn2policy == nil {
                urn2policy = [String:String?]()
            }
            urn2policy![deviceModelUrn] = policyId
            policiesByDeviceId[deviceId] = urn2policy!
        }
    }
    
    /**
     * - returns the policyId mapped to `deviceId` and `deviceModelUrn` or `nil`
     *   if there is no mapping
     */
    func get(deviceId:String, deviceModelUrn:String) -> String? {
        // Don't care what the value is
        var urnExists:Bool = false
        return get(deviceId:deviceId, deviceModelUrn:deviceModelUrn, urnExists: &urnExists)
    }
    
    /**
     * Sets `urnExists` to true if there is a mapping for `deviceModelUrn` in
     * `policiesByDeviceId`
     *
     * - returns the policyId mapped to `deviceId` and `deviceModelUrn` or `nil`
     *   if there is no mapping
     */
    func get(deviceId:String, deviceModelUrn:String, urnExists: inout Bool) -> String? {
        return mapQueue.sync(execute: {
            urnExists = false
            var policyId:String? = nil
            // policiesByDeviceId looks like { <device-id> : { <device-model-urn> : <policy-id> }}
            // If devicePolicies is nil, we drop through and do the lookup.
            if let urn2policy:[String:String?] = self.policiesByDeviceId[deviceId] {
                // If the deviceModelURN is not found in the map, we drop through
                // and do a lookup. There may be a mapping for the device model
                // urn, but the value may be nil, which means that there is no
                // policy for the combination of device model and device.
                urnExists = urn2policy.keys.contains(deviceModelUrn)
                if urnExists {
                    // Force unwrap of the possibly non existent key entry,
                    // but we know it is there because of the contains.
                    // Remember to insert a "nil" String? value it must be
                    // v[k] = nil as String?
                    policyId = urn2policy[deviceModelUrn]!
                }
            }
            return policyId
        })
    }
    
    // HACK: This should probably also remove the association since this
    // method doesn't know the context of the caller. But this method was
    // created for IOT-48979 and it just needs to remove the policy so that
    // it will be loaded again. And should call removePolicyFromLocalStore to
    // do that.
    func removeDevicePolicy(devicePolicyId:String) {
        mapQueue.sync(){
            if let _ = policiesByPolicyId.removeValue(forKey:devicePolicyId) {
                deleteFileFromLocalStore(filename:String(
                    format:DEVICEPOLICY_FILENAME_FORMAT, devicePolicyId))
            }
        }
    }
    
    // TODO: Should associations always be removed ? there was one instance
    // of this code that did not in policyChanged.
    func removeDevicePolicy(deviceModelUrn:String, devicePolicyId:String) {
        mapQueue.sync() {
            // May not be there
            policiesByPolicyId.removeValue(forKey:devicePolicyId)
            
            // policies { <policyId>: { <deviceId>... } }
            var policies:[String:Set<String>]? = policiesByDeviceModelUrn[deviceModelUrn]
            let assignedDevices:Set<String>? = policies?.removeValue(forKey:devicePolicyId)
            if policies != nil && assignedDevices != nil {
                // restore modified map
                policiesByDeviceModelUrn[deviceModelUrn] = policies
            }
            let associations:PolicyAssociation = PolicyAssociation()
            // assignedDevices == { <deviceId>... }
            if assignedDevices != nil {
                for deviceId in assignedDevices! {
                    if var devicePolicies:[String:String?] = policiesByDeviceId[deviceId] {
                        devicePolicies.removeValue(forKey:devicePolicyId)
                        policiesByDeviceId[deviceId] = devicePolicies
                    }
                    associations.removeAssociation(deviceId:deviceId,
                                                   deviceModelUrn:deviceModelUrn,
                                                   devicePolicyId:devicePolicyId)
                }
            } else {
                // There are no devices assigned to this policy
                // { <deviceId>, { <modelurn>, <policyId> }}
                for (deviceId,var urn2policy) in policiesByDeviceId {
                    // I think policyId always equals devicePolicyId if there is
                    // a deviceModelUrn entry
                    if let policyId:String = urn2policy[deviceModelUrn]!,
                            policyId == devicePolicyId {
                        urn2policy[deviceModelUrn] = nil as String?
                        policiesByDeviceId[deviceId] = urn2policy
                        associations.removeAssociation(deviceId:deviceId,
                                                       deviceModelUrn:deviceModelUrn,
                                                       devicePolicyId:devicePolicyId)
                    }
                }
            }
        }
    }
    
    func getAssignedDevices(deviceModelUrn:String, devicePolicyId:String) -> Set<String>? {
        return mapQueue.sync(execute: {
            return policiesByDeviceModelUrn[deviceModelUrn]?[devicePolicyId]
        })
    }

    func isAssigned(deviceId:String, deviceModelUrn:String? = nil,
                    devicePolicyId:String? = nil,
                    assignedPolicyId: inout String?) -> Bool {
        if let deviceModelUrn = deviceModelUrn, let devicePolicyId = devicePolicyId {
            guard let m2p:[String:String?] = policiesByDeviceId[deviceId],
                    let _assignedPolicyId = m2p[deviceModelUrn] else {
                return false
            }
            assignedPolicyId = _assignedPolicyId
            return assignedPolicyId == devicePolicyId
        }
        // Do a long lookup via policiesByDeviceModelUrn
        // { <device-model-urn> : { <policy-id> : [ <device-id>...] }
        for (_,p2d) in policiesByDeviceModelUrn {
            for (pid,dids) in p2d {
                if dids.index(of:deviceId) != nil {
                    assignedPolicyId = pid
                    return true
                }
            }
        }
        return false
    }
    
    func sync<V>(execute: () throws ->V?) rethrows -> V? {
        return try mapQueue.sync(execute: execute)
    }
    
    func sync(execute: ()->()) {
        mapQueue.sync(execute: execute)
    }
}

fileprivate let policyAssociationsQueue:DispatchQueue =
    DispatchQueue(label:"policyAssociationsQueue", qos:.utility)

class PolicyAssociation {
    
    // Associations should look like this:
    // {
    //  "devicePolicyIdsToEndpointIds": {
    //    "devicePolicyId1": [
    //      "endpoint1",
    //      "endpoint2"
    //    ],
    //    "devicePolicyId2", [
    //      "endpointId3",
    //      "endpointId4"
    //    ]
    //  },
    //  "deviceModelUrnsToDevicePolicies": {[
    //    "deviceModelUrn1": [
    //      "devicePolicyId1",
    //      "devicePolicyId2"
    //    ],
    //    "deviceModelUrn2": [
    //      "devicePolicyId3",
    //      "devicePolicyId4"
    //    ]
    //  }
    // }
    let DEVICE_ASSOCIATIONS_FILE:String = "device-associations.json"
    
    var associations:[String:JSON]! = nil
    
    init() {
        guard let _associations = getPolicyAssociations() else {
            associations = [String:JSON]()
            return
        }
        associations = _associations
    }
    
    func getModelsToPolicies() -> JSON? {
        return associations["deviceModelUrnsToDevicePolicies"]
    }
    
    func getPoliciesToDevices() -> JSON? {
        return associations["devicePolicyIdsToEndpointIds"]
    }
    
    func addAssociation(deviceId:String, deviceModelUrn:String,
                        devicePolicyId:String) {
        // Find the device model associated with the policy and remove
        // the association.
        var modified:Bool = false
        
        // { <urn> : [ <policyid>, ... ]}
        var modelToPolicies:[String:JSON]! =
            associations["deviceModelUrnsToDevicePolicies"]?.dictionaryValue
        
        // Special case no mappings
        if modelToPolicies == nil {
            modelToPolicies = JSON().dictionaryValue
            var modelPolicies:[JSON] = JSON().arrayValue
            modelPolicies.append(JSON(devicePolicyId))
            modelToPolicies[deviceModelUrn] = JSON(modelPolicies)
            associations["deviceModelUrnsToDevicePolicies"] =
                JSON(modelToPolicies)
            modified = true
        } else {
            // [ <policyid>, ... ]
            var modelPolicies:[JSON]! = modelToPolicies[deviceModelUrn]?.arrayValue
            if modelPolicies == nil {
                modelPolicies = JSON().arrayValue
            }
            // TODO: Consider using JSON("a") == JSON("a") and then
            // index(of:x)
            var index:Int? = nil
            for n in 0..<modelPolicies.endIndex {
                if devicePolicyId == modelPolicies[n].stringValue {
                    index = n
                    break
                }
            }
            // Not in the map
            if index == nil {
                modelPolicies.append(JSON(devicePolicyId))
                modified = true
                modelToPolicies[deviceModelUrn] = JSON(modelPolicies)
                associations["deviceModelUrnsToDevicePolicies"] =
                    JSON(modelToPolicies)
                modified = true
            }
        }
        
        // associations to the endpoint ids for policy
        // { <policyid> : [ <endpointid>,,, ]}
        var policyToDevices:[String:JSON]! =
            associations["devicePolicyIdsToEndpointIds"]?.dictionaryValue
        // special case no mappings
        if policyToDevices == nil {
            policyToDevices = JSON().dictionaryValue
            var endpointIds:[JSON] = JSON().arrayValue
            endpointIds.append(JSON(deviceId))
            policyToDevices[devicePolicyId] = JSON(endpointIds)
            associations["devicePolicyIdsToEndpointIds"] = JSON(policyToDevices)
            modified = true
        } else {
            var endpointIds:[JSON]! = policyToDevices[devicePolicyId]?.arrayValue
            if endpointIds == nil {
                endpointIds = JSON().arrayValue
            }
            var index:Int? = nil
            for n in 0..<endpointIds.endIndex {
                if deviceId == endpointIds[n].stringValue {
                    index = n
                    break
                }
            }
            if index == nil {
                endpointIds.append(JSON(deviceId))
                modified = true
                policyToDevices[devicePolicyId] = JSON(endpointIds)
                associations["devicePolicyIdsToEndpointIds"] =
                    JSON(policyToDevices)
                modified = true
            }
        }
        if modified {
            putPolicyAssociations(policyAssociations:JSON(associations))
        }
    }
    
    // TODO: If there no more endpoints associated with a policy, should the
    // policy be unmapped from the modelurn map ?
    // removes only from associations["devicePolicyIdsToEndpointIds"]
    func removeAssociation(deviceId:String, deviceModelUrn:String, devicePolicyId:String) {
        var modified:Bool = false
        if var policyToDevices:[String:JSON] =
                associations["devicePolicyIdsToEndpointIds"]?.dictionaryValue {
            if var deviceIds:[JSON] = policyToDevices[devicePolicyId]?.arrayValue {
                for n in 0..<deviceIds.endIndex {
                    if deviceId == deviceIds[n].stringValue {
                        deviceIds.remove(at:n)
                        modified = true
                        break
                    }
                }
                if modified {
                    if deviceIds.count == 0 {
                        policyToDevices.removeValue(forKey:devicePolicyId)
                    } else {
                        policyToDevices[devicePolicyId] = JSON(deviceIds)
                    }
                    associations["devicePolicyIdsToEndpointIds"] = JSON(policyToDevices)
                }
            }
        }
        if modified {
            putPolicyAssociations(policyAssociations:JSON(associations))
        }
        
    }
    
    // called from removePolicyFromLocalStore, but that method is never called.
    // updates both "deviceModelUrnsToDevicePolicies" and "devicePolicyIdsToEndpointIds"
    func removeAssociation(deviceModelUrn:String, devicePolicyId:String) {
        // Find the device model associated with the policy and remove
        // the association.
        var modified:Bool = false
        if var modelToPolicies:[String:JSON] =
                associations["deviceModelUrnsToDevicePolicies"]?.dictionaryValue {
            if var modelPolicies:[JSON] = modelToPolicies[deviceModelUrn]?.arrayValue {
                var index:Int? = nil
                for n in 0..<modelPolicies.endIndex {
                    if devicePolicyId == modelPolicies[n].stringValue {
                        index = n
                        break
                    }
                }
                if index != nil {
                    modelPolicies.remove(at:index!)
                    modelToPolicies[deviceModelUrn] = JSON(modelPolicies)
                    associations["deviceModelUrnsToDevicePolicies"] =
                        JSON(modelToPolicies)
                    modified = true
                }
                // If there are no more policies for deviceModelUrn
                // remove the empty array
                if modelPolicies.count == 0 {
                    modelToPolicies.removeValue(forKey:deviceModelUrn)
                    associations["deviceModelUrnsToDevicePolicies"] =
                        JSON(modelToPolicies)
                    modified = true
                }
            }
        }
        
        // Remove all associations to the endpoint ids for policy
        // that was removed.
        if var policyToDevices:[String:JSON] =
                associations["devicePolicyIdsToEndpointIds"]?.dictionaryValue {
            if policyToDevices.removeValue(forKey:devicePolicyId) != nil {
                associations["devicePolicyIdsToEndpointIds"] = JSON(policyToDevices)
                modified = true
            }
        }
        if modified {
            putPolicyAssociations(policyAssociations:JSON(associations))
        }
    }
    
    func getPolicyAssociations() -> [String:JSON]? {
        guard LOCAL_STORE else {
            return nil
        }
        
        guard let associationsFile:String =
                policyAssociationsQueue.sync(execute: {
                    return getFileFromLocalStore(filename:DEVICE_ASSOCIATIONS_FILE)}) else {
            return nil
        }
        
        guard let jsonAssociationData:Foundation.Data =
                associationsFile.data(using:.utf8,
                                  allowLossyConversion:false),
                !jsonAssociationData.isEmpty  else {
            // Bad data delete it
            log(type:.fault,
                message: "Device associations file %@ corrupted.",
                args: DEVICE_ASSOCIATIONS_FILE)
            policyAssociationsQueue.sync() {
                deleteFileFromLocalStore(filename:DEVICE_ASSOCIATIONS_FILE)
            }
            return nil
        }
        
        let jsonAssociations:JSON = JSON(jsonAssociationData)
        guard !jsonAssociations.isEmpty else {
            policyAssociationsQueue.sync() {
                deleteFileFromLocalStore(filename:DEVICE_ASSOCIATIONS_FILE)
            }
            return nil
        }
        return jsonAssociations.dictionaryValue
    }

    private func putPolicyAssociations(policyAssociations:JSON) {
        guard LOCAL_STORE else {
            return
        }
        policyAssociationsQueue.sync() {
            let content:String = policyAssociations.rawString()!
            if !saveFileToLocalStore(filename:DEVICE_ASSOCIATIONS_FILE,
                                         content:content) {
                log(type:.fault,
                    message:"%@: cannot save policy associations %@",
                    args: "DevicePolicyManager.putPolicyAssociations()",
                    DEVICE_ASSOCIATIONS_FILE)
            }
        }
    }
}

fileprivate var LOCAL_STORE:Bool = false
// The device policy filename format
fileprivate let DEVICEPOLICY_FILENAME_FORMAT:String = "dp-%@.json"
// Needed for Messaging Applications
public let DEVICE_POLICY_CAPABILITY_URN:String =
    "urn:oracle:iot:dcd:capability:device_policy"

class DevicePolicyManager {

    
    let DEVICE_ASSOCIATIONS_FILE:String = "device-associations.json"
    
    public static let CAPABILITY_URN:String = DEVICE_POLICY_CAPABILITY_URN
    public static let PATH:String = "deviceModels/" + CAPABILITY_URN + "/policyChanged"
    
    // Have to use lock between getPolicy and policyChanged, not synchronized method,
    // because policyChanged will call notifyPolicyAssigned/Unassigned, which will cause
    // the virutual device impl or the messaging policy impl to call getPolicy. If
    // we use syncrhonization at the method level, policyChanged will block getPolicy.
    private var policyChangeLock:DispatchQueue = DispatchQueue(label:"policyChangeLock",
                                                               qos:.utility) //new ReentrantLock();
    private var listenersQueue:DispatchQueue =
        DispatchQueue(label:"DevicePolicyManagerChangeListener queue", qos:.utility)

    private var changeListeners:[/*WeakReference*/ DevicePolicyManagerChangeListener] =
        [/*WeakReference*/DevicePolicyManagerChangeListener]()

    private var devicePolicyMap:DevicePolicyMap!

    private var /*WeakReference*/ secureConnection:SecureConnection? = nil
    
    
    // A list of device id's that have obtained a policy manager and may
    // have had policies loaded locally. They will be notified when
    // their change listener has been added.
    // [ <deviceId>:<policyId> ]
    private var pendingNotifications:[String:String] = [String:String]()
    
    let initializeSemaphore:DispatchSemaphore = DispatchSemaphore(value:0)
    init(endpointId:String, secureConnection:SecureConnection) {
        self.secureConnection = /*WeakReference*/ secureConnection
        
        //TODO: Is this needed ?
        // Yes. In JCL there is a "PersistentStore" which holds a "bunch"
        // of references. Until iOS has a generalized store, this is needed
        //syncQueue.sync() {
        //    devicePolicyManagerMap[endpointId] = self
        //}
        addDevicePolicyManager(devicePolicyManager:self,
                               endpointId:endpointId)
        
        LOCAL_STORE = haveLocalStore()
        self.devicePolicyMap = self.initializeFromPersistedData()
        var assignedPolicyId:String?
        if self.devicePolicyMap.isAssigned(deviceId:endpointId,
                                        assignedPolicyId: &assignedPolicyId) {
            pendingNotifications[endpointId] = assignedPolicyId
        }
    }

    
    /**
     * Add a listener for receiving notification of policies being assigned or
     * unassigned.
     *
     * - parameter changeListener: the `DevicePolicyManagerChangeListener` to add
     */
    public func addChangeListener(changeListener:DevicePolicyManagerChangeListener) {
        listenersQueue.sync() {
            // Don't add a listener twice
            if self.changeListeners.index(of:changeListener) != nil {
                return
            }
            self.changeListeners.append(/*WeakReference*/changeListener)
            // HACK to see if this change listener needs to be notifiedappend
            // of locally added policies
            guard let endpointId:String =
                    (changeListener as? VirtualDeviceImpl)?.getEndpointId() else {
                return
            }
            if let policyId:String = pendingNotifications.removeValue(forKey:endpointId) {
                do {
                    if let devicePolicy:DevicePolicy =
                            devicePolicyMap.getDevicePolicy(devicePolicyId:policyId) {
                        try changeListener.policyAssigned(devicePolicy:devicePolicy,
                                                      assignedDevices: [endpointId])
                    }
                } catch {
                    // SEVERE listener may throw exception.
                    log(type: .error,
                        message:"%@: notify for local policies threw exception %@",
                        args: "DevicePolicyManager.addChangeListener():",
                        "\(error)")
                }
                
            }
        }
    }
    
    /**
     * Remove a listener from receiving notification of policies being added or
     * removed.
     *
     * - parameter changeListener: the `DevicePolicyManagerChangeListener` to remove
     */
    public func removeChangeListener(changeListener:DevicePolicyManagerChangeListener) {
        listenersQueue.sync() {
            if let index = self.changeListeners.index(of:changeListener) {
                self.changeListeners.remove(at: index)
            }
        }
    }
    
    /**
     * Invoke `policyAssigned` method on change listeners.
     *
     * - parameter devicePolicy: the assigned policy
     *
     * - parameter assignedDevices: the devices to which the policy was assigned.
     */
    private func notifyPolicyAssigned(devicePolicy:DevicePolicy,
                                      assignedDevices:Set<String>) {
    
        if assignedDevices.isEmpty {
            return
        }
        
        listenersQueue.sync() {
            for changeListener in changeListeners {
                do {
                    try changeListener.policyAssigned(devicePolicy:devicePolicy,
                                            assignedDevices: assignedDevices)
                } catch {
                    // SEVERE listener may throw exception.
                    log(type: .error,
                        message:"%@: listener threw exception %@",
                        args: "DevicePolicyManager.notifyPolicyAssigned():",
                              "\(error)")
                }
            }
        }
    }
    
    /**
     * Invoke `policyUnassigned` method on change listeners.
     *
     * - parameter devicePolicy: the assigned policy
     *
     * - parameter unassignedDevices: the devices to which the policy was assigned.
     */
    private func notifyPolicyUnassigned(devicePolicy:DevicePolicy,
                                        unassignedDevices:Set<String>) {
    
        if unassignedDevices.isEmpty {
            return
        }
    
        listenersQueue.sync() {
            for changeListener in changeListeners {
                do {
                    try changeListener.policyUnassigned(devicePolicy:devicePolicy,
                                            unassignedDevices:unassignedDevices)
                } catch {
                    // SEVERE listener may throw exception.
                    log(type: .error,
                        message:"%@: listener threw exception %@",
                        args: "DevicePolicyManager.notifyPolicyUnassigned():",
                              "\(error)")
                }
            }
        }
    }
    
    public static func getDevicePolicyManager(dcd:AdvancedDirectlyConnectedDevice) -> DevicePolicyManager? {
        // TODO: JCL uses "PersistentStore" an "in memory store"
        return devicePolicyManagerMapQueue.sync(execute: {
            guard let dpm:DevicePolicyManager =
                    devicePolicyManagerMap[dcd.getEndpointId()] else {
                // SEVERE
                log(type:.error,
                    message:"%@: cannot access DevicePolicyManager for %@",
                    args:"DevicePolicyManager.getDevicePolicyManager()",
                         dcd.getEndpointId())
                return nil
            }
            return dpm
        })
    }

    /* GET RestAPi.getPrivateRoot()/deviceModels/<urn>/devicePolicies
     * {
     *    "offset":0,
     *    "count":1,
     *    "limit":200,
     *    "hasMore":false,
     *        "items":[{
     *            "pipelines":[{
     *            "attributeName":"humidity"
     *            "pipeline":[
     *                {
     *                    "id":"mean",
     *                    "parameters":{
     *                        "window":15000
     *                }
     *                },
     *                {
     *                    "id":"alertCondition",
     *                    "parameters":{
     *                        "filter":true,
     *                        "condition":"$(humidity) > $(maxThreshold)",
     *                        "alert":{
     *                            "urn":"urn:com:oracle:iot:device:humidity_sensor:too_humid",
     *                            "severity":"SIGNIFICANT",
     *                            "fields":{
     *                                "humidity":"$(humidity)",
     *                                "threshold":"$(maxThreshold)"
     *                            }
     *                        }
     *                    }
     *                },
     *                {
     *                    "id":"eliminateDuplicates",
     *                    "parameters":{
     *                        "window":60000
     *                    }
     *                }
     *            ],
     *        }],
     *        "createdAsString":"Mon Dec 04 10:37:48 EST 2017",
     *        "lastModifiedAsString":"Mon Dec 04 10:37:48 EST 2017",
     *        "created":1512401868143,
     *        "name":"Humidity-Sensor-Policy-for-Client-Library-Samples",
     *        "description":"Send average humidity over 15 seconds,
     *        or send an alert if the average humidity exceeds the threshold,
     *        finally do not send duplicate data within 1 minute",
     *        "userLastModified":"mockserver",
     *        "id":"0f4112b2-a66e-418f-9c9e-4574638e5e35",
     *        "lastModified":1512401868143
     *    }]
     * }
     */
    
    // TODO: Temporary declaration to allow refactoring, until async can be
    // implemented
    /**
     * Get the `DevicePolicy` for the given device model, and device id.
     *
     * - parameter deviceModelURN: the device model URN
     *
     * - parameter deviceId: the device id
     *
     * - throws: `ClientError.network` if policies cannot be loaded
     *     `ClientError.security` if there is a security exception trying to load
     *                            policies.
     *
     * - returns: the `DevicePolicy` or `nil`
     */
    public func getPolicy(deviceModelURN:String,
                          deviceId:String) throws -> DevicePolicy? {
        
        let waitForPolicies:DispatchSemaphore = DispatchSemaphore(value:0)
        var devicePolicy:DevicePolicy? = nil
        do {
            try self.getPolicy(deviceModelURN:deviceModelURN,
                               deviceId:deviceId,
                               callback: {(policy, error) in
                if let error = error {
                    // WARNING
                    log(type:.fault,
                        message:"%@: deviceModelUrn:%@, deviceId:%@, returned error. %@",
                        args:"DevicePolicyManager.getPolicy()",
                        deviceModelURN, deviceId, "\(error)")
                }
                devicePolicy = policy
                waitForPolicies.signal()
            })
        } catch {
            return devicePolicy
        }
        // TODO: Should use timeout
        //print("getPolicy: Waiting for getPolicy")
        waitForPolicies.wait()
        //print("getPolicy: Return from waiting for getPolicy")
        return devicePolicy
        
    }
    
    /**
     * Get the `DevicePolicy` for the given device model, and device id.
     *
     * - parameter deviceModelURN: the device model URN
     *
     * - parameter deviceId: the device id
     *
     * - throws: `ClientError.network` if policies cannot be loaded
     *     `ClientError.security` if there is a security exception trying to load
     *                            policies.
     *
     * - returns: the `DevicePolicy` or `nil`
     */
    private func getPolicy(deviceModelURN:String, deviceId:String,
            callback: @escaping (DevicePolicy?, ClientError?)->()) throws {
        
        // If we already have the policy for this device model and device, return it.
        // Otherwise, look it up. If there isn't an entry in policiesByDeviceId
        // for the device model urn and device id when this method is called,
        // there will be by the time this method completes.
        //
        
        // Set to true if there is a modelurn mapping, but policyId is nil
        var urnExists:Bool = false
        var devicePolicy:DevicePolicy? = nil
        // If the policyId is not nil use it
        if let devicePolicyId:String =
                devicePolicyMap.get(deviceId:deviceId,
                                    deviceModelUrn:deviceModelURN,
                                    urnExists:&urnExists) {
            // Assumes that if there is mapped policyId then the policy
            // exists in the policies map
            // Very important that the client has gotten
            // the policy before we get here!
            devicePolicy = devicePolicyMap.getDevicePolicy(devicePolicyId: devicePolicyId)
            callback(devicePolicy, nil)
            return
        } else if urnExists {
            // If the deviceModelURN was not found in the map, then devicePolicyId
            // will be nil and we drop through and do the lookup.
            // There may be a mapping for the device model urn, but the policyId may
            // be nil (urnExists == true),
            // which means that there is no policy for the combination of device
            // model and device.
            callback(nil, nil)
            return
        }
        
        // stop policyChanged while doing this lookup.
        // TODO: May need to be reentrant lock
        try policyChangeLock.sync() {
            do {
                //defer {
                //    policyChangeLock.unlock()
                //}
                
                // If we get to here, then there was no mapping for the deviceId
                // in policiesByDeviceId, or there was a mapping for the deviceId,
                // but not for the device model. So we need
                // to do a lookup and update policiesByDeviceId
                let devicePolicy:DevicePolicy? =
                    try self.lookupPolicyForDevice(deviceModelUrn:deviceModelURN,
                                                   deviceId:deviceId)
                
                // Add the mapping so the code doesn't try to fetch a policy for this
                // device again. The only way the device will get a policy after this
                // is from an "assigned" policyChanged, or when the client restarts.

                devicePolicyMap.update(deviceId: deviceId,
                                    deviceModelUrn: deviceModelURN,
                                    devicePolicyId: devicePolicy?.getId())
                
                if let devicePolicy = devicePolicy {
                    var assignedDevices:Set<String> = Set<String>()
                    assignedDevices.insert(deviceId)
                    self.notifyPolicyAssigned(devicePolicy:devicePolicy,
                                              assignedDevices:assignedDevices)
                }
                
                callback(devicePolicy, nil)
            }
        }
    }
    
    /**
     * Handle `deviceModels/urn:oracle:iot:dcd:capability:device_policy/policyChanged`
     *
     * - parameter requestMessage: the RequestMessage from the server
     *
     * - throws: `ClientError.network` if there is an exception from handling the request
     */
    public func policyChanged(directlyConnectedDevice:AdvancedDirectlyConnectedDevice,
                              requestMessage:RequestMessage?) throws -> ResponseMessage {

        //
        // The server will not send a policyChanged to a device for which the policy is not intended.
        // If this is a DCD, then the policy  is meant for this DCD.
        // If this is a GW, then the policy is meant for one or more of its ICDs.
        //
        // RequestMessage body looks like:
        // [{
        //    "deviceModelUrn": "urn:com:oracle:iot:device:humidity_sensor",
        //    "id": "547B66F3-5DC8-4F60-835F-7B7773C8EE7A",
        //    "lastModified": 1511805927387,
        //    "op": "changed"
        // }]
        //
        // Where op is:
        //   "changed"    - The policy pipeline was changed. The client needs to GET the policy.
        //   "assigned"   - The policy was assigned to device(s). The policy pipeline itself
        //                  has not changed. The server will not send this to the client
        //                  unless the client has the device(s). A gateway client needs to get
        //                  a list of devices the policy applies to, but a directly connected
        //                  device can assume the policy is for it. If necessary, the client
        //                  will GET the policy.
        //   "unassigned" - The policy was unassigned from device(s). The policy pipeline itself
        //                  has not changed. The server will not send this to the client
        //                  unless the client has the device(s). A gateway client needs to get
        //                  a new list of devices the policy applies to, but a directly connected
        //                  device can assume the policy is for it.
        //
        
        let dcdIsGatewayDevice:Bool = directlyConnectedDevice is AdvancedGatewayDevice
        let endpointId:String = directlyConnectedDevice.getEndpointId()
        let body:Foundation.Data = requestMessage!.getBody()
        
        // block second half of getPolicy while doing an update.
        try policyChangeLock.sync() {
            
            let items:[JSON] = JSON(body).arrayValue
            for item in items {
                let op:String = item["op"].stringValue
                let deviceModelUrn:String = item["deviceModelUrn"].stringValue
                let policyId:String = item["id"].stringValue
                let lastModified:Int64 = item["lastModified"].int64Value
                
                // debug
                log(type:.debug,
                    message:"%@: op:%@, deviceModelUrn:%@, policyId:%@, lastModified:%@",
                    args: "DevicePolicyManager.policyChanged()",
                          op, deviceModelUrn, policyId,
                          ISO8601DateFormatter().string(from:Date(
                            timeIntervalSince1970: Double(lastModified) / 1000.0)))

                switch(op) {
                case "unassigned":
                    // TODO: JCL does not throw
                    try processUnassign(deviceModelUrn:deviceModelUrn,
                                    policyId:policyId,
                                    endpointId:endpointId,
                                    isGateway:dcdIsGatewayDevice,
                                    lastModified:lastModified)
                break
                case "assigned":
                    try processAssign(deviceModelUrn:deviceModelUrn,
                                    policyId:policyId,
                                    endpointId:endpointId,
                                    isGateway:dcdIsGatewayDevice,
                                    lastModified:lastModified)
                    break
                case "changed":
                    // If device policy is nil, then something is wrong in our mappings.
                    // Remove the references to this device model urn and policy id.
                    guard let policyBeforeChange:DevicePolicy =
                            devicePolicyMap.getDevicePolicy(devicePolicyId:policyId) else {
                        devicePolicyMap.removeDevicePolicy(deviceModelUrn:deviceModelUrn,
                                                          devicePolicyId:policyId)
                        break
                    }
                    
                    // Before updating the policy, notify the devices the policy is unassigned.
                    // This gives the code in VirtualDeviceImpl or MessagingPolicyImpl a chance
                    // to clean up the exisiting pipeline before the new pipeline comes in.
                    let assignedDevices =
                            devicePolicyMap.getAssignedDevices(deviceModelUrn:deviceModelUrn,
                                                               devicePolicyId:policyId)
                    // The devices we are going "unassign" are the currently "assigned" devices
                    if let unassignedDevices = assignedDevices {
                        notifyPolicyUnassigned(devicePolicy:policyBeforeChange,
                                               unassignedDevices:unassignedDevices)
                    }
                    
                    try processPipelineChanged(directlyConnectedDevice:directlyConnectedDevice,
                                           deviceModelUrn:deviceModelUrn,
                                           policyId:policyId,
                                           lastModified:lastModified)
                    
                    if let assignedDevices = assignedDevices {
                        // Get the new devicePolicy
                        guard let policyAfterChange:DevicePolicy =
                                devicePolicyMap.getDevicePolicy(devicePolicyId:policyId) else {
                            log(type:.fault,
                                message:"%@: changed policy not found, op:changed, " +
                                    "policyId:%@, deviceModelUrn:%@, endpointId:%@",
                                args:"DevicePolicyManager.policyChanged()",
                                policyId, deviceModelUrn, endpointId)
                            //return
                            break
                        }
                        notifyPolicyAssigned(devicePolicy:policyAfterChange,
                                             assignedDevices:assignedDevices)
                    }
                    break
                default:
                    // WARNING
                    log(type:OSLogType.fault, message: "%@ invalid operation %@",
                        args: requestMessage!.getURL(), op)
                }
            }
        }

        do {
            let responseMessageBuilder:ResponseMessageBuilder =
                ResponseMessageBuilder(requestMessage: requestMessage!)
            responseMessageBuilder.setStatusCode(statusCode: StatusCode.OK)
            return try responseMessageBuilder.build()
        } catch {
            print("\(error)")
            throw error
        }
    }

    /**
     * Process the "change" operation from policyChanged.
     * The method needs to fetch the policy.
     *
     * - parameter deviceModelUrn: the device model urn of the policy that is
     *                             unassigned
     *
     * - parameter policyId: the id of the policy that is unassigned.
     *
     * - throws: `ClientError.state` if there is an error parsing the JSON
     *       `ClientError.networkd` if thrown from SecureConnection API
     *       `ClientError.security` if thrown from SecureConnection API
     */
    private func processPipelineChanged(directlyConnectedDevice:AdvancedDirectlyConnectedDevice,
                                        deviceModelUrn:String, policyId:String,
                                        lastModified:Int64) throws {
        // first, check to see if we have a copy,
        // and if so, whether or not it is more recent.
        if let currentDevicePolicy:DevicePolicy =
                devicePolicyMap.getDevicePolicy(devicePolicyId:policyId),
                Double(lastModified) < currentDevicePolicy.getLastModified() {
            // our copy is more recent, return.
            return
        }
        
        // Download the policy
        // block getPolicy from accessing policiesByDeviceId while the policy is being updated.
        let devicePolicy:DevicePolicy! =
            try downloadPolicy(deviceModelUrn:deviceModelUrn, policyId:policyId)
        
        // Assumption is that other maps have the policyId mapped
        devicePolicyMap.putDevicePolicy(devicePolicy:devicePolicy)

        // FINE
        if devicePolicy != nil  {
            log(type:.debug,
                message:"%@: %@ : PolicyChanged : \"%@\"",
                args:"DevicePolicyManager.processPipelineChanged()",
                     directlyConnectedDevice.getEndpointId(),
                     devicePolicy.toString())
            
        }
        // Nothing else to do here...
    }
    
    /**
     * Process the "assign" operation from policyChanged.
     * The method needs to notify listeners that the policy was assigned,
     * then update data structures and persistence to add the association
     * of the policy to the device.
     *
     * - parameter deviceModelUrn: the device model urn of the policy that is
     *                            unassigned
     *
     * - parameter policyId: the id of the policy that is unassigned.
     *
     * - parameter endpointId: the endpointId of the device that called the
     *                         policyChanged method
     *
     * - parameter isGateway: whether or not that device is a gateway device
     */
    private func processAssign(deviceModelUrn:String,
                               policyId:String,
                               endpointId:String,
                               isGateway:Bool,
                               lastModified:Int64) throws {
        
        // get the set of devices to which this policy is assigned.
        // If the directly connected device is a gateway, then get
        // the assigned devices from the server. Otherwise, the
        // assigned device is the directly connected device with endpointId.
        //
        // Note that if the call to get the ICD ids that are assigned
        // to the policy throws an exception, unassign every device
        // with that device model urn. In such a case, we  remove the
        // mapping from device id the device-model urn from
        // policiesByDeviceId. If there is no mapping, getPolicy
        // will try to create a new mapping by getting the policy
        // for the device from the server.
        var assignedDevices:Set<String>! = nil
        if isGateway {
            var couldNotGetIcds:Bool = true
            do {
                assignedDevices =
                    try getIndirectlyConnectedDeviceIdsForPolicy(deviceModelUrn:deviceModelUrn,
                                                             policyId: policyId,
                                                             directlyConnectedOwner:endpointId)
                
                couldNotGetIcds = false
                
            } catch {
                // ignore
                // debug
                log(type:.debug,
                    message:"%@: deviceModelUrn:%@, policyId:%@, endpointId:%@, " +
                        "getIndirectlyConnectedDeviceIdsForPolicy() threw exception %@",
                    args: "DevicePolicyManager.processAssign():",
                        deviceModelUrn, policyId, endpointId, "\(error)")
            }
            // finally but always gets here anyway
            if couldNotGetIcds {
                // remove the mappings for all devices that reference this policy
                // since we no longer know which policy they refer to. On next call
                // to getPolicy, this should self-correct.
                devicePolicyMap.removeDevicePolicy(deviceModelUrn: deviceModelUrn,
                                             devicePolicyId:policyId)
                return
            }
        } else {
            assignedDevices = Set<String>()
            assignedDevices.insert(endpointId)
        }
        
        if assignedDevices == nil || assignedDevices.isEmpty {
            return
        }
        // removing policy from policiesByPolicyId map will
        // cause the policy to be downloaded again. The reason
        // we have to download again on assign is that the
        // policy may have been modified while it was not
        // assigned to this device or ICDs.
        devicePolicyMap.removeDevicePolicy(devicePolicyId:policyId)
        
        for deviceId in assignedDevices {
            // TODO: JCL Does not throw
            try assignPolicyToDevice(
                deviceModelUrn:deviceModelUrn, policyId:policyId,
                deviceId:deviceId, lastModified:lastModified)
        }

        if let devicePolicy:DevicePolicy =
                devicePolicyMap.getDevicePolicy(devicePolicyId:policyId) {
            notifyPolicyAssigned(devicePolicy:devicePolicy,
                                     assignedDevices:assignedDevices)
        }
        
    }

    private func assignPolicyToDevice(deviceModelUrn:String,
                                      policyId:String,
                                      deviceId:String,
                                      lastModified:Int64) throws {
        // if the current policy is the same as the policy that is being
        // assigned, no need to continue.
        var currentPolicyId:String? = nil
        if devicePolicyMap.isAssigned(deviceId: deviceId,
                                      deviceModelUrn: deviceModelUrn,
                                      devicePolicyId:policyId,
                                      assignedPolicyId:&currentPolicyId) {
            return
        }
        // Make sure we have the policy
        // before assigning the new policy.
        var devicePolicy:DevicePolicy! =
            devicePolicyMap.getDevicePolicy(devicePolicyId:policyId)
        if devicePolicy == nil {
            devicePolicy = loadLocalDevicePolicy(deviceModelUrn: deviceModelUrn,
                                                 devicePolicyId: policyId)
            if devicePolicy == nil {
                devicePolicy = try downloadPolicy(deviceModelUrn: deviceModelUrn,
                                                  policyId: policyId)
            }
        }
            
        if devicePolicy != nil {
            devicePolicyMap.putDevicePolicy(devicePolicy:devicePolicy,
                                            deviceId:deviceId,
                                            deviceModelUrn:deviceModelUrn)
        }
        // debug
        if devicePolicy != nil {
            log(type:.debug,
                message:"%@: deviceId:%@, policyId:%@, devicePolicy:%@",
                args: "DevicePolicyManager.assignPolicyToDevice()",
                deviceId, policyId, devicePolicy!.toString())
        }
        
        if currentPolicyId != nil {
            removePersistedAssociation(deviceModelURN: deviceModelUrn,
                                       policyId: currentPolicyId!,
                                       deviceId: deviceId)
        }
        persistAssociation(deviceModelURN: deviceModelUrn,
                           policyId: policyId,
                           deviceId: deviceId)
    }
    
    /**
     * Process the "unassign" operation from policyChanged.
     * The method updates the data structures and persistence to remove the
     * association of the policy to the device.
     *
     * - parameter deviceModelUrn: the device model urn of the policy that is
     *                             unassigned
     * - parameter policyId: the id of the policy that is unassigned.
     *
     * - parameter endpointId: the endpointId of the device that called the
     *                         policyChanged method
     * - parameter isGateway: whether or not that device is a gateway device
     *
     * - parameter lastModified: is the time the policy was last modified on
     *                           the server.
     * - returns: the set of devices affected by the unassign
     */
    private func processUnassign(deviceModelUrn:String,
                                 policyId:String,
                                 endpointId:String,
                                 isGateway:Bool,
                                 lastModified:Int64) throws {
        // Get the set of devices to which this policy is assigned. This
        // This will be the difference of the set of devices the client
        // says are assigned and the set of devices the server says are
        // assigned (the server doesn't say who was unassigned, we can
        // only ask who is assigned).

        // { <endpointId>... }
        // Get currently assigned devices, if there aren't any return
        guard var unassignedDevices:Set<String> =
                devicePolicyMap.getAssignedDevices(deviceModelUrn: deviceModelUrn,
                                               devicePolicyId: policyId) else {
            return
        }
        
        // The flag "couldNotGetIcds" lets us know that
        // getIndirectlyConnectedDeviceIdsForPolicy failed.
        var couldNotGetIcds:Bool = isGateway
        
        // If we aren't a gateway device, then make sure the
        // assigned devices contains the directly connected device
        // endpoint id, and ensure that the only element of
        // unassignedDevices is the directly connected device
        // endpont id.
        // If the directly connected device is not a gateway, then we
        // know that the subject of the unassign is the directly connected
        // device and there is no need to make a call to the server for
        // the assigned devices.
        //
        if !isGateway {
            if !unassignedDevices.contains(endpointId) {
                // this endpoint is not currently assigned to the policy
                return
            }
            // TODO: This doesn not seem to be needed given the logic above
            // and given that this is a DCD, there shouldn't be any more content
            // and since contains is true, then it is in the Set
            //unassignedDevices.removeAll()
            //unassignedDevices.insert(endpointId)
        } else {
        
            // If the device is a gateway...
            // Now get the set of devices to which this policy is assigned,
            // according to the server. Remove the set of server assigned
            // devices from the client assigned devices so that
            // unassignedDevices is now the set of devices that have
            // been unassigned from this policy.
            //

            // Note that if the call to get the set of ICD ids that are assigned
            // to the policy might fail, throwing an exception. In this case,
            // there is no way to tell what devices belong to the policy or not.
            // To handle this situation, every device on the client that has
            // this policy will be will unassigned from the policy _and_
            // the device's mapping to the device model urn in policiesByDeviceId
            // will be removed. Removing the mapping ensures that getPolicy
            // will fetch the policy anew and the mapping will self correct.

           //if isGateway {
            do {
                let serverAssignedDevices:Set<String> =
                    try getIndirectlyConnectedDeviceIdsForPolicy(
                        deviceModelUrn: deviceModelUrn, policyId: policyId,
                        directlyConnectedOwner: endpointId)
                
                // returned without exception, couldNotGetIcds is false.
                couldNotGetIcds = false
                // remove the unassigned icds
                // TODO: update maps
                unassignedDevices.subtract(serverAssignedDevices)
                
                // if unassignedDevices is empty now that we removed
                // all the ICD ids from the server, we should return
                // since there are no devices on the client affected
                // by the change.
                //if unassignedDevices.isEmpty {
                //    return
                //}
            } catch {
                // ignored
                log(type:.debug,
                    message:"%@: deviceModelUrn:%@, policyId:%@, endpointId:%@, " +
                    "isGateway:%@, lastModified:%d, %@ threw exception %@",
                    args:"DevicePolicyManager.processUnassign()",
                    deviceModelUrn, policyId, endpointId,
                    "\(isGateway)", lastModified,
                    "getIndirectlyConnectedDeviceIdsForPolicy",
                    "\(error)")
            }
        }
        // if unassignedDevices is empty now that we removed
        // all the ICD ids from the server, we should return
        // since there are no devices on the client affected
        // by the change.
        // TODO: But what about updating maps/persistence ?
        if !unassignedDevices.isEmpty {
            // JCL assert  devicePolicy != nil
            // TODO: What should iOS do ?
            guard let devicePolicy:DevicePolicy =
                    devicePolicyMap.getDevicePolicy(devicePolicyId:policyId) else {
                log(type:.error,
                    message:"%@: deviceModelUrn:%@, policyId:%@, endpointId:%@, " +
                        "isGateway:%@, lastModified:%d, " +
                        "assert failed %@ is nil",
                    args:"DevicePolicyManager.processUnassign()",
                        deviceModelUrn, policyId, endpointId,
                        "\(isGateway)", lastModified,
                        "policiesByPolicyId[\(policyId)]")
                // TODO: return or throw ? update maps ?
                return
            }
            notifyPolicyUnassigned(devicePolicy: devicePolicy,
                                   unassignedDevices: unassignedDevices)
        }

        // Now unassignedDevices is the set of device ids that
        // have been unassigned from this policy.
        for deviceId in unassignedDevices {
            
            _ = unassignPolicyFromDevice(deviceModelUrn: deviceModelUrn,
                                     policyId: policyId, deviceId: deviceId,
                                     lastModified: lastModified,
                                     couldNotGetIcds:couldNotGetIcds)
        }
    }
    // TODO: return value is unused
    /**
     * Handle the logic for unassigning a policy from a device. The only reason
     * for this  method to return false is if the client has a more recent
     * change than what it was told by the server.
     *
     * - parameter deviceModelUrn: the device model urn from which the policy
     *                             is unassigned
     *
     * - parameter policyId: the policy id of the policy that is unassigned
     *
     * - parameter deviceId: the device from which the policy is unassigned
     *
     * - parameter lastModified: the lastModification time from the change
     *                           request on the server
     *
     * - returns: whether or not the policy was unassigned.
     */
    private func unassignPolicyFromDevice(deviceModelUrn:String,
                                          policyId:String,
                                          deviceId:String,
                                          lastModified:Int64,
                                          couldNotGetIcds:Bool) -> Bool {
        
        // Sanity check... does this device have the unassigned policy?
        // return true if not.

        // Get the current policyId mapped to this device/modelurn
        
        // isAssigned returns true if the device is assigned to policyId and
        // false if not.
        // currentPolicyId will contain the policy the that is assigned to the
        // the device. It will be nil if no policy is assigned.
        
        var currentPolicyId:String? = nil
        let isAssigned = devicePolicyMap.isAssigned(deviceId:deviceId,deviceModelUrn:deviceModelUrn,
                                      devicePolicyId:policyId,
                                      assignedPolicyId:&currentPolicyId)
        // No policy assigned
        if currentPolicyId == nil {
            return true
        }

        if isAssigned {
            // If nothing is wrong in our mapping, then
            // set the current policy for this device and
            // device model urn to null. This state causes
            // getPolicy to return null without futher lookup.
            //policies.put(deviceModelUrn, null);

            devicePolicyMap.update(deviceId:deviceId, deviceModelUrn:deviceModelUrn,
                                   devicePolicyId: nil, removeUrn:couldNotGetIcds)
            removePersistedAssociation(deviceModelURN: deviceModelUrn,
                                       policyId: policyId, deviceId: deviceId)
            return true
            
        }
        // There is something wrong with the mapping
        // policyId does not == currentPolicyId
        
        // server is telling me to unassign a policy id
        // that the client doesn't have assigned. If
        // the policy that is assigned is newer than
        // lastModified, then the client is right and
        // we move on. Otherwise, unassign whatever
        // policy the device has and let the state
        // auto-correct on the next call to getPolicy.
        
        if let devicePolicy:DevicePolicy =
            devicePolicyMap.getDevicePolicy(devicePolicyId:currentPolicyId!) {
            if devicePolicy.getLastModified() > Double(lastModified) {
                // client info is newer, move on to the next device id
                return false
            }
        }
        
        // if badMapping is set to true, the policiesByDeviceId entry for
        // the device-model urn of this device is removed. On the next
        // call to getPolicy, the map will auto-correct.

        // If we get here it is a bad mapping clean up maps
        // server info is newer and
        // this device has a bad mapping
        
        // if there was something bad in our mapping,
        // the remove the deviceModelUrn entry altogether.
        // On the next call to getPolicy for this device
        // and device model, the map will be corrected.
        
        // Oh my. The device maps to some policy that
        // the client doesn't know about. Remove the mapping
        // and policiesByPolicyId will self correct for this
        // device the next time getPolicy is called.
        // Note that since devicePolicy is null, getPolicy
        // will return null for this device and device model anyway,
        // so taking an axe to policiesByPolicyId is okay here.
        //
        // policiesByDeviceId is { <device-id> : { <device-model-urn> : <policy-id> } }
        devicePolicyMap.removeDevicePolicy(deviceModelUrn: deviceModelUrn,
                                           devicePolicyId: currentPolicyId!)
        removePersistedAssociation(deviceModelURN:deviceModelUrn,
                                   policyId:currentPolicyId!,
                                   deviceId:deviceId)
        return true
    }
    
    /**
     * GET `RestApi.getPrivateRoot()/deviceModels/{urn}/devicePolicies/{policyId}`.
     * The downloaded policy is persisted, the policiesByPolicyId map
     * is updated, and an entry is made (if necessary) in the
     * `policiesByDeviceModelUrn` map. The downloaded policy is returned
     * via the callback
     *
     * - parameter deviceModelUrn: the device model urn
     *
     * - parameter policyId: the policy id
     *
     * - parameter callback: a closure to be called when policies are loaded
     *
     * - throws: `ClientError.network` from server communication
     *      `ClientError.security` from accessing trusted assets
     */
    private func downloadPolicy(deviceModelUrn:String,
                                policyId:String) throws -> DevicePolicy? {
        // data is the device policy JSON bytes.
        // We either get it from the local store, or from the server.
        // loadLocalDevicePolicies returns null if there is no local store,
        // or no device policies file for the device model urn
        
        let fields:String = "id,pipelines,enabled,lastModified"
        let query:String = "/" + policyId + "?fields=" + fields
        // There is only one if any policy returned
        guard let jsonPolicy:JSON =
                try get(query:query, deviceModelUrn:deviceModelUrn) else {
            return nil
        }

        let devicePolicy:DevicePolicy =
                self.devicePolicyFromJSON(deviceModelUrn:deviceModelUrn,
                                      jsonObject:jsonPolicy)
        // persist the policy
        persistPolicy(policy:jsonPolicy)
        // maps updated by caller

        // debug
        log(type:.debug,
             message:"%@: deviceModelUrn:%@, policyId:%@, devicePolicy:%@",
             args:"DevicePolicyManager.downloadPolicy()",
             deviceModelUrn, policyId, devicePolicy.toString())
        
        return devicePolicy
    }
    
    /**
     * Lookup the policy for the combination of device model and device id on
     * the server. Should only be called from getPolicy when there is no policy
     * for the device.
     *
     * - parameter deviceModelUrn: the device model urn
     *
     * - parameter deviceId: the device id to query for
     *
     * - returns: the JSON policy, or `nil` if there is no policy for the
     *            combination of deviceModelUrn and deviceId
     *
     * - throws: `ClientError.network` thrown by SecureConnection if there is a
     *           network error or unexpected HTTP response
     *      `ClientError.security` thrown by SecureConnection if there is an
     *           error from trusted assets
     */
    private func lookupPolicyForDevice(deviceModelUrn:String,
                                       deviceId:String) throws  -> DevicePolicy? {
        // Do we already have the policy?
        if let devicePolicy:DevicePolicy =
            devicePolicyMap.getDevicePolicy(deviceId:deviceId,
                                            deviceModelUrn:deviceModelUrn) {
            return devicePolicy
        }
        
        // data is the device policy JSON bytes.
        // We either get it from the local store, or from the server.
        // loadLocalDevicePolicies returns null if there is no local store,
        // or no device policies file for the device model urn
        let devicesDotId:String = "{\"devices.id\":\"" + deviceId + "\"}"
        let fields:String = "id,pipelines,enabled,lastModified"
        let query:String = "/?q=" + devicesDotId + "&fields=" + fields
        guard let jsonItems:JSON =
                    try get(query:query,deviceModelUrn:deviceModelUrn),
                let jsonPolicy:JSON = getJsonItem(items:jsonItems) else {
            return nil
        }
        
        let devicePolicy:DevicePolicy =
            self.devicePolicyFromJSON(deviceModelUrn:deviceModelUrn,
                                      jsonObject:jsonPolicy)
        // persist the policy
        persistPolicy(policy:jsonPolicy)
        
        let policyId:String = devicePolicy.getId()
        
        // Only put the policy in policiesByPolicyId if it isn't there already.
        // This prevents putting a changed policy, which should be processed
        // through policyChanged.
        if devicePolicyMap.update(devicePolicy:devicePolicy, force:false) {
            // debug
            log(type:.debug,
                message:"%@: deviceModelUrn:%@, deviceId:%@, policy:%@",
                args:"DevicePolicyManager.lookupPolicyForDevice()",
                deviceModelUrn, deviceId, devicePolicy.toString())
        }
        
        // remember this policy maps to this device model.
        // Do not add the device id to the set of device ids here.
        // Do that in getPolicy (just to keep all of the device
        // id state updates in one place)
        // TODO: Will do it all here, JCL does not. Not sure if this will be
        // an issue later on.
        devicePolicyMap.update(deviceId:deviceId, deviceModelUrn:deviceModelUrn,
                            devicePolicyId: policyId)
        
        persistAssociation(deviceModelURN: deviceModelUrn,
                           policyId: policyId,
                           deviceId: deviceId)
        
        return devicePolicy
    }
    
    /**
     * Get the ids of indirectly connected devices that are assigned to the
     * policy.
     *
     * - parameter deviceModelUrn: the device model urn
     *
     * - parameter policyId: the policy id to query for
     *
     * - returns: the set of indirectly connected device ids for this policy
     * - throws: `ClientError.network` is thrown by SecureConnection if there
     *                         is a network error or unexpected HTTP response
     *       `ClientError.security` is thrown by SecureConnection if there is
     *                              an error from trusted assets
     */
    private func getIndirectlyConnectedDeviceIdsForPolicy(deviceModelUrn:String,
                                policyId:String, directlyConnectedOwner:String)
                                throws  -> Set<String> {
                                    
        // GET RestApi.getPrivateRoot()/deviceModels/{urn}/devicePolicies/{id}/devices?
        //     q={"directlyConnectedOwner" : "GW-endpoint-id"}
        let icdFilter:String = "{\"directlyConnectedOwner\":\"" + directlyConnectedOwner + "\"}"
        let query:String = "/\(policyId)/devices/?q=" + icdFilter + "&fields=id"
        var jsonItems:JSON! = nil
        do {
            jsonItems = try get(query:query, deviceModelUrn:deviceModelUrn)
        } catch {
            // If there are any exceptions, no data, non OK status, return
            // empty set, exception will have been logged in get()
            return Set<String>()
        }
        var icdIds:Set<String> = Set<String>()
        if let jsonIcds:[JSON] = jsonItems.dictionaryValue["items"]?.arrayValue {
            // elements are dictionaries
            for jsonIcd in jsonIcds {
                let icd:[String:JSON] = jsonIcd.dictionaryValue
                if let id = icd["id"]?.stringValue {
                    icdIds.insert(id)
                }
            }
        }
        return icdIds
    }
    
    
    /**
     * Write the JSON policy to local persistence.
     *
     * - parameter policy: the policy
     */
    private func persistPolicy(policy:JSON) {
        
        guard LOCAL_STORE else {
            return
        }
        guard let devicePolicyId:String =
                policy.dictionaryValue["id"]?.stringValue else {
            // WARNING
            log(type:.fault,
                message:"%@: policy not persisted, no policy id",
                args: "DevicePolicyManager.persistPolicy()")
            return
        }
        // TODO: Need local store sync
        let filename:String = String(format:DEVICEPOLICY_FILENAME_FORMAT, devicePolicyId)
        if let policyJson:String = policy.rawString() {
            if !saveFileToLocalStore(filename:filename, content:policyJson) {
                log(type:.fault,
                    message: "%@: cannot persist device policy %@",
                    args:"DevicePolicyManager.persistPolicy",
                         devicePolicyId)
            }
        }
    }
    
    /**
     * Update device-associations.json by adding an association between the
     * device model urn and the policy,
     * and also between the policy and device id. A device can only be associated with one policy;
     * if an association for the device already exists, it is removed and the new association is added.
     *
     * @param deviceModelURN the device model urn
     * @param policyId the policy id
     * @param deviceId the device id
     */
    // synchronized
    private func persistAssociation(deviceModelURN:String, policyId:String, deviceId:String) {

        let associations:PolicyAssociation = PolicyAssociation()
        associations.addAssociation(deviceId:deviceId,
                                    deviceModelUrn:deviceModelURN,
                                    devicePolicyId:policyId)
    }
    
    /**
     * Update device-associations.json by removing the association between the
     * policy and device id.
     *
     * - parameter deviceModelURN: the device model urn
     *
     * - parameter policyId: the policy id
     *
     * - parameter deviceId: the device id
     */
    // synchronized
    private func removePersistedAssociation(deviceModelURN:String,
                                            policyId:String, deviceId:String) {
        let associations:PolicyAssociation = PolicyAssociation()
        associations.removeAssociation(deviceId:deviceId,
                                       deviceModelUrn:deviceModelURN,
                                       devicePolicyId:policyId)
        
    }
    
    /**
     * Initialize policy maps from persisted data.
     */
    private func initializeFromPersistedData() -> DevicePolicyMap {
        guard LOCAL_STORE else {
            return DevicePolicyMap()
        }
    
        // Associations should look like this:
        // {
        //  "devicePolicyIdsToEndpointIds": {
        //    "devicePolicyId1": [
        //      "endpoint1",
        //      "endpoint2"
        //    ],
        //    "devicePolicyId2", [
        //      "endpointId3",
        //      "endpointId4"
        //    ]
        //  },
        //  "deviceModelUrnsToDevicePolicies": {
        //    "deviceModelUrn1": [
        //      "devicePolicyId1",
        //      "devicePolicyId2"
        //    ],
        //    "deviceModelUrn2": [
        //      "devicePolicyId3",
        //      "devicePolicyId4"
        //    ]
        //  }
        // }
        // If there are model to policy associations load them
        
        let associations:PolicyAssociation = PolicyAssociation()
        
        var policiesByPolicyId:[String:DevicePolicy] = [String:DevicePolicy]()
        // { <device-id> : { <device-model-urn> : <policy-id> }}
        var policiesByDeviceId:[String:[String:String?]] =
            [String:[String:String?]]()
        // { <device-model-urn> : { <policy-id> : [ <device-id>...] }}
        var policiesByDeviceModelUrn:[String:[String:Set<String>]] =
            [String:[String:Set<String>]]()
        
        let m2PJson:JSON = associations.getModelsToPolicies() ?? JSON()
        let p2dJson:JSON = associations.getPoliciesToDevices() ?? JSON()
        // (String,JSON)
        for (modelUrn,policyIdsJson) in m2PJson.dictionaryValue {
            for policyIdJson in policyIdsJson.arrayValue {
                let policyId:String = policyIdJson.stringValue
                var devicePolicy:DevicePolicy? =
                    loadLocalDevicePolicy(deviceModelUrn:modelUrn,
                                          devicePolicyId:policyId)
                if devicePolicy == nil {
                    do {
                        devicePolicy = try downloadPolicy(deviceModelUrn:modelUrn,
                                                          policyId:policyId)
                    } catch {
                        // log
                        continue
                    }
                }
                if devicePolicy == nil {
                    continue
                }
                policiesByPolicyId[policyId] = devicePolicy
                
                // debug
                log(type:.debug,
                    message:"%@: deviceModelUrn:%@, policyId:%@, policy:%@",
                    args:"DevicePolicyManager.initializeFromPersistedData()",
                    modelUrn, policyId, devicePolicy!.toString())

                // There should be a one to one relationship to policyId's found
                // in deviceModelUrnsToDevicePolicies
                // { <policyId> : { <deviceId>... }}
                let devicesByPolicyIdJson:[String:JSON] = p2dJson.dictionaryValue
                let deviceIdsJson:[JSON] =
                    devicesByPolicyIdJson[policyId]?.arrayValue ?? [JSON]()
                
                var m2P:[String:String?] = [String:String?]()
                m2P[modelUrn] = policyId
                // { <policyId> : { <deviceId>... }}
                var assignedDevices:[String:Set<String>] = [String:Set<String>]()
                var devices:Set<String> = Set<String>()
            
                for deviceIdJson in deviceIdsJson {
                    let deviceId:String = deviceIdJson.stringValue
                    // Add this device to the set of devices that have this policy
                    devices.insert(deviceId)
                    // { <deviceid> : { <deviceModelUrn : <policyId> }
                    policiesByDeviceId[deviceId] = m2P
                }
                // { <policy-id> : [ <device-id>...] }
                assignedDevices[policyId] = devices
                // { <device-model-urn> : { <policy-id> : [ <device-id>...] }}
                policiesByDeviceModelUrn[modelUrn] = assignedDevices
            }
        }
        return DevicePolicyMap(policiesByPolicyId:policiesByPolicyId,
                               policiesByDeviceId:policiesByDeviceId,
                               policiesByDeviceModelUrn:policiesByDeviceModelUrn)
    }
    
    /**
     * Remove the policy from the local store.
     * TODO: NOT USED AT THIS TIME. LEAVE HERE FOR FUTURE.
     *
     * - parameter devicePolicy: policy to remove
     */
    //private
    func removePolicyFromLocalStore(devicePolicy:DevicePolicy) {
        // Need policy id and device model urn to update
        // device-associations.json
        let policyId:String = devicePolicy.getId()
        let deviceModelUrn:String = devicePolicy.getDeviceModelURN()
        removePolicyFromLocalStore(policyId:policyId,
                                   deviceModelUrn:deviceModelUrn)
    }
    
    func removePolicyFromLocalStore(policyId:String,
                                    deviceModelUrn:String) {
        guard LOCAL_STORE else {
            return
        }
        deleteFileFromLocalStore(filename:String(
            format:DEVICEPOLICY_FILENAME_FORMAT, policyId))
        let associations:PolicyAssociation = PolicyAssociation()
                associations.removeAssociation(deviceModelUrn:deviceModelUrn,
                                       devicePolicyId:policyId)
    }
    
    // returns null if LOCAL_STORE is null, or there is no file for the device policy id
    func loadLocalDevicePolicy(deviceModelUrn:String, devicePolicyId:String) -> DevicePolicy? {
        guard let policyJson:JSON =
            loadLocalDevicePolicy(devicePolicyId: devicePolicyId) else {
                return nil
        }
        // This JSON is in the "items : [ ]" format
        // We know that this array is not empty
        //guard let policyJson:JSON = policyItemsJSON.dictionaryValue["items"],
        //    !policyJson.arrayValue.isEmpty else {
        //        return nil
        //}
        return devicePolicyFromJSON(deviceModelUrn: deviceModelUrn,
                                    jsonObject: policyJson) //.arrayValue[0])
    }
    
/*
{ "items" : [
    { // policy
        "id": "A unique ID for this policy configuration",
        "name": "This is the name for this policy configuration",
        "description": "This is the description for this policy",
        "enabled": true,
        "created" : 1512091219189,
        "createdAsString" : "2017-12-01T01:20:19Z",
        "lastModified" : 1512091352324,
        "lastModifiedAsString" : "2017-12-01T01:22:32Z",
        "lastModifiedBy" : "iot",
        "pipelines" : [{
            "attributeName": "name of the attribute", // if omitted or * it applies to all attributes
            "description" "description of this pipeline",
            "pipeline": [{ // array of functions
                "id": "This is the policy function",
                "description": "This is the pipeline description",
                "parameters": {
                    "parameter-name": "parameter-value",
                    ...
                }
            },
            {... more functions ...}
            ]},
            {[... more pipelines ]}
        ],
        "deviceModel": { // exclude for DEVICE role
            "links": [{
                "href": "https://iotserver:443/iot/privateapi/v2/deviceModels/{urn}",
                "rel": "deviceModel"
            }]
         },
         "devices": {
            "links": [{
                "href": "https://iotserver:443/iot/privateapi/v2/deviceModels/{urn}/devicePolicies/{id}/devices",
                "rel": "self"
            }]
         },
         "links": [{
            "href": "https://iotserver:443/iot/privateapi/v2/deviceModels/{urn}/devicePolicies/{id}",
            "rel": "self"
         }]
     },
     { ...more policies...}
],
"links": [{}..{}],
"offset" : 0,
"limit" : 10,
"count" : 3,
"hasMore" : false
}
*/
    
    // GET RestApi.getPrivateRoot()/deviceModels/{urn}/devicePolicies/{id}
    // {
    //     "id": "A unique ID for this policy configuration",
    //         "name": "This is the name for this policy configuration",
    //         "description": "This is the description for this policy",
    //         "pipelines" : [
    //     {
    //         "attribute": "name of the attribute", // if omitted or * it applies to all attributes
    //             "description" "description of this pipeline",
    //             "pipeline": [
    //         {
    //             "id": "This is the policy function",
    //                 "description": "This is the pipeline description",
    //                 "parameters": {
    //             "parameter-name": "parameter-value",
    //                 ...
    //         }
    //             ...
    //         ],
    //         },
    //     ...
    // ]
    //private static
    // If there are json failures they will occur in the "get" code
    // This method assumes all calls use that get data when calling this method.
    // No hand built json.
    func devicePolicyFromJSON(deviceModelUrn:String,
                              jsonObject:JSON) -> DevicePolicy {
        
        let id:String = jsonObject["id"].stringValue
        let description:String = jsonObject["description"].stringValue
        let lastModified:Double = jsonObject["lastModified"].doubleValue
        let enabled:Bool = jsonObject["enabled"].boolValue
        
        // If no "pipelines" old policy
        guard let pipelines:[JSON] = jsonObject["pipelines"].array else {
            return DevicePolicy(id:"", deviceModelURN:"",
                                pipelines:[:], description:"",
                                lastModified:0.0, enabled:false)
        }
        var policyPipelines:[String:[DevicePolicy.Function]] =
            Dictionary<String,[DevicePolicy.Function]>()
        
        for attributePipeline in pipelines {
            let attributeName:String = attributePipeline["attributeName"].string ??
                DevicePolicy.ALL_ATTRIBUTES()
            let pipeline:[JSON] = attributePipeline["pipeline"].arrayValue
            
            var functions:[DevicePolicy.Function] = Array<DevicePolicy.Function>()
            
            for function in pipeline {
                let functionId:String = function["id"].stringValue
                var parameterMap:[String:Any] = [String:Any]()
                let parameters:[String:JSON] = function["parameters"].dictionaryValue
                for (k, v) in parameters {
                    switch(k) {
                    case "action":
                        let actionName:String = v["name"].stringValue
                        if !actionName.isEmpty {
                            parameterMap["name"] = actionName
                        }
                        
                        let actionArguments:[JSON] = v["arguments"].arrayValue
                        if !actionArguments.isEmpty {
                            var argumentList:[AnyObject] = [AnyObject]()
                            for arg in actionArguments {
                                argumentList.append(arg.rawValue as AnyObject)
                            }
                            // Add after modifications
                            parameterMap["arguments"] = argumentList
                        }
                        break
                    case "alert":
                        let urn:String = v["urn"].stringValue
                        if !urn.isEmpty {
                            parameterMap["urn"] = urn
                        }
                        let fields:[String:JSON] = v["fields"].dictionaryValue
                        var fieldList:[String:String] = [String:String]()
                        for (fieldName,fieldValue) in fields {
                            fieldList[fieldName] = fieldValue.stringValue
                        }
                        // Add after modification
                        parameterMap["fields"] = fieldList
                        let severity = v["severity"].stringValue
                        if !severity.isEmpty {
                            parameterMap["severity"] = severity
                        }
                    default:
                        parameterMap[k] = v.rawValue as Any
                        break
                    }
                }
                functions.append(DevicePolicy.Function(id:functionId,
                                                       parameters:parameterMap))
            }
            policyPipelines[attributeName] = functions
        }
        
        return DevicePolicy(id:id, deviceModelURN:deviceModelUrn,
                            pipelines:policyPipelines, description:description,
                            lastModified:lastModified, enabled:enabled)
    }

    private let getQueue:DispatchQueue =
        DispatchQueue(label:"DevicePolicyManager_get_queue", qos:.utility)
    private func get(query:String, deviceModelUrn:String) throws -> JSON? {
        
        let getSemaphore:DispatchSemaphore = DispatchSemaphore(value:0)
        var clientError:ClientError? = nil
        var jsonObject:JSON? = nil
        
        getQueue.async(execute: {

            // Gets encoded before Alamofire request
            let restApi:String = RestApi.V2.getPrivateRoot() +
                "/deviceModels/" + deviceModelUrn + "/devicePolicies" + query
                            
            log(type:.debug,
                message:"%@: %@",
                args:"DevicePolicyManager.get()", restApi)
            
            do {
                try self.secureConnection?.get(restApi:restApi,
                                               response: {(response, error) in
                    defer {
                        getSemaphore.signal()
                    }
                    
                    if error != nil {
                        let err:String = logMsg(type:.fault,
                                                message: "%@: GET %@ : %@",
                                                args: ["DevicePolicyManager:get()",
                                                      "\(restApi)", "\(error!)"])
                        log(type:.fault, message: err)
                        clientError = ClientError.network(err)
                        return
                    }
                    guard let status:Int = response?.getStatus() else {
                        let err:String = logMsg(type:.fault,
                                                message:"%@: GET %@ : no status",
                                                args:["DevicePolicyManager:get()",
                                                      "\(restApi)"])
                        log(type:.fault, message: err)
                        clientError = ClientError.network(err)
                        return
                    }
                    
                    guard status == StatusCode.OK.getCode() else {
                        let err:String = logMsg(type:.fault,
                                                message:"%@: GET %@ : status %@",
                                                args:["DevicePolicyManager:get()",
                                                      "\(restApi)", "\(status)"])
                        log(type:.fault, message: err)
                        clientError = ClientError.network(err)
                        return
                    }
                    
                    let data:Foundation.Data = response!.getData()
                    guard data.count > 2 else {
                        let err:String = logMsg(type:.fault,
                                                message:"%@: GET %@ : no data.",
                                                args:["DevicePolicyManager:get()",
                                                      "\(restApi)"])
                        log(type:.fault, message:err)
                        clientError = ClientError.network(err)
                        return
                    }
                    do {
                        jsonObject = try JSON(data:data)
                        return
                    } catch {
                        let err:String = logMsg(type:.fault,
                                                message:"%@: GET %@ : %@",
                                                args:["DevicePolicyManager:get()",
                                                      "\(restApi)", "\(error)"])
                        log(type:.fault, message: err)
                        //callback(nil, ClientError.network(err))
                        clientError = ClientError.network(err)
                        return
                    }
                })

            } catch {
                getSemaphore.signal()
                let err:String = logMsg(type:.fault,
                                        message:"%@: GET %@ : %@",
                                        args: ["DevicePolicyManager.get()",
                                               "\(restApi)", "\(error)"])
                log(type:.fault, message: err)
                clientError = ClientError.network(err)
            }
        })
        // TODO: Check for timeout
        getSemaphore.wait()
        if clientError != nil {
            throw clientError!
        }
        return jsonObject
    }
    
    private func getJsonItem(items:JSON) -> JSON? {
        guard let jsonItems:[JSON?] = items.dictionaryValue["items"]?.arrayValue,
                !jsonItems.isEmpty,
                let jsonItem:JSON = jsonItems[0] else {
                return nil
        }
        return jsonItem
    }
    
    // TODO: Need to sync writes to policy files when writing the same
    // policy
    
    // TODO: replace original function based on model urn to this one based
    // on device policy
    
    // TODO: Do we never use the "items:[]" format any more ? JCL doesn't appear
    // to be using it any more in "loadLocalDevicePolicy"
    
    // returns nil if LOCAL_STORE is false, or there is no file for the device
    // policy id
    func loadLocalDevicePolicy(devicePolicyId:String) -> JSON? {
        // TODO: Need to allow configurable local storage predicate
        guard LOCAL_STORE else {
            return nil
        }
        
        let filename:String = String(format:DEVICEPOLICY_FILENAME_FORMAT, devicePolicyId)
        guard let jsonPolicy:String = getFileFromLocalStore(filename:filename) else {
            return nil
        }
        
        guard let jsonPolicyData:Foundation.Data =
            jsonPolicy.data(using:.utf8, allowLossyConversion:false) else {
                // Bad data delete it
                deleteFileFromLocalStore(filename:filename)
                return nil
        }
        let localPolicy:JSON = JSON(jsonPolicyData)
        //let items:[String:[JSON]] = ["items": [localPolicy]]
        
        return localPolicy
    }
}
