/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

// oracle.iot.client.device
/**
 * A directly-connected device is able to send messages to, and receive messages
 * from, the IoT server. A directly-connected device has a physical-device
 * identifier that is registered with the server. When the directly-connected
 * device is activated on the server, the server assigns a logical-endpoint
 * identifier. This logical-endpoint identifier is required for sending messages
 * to, and receiving messages from, the server.
 */
public class DirectlyConnectedDevice: Client {
    
    var msgDcd:AdvancedDirectlyConnectedDevice!
    
    /**
     * Constructs a new `DirectlyConnectedDevice` instance.
     */
    convenience init() throws {
        try self.init(msgDcd: nil)
    }
    
    /**
     * Constructor used by `GatewayDevice`. This method is used internally and
     * is not intended for general use.
     *
     * - parameter context: a platform specific object (e.g. application 
     *                      context), that needs to be associated with this 
     *                      client, if any.
     *
     * - parameter msgDcd: an `AdvancedDirectlyConnectedDevice`
     *
     */
    init(msgDcd: AdvancedDirectlyConnectedDevice?) throws {
        if msgDcd != nil {
            self.msgDcd = msgDcd
        } else {
            self.msgDcd = try AdvancedDirectlyConnectedDevice()
        }
        try bootstrapMessageDispatcher()
    }
    
    /**
     * Constructs a new `DirectlyConnectedDevice` instance that will use the 
     * given filename, file-extension, subdirectory and password to provision a 
     * `DirectlyConnectedDeviceImpl` instance
     */
    init(file: String, fileExtension: String? = nil,subdirectory: String? = nil,
                                                    password: String) throws {
        let msgDcdlo:AdvancedDirectlyConnectedDevice =
            try AdvancedDirectlyConnectedDevice(file: file,
                                                fileExtension: fileExtension,
                                                subdirectory: subdirectory,
                                                password: password)
        self.msgDcd = msgDcdlo
        try bootstrapMessageDispatcher()
    }
    
    /**
     * Constructs a new `DirectlyConnectedDevice` instance that will load the 
     * device configuration from the given path and password.
     *
     * - parameter path: the path to the provisioning file
     *
     * - parameter password: the password to protect the provisioning file 
     *                       during the creation of the file
     *
     * - throws: 
     *       - `ClientError.file` if the provisioning file is not present
     *       - `ClientError.security` if the provisioning file can't be
     *         decrypted or there are invalid information in the file
     *
     */
    public convenience init(path: String, password: String) throws {
        try self.init(file:path, password:password)
    }
    
    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns an endpoint identifier to the device.
     *
     * If the device is already activated, this method will throw an
     * `ClientError.state`. The user should call the `isActivated()`
     * method prior to calling activate.
     *
     * The `callback` closure can run any code that is required after
     * device activation, such as device registration (indirect enrollment).
     *
     * - parameter deviceModels: should contain the device model type URNs of 
     *                           this directly connected device. The device is 
     *                           activated with the given device models.
     *                           The `deviceModels` parameter is zero or more, 
     *                           comma separated, device model URNs.
     *
     * - parameter callback: The function to be executed after completion of
     *                      activation. If activation is not successful then
     *                      the object will be `nil` and an error object is
     *                      passed in the form `callback(endpoint, error)` and
     *                      the reason can be taken from error
     *
     * - throws: `ClientError.state` if the device is already activated or if
     *              `close` has been called.
     *
     * - seealso: `isActivated`
     */
    public func activate(deviceModels: String...,
            callback: @escaping (String?, ClientError?) -> ()) throws {
        try activate(deviceModels:deviceModels, callback:callback)
    }
    
    public func activate(deviceModels: [String],
        callback: @escaping (String?, ClientError?) -> ()) throws {
        if self.msgDcd.isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        // Note: JCL orders these differently. The "canned" urns come first.
        let len = (deviceModels.count > 0 ? (deviceModels.count + 3) : 3)
        var dmArray:[String] = [String](repeating: "", count: len)
        for i in 0..<deviceModels.count {
            dmArray[i] = deviceModels[i]
        }
        // Make sure to add diagnostics and message_dispatcher capability URNs.
        dmArray[len - 3] = DevicePolicyManager.CAPABILITY_URN
        dmArray[len - 2] = ActivationManager.DIAGNOSTICS_URN
        dmArray[len - 1] = ActivationManager.MESSAGE_DISPATCHER_URN
        try self.msgDcd.activate(deviceModels: dmArray, callback: {
                (endpointId, error) in
            if error != nil {
                callback(endpointId, error)
                return
            }
            var msgDispError:ClientError? = nil
            do {
                try self.bootstrapMessageDispatcher()
            } catch {
                msgDispError = ClientError.network("\(error)")
            }
            callback(endpointId, msgDispError)
        })
    }
    
    /**
     * Returns whether the device has been activated against IOT CS server.
     *
     * - returns: whether the device has been activated against IOT CS server.
     */
    public func isActivated() -> Bool {
        return self.msgDcd.isActivated()
    }
    
    /**
     * Return the logical-endpoint identifier of this directly-connected
     * device. The logical-endpoint identifier is assigned by the server
     * as part of the activation process.
     *
     * If `close` has been called, the return value is undefined.
     *
     * - returns: the logical-endpoint identifier of this directly-connected
     *            device.
     *
     * - seealso: `activate(callback:deviceModels:)`
     */
    public func getEndpointId() -> String {
        return self.msgDcd.getEndpointId()
    }
    
    /**
     * Create an `AbstractVirtualDevice` instance with the given device model
     * for the given device identifier. This method creates a new `VirtualDevice`
     * instance for the given parameters. The client library does not
     * cache previously created VirtualDevice objects.
     *
     * - parameter deviceId: The device identifier of the device being modeled
     *
     * - parameter deviceModel: The device model URN that this device implements
     *
     * - returns: a new `VirtualDevice`
     *
     * - throws: 
     *       - `ClientError.network` if a `VirtualDevice` cannot be created.
     *       - `ClientError.state` if `close` has been called.
     */
    public func createVirtualDevice(deviceId: String,
                            deviceModel: DeviceModel) throws -> VirtualDevice {
        if (self.msgDcd.isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed", comment: ""))
        }
        return try VirtualDeviceImpl(msgDirectlyConnectedDevice: msgDcd,
                                     endpointId: deviceId,
                                     deviceModel: deviceModel)
    }
    
    /**
     * Get the `DeviceModel` for the device model urn. This method may
     * return `nil` if there is no device model for the URN. `nil` may also be
     * returned if the device model is a "draft" and the property
     * `Config.ALLOW_DRAFT_DEVICE_MODELS` is set to
     * `false`, which is the default.
     *
     * - parameter deviceModelUrn: The URN of the device model
     *
     * - parameter callback: This function is called with the retrieved device 
     *                       model.If retrieving the device model is not 
     *                       successful then the object will be `nil` and an
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the reason can be 
     *                       taken from error
     *
     * - throws:
     *      - `ClientError.state` if `close` has been called
     *
     * On error `callback` receives a `ClientError.network` error if:
     *    the object is not found (`StatusCode.NOT_FOUND`),
     *    the request fails (`!StatusCode.OK`), or
     *    if the returned data is `nil` or JSON parsing fails
     */
    public func getDeviceModel(deviceModelUrn: String,
            callback: @escaping (DeviceModel?, ClientError?) -> Void) throws {
        if self.msgDcd.isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed.",
                                                      comment:""))
        }
        try self.msgDcd.getDeviceModel(deviceModelUrn: deviceModelUrn,
                                               callback: callback)
    }

    /**
     *  Release the resources held by the client
     *  This call is idempotent
     *
     * - throws: `ClientError.network` if a `MessageDispatcher` cannot be 
     *           obtained.
     */
    public func close() throws {
        if !isActivated() || self.msgDcd.isClosed {
            return
        }
        if let messageDispatcher:MessageDispatcher =
                MessageDispatcher.removeMessageDispatcher(msgDirectlyConnectedDevice:self.msgDcd) {
            messageDispatcher.close(callback: {
                self.msgDcd.close()
            })
        }
    }
    
    private func bootstrapMessageDispatcher() throws {
        if !isActivated() {
            return
        }
        let messageDispatcher:MessageDispatcher =
            try MessageDispatcherImpl(deviceClient:self.msgDcd)
        MessageDispatcher.setMessageDispatcher(
            msgDirectlyConnectedDevice:self.msgDcd,
            messageDispatcher:messageDispatcher)
    }
    
    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the device's endpoint ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX`
     * to `true`.
     * <p>
     * If `contentType` is nil, the mime-type defaults to
     * "application/octet-stream".
     *
     * - parameter name: the unique name to be used to reference the content in
     *                   storage
     * - parameter contentType: The mime-type of the content or `nil`
     *
     * - returna a `StorageObject`
     *
     * - throws: 
     *      - `ClientError.network if there is an exception raised by the
     *                                runtime, or an abnormal response from the 
     *                                storage cloud
     *      - `ClientError.security` if there is an exception establishing
     *                              a secure connection to the storage cloud
     *      - `ClientError.state` if `close` has been called
     */
    public func createStorageObject(name:String,
            contentType:String? = StorageObject.OCTET_STREAM) throws  -> StorageObject {
        return try createStorageObject(name:name,
                                       contentType:contentType ?? StorageObject.OCTET_STREAM,
                                       mutable:true)
    }
    
    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the device's endpoint ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX`
     * to `true`.
     * <p>
     * If `contentType` is nil, the mime-type defaults to
     * "application/octet-stream".
     *
     * - parameter name: the unique name to be used to reference the content in
     *                   storage
     * - parameter contentType: The mime-type of the content or `nil`
     *
     * - parameter mutable: mutable `true` means mutable, `false` means immutable
     *        If `mutable` == `false`, then the AdvancedStorageObject will be made
     *        immutable by setting it’s *Write-Once-Read-Many* (WORM) policy when
     *        uploading it to the container to prevent the users
     *        from deleting it for a 7 days or modifying it.
     *        Number of days is 7 by default and can be overridden by setting property
     *        `Config.STORAGE_WORM_DAYS`. If the property is "0" then no WORM policy
     *        will be applied.
     *
     * - returna a `StorageObject`
     *
     * - throws:
     *      - `ClientError.network if there is an exception raised by the
     *                                runtime, or an abnormal response from the
     *                                storage cloud
     *      - `ClientError.security` if there is an exception establishing
     *                              a secure connection to the storage cloud
     *      - `ClientError.state` if `close` has been called
     */
    public func createStorageObject(name:String,
                                    contentType:String? = StorageObject.OCTET_STREAM,
                                    mutable:Bool) throws  -> StorageObject {
        if self.msgDcd.isClosed {
            throw ClientError.state(NSLocalizedString("Client is closed.",
                                                      comment:""))
        }
        let delegate: AdvancedStorageObject =
            try msgDcd.createStorageObject(name: name,
                                            contentType: contentType ?? StorageObject.OCTET_STREAM,
                                            mutable:mutable)
        return StorageObjectImpl(dcd: msgDcd, delegate:delegate)
    }
    
    public func offer(message:Message) throws {
        
    }
}
