/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device

import Foundation

enum FormulaParserNodeOperation {
    case unaryPlus
    case unaryMinus
    case plus
    case minus
    case mul
    case div
    case mod
    case and
    case or
    case eq
    case like
    case neq
    case lt
    case lte
    case gt
    case gte
    case ternary
    case alternative
    case not
    case lower
    case upper
    case function
    case group
    case terminal

    func getPrec() -> Int {
        var result:Int = 0
        switch (self) {
        case .unaryPlus: result = 6 ; break
        case .unaryMinus: result = 6 ; break
        case .plus: result = 3 ; break
        case .minus: result = 3 ; break
        case .mul: result = 4 ; break
        case .div: result = 4 ; break
        case .mod: result = 4 ; break
        case .and: result = 1 ; break
        case .or: result = 1 ; break
        case .eq: result = 2 ; break
        case .like: result = 2; break
        case .neq: result = 2 ; break
        case .lt: result = 2 ; break
        case .lte:result = 2 ; break
        case .gt: result = 2 ; break
        case .gte: result = 2 ; break
        // this is for the logical part of ?:, LHS is the logical, RHS is the alternatives
        case .ternary: result = 0 ; break
        // this is for the alternatives part of ?:, RHS is true choice, LHS is false choice
        case .alternative: result = 0 ; break
        // ! has only LHS, no RHS. LHS is an equality expression or numeric expression
        case .not: result = 6 ; break
        // has only LHS, no RHS. LHS is an string expression
        case .lower: result = 6 ; break
        // has only LHS, no RHS. LHS is an string expression
        case .upper: result = 6 ; break
        // function LHS is function name. args, if any, chain to rhs
        case .function: result = 6 ; break
        // group LHS is the enclosed arithmetic expression
        case .group: result = -1 ; break
        // terminal is a number or attribute, LHS is a FormulaParserTerminal, no RHS
        case .terminal: result = -1 ; break
        }
        return result
    }
}

//
// AST part
//
class FormulaParserNode : Hashable {

    public var hashValue:Int {
        get {
            let rhshv:Int = rightHandSide?.hashValue ?? 0
            let lhshv:Int = leftHandSide?.hashValue ?? 0
            return operation.hashValue + lhshv + rhshv
        }
    }

    private var operation:FormulaParserNodeOperation!
    private var leftHandSide:FormulaParserNode? = nil
    private var rightHandSide:FormulaParserNode? = nil

    init(operation:FormulaParserNodeOperation, leftHandSide:FormulaParserNode?) {
        self.operation = operation
        self.leftHandSide = leftHandSide
    }

    func getNodeOperation() -> FormulaParserNodeOperation {
        return operation
    }

    func getLeftHandSide() -> FormulaParserNode? {
        return leftHandSide
    }

    func setLeftHandSide(leftHandSide:FormulaParserNode?) {
        self.leftHandSide = leftHandSide
    }

    func getRightHandSide() -> FormulaParserNode? {
        return rightHandSide
    }

    func setRightHandSide(rightHandSide:FormulaParserNode?) {
        self.rightHandSide = rightHandSide
    }

    static func ==(lhs:FormulaParserNode, rhs:FormulaParserNode) -> Bool {
        return lhs === rhs ||
            (lhs.leftHandSide == rhs.leftHandSide &&
                lhs.rightHandSide == rhs.rightHandSide)
    }
}

enum TokenType : Int {
    case AND = 0    // &&
    case COLON  // :
    case COMMA  // ,
    case DIV    // \
    case DOLLAR // $
    case EQ     // =
    case FUNCTION // IDENT '('
    case ATTRIBUTE // '$(' IDENT ')'
    case GT     // >
    case GTE    // >=
    case IDENT  // [_a-zA-Z]+ [_a-zA-Z0-9\-]*
    case LPAREN // (
    case LT     // <
    case LTE    // <=
    case MINUS  // -
    case MOD    // %
    case MUL    // *
    case NEQ    // !=
    case NOT    // !
    case NUMBER // [0-9]+|[0-9]*"."[0-9]+
    case OR     // ||
    case PLUS   // +
    case QUESTION_MARK // ?
    case RPAREN // )
    case STRING // "
    case WS     // whitespace is not significant and is consumed
    case NIL
}

//
// Lexer
//
class FormulaParserToken : Hashable {

    private let type:TokenType!
    private let pos:Int!
    private let length:Int!

    var hashValue:Int {
        get {
            return type.hashValue +
                   pos.hashValue +
                   length.hashValue
        }
    }

    init(type:TokenType, pos:Int, length:Int) {
        self.type = type
        self.pos = pos
        self.length = length
    }

    func getTokenType() -> TokenType {
        return type
    }

    func getPos() -> Int {
        return pos
    }

    func getLength() -> Int {
        return length
    }

    static func ==(lhs:FormulaParserToken, rhs:FormulaParserToken) -> Bool {
        return lhs === rhs ||
            (lhs.type == rhs.type &&
                lhs.pos == rhs.pos &&
                    lhs.length == rhs.length)
    }
}

enum TerminalType {
    case inprocessAttribute
    case currentAttribute
    case number
    case ident
    case string
}

// A terminal is either an attribute, number, or no-arg function.
// It is a FormulaParserNode, but has no FormulaParserNodeOperation, LHS, or RHS
class FormulaParserTerminal : FormulaParserNode {
    
    override
    var hashValue:Int {
        get {
            return type.hashValue + value.hashValue
        }
    }

    let type:TerminalType!
    let value:String!

    init(type:TerminalType, value:String) {
        self.type = type
        self.value = value
        super.init(operation:.terminal, leftHandSide:nil)
    }

    func getValue() -> String {
        return value
    }
    
    func getTerminalType() -> TerminalType {
        return type
    }

    static func  ==(lhs:FormulaParserTerminal, rhs:FormulaParserTerminal) -> Bool {
        return lhs === rhs ||
            (lhs.type == rhs.type && lhs.value == rhs.value)
    }
}

fileprivate func isWhitespace(char:Character) -> Bool {
    return CharacterSet.whitespacesAndNewlines.contains(UnicodeScalar(String(char))!)
}

fileprivate func isDigit(char:Character) -> Bool {
    return CharacterSet.decimalDigits.contains(UnicodeScalar(String(char))!)
}

/**
 * FormulaParser
 */
class FormulaParser {

    static func tokenize(formula:String) throws -> [FormulaParserToken] {

        var tokens:[FormulaParserToken] = [FormulaParserToken]()
        var pos:String.Index = formula.startIndex
        var tokenType:TokenType = .NIL

        var i:String.Index = formula.startIndex
        while i < formula.endIndex {
            var type:TokenType = tokenType
            // This is the length of the previously parsed token
            let length:Int = formula.distance(from:pos, to:i)
            let ch:Character = formula[i]
            switch (ch) {
                case "(":
                    type = .LPAREN
                    break
                case ")":
                    type = .RPAREN
                    break
                case ",":
                    type = .COMMA
                    break
                case "?":
                    type = .QUESTION_MARK
                    break
                case ":":
                    type = .COLON
                    break
                case "+":
                    type = .PLUS
                    break
                case "-":
                    if tokenType != .IDENT {
                        type = .MINUS
                    }
                    break
                case "*":
                    type = .MUL
                    break
                case "/":
                    type = .DIV
                    break
                case "%":
                    type = .MOD
                    break
                case "=":
                    type = .EQ
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        // be forgiving of "=="
                        if peekChar == "=" {
                            i = formula.index(after:i)
                        }
                    }
                    break
                case "!":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "=" {
                            type = .NEQ
                            i = formula.index(after:i)
                        } else {
                            type = .NOT
                        }
                    }
                    break
                case ">":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "=" {
                            type = .GTE
                            i = formula.index(after:i)
                        } else {
                            type = .GT
                        }
                    }
                    break
                case "<":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "=" {
                            type = .LTE
                            i = formula.index(after:i)
                        } else {
                            type = .LT
                        }
                    }
                    break
                case "|":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "|" {
                            type = .OR
                            i = formula.index(after:i)
                        }
                    }
                break
                case "&":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "&" {
                            type = .AND
                            i = formula.index(after:i)
                        }
                    }
                break
                case "$":
                    if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                        if peekChar == "(" {
                            type = .ATTRIBUTE
                            i = formula.index(after:i) // the default increment will point after (
                        } else {
                            type = .DOLLAR
                        }
                    }
                    break
                default:
                    if isWhitespace(char:ch) {
                        type = .WS
                    } else if tokenType != .IDENT {
                        if isDigit(char:ch) {
                            type = .NUMBER
                        } else if ch == "." {
                            // [0-9]+|[0-9]*"."[0-9]+
                            if tokenType != .NUMBER {
                                if let peekChar:Character = peek(str:formula, offset:formula.index(after:i)) {
                                    if isDigit(char:peekChar) {
                                        type = .NUMBER
                                        i = formula.index(after:i) // the default increment will point after digit
                                    } else {
                                        throw ClientError.argument(String(format:NSLocalizedString(
                                            "Found %@ @ %@ + 1 : expected [0-9]", comment:""),
                                                                          "\(peekChar)", "\(i)"))
                                    }
                                }
                            }
                        } else {
                            type = .IDENT
                        }
                    }
                    break
            }
        
            // Add previous token when lexer hits a new token.
            if tokenType != type {
                if tokenType == .IDENT {
                    let token:String = String(formula[pos..<formula.index(pos, offsetBy:length)])
                    if "AND".caseInsensitiveCompare(token) == .orderedSame {
                        tokenType = .AND
                    } else if "OR".caseInsensitiveCompare(token) == .orderedSame {
                        tokenType = .OR
                    } else if "NOT".caseInsensitiveCompare(token) == .orderedSame {
                        tokenType = .NOT
                    } else if type == .LPAREN {
                        tokenType = .FUNCTION
                        type = tokenType
                        // Make sure i is pointing to the next character
                        i = formula.index(after:i)
                        continue
                    }
                }
                if tokenType != .NIL  {
                    if tokenType != .WS {
                        // Convert pos to integer, and length to Int
                        // pos == ch position where the current parse started
                        let posInt:Int = formula.distance(from:formula.startIndex, to:pos)
                        // length is the number of characters in the previously parsed token
                        tokens.append(FormulaParserToken(type:tokenType, pos:posInt, length:length))
                    }
                    // update pos to one after the previous token
                    //pos += length
                    pos = formula.index(pos, offsetBy: length)
                }
                // previous token is now current token
                tokenType = type
            }
            // Make sure i is pointing to the next character
            i = formula.index(after:i)
        }
    
        // add the last token
        // Must calculate pos from Index
        // Convert pos to integer, and length to Int
        let posInt:Int = formula.distance(from:formula.startIndex, to:pos)
        let length:Int = formula.distance(from:pos, to:formula.endIndex)
        if tokenType != .WS {
            tokens.append(FormulaParserToken(type:tokenType, pos:posInt, length: length))
        }

        return tokens
    }

    static func peek(str:String, offset:String.Index) -> Character? {
        return offset < str.endIndex ? str[offset] : nil
    }

    public static func comparePrec(lhs:FormulaParserNode, rhs:FormulaParserNode) -> Int {
        return lhs.getNodeOperation().getPrec() - rhs.getNodeOperation().getPrec()
    }

    // formula
    //    : numericExpression
    //    | ternaryExpression
    //    
    //
    // returns the root of the AST
    static func parseFormula(tokens:[FormulaParserToken],
                             formula:String) throws -> FormulaParserNode {

        var stack:Stack<FormulaParserNode> = Stack<FormulaParserNode>()
        var index:Int = -1

        do {
            index = try parseNumericExpression(stack: &stack, tokens:tokens,
                                               formula:formula, index:0)
        } catch {
            // drop through = try as conditional expression
        }

        if index < tokens.count {
            stack.removeAll()
            index = try parseTernaryExpression(stack: &stack, tokens:tokens,
                                               formula:formula, index:0)
        }

        if index < tokens.count {
            let lastToken:FormulaParserToken = tokens[index]
            throw ClientError.argument(String(format:NSLocalizedString(
                "formula: parser bailed @ %d", comment: ""),
                                              lastToken.getPos()))
        }

        return stack[0]
    }
    
    // ternaryExpression
    //     : conditionalOrExpression QUESTION_MARK additiveExpression COLON additiveExpression
    //     
    //
    // returns the index of the next token to be processed.
    static func parseTernaryExpression(stack: inout Stack<FormulaParserNode>,
                                       tokens:[FormulaParserToken],
                                       formula:String, index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseConditionalOrExpression(stack: &stack,
                                                          tokens:tokens,
                                                          formula:formula,
                                                          index:index)
        if _index >= tokens.count {
            return tokens.count
        }
        
        var token:FormulaParserToken = tokens[_index]
        if token.getTokenType() != .QUESTION_MARK {
            throw ClientError.argument(String(format:NSLocalizedString(
                    "parseTernaryExpression: found %@ expected QUESTION_MARK",
                    comment:""), "\(token)"))
        }

        let ternary:FormulaParserNode =
            FormulaParserNode(operation:.ternary,
                              leftHandSide:stack.popLast())

        _index = try parseAdditiveExpression(stack: &stack, tokens:tokens,
                                             formula:formula, index:_index + 1)
        token = tokens[_index]
        if token.getTokenType() != .COLON {
            throw ClientError.argument(String(format:NSLocalizedString(
                    "parseTernaryExpression: found %@ expected COLON",
                    comment: ""), "\(token)"))
        }

        let aternatives:FormulaParserNode =
            FormulaParserNode(operation:.alternative,
                              leftHandSide:stack.popLast())
        ternary.setRightHandSide(rightHandSide:aternatives)

        _index = try parseAdditiveExpression(stack: &stack, tokens:tokens,
                                             formula:formula, index:_index + 1)
        aternatives.setRightHandSide(rightHandSide:stack.popLast())

        stack.append(ternary)

        return _index
    }

    // conditionalOrExpression
    //     : conditionalAndExpression ( OR conditionalAndExpression )*
    //     
    //
    // returns the index of the next token to be processed.
    static func parseConditionalOrExpression(stack: inout Stack<FormulaParserNode>,
                                             tokens:[FormulaParserToken], formula:String,
                                             index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseConditionalAndExpression(stack: &stack,
                                                           tokens:tokens,
                                                           formula:formula,
                                                           index:index)
        if _index >= tokens.count {
            return tokens.count
        }

        let token:FormulaParserToken = tokens[_index]
        let lhs:FormulaParserNode!
        switch (token.getTokenType()) {
            case .OR:
                lhs = FormulaParserNode(operation:.or, leftHandSide:stack.popLast())
                _index += 1
                break
            default:
                return _index
        }

        _index = try parseConditionalOrExpression(stack: &stack, tokens:tokens,
                                                  formula:formula, index:_index)
        // Assume we can't get here if lhs and rhs are non nil
        // force unwrap of prioritize
        stack.append(prioritize(lhs: lhs, rhs: stack.popLast())!)
        return _index
    }
    
    //
    // conditionalAndExpression
    //     : valueLogical ( AND valueLogical )*
    //     
    //
    // returns the index of the next token to be processed.
    static func parseConditionalAndExpression(stack: inout Stack<FormulaParserNode>,
                                              tokens:[FormulaParserToken], formula:String,
                                              index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseValueLogical(stack: &stack, tokens:tokens, formula:formula, index:index)
        if _index >= tokens.count {
            return tokens.count
        }

        let token:FormulaParserToken = tokens[_index]
        let lhs:FormulaParserNode!
        switch (token.getTokenType()) {
            case .AND:
                lhs = FormulaParserNode(operation:.and,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            default:
                return _index
        }

        _index = try parseConditionalAndExpression(stack:&stack, tokens:tokens,
                                                   formula:formula, index:_index)
        // Assume we can't get here unless there is a non-nil lhs and rsh
        // so force unwrap of prioritize
        stack.append(prioritize(lhs: lhs, rhs: stack.popLast())!)
        return _index
    }

    //
    // valueLogical
    //     : relationalExpression
    //     
    //
    // returns the index of the next token to be processed.
    static func parseValueLogical(stack: inout Stack<FormulaParserNode>,
                                  tokens:[FormulaParserToken], formula:String,
                                  index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }
        return try parseRelationalExpression(stack: &stack, tokens:tokens,
                                             formula:formula, index:index)
    }
    
    //
    // relationalExpression
    //     : numericExpression (EQ numericExpression | NEQ numericExpression | 
    //     LT numericExpression | GT numericExpression | LTE numericExpression | 
    //     GTE numericExpression )?
    //
    // returns the index of the next token to be processed.
    static func parseRelationalExpression(stack: inout Stack<FormulaParserNode>,
                                          tokens:[FormulaParserToken], formula:String,
                                          index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseNumericExpression(stack: &stack, tokens:tokens,
                                                    formula:formula, index:index)
        if _index >= tokens.count {
            return tokens.count
        }

        let token:FormulaParserToken = tokens[_index]
        let lhs:FormulaParserNode!
        switch (token.getTokenType()) {
            case .EQ:
                lhs = FormulaParserNode(operation:.eq,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .NEQ:
                lhs = FormulaParserNode(operation:.neq,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .LT:
                lhs = FormulaParserNode(operation:.lt,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .LTE:
                lhs = FormulaParserNode(operation:.lte,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .GT:
                lhs = FormulaParserNode(operation:.gt,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .GTE:
                lhs = FormulaParserNode(operation:.gte,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            default:
                return _index
        }

        _index = try parseRelationalExpression(stack: &stack, tokens:tokens,
                                               formula:formula, index:_index)
        // Assume we can't get here if lhs and rhs are non nil
        // force unwrap of prioritize
        stack.append(prioritize(lhs: lhs, rhs: stack.popLast())!)
        return _index
    }
    
    //
    // numericExpression
    //     : additiveExpression
    //     
    //
    // returns the index of the next token to be processed.
    static func parseNumericExpression(stack: inout Stack<FormulaParserNode>,
                                       tokens:[FormulaParserToken], formula:String,
                                       index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }
        return try parseAdditiveExpression(stack: &stack, tokens:tokens,
                                           formula:formula, index:index)
    }
    
    //
    // additiveExpression
    //     : multiplicativeExpression (PLUS multiplicativeExpression | MINUS multiplicativeExpression )*
    //     
    //
    // returns the index of the next token to be processed.
    static func parseAdditiveExpression(stack: inout Stack<FormulaParserNode>,
                                        tokens:[FormulaParserToken], formula:String,
                                        index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseMultiplicativeExpression(stack: &stack,
                                                           tokens:tokens,
                                                           formula:formula,
                                                           index:index)
        if _index >= tokens.count {
            return tokens.count
        }

        let token:FormulaParserToken = tokens[_index]
        let lhs:FormulaParserNode!
        switch (token.getTokenType()) {
            case .PLUS:
                lhs = FormulaParserNode(operation:.plus,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .MINUS:
                lhs = FormulaParserNode(operation:.minus,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            default:
                return _index
        }

        _index = try parseAdditiveExpression(stack: &stack, tokens:tokens,
                                             formula:formula, index:_index)
        // Assume we can't get here if lhs and rhs are non nil
        // force unwrap of prioritize
        stack.append(prioritize(lhs: lhs, rhs: stack.popLast())!)
        return _index
    }
    
    //
    // multiplicativeExpression
    //     : exponentiationExpression (MUL exponentiationExpression | 
    //     DIV exponentiationExpression | MOD exponentiationExpression)*
    //     
    //
    // returns the index of the next token to be processed.
    static func parseMultiplicativeExpression(stack: inout Stack<FormulaParserNode>,
                                              tokens:[FormulaParserToken], formula:String,
                                              index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = try parseUnaryExpression(stack: &stack, tokens:tokens,
                                                  formula:formula, index:index)
        if _index >= tokens.count {
            return tokens.count
        }

        let token:FormulaParserToken = tokens[_index]
        let lhs:FormulaParserNode
        switch (token.getTokenType()) {
            case .MUL:
                lhs = FormulaParserNode(operation:.mul,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .DIV:
                lhs = FormulaParserNode(operation:.div,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            case .MOD:
                lhs = FormulaParserNode(operation:.mod,
                                        leftHandSide:stack.popLast())
                _index += 1
                break
            default:
                return _index
            //break
        }

        _index = try parseMultiplicativeExpression(stack: &stack, tokens:tokens,
                                                   formula:formula, index:_index)
        // Assume we can't get here if lhs and rhs are non nil
        // force unwrap of prioritize
        stack.append(prioritize(lhs: lhs, rhs: stack.popLast())!)
        return _index
    }
    
    //
    // unaryExpression
    //     : NOT primaryExpression
    //     | PLUS primaryExpression
    //     | MINUS primaryExpression
    //     | primaryExpression
    //     
    //
    // returns the index of the next token to be processed.
    static func parseUnaryExpression(stack: inout Stack<FormulaParserNode>,
                                     tokens:[FormulaParserToken], formula:String,
                                     index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index
        let token:FormulaParserToken = tokens[_index]
        switch (token.getTokenType()) {
            case .NOT:
                _index = try parsePrimaryExpression(stack: &stack, tokens:tokens,
                                                    formula:formula,
                                                    index:_index + 1)
                stack.append(FormulaParserNode(operation: .not,
                                               leftHandSide: stack.popLast()))
                break
            case .PLUS:
                _index = try parsePrimaryExpression(stack: &stack, tokens:tokens,
                                                    formula:formula,
                                                    index:_index + 1)
                stack.append(FormulaParserNode(operation:.unaryPlus,
                                               leftHandSide:stack.popLast()))
                break
            case .MINUS:
                _index = try parsePrimaryExpression(stack: &stack, tokens:tokens,
                                                    formula:formula,
                                                    index:_index + 1)
                stack.append(FormulaParserNode(operation:.unaryMinus,
                                               leftHandSide:stack.popLast()))
                break
            default:
                _index = try parsePrimaryExpression(stack: &stack, tokens:tokens,
                                                    formula:formula, index:_index)
                break
        }
        return _index
    }
    
    //
    // primaryExpression
    //     : brackettedExpression
    //     | functionElement
    //     | expressionElement
    //     
    //
    // returns the index of the next token to be processed.
    static func parsePrimaryExpression(stack: inout Stack<FormulaParserNode>,
                                       tokens:[FormulaParserToken], formula:String,
                                       index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var newIndex:Int = try parseBrackettedExpression(stack: &stack,
                                                         tokens:tokens,
                                                         formula:formula,
                                                         index:index)
        if newIndex == index {
            newIndex = try parseFunctionElement(stack: &stack, tokens:tokens,
                                                formula:formula, index:index)
            if newIndex == index {
                newIndex = try parseExpressionElement(stack: &stack, tokens:tokens,
                                                      formula:formula, index:index)
            }
        }

        if newIndex == index {
            throw ClientError.argument(NSLocalizedString(
                "parsePrimaryExpression: expected " +
                "[brackettedExpression|functionElement|expressionElement]",
                comment: ""))
        }
        return newIndex
    }
    
    //
    // brackettedExpression
    //     : LPAREN conditionalOrExpression RPAREN
    //     
    //
    // returns the index of the next token to be processed.
    static func parseBrackettedExpression(stack: inout Stack<FormulaParserNode>,
                                          tokens:[FormulaParserToken], formula:String,
                                          index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index
        let token:FormulaParserToken = tokens[_index]
        switch (token.getTokenType()) {
            case .LPAREN:
                _index = try parseConditionalOrExpression(stack: &stack,
                                                          tokens:tokens,
                                                          formula:formula,
                                                          index:_index + 1)
                let current:FormulaParserToken = peek(tokens:tokens, offset:_index)
                if current.getTokenType() != .RPAREN {
                    throw ClientError.argument(String(format:NSLocalizedString(
                            "term: Found %@ @ %d expected RPAREN", comment:""),
                            current.getTokenType() as! CVarArg, current.getPos()))
                }
                stack.append(FormulaParserNode(operation:.group,
                                                leftHandSide:stack.popLast()))
                _index += 1; // consume RPAREN
            break
            default:
            break
        }
        return _index
    }
    
    //
    // functionElement
    //     : FUNCTION (args)? RPAREN
    //     
    //
    // returns the index of the next token to be processed.
    static func parseFunctionElement(stack: inout Stack<FormulaParserNode>,
                                     tokens:[FormulaParserToken], formula:String,
                                     index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index
        let token:FormulaParserToken = tokens[_index]
        switch (token.getTokenType()) {
            case .FUNCTION:
                let next:FormulaParserToken = peek(tokens:tokens, offset:_index + 1)
                // token.getLength()-1 to strip off LPAREN
                //let value:String = formula.substring(token.getPos(), 
                // token.getPos() + token.getLength() - 1)
                // TODO: make sure not off by 1
                let pos:String.Index = formula.index(formula.startIndex,
                                                     offsetBy: token.getPos())
                let len:String.Index = formula.index(pos,
                                                     offsetBy:token.getLength())
                let value:String = String(formula[pos..<len])
                // FUNCTION operation has function name on LHS, args chaining 
                // from RHS to RHS
                let function:FormulaParserNode =
                    FormulaParserNode(operation:.function,
                        leftHandSide:FormulaParserTerminal(type: .ident, value:value))

                if next.getTokenType() != .RPAREN {
                    // FUNCTION arg [, arg]* )
                    _index = try parseArgs(stack: &stack, tokens:tokens,
                                           formula:formula, index:_index + 1)
                    function.setRightHandSide(rightHandSide:stack.popLast())

                    let current:FormulaParserToken = peek(tokens:tokens, offset:_index)
                    if current.getTokenType() != .RPAREN {
                        throw ClientError.argument(String(format:NSLocalizedString(
                                "term: Found %@ @ %d. Expected RPAREN", comment:""),
                                current.getTokenType() as! CVarArg, current.getPos()))
                    }
                    _index += 1
                }
                // else {
                // no-arg function
                //}
                
                stack.append(function)
                _index += 1; // consume RPAREN
                break
            default:
            break
        }
        return _index
    }
    
    //
    // expressionElement
    //     : IDENT | NUMBER | propertyRef
    //     
    //
    // returns the index of the next token to be processed.
    static func parseExpressionElement(stack: inout Stack<FormulaParserNode>,
                                       tokens:[FormulaParserToken], formula:String,
                                       index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index
        let token:FormulaParserToken = tokens[_index]
        switch (token.getTokenType()) {
            case .STRING:
                // Do not include quotes in the value
                let pos:String.Index = formula.index(
                    formula.index(after: formula.startIndex),
                    offsetBy: token.getPos())
                let len:String.Index = formula.index(pos,
                                                     offsetBy:token.getLength() - 1)
                let value:String = String(formula[pos..<len])
                
                stack.append(FormulaParserTerminal(type: .string, value:value))
                _index += 1; // consume STRING
                break;
            
            case .IDENT:
                // TODO: make sure not off by 1
                let pos:String.Index = formula.index(formula.startIndex,
                                                     offsetBy: token.getPos())
                let len:String.Index = formula.index(pos,
                                                     offsetBy:token.getLength())
                let value:String = String(formula[pos..<len])
                
                //let value:String = formula.substring(token.getPos(),
                //                            token.getPos() + token.getLength())
                stack.append(FormulaParserTerminal(type: .ident, value: value))
                _index += 1; // consume ident
                break
            case .NUMBER:
                // TODO: make sure not off by 1
                let pos:String.Index = formula.index(formula.startIndex,
                                                     offsetBy: token.getPos())
                let len:String.Index = formula.index(pos,
                                                     offsetBy:token.getLength())
                let value:String = String(formula[pos..<len])
                //let value:String = formula.substring(token.getPos(),
                //                            token.getPos() + token.getLength())
                stack.append(FormulaParserTerminal(type:.number, value: value))
                _index += 1; // consume number
                break
            default:
                _index = try parsePropertyRef(stack: &stack, tokens:tokens,
                                          formula:formula, index:_index)
                break
        }
        return _index
    }

    //
    // args
    //     : conditionalOrExpression
    //     | conditionalOrExpression COMMA args
    //     
    //
    // returns the index of the next token to be processed.
    static func parseArgs(stack: inout Stack<FormulaParserNode>,
                          tokens:[FormulaParserToken], formula:String,
                          index:Int) throws -> Int {

        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index
        var previous:FormulaParserNode? = nil
        while _index < tokens.count {
            _index = try parseConditionalOrExpression(stack: &stack,
                                                      tokens:tokens,
                                                      formula:formula,
                                                      index:_index)
            let arg:FormulaParserNode? = previous == nil ?
                stack.last /*stack.peek()*/ : stack.popLast()
            if let previous = previous {
                previous.setRightHandSide(rightHandSide:arg)
            }
            previous = arg
            let current:FormulaParserToken = tokens[_index]
            switch (current.getTokenType()) {
                case .COMMA:
                    _index += 1
                    break
                default:
                    return _index
                //break
            }
        }
        return _index

    }

    //
    // propertyRef
    //     : DOLLAR? ATTRIBUTE IDENT RPAREN
    //     
    //
    // returns the index of the next token to be processed.
    static func parsePropertyRef(stack: inout Stack<FormulaParserNode>, 
                                 tokens:[FormulaParserToken], formula:String,
                                 index:Int) throws -> Int {
        if index >= tokens.count {
            return tokens.count
        }

        var _index:Int = index

        let token:FormulaParserToken = tokens[_index]
        switch (token.getTokenType()) {
        case .ATTRIBUTE, .DOLLAR:
                var current:FormulaParserToken = token
                // Handle attribute, which is $? $( IDENT )
                var dollarCount:Int = 0
                while current.getTokenType() == .DOLLAR {
                    dollarCount += 1
                    if dollarCount > 1 {
                        throw ClientError.argument(String(format:NSLocalizedString(
                                "term: %@  @ %d not expected", comment:""),
                                current.getTokenType() as! CVarArg, current.getPos()))
                    }
                    _index += 1
                    current = peek(tokens:tokens, offset:_index)
                }

                //var attrType:TerminalType = TerminalType.values()[dollarCount]
                let attrType:TerminalType = dollarCount == 0 ?
                    .inprocessAttribute : .currentAttribute

                if current.getTokenType() != .ATTRIBUTE {
                    throw ClientError.argument(String(format:NSLocalizedString(
                            "term: Found %@ @ %d. Expected ATTRIBUTE", comment: ""),
                            current.getTokenType() as! CVarArg, current.getPos()))
                }

                _index += 1
                current = peek(tokens:tokens, offset:_index)
                if current.getTokenType() != .IDENT {
                    throw ClientError.argument(String(format:NSLocalizedString(
                            "term: Found %@ @ %d . Expected IDENT", comment: ""),
                            current.getTokenType() as! CVarArg, current.getPos()))
                }
                // TODO: make sure not off by 1
                let pos:String.Index = formula.index(formula.startIndex,
                                                offsetBy: current.getPos())
                let len:String.Index = formula.index(pos,
                                                offsetBy:current.getLength())
                let value:String = String(formula[pos..<len])
                //let value:String = formula.substring(current.getPos(),
                //        current.getPos() + current.getLength())

                _index += 1
                current = peek(tokens:tokens, offset:_index)
                if current.getTokenType() != .RPAREN {
                    throw ClientError.argument(String(format:NSLocalizedString(
                            "term: Found %@ @ . Expected RPAREN", comment: ""),
                            current.getTokenType() as! CVarArg, current.getPos()))
                }

                stack.append(FormulaParserTerminal(type:attrType, value:value))
                _index += 1 // consume RPAREN
                break
            default:
            break
        }
        return _index
    }

    static func peek(tokens:[FormulaParserToken], offset:Int) -> FormulaParserToken {
        let index:Int = 0 <= offset && offset <= (tokens.count - 1) ? 
            offset : tokens.count - 1
        return tokens[index]
    }
    
    // left hand side needs to have higher precedence than right hand side
    // so that post-fix traversal does higher precedence operations first.
    // The swap on compare == 0 ensures the remaining operations are left-to-right.
    static func prioritize(lhs:FormulaParserNode?, rhs:FormulaParserNode?) -> FormulaParserNode? {
        guard let lhs = lhs else {
            return rhs
        }
        guard let rhs = rhs else {
            return lhs
        }
        if rhs.getNodeOperation() != .terminal {
            let c:Int = comparePrec(lhs:lhs, rhs:rhs)
            if c == 0 {
                lhs.setRightHandSide(rightHandSide:rhs.getLeftHandSide()!)
                let rightHandSide:FormulaParserNode = rhs.getRightHandSide()!
                rhs.setLeftHandSide(leftHandSide:lhs)
                rhs.setRightHandSide(rightHandSide:rightHandSide)
                return rhs
            } else if c > 0 {
                let leftHandSide:FormulaParserNode = rhs.getLeftHandSide()!
                rhs.setLeftHandSide(leftHandSide:lhs)
                lhs.setRightHandSide(rightHandSide:leftHandSide)
                return lhs
            } else {
                lhs.setRightHandSide(rightHandSide:rhs)
                return lhs
            }
        } else {
            lhs.setRightHandSide(rightHandSide:rhs)
            return lhs
        }
    }
    
    static func dump(node:FormulaParserNode?) -> String {
        guard let node = node else {
            return ""
        }
    
        if node is FormulaParserTerminal {
            let terminal:FormulaParserTerminal = node as! FormulaParserTerminal
            var s:String = terminal.getValue()
            if terminal.type == .inprocessAttribute {
                s = "$(\(s))"
            } else if terminal.type == .currentAttribute {
                s = "$$(\(s))"
            }
            return s
        }

        let lhs:String = node.getLeftHandSide() == nil ?
            "lhs is nil" : dump(node:node.getLeftHandSide())
        let rhs:String = node.getRightHandSide() == nil ?
            "rhs is nil" : dump(node:node.getRightHandSide())
        let operation:FormulaParserNodeOperation = node.getNodeOperation()
        
        return "[\(operation)|\(lhs)|\(rhs)]"
    }
}
