/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

/**
 * The MessageDispatcher queues messages for automatic dispatching to the
 * for the Oracle IoT Cloud Service.
 *
 * The MessageDispatcher prioritizes message dispatching to ensure
 * high priority messages are dispatched ahead of lower priority messages.
 * There is one MessageDispatcher instance per client.
 */
public class MessageDispatcher {
    
    static let dispatcherMapQueue:DispatchQueue =
        DispatchQueue(label:"MessageDispatcher dispatchMap queue",
                      qos:.utility)
     // TODO: this leaks. MessageDispatcherImpl holds DirectlyConnectedDevice.
    // Maybe move all this to impl.
    internal static var dispatcherMap: [AdvancedDirectlyConnectedDevice : MessageDispatcherImpl] = [:]

    /**
     * Get the instance of a `MessageDispatcher` for the `DirectlyConnectedDevice`.
     * This is an implementation detail and is called from the implementation of
     * the virtualization API.
     *
     * - parameter msgDirectlyConnectedDevice: is a gateway or directly
     *                                         connected device
     *
     * - returns: A MessageDispatcher for the `DirectlyConnectedDevice`.
     *
     * - throws: `ClientError.network` if `MessageDispatcherImpl.init()` cannot
     *            register resources
     */
    public static func getMessageDispatcher(msgDirectlyConnectedDevice:
            AdvancedDirectlyConnectedDevice) -> MessageDispatcher? {
        return MessageDispatcher.dispatcherMapQueue.sync(execute: {
            if let messageDispatcher:MessageDispatcherImpl =
                MessageDispatcher.dispatcherMap[msgDirectlyConnectedDevice] {
                return messageDispatcher
            }
            do {
            let messageDispatcher:MessageDispatcherImpl =
                try MessageDispatcherImpl(deviceClient: msgDirectlyConnectedDevice)
            dispatcherMap[msgDirectlyConnectedDevice] = messageDispatcher
            return messageDispatcher
            } catch {
                log(type:.fault,
                    message:"%@: cannot obtain MessageDispatcher for %@. %@",
                    args: "MessageDispatcher.getMessageDispatcher()",
                          msgDirectlyConnectedDevice.getEndpointId(),
                          "\(error)")
                return nil
            }
        })
    }
    
    /**
     * Set the instance of a `MessageDispatcher` for the `DirectlyConnectedDevice`.
     * This is an implementation detail and is called from the implementation of
     * the virtualization API.
     *
     * - parameter msgDirectlyConnectedDevice: is a gateway or directly connected
     *                                         device
     * - parameter messageDispatcher: is the `MessageDispatcher for the directly
     *                                 connected device
     */
    public static func setMessageDispatcher(
            msgDirectlyConnectedDevice:AdvancedDirectlyConnectedDevice,
            messageDispatcher:MessageDispatcher) {
        MessageDispatcher.dispatcherMapQueue.sync() {
            MessageDispatcher.dispatcherMap[msgDirectlyConnectedDevice] =
                messageDispatcher as? MessageDispatcherImpl
        }
    }
    
    /**
     * Remove the instance of a `MessageDispatcher` for the `DirectlyConnectedDevice`.
     * After calling `removeMessageDispatcher`, a call to `getMessageDispatcher`
     * will return `nil`.
     * This is an implementation detail and is called from the implementation of
     * the virtualization API.
     *
     * - parameter msgDirectlyConnectedDevice: is a gateway or directly connected device
     *
     * - returns: the `MessageDispatcher` for the `DirectlyConnectedDevice`
     */
    public static func removeMessageDispatcher(
            msgDirectlyConnectedDevice:AdvancedDirectlyConnectedDevice) ->
                MessageDispatcher? {
        return MessageDispatcher.dispatcherMapQueue.sync(execute: {
            return MessageDispatcher.dispatcherMap.removeValue(forKey:msgDirectlyConnectedDevice)
        })
    }

    /**
     * A callback interface for successful delivery of messages.
     *
     * - (Abstract class)
     *
     */
    open class DeliveryCallback {
        public init() {
        }
        
        /**
         * Notify that messages have been successfully dispatched. This callback
         * indicates that the messages have been delivered to, and accepted by,
         * the server and are no longer queued.
         *
         * - (Abstract method)
         *
         * - parameter messages: the messages that were delivered
         */
        open func delivered(messages: [Message]) {
            fatalError("Subclass need to implement the 'delivered()' method.")
        }
    }
    
    /**
     * A callback interface for errors in delivery of messages.
     *
     * - (Abstract class)
     *
     */
    open class ErrorCallback {
        public init() {
        }
        
        /**
         * Notify that an error occurred when attempting to deliver messages
         * to the server. This callback indicates that the messages have
         * not been delivered to the server and are no longer queued.
         *
         * - (Abstract method)
         *
         * - parameter messages: the messages that were not delivered
         */
        open func failed(messages: [Message], error:ClientError? = nil) throws {
            fatalError("Subclass need to implement the 'failed()' method.")
        }
    }
    
    /**
     * Get the RequestDispatcher that is used by this MessageDispatcher for
     * handling RequestMessages.
     *
     * - (Abstract method)
     *
     * - returns: a RequestDispatcher
     *
     * - see: RequestDispatcher
     */
    public func getRequestDispatcher() -> RequestDispatcher {
        fatalError("Subclass need to implement the 'getRequestDispatcher()' method.")
    }
    
    /**
     * Add the messages to the outgoing message queue if it is possible to
     * do so without violating capacity restrictions.
     *
     * - parameter messages: The messages to be queued
     *
     * - throws: `ClientError.state` if all the messages cannot be added to the queue
     *      `ClientError.argument` `messages` is `nil` or empty
     */
    public func queue(messages:[Message]) throws {
        fatalError("Subclass need to implement the 'queue(messages:)' method.")
    }
    
    /**
     * Add the message to the outgoing message queue if it is possible to
     * do so without violating capacity restrictions.
     *
     * - (Abstract method)
     *
     * - parameter message: The message to be queued
     */
    public func queue(message: Message) throws {
        fatalError("Subclass need to implement the 'queue(message:)' method.")
    }
    
    /**
     * Offer messages to be queued. Depending on the policies, if any, the 
     * messages will be queued if it is possible to do so without violating 
     * capacity restrictions.
     *
     * - parameter messages: The messages to be queued
     * - throws: `ClientError.state` if all the messages cannot be added to the queue
     *      `ClientError.argument` if `messages` is empty
     */
    public func offer(messages:[Message]) throws {
        fatalError("Subclass need to implement the 'offer(messages:)' method.")
    }
    
    /**
     * Offer a message to be queued. Depending on the policies, if any, the 
     * message will be queued if it is possible to do so without violating 
     * capacity restrictions.
     *
     * - parameter message: The message to be queued
     * - throws `ClientError.state` if the message cannot be added to the queue
     */
    public func offer(message:Message) throws {
        try offer(messages:[message])
    }
    
    /**
     * Set a callback to be notified if there is an error in dispatching.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func setOnError(callback: ErrorCallback) {
        fatalError("Subclass need to implement the 'setOnError(ErrorCallback)' method.")
    }
    
    /**
     * Set a callback to be notified if message is successfully delivered.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func setOnDelivery(callback: DeliveryCallback) {
        fatalError("Subclass need to implement the 'setOnDelivery(DeliveryCallback)' method.")
    }
    
    /**
     * Closes the MessageDispatcher.
     * The callback will not be invoked on the main thread.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func close(callback: @escaping () -> Void) {
        fatalError("Subclass need to implement the 'close()' method.")
    }
    
    
}
