/*
 *  Copyright © 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

fileprivate let timed_policy_thread_count_queue:DispatchQueue =
    DispatchQueue(label:"time_policy_thread_count queue", qos:.utility)
fileprivate var _timed_policy_thread_count:Int = 0
fileprivate var timed_policy_thread_count:Int {
    get {
        return timed_policy_thread_count_queue.sync(execute: {
            return _timed_policy_thread_count
        })
    }
    set {
        timed_policy_thread_count_queue.sync() {
            _timed_policy_thread_count += newValue
        }
    }
}

// TODO: can we have only one of these threads for all virtual devices?
class TimedPolicyThread {

    let timedThreadRunQueue:DispatchQueue = DispatchQueue(label:"timedThreadRunQueue",
                                                          qos:.userInitiated)
    let scheduledPolicyDataSemaphore:DispatchSemaphore = DispatchSemaphore(value:0)
    // TODO: Why was this a global queue ?
    //let scheduledPolicyDataQueue:DispatchQueue =
    //   DispatchQueue.global(qos:.userInitiated)
    let scheduledPolicyDataQueue:DispatchQueue =
        DispatchQueue(label:"scheduledPolicyDataQueue", qos:.userInitiated)
    private var scheduledPolicyData:[ScheduledPolicyData] = [ScheduledPolicyData]()
    
    let cancelledQueue:DispatchQueue = DispatchQueue.global(qos:.userInitiated)
    var /*transient*/ _cancelled:Bool = false
    var isCancelled:Bool {
        get {
            return cancelledQueue.sync() {
                return _cancelled
            }
        }
        set {
            cancelledQueue.sync() {
                _cancelled = newValue
            }
        }
    }

    var _isAlive:Bool = false
    var isAlive:Bool {
        get {
            return cancelledQueue.sync() {
                return _isAlive
            }
        }
        set {
            cancelledQueue.sync() {
                _isAlive = newValue
            }
        }
    }

    var virtualDevice:VirtualDeviceImpl!
    var name:String!
    
    init(virtualDevice:VirtualDeviceImpl) {
        self.virtualDevice = virtualDevice
        // Always set to 1 it will increment by one
        timed_policy_thread_count = 1
        self.name = "timed-policy-thread_\(timed_policy_thread_count)"
        //self.run()
    }

    func addTimedPolicyData(data:ScheduledPolicyData) {
        scheduledPolicyDataQueue.sync() {
            let index = scheduledPolicyData.index(of:data)
            if let index = index, !scheduledPolicyData.isEmpty {
                scheduledPolicyData[index] = data
            } else {
                scheduledPolicyData.append(data)
            }
            scheduledPolicyDataSemaphore.signal()
        }
    }

    func removeTimedPolicyData(data:ScheduledPolicyData) {
        scheduledPolicyDataQueue.sync() {
            let index = scheduledPolicyData.index(of:data)
            if let index = index {
                scheduledPolicyData.remove(at:index)
            }
            scheduledPolicyDataSemaphore.signal()
        }
    }

    // TODO: never used. Do we need cancelled and cancel()?
    func cancel() {
        isCancelled = true
        //this.interrupt()
        scheduledPolicyDataSemaphore.signal()
    }

    public func start() {

        timedThreadRunQueue.async() {
            if !self.isAlive {
                self.isAlive = true
            } else {
                return
            }
            
        var delay:Int64 = -1
        while !self.isCancelled {
            delay = self.scheduledPolicyDataQueue.sync() {
                // Wait for a timed function to be added
                
                if !self.isCancelled && self.scheduledPolicyData.isEmpty {
                    return -1
                }

                let now:Int64 = getCurrentTimeMillis()
                
                // Must be done here pushes expired policies down the list
                self.scheduledPolicyData.sort(by: { (o1, o2) in
                    let x:Int64 = o1.getDelay(now:now)
                    let y:Int64 = o2.getDelay(now:now)
                    // Don't do anything if equal
                    return (x < y) ? true : false //((x == y) ? false : false)
                })
                
                // get the shortest delay
                let _delay:Int64 = self.scheduledPolicyData[0].getDelay(now:now)

                // TODO: This is a wait in JCL. When it returns it does not
                // sort again. In this case when the loop is entered it sorts
                // again. Is this an issue
                
                // May have been cancelled while waiting for data to be
                // added to scheduledPolicyData
                if !self.isCancelled && _delay > 0 {
                    return _delay
                }

                let currentTimeMillis:Int64 = getCurrentTimeMillis()

                var updatedAttributes:[Pair<VirtualDeviceAttribute,AnyObject>] =
                    [Pair<VirtualDeviceAttribute,AnyObject>]()
                for policyData in self.scheduledPolicyData {
                    if policyData.getDelay(now: currentTimeMillis) <= delay {
                        do {
                            try policyData.processExpiredFunction(
                                virtualDevice:self.virtualDevice,
                                updatedAttributes: &updatedAttributes,
                                timeZero:currentTimeMillis/*, no slide
                                devicePolicy: self.getDevicePolicy()*/)
                        } catch {
                            log(type:.fault,
                                message:"TimedPolicyThread:run(): " +
                                    "Unexpected exception from processExpiredFunction, virtualDevice %@ %@",
                                args: self.virtualDevice.getDeviceModel().getURN(),
                                      self.virtualDevice.getEndpointId())
                        }
                    }
                }
                if !updatedAttributes.isEmpty {
                    //
                    // Call updateFields to ensure the computed metrics get run,
                    // and will put all attributes into one data message.
                    //
                    do {
                        try self.virtualDevice.adapter!.updateFields(updatedAttributes:&updatedAttributes)
                    } catch {
                        log(type:.fault,
                            message:"TimedPolicyThread:run(): " +
                            "Unexpected exception from updateFields, virtualDevice %@ %@",
                            args: self.virtualDevice.getDeviceModel().getURN(),
                                  self.virtualDevice.getEndpointId())
                    }
                }
                // This should never return < 0
                // if delay is less than zero then the list of policies was
                // exhausted. set delay to the top most delay
                // if it is zero it will re-enter the expire loop and resort the
                // the list of scheduled policies
                return _delay < 0 ? self.scheduledPolicyData[0].getDelay(now:getCurrentTimeMillis()) : _delay
            }
            if delay < 0 {
                self.scheduledPolicyDataSemaphore.wait()
            } else {
                let _ = self.scheduledPolicyDataSemaphore.wait(timeout:DispatchTime.now() +
                    DispatchTimeInterval.milliseconds(Int(delay)))
            }
        }
        self.isAlive = false
            
        } // end async
    }
    
    private func getDevicePolicy() -> DevicePolicy? {
        do {
            guard let devicePolicy:DevicePolicy =
                try virtualDevice.devicePolicyManager!.getPolicy(
                    deviceModelURN:virtualDevice.getDeviceModel().getURN(),
                    deviceId:virtualDevice.getEndpointId()) else {
                        // SEVERE
                        log(type:OSLogType.error,
                            message: "TimedPolicyThread:getDevicePolicy: " +
                                "no policy for %@, %@",
                            
                            args: virtualDevice.getDeviceModel().getURN(),
                                  virtualDevice.getEndpointId())
                        return nil
            }
            return devicePolicy
        } catch {
            // SEVERE
            log(type:.error,
                message: "TimedPolicyThread:getDevicePolicy: " +
                    "caught exception getting policy for %@, %@ : %@",
                args: virtualDevice.getDeviceModel().getURN(),
                      virtualDevice.getEndpointId(), "\(error)")
        }
        return nil
    }
}
//
// Window based policy support (as in "window", not Windows OS).
// Have one scheduled task for each policy "window".
// When the window expires, the runnable will call back each
// VirtualDeviceAttribute that has a policy for that window.
//

// [attributeName:pipelineIndex]
fileprivate let pipelineIndexQueue:DispatchQueue =
    DispatchQueue(label:"pipeline index queue", qos:.utility)

class ScheduledPolicyData : Hashable {
    
    class Key : Hashable {
        private var window:Int64!
        private var slide:Int64!
        
        public var hashValue:Int {
            get {
                // Assumes slide and window are always positive and unsigned
                return Int((window ^ (window >> 32)) + (slide ^ (slide >> 32)))
            }
        }
        
        init(window:Int64, slide:Int64) {
            self.window = window
            self.slide = slide
        }

        public static func ==(lhs:ScheduledPolicyData.Key,
                              rhs:ScheduledPolicyData.Key) -> Bool {
            return lhs === rhs || (lhs.window == rhs.window &&
                                  lhs.slide == rhs.slide)
        }
    }
    
    // Millisecond time in the future at which the policy value should be computed
    private var expiry:Int64!
    
    // How much the window moves is used to calculate expiry.
    private var slide:Int64!
    
    // no slide
    //private var window:Int64!

    private var pipelineIndices:[String:Int]!
    /* no slide
    // last time this policy was run
    private var timeZero:Int64!
    */
    
    // TODO: JCL only uses "window"
    public var hashValue:Int {
        get {
            // Assumes slide is positive and unsigned
            return Int(slide^(slide>>32))
            /* No slide
            var hc:Int = Int(window^(window>>32))// as! Int
            //hc = 31 &* hc &+ attributeName.hashValue
            //hc = 31 &* hc &+ pipelineIndex
            hc += 31 &* hc &+ pipelineIndexQueue.sync() {
                return NSDictionary(dictionary:pipelineIndices).hashValue
            }
            return hc
            */
        }
    }
   
    public static func ==(lhs:ScheduledPolicyData, rhs:ScheduledPolicyData) -> Bool {
        return lhs === rhs ||
            /* no slide
            // TODO: JCL only uses "window"
            (lhs.window == rhs.window &&
                pipelineIndexQueue.sync() {
                    return NSDictionary(dictionary:lhs.pipelineIndices).isEqual(to: rhs.pipelineIndices)
                })
                / *&&
                lhs.pipelineIndex == rhs.pipelineIndex &&
                lhs.attributeName == rhs.attributeName)
                * /
            */
           lhs.slide == rhs.slide
    }
    
    var isEmpty:Bool {
        get {
            return pipelineIndexQueue.sync() {
                return pipelineIndices.isEmpty
            }
        }
    }

    init(window:Int64, slide:Int64, timeZero:Int64) {
        // No slide
        //self.window = window
        self.slide = slide
        self.pipelineIndices = [String:Int]()
        /* no slide
        // tenth of a millisecond resolution helps group
        // intersecting windows (10 second and 20 second,
        // for example).
        self.timeZero = (timeZero/10)*10
        */
        // Initial expiry is window milliseconds past time zero.
        // Tenth of a millisecond resolution helps group
        // intersecting slide values (10 second and 20 second,
        // for example).
        self.expiry = ((window + timeZero) / 10) * 10
    }
    
    func addAttribute(attributeName:String, pipelineIndex:Int) {
        pipelineIndexQueue.sync() {
            pipelineIndices[attributeName] = pipelineIndex
        }
    }
    
    func removeAttribute(attributeName:String, pipelineIndex:Int) {
        pipelineIndexQueue.sync() {
            _ = pipelineIndices.removeValue(forKey:attributeName)
        }
    }

    func getDelay(now:Int64) -> Int64 {
        /* no slide
        let delay = window - (now - timeZero)
        */
        let delay:Int64 = expiry - now
        return delay > 0 ? delay : 0
    }

    func processExpiredFunction(virtualDevice:VirtualDeviceImpl,
            updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>],
                                timeZero:Int64/*, no slide
                                devicePolicy:DevicePolicy?*/) throws {
        defer {
            /* no slide
            // ensure timeZero is reset
            // tenth of a millisecond resolution
            self.timeZero = (timeZero/10) * 10
            */
            // ensure expiry is reset
            // tenth of a millisecond resolution
            self.expiry = ((slide + timeZero) / 10) * 10
        }

        try handleExpiredFunction(virtualDevice:virtualDevice,
                                  updatedAttributes:&updatedAttributes/*, no slide
                                  devicePolicy:devicePolicy*/)
    }

    func handleExpiredFunction(virtualDevice:VirtualDeviceImpl,
            updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>]/*, no slide
                               devicePolicy:DevicePolicy?*/) throws {
        /* no slide
        guard let devicePolicy = devicePolicy else {
            // SEVERE
            // TODO: better log message here
            log(type:.error,
                message:"ScheduledPolicyData.handleExpiredFunction(): " +
                    "passed nil devicePolicy for %@ %@",
                args: virtualDevice.getDeviceModel().getURN(),
                      virtualDevice.getEndpointId())
            return
        }
        let pipelineIndicesCopy:[String:Int] =
            pipelineIndexQueue.sync() {
                return pipelineIndices
        }
        try handleExpiredFunction(updatedAttributes:&updatedAttributes,
                              devicePolicy:devicePolicy,
                              pipelineIndices:pipelineIndicesCopy,
                              virtualDevice:virtualDevice)
        */
        guard let devicePolicy =
            try virtualDevice.devicePolicyManager?.getPolicy(
                deviceModelURN:virtualDevice.getDeviceModel().getURN(),
                deviceId:virtualDevice.getEndpointId()) else {
            // SEVERE
            // TODO: better log message here
            log(type:.error,
                message:"%@: could not find %@, %@, in policy configuration",
                args: "ScheduledPolicyData.handleExpiredFunction()",
                    virtualDevice.getDeviceModel().getURN(),
                    virtualDevice.getEndpointId())
            return
        }
        // structs passed by value
        // TODO: But what about the internal references ?
        let pipelineIndicesCopy:[String:Int] =
            pipelineIndexQueue.sync() {
                return pipelineIndices
        }
        try handleExpiredFunction(virtualDevice:virtualDevice,
                                  updatedAttributes:&updatedAttributes,
                                  devicePolicy:devicePolicy,
                                  pipelineIndices:pipelineIndicesCopy/*, no slide
                                  virtualDevice:virtualDevice*/)
    }
    
    func handleExpiredFunction(virtualDevice:VirtualDeviceImpl,
            updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>],
                               devicePolicy:DevicePolicy,
                               pipelineIndices:[String:Int]) throws {
    
        // TODO: Can attributeName be nil here ?
        // Don't think so since it is a key.
        // Need to see if there is still a "MODEL_POLICY_TOKEN" needed
        //for (Map.Entry<String, Integer> entry : _pipelineIndices.entrySet()) {
        for (attributeName,pipelineIndex) in pipelineIndices {
            try handleExpiredFunction(virtualDevice:virtualDevice,
                                  updatedAttributes:&updatedAttributes,
                                  devicePolicy:devicePolicy,
                                  attributeName:attributeName,
                                  pipelineIndex:pipelineIndex)
        }
    }

    func handleExpiredFunction(virtualDevice:VirtualDeviceImpl,
            updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>],
                               devicePolicy:DevicePolicy,
                               attributeName:String,
                               pipelineIndex:Int) throws {

        guard let pipeline:[DevicePolicy.Function] =
            devicePolicy.getPipeline(attributeName:attributeName), !pipeline.isEmpty else {
            return
        }
        if pipeline.count < pipelineIndex {
            // SEVERE
            log(type:OSLogType.error,
                message: "%@: pipeline for %@ does not match configuration " +
                    "policyId: %@, pipeline.count: %@, pipelineIndex: %@",
                args: "ScheduledPolicyData.handleExpiredFunction()",
                    attributeName, devicePolicy.getId(),
                    "\(pipeline.count)", "\(pipelineIndex)")
            return
        }
        // If attributeName is nil, the MODEL_PIPELINE_TOKEN will be used.
        var pipelineData:[Dictionary<String,AnyObject>] =
            try virtualDevice.getPipelineData(attribute: attributeName)
        if pipelineData.count <= pipelineIndex {
            // SEVERE
            log(type:OSLogType.error,
                message: "%@: pipeline data for %@ does not match configuration " +
                "policyId: %@, pipelineData.count: %@, pipelineIndex: %@",
                args: "ScheduledPolicyData.handleExpiredFunction()",
                    attributeName, devicePolicy.getId(),
                    "\(pipelineData.count)", "\(pipelineIndex)")
            return
            
        }

        let remainingPipelineConfigs:[DevicePolicy.Function] =
                Array(pipeline[pipelineIndex..<pipeline.count])

        var remainingPipelineData:[Dictionary<String,AnyObject>] =
                Array(pipelineData[pipelineIndex..<pipelineData.count])

        //if let attributeName = attributeName {
        if attributeName != DevicePolicy.ALL_ATTRIBUTES() {
            virtualDevice.processExpiredFunction(
                    updatedAttributes:&updatedAttributes,
                    attributeName:attributeName,
                    pipeline:remainingPipelineConfigs,
                    pipelineData:&remainingPipelineData)
        } else {
            virtualDevice.processExpiredFunction(
                pipeline:remainingPipelineConfigs,
                pipelineData:&remainingPipelineData)
        }
        // Update the virtual device with modified values
        pipelineData.replaceSubrange(pipelineIndex..<pipeline.count,
                                     with: remainingPipelineData)
        virtualDevice.putPipelineData(attribute:attributeName,
                                      pipelineData: pipelineData)
    }
}

