/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * `RequestMessageBuilder` extends the `MessageBuilder` class. The 
 * `RequestMessage` class is immutable.
 * A builder is required when creating a `RequestMessage`. `RequestMessage` uses
 * the "builder" design pattern.
 */
 public class RequestMessageBuilder: MessageBuilder {

    /** HTTP request message header. Header is pair of name and list of values. 
     */
    fileprivate var headers: [String : [String]] = [ : ]
    /** HTTP request message parameter */
    fileprivate var params: [String : String] = [ : ]
    /** Method of the HTTP request message */
    fileprivate var method: String?
    /** URL of the HTTP request message */
    fileprivate var url: String?
    /** Body of the HTTP request message */
    fileprivate var body: Foundation.Data?
 
    public override init() {
    }
    
    private func isAsciiPrintable(text: String) -> Bool {
        for i in text {
            let someString = String(i)
            for myChar in someString.unicodeScalars {
                if (myChar.value < 32) || (myChar.value > 126) {
                    return false
                }
            }
        }
        return true
    }
    
    // TODO: name can really never be nil, so the name parameter should not
    // require '?'. Code inconsistent with doc, says value can be nil but
    // is coded so that it cannot be nil
    /**
     * Add Http message header. The name cannot be `nil`. `value` can be
     * `nil`. The `value` array of values cannot contain `nil`s.
     * `name` and `value` elements must contain ASCII-printable characters only.
     * Note that the name of the header is always converted to lower case.
     *
     * - parameter name: Name of the http header.
     *
     * - parameter values: Http header values.
     *
     * - throws: `ClientError.argument` if `name` or `value` is `nil`, or are 
     *           not printable ascii.
     */
    public func addHeader(name: String?, value: [String]?) throws {
        guard let name = name else {
            throw ClientError.argument(NSLocalizedString(
                "Request Message header name must not be nil", comment: ""))

        }
        guard let value = value else {
            throw ClientError.argument(NSLocalizedString(
                "Request Message header value must not be nil", comment: ""))
        }
        if !isAsciiPrintable(text: name) {
            throw ClientError.argument(NSLocalizedString(
                "Request Message header name contains non ASCII printable " +
                "characters", comment: ""))
        }
        for v in value {
            if !isAsciiPrintable(text: v) {
                throw ClientError.argument(NSLocalizedString(
                    "Request Message header value contains non ASCII " +
                    "printable characters", comment: ""))
            }
        }

        let lowerName:String = name.lowercased()
        if headers[lowerName] == nil {
            let newValues: [String] = []
            headers[lowerName] = newValues
        }
        for v in value {
            headers[lowerName]!.append(v)
        }
    }

    // TODO: name can really never be nil, so the name parameter should not
    // require '?'.
    
    /**
     * Add Http parameter. Name of the param cannot be `nil`.
     *
     * - parameter name: Http parameter name.
     *
     * - parameter value: Http parameter value.
     *
     * - throws: `ClientError.argument` if `name` is `nil`
     */
    public func addParam(name: String?, value: String) throws {
        guard let name = name else {
            throw ClientError.argument(NSLocalizedString(
                "Request Message parameter name must not be nil", comment: ""))
        }
        self.params[name] = value
    }

    /**
     * Set message URL. The URL is not checked for validity.
     *
     * - parameter url: URL of the HTTP request.
     */
    public func setUrl(url: String) {
        self.url = url
    }

    /**
     * Set message body.
     *
     * - parameter body: Body of the HTTP message, should not be `nil`
     */
    public func setBody(body: Foundation.Data) {
        self.body = body
    }

    /**
     * Set message body using a `String`. The encoding for the body is UTF-8.
     *
     * - parameter body: Body of the HTTP message
     */
    public func setBody(body: String) {
        setBody(body: body.data(using: .utf8)!)
    }

    /**
     * Set Http method. Tha name of the HTTP method is always converted to 
     * lower case.
     *
     * - parameter method: Method of the HTTP.
     */
    public func setMethod(method: String?) {
        if let method = method {
            self.method = method.lowercased()
        } else {
            // Set it to nil
            self.method = method
        }
    }
    
    /**
     * Method to read a JSON object to a `RequestMessageBuilder`.
     *
     * - parameter jsonObject: the JSON Object as `Foundation.Data`.
     *
     * - returns: the extracted item from `MessageBuilder`
     *
     * - throws: `ClientError.state` if `clientId` is `nil` or empty, `payload`
     *           is `nil`, or if `method` or `url` in the payload are `nil` or
     *           empty.
     */
    override public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String,Any> {
        
        let jsonDict:Dictionary<String,Any> = try super.fromJSON(data: data)
        guard let _clientId:String = self.clientId, !_clientId.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Illegal Request Message client ID", comment: ""))
        }
        var headers: Dictionary<String,Any>? = nil
        var params: Dictionary<String,Any>? = nil
        var body: Foundation.Data? = nil
        var method: String? = nil
        var url: String? = nil
        
        // TODO: according to the declarations, jsonDict:Dictionary<String,Any>
        // value can never be nil, not need to check if for nil in code
        for (key, value) in jsonDict {
            if key != "payload" {
                continue
            }
            guard let payload:Dictionary<String,Any> =
                                    (value as? Dictionary<String,Any>) else {
                throw ClientError.state(NSLocalizedString(
                    "Request Message payload must not be nil", comment: ""))
            }
            for (key1, value1) in payload {
                switch(key1) {
                case "method":
                    guard let m = (value1 as? String), !m.isEmpty else {
                        throw ClientError.state(NSLocalizedString(
                            "Illegal Request Message method", comment: ""))
                    }
                    method = m
                    break
                case "url":
                    guard let u = (value1 as? String), !u.isEmpty else {
                        throw ClientError.state(NSLocalizedString(
                            "Illegal Request Message URL", comment: ""))
                    }
                    url = u
                    break
                case "body":
                    if let v:String = value1 as? String,
                        let b:Foundation.Data = Foundation.Data(base64Encoded: v,
                            options: Foundation.Data.Base64DecodingOptions(rawValue: 0)) {
                        body = b
                    }
                break
                case "headers":
                    headers = value1 as? Dictionary<String,Any>
                break
                case "params":
                    params = value1 as? Dictionary<String,Any>
                break
                default:
                break
                }
            }
        }
        setMethod(method: method)
        setUrl(url: url!)
        setBody(body: body!)
        
        if let headers = headers {
            for (key, value) in headers {
                if let vals = value as? Array<Any> {
                    var vaList: [String] = []
                    for val in vals {
                        if let v:String = val as? String {
                            vaList.append(v)
                        }
                    }
                    try addHeader(name: key, value: vaList)
                }
            }
        }
        if let params = params {
            for (key, value) in params {
                if let v:String = value as? String {
                    try addParam(name: key, value: v)
                }
            }
        }
        
        return jsonDict
    }
    /**
     * Creates new instance of `RequestMessage` using values from 
     * `RequestMessageBuilder`.
     *
     * - returns: Instance of `RequestMessage`
     *
     * - throws: WrongSource
     */
    override public func build() throws -> RequestMessage {
        return try RequestMessage(builder: self)
    }
}

/**
 * Implementation of protocol Equatable
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: *true* if equal, *false* otherwise
 */
public func ==(lhs: RequestMessage, rhs: RequestMessage) -> Bool {
    if lhs === rhs {
        return true
    }
    if lhs.body.count != rhs.body.count {
        return false
    }
    if lhs.body != rhs.body {
        return false
    }
    if !(NSDictionary(dictionary: lhs.headers).isEqual(to: rhs.headers)) {
        return false
    }
    if lhs.method != rhs.method {
        return false
    }
    if !(NSDictionary(dictionary: lhs.params).isEqual(to: rhs.params)) {
        return false
    }
    return lhs.url == rhs.url
}

/**
 * `RequestMessage` extends Message class. It stores HTTP headers, parameters,
 * method, URL and body. The body is in byte array. Default encoding for string
 * body is UTF-8.
 * This class is immutable.
 */
public class RequestMessage: Message {
    /**
     * Implementation of protocol `Hashable`
     *
     * - returns: the hash value
     */
    override public var hashValue : Int {
        // TODO: How create a Dictionary from a Dictionary ?
        var result: Int = super.hashValue
        result = 31 &* result &+ NSDictionary(dictionary: headers).hashValue
        result = 31 &* result &+ NSDictionary(dictionary: params).hashValue
        result = 31 &* result &+ method.hashValue
        result = 31 &* result &+ url.hashValue
        // DJB hash function for arrays
        let bodyArray: [UInt8] = [UInt8](repeating: 0, count: body.count)
        body.copyBytes(to: UnsafeMutablePointer(mutating: bodyArray),
                       count: body.count)
        result = 31 &* result &+ bodyArray.reduce(5381) { ($0 << 5) &+ $0 &+ Int($1) }
        return result
    }
    
    /** HTTP request message header. Header is pair of name and list of values. */
    internal var headers: [String : [String]]
    
    /** HTTP request message parameter */
    internal var params: [String : String]
    
    /** Method of the HTTP request message */
    internal var method: String
    
    /** URL of the HTTP request message */
    internal var url: String
    
    /** Body of the HTTP request message */
    internal var body: Foundation.Data
    
    /**
     * `RequestMessage` constructor takes a  `RequestMessageBuilder` and sets values to
     * each field. If the value is `nil`, set a default value. Url cannot be `nil` or empty.
     *
     * - parameter builder: generic type of message builder
     *
     * - throws: `ClientError.state` if `builder.source` is `nil` or is the
     *                                empty string.
     */
    init(builder: RequestMessageBuilder) throws {

        self.headers = builder.headers
        self.params = builder.params
        guard let url:String = builder.url, !url.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Requested url cannot be nil or empty.", comment: ""))
        }
        self.url = url
        self.body = builder.body ??
            Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0)
        self.method = builder.method ?? "get"
        try super.init(builder: builder)
    }

    /**
     * Get Http request message headers.
     *
     * - returns: dictionary of pair String and array of String representing 
     *            headers, never `nil`
     */
    public func getHeaders() -> [String : [String]] {
        return self.headers
    }
    
    // TODO: name can really never be nil, so the name parameter should not
    // require '?'.

    /**
     * Get Http request message header values for given header name.
     *
     * - parameter name: header name, should not be `nil`
     *
     * - returns: array of header values, never `nil`
     */
    public func getHeaderValues(name: String?) -> [String] {
        guard let name = name else {
            return []
        }
        let lname = name.lowercased()
        return self.headers[lname] ?? []
    }
    
    // TODO: name can really never be nil, so the name parameter should not
    // require '?'.
    
    /**
     * Get request message header value at given index.
     *
     * - parameter name: header name, should not be `nil`
     *
     * - parameter index: index in the array of header values.
     *
     * - returns: header value, may return `nil` if header name does not exist 
     *            or index is out of range.
     */
    public func getHeaderValue(name: String?, index: Int) -> String? {
        guard let name = name else {
            return nil
        }
        let lname = name.lowercased()
        guard let values = self.headers[lname], values.count > index else {
            return nil
        }
        return values[index]
    }
    // TODO: name can really never be nil, so the name parameter should not
    // require '?'.
    /**
     * Get first value from request message header.
     *
     * - parameter name: header name, should not be `nil`
     *
     * - returns: first header value, may return `nil` if the header name does 
     *            not exist or no values were set for given header name.
     */
    public func getHeaderValue(name: String?) -> String? {
        guard let name = name else  {
            return nil
        }
        let lname = name.lowercased()
        guard let values = self.headers[lname], !values.isEmpty else {
            return nil
        }
        return values[0]
    }

    /**
     * Get Http request message parameters.
     *
     * - returns: dictionary of pair String and String representing parameters
     */
    public func getParams() -> [String : String] {
        return self.params
    }

    /**
     * Get parameter value for given parameter name.
     *
     * - parameter name: parameter name
     *
     * - returns: parameter value, may return `nil` if the parameter name does 
     *            not exist.
     */
    public func getParam(name: String) -> String? {
        return self.params[name]
    }

    /**
     * Get request message URL.
     *
     * - returns: URL
     */
    public func getURL() -> String {
        return self.url
    }

    /**
     * Get HTTP request message body. If message body contains string value, 
     * default encoding is UTF-8.
     *
     * - returns: body, never `nil`
     */
    public func getBody() -> Foundation.Data {
        return self.body
    }

    /**
     * Get HTTP request message body as String. Return the inner representation 
     * of message body to String.
     *
     * - returns: Http message body as String
     *
     */
    public func getBodyString() throws -> String {
        // TODO: What is thrown ?
        return String(data: getBody(), encoding: .utf8)!
    }

    /**
     * Get request message method. Note that the returned String is always in 
     * lower case.
     *
     * - returns Http method, never `nil`
     */
    public func getMethod() -> String? {
        return self.method.lowercased()
    }

    /**
     * Get message type.
     *
     * - returns: type
     */
    override public func getType() -> MsgType {
        return .REQUEST
    }

    /**
     * Exports data from `RequestMessage` to String using JSON interpretation
     * of the message.
     *
     * - returns: JSON interpretation of the message as String.
     *
     */
    override public func toString() -> String {
        return toJsonString()
    }

    /**
     * Exports Http request message as JSON string.
     *
     * - returns: request message as JSON string
     *
     */
    override public func toJsonString() -> String {
        // TODO: What is thrown ?
        var result: String
        var addings: String = ""
        
        addings = addings + "\"payload\":{"
        addings = addings + "\"method\":\"" + method + "\","
        addings = addings + "\"url\":\"" + url + "\","
        addings = addings + "\"headers\":{"
        let headerKeys: [String] = Array(self.headers.keys)
        var z: Int = 0
        for k in headerKeys {
            addings = addings + "\"" + k + "\":["
            var zz: Int = 0
            if headers[k] != nil {
                for s in headers[k]! {
                    addings = addings + "\"" + s + "\""
                    zz += 1
                    if zz < headers[k]!.count {
                        addings = addings + ","
                    }
                }
                addings = addings + "]"
            }
            z += 1
            if z < headers.count {
                addings = addings + ","
            }
        }
        addings = addings + "},"
        addings = addings + "\"params\":{"
        let paramKeys: [String] = Array(self.params.keys)
        z = 0
        for p in paramKeys {
            addings = addings + "\"" + p + "\":\"" + params[p]! + "\""
            z += 1
            if z < params.count {
                addings = addings + ","
            }
        }
        addings = addings + "},"
        addings = addings + "\"body\":\"" + body.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0)) + "\""
        addings = addings + "}"
        
        result = toJsonString(expand: false, additionalItems: addings)
        
        return result
    }
}
