/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func > <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l > r
  default:
    return rhs < lhs
  }
}


/**
 * *ResponseMessageBuilder* extends *MessageBuilder* class. *ResponseMessage* 
 * class is immutable.
 * A builder is required when creating a *ResponseMessage*. *ResponseMessage*
 * uses Builder design pattern.
 */
public class ResponseMessageBuilder: MessageBuilder {
    
    /** HTTP response message header. Header is a name and list of values pair. 
     */
    fileprivate var headers: [String : [String?]?] = [ : ]
    /** Status code of the HTTP response message */
    fileprivate var statusCode: StatusCode?
    /** URL of the HTTP response message */
    fileprivate var url: String?
    /** ID of the request. Can be used for pairing request-response */
    fileprivate var requestId: String?
    /** Body of the HTTP response message */
    fileprivate var body: Foundation.Data?
    
    public override init() {
    }
    
    /**
     * Constructs *ResponseMessage* from a *RequestMessage*.
     *
     * - parameter requestMessage: the *RequestMessage*
     */
    public convenience init(requestMessage: RequestMessage) {
        self.init()
        self.source = requestMessage.destination;
        self.destination = requestMessage.source;
        self.requestId = requestMessage.getId()
        self.sender = self.source
    }
    
    /**
     * Add message header. The name cannot be `nil`. The values can be 
     * `nil`. The array of values cannot contain `nil`. Name and values must 
     * contain ASCII-printable characters only. Note that
     * the name of the header is always converted to lower case.
     *
     * - parameter name: header name
     *
     * - parameter values: header values
     *
     * - throws: `ClientError.argument` if `name` or a `value` is `nil` or not 
     *           ascii printable
     */
    public func addHeader(name: String?, values: [String?]?) throws {
        var vals = values
        if name == nil {
            throw ClientError.argument(NSLocalizedString(
                "Response Message header name must not be nil", comment: ""))
        }
        if !isAsciiPrintable(text: name!) {
            throw ClientError.argument(NSLocalizedString(
                "Response Message header name must be ASCII-printable",
                comment: ""))
        }
        if vals == nil {
            vals = []
        } else {
            for v in vals! {
                if v == nil {
                    throw ClientError.argument(NSLocalizedString(
                        "Response Message header value must not be nil",
                        comment: ""))
                }
                if !isAsciiPrintable(text: v!) {
                    throw ClientError.argument(NSLocalizedString(
                        "Response Message header value must be ASCII-printable",
                        comment: ""))
                }
            }
        }

        let lowerName: String = (name?.lowercased())!
        
        if headers[lowerName] != nil {
            var newValues: [String?]? = headers[lowerName]!
            for v in vals! {
                newValues!.append(v!)
            }
            headers[lowerName] = newValues
        } else {
            headers[lowerName] = []
            for v in vals! {
                headers[lowerName]!!.append(v!)
            }
        }
    }
    
    
    private func isAsciiPrintable(text: String) -> Bool {
        for i in text {
            let someString = String(i)
            let someInt = someString.utf8.first
            if (someInt < 32) || (someInt > 126) {
                return false
            }
        }
        return true
    }
    
    /**
     * A convenience method for setting the "Content-Type" header to the 
     * specified value.
     *
     * - parameter contentType: The content type to use, cannot be `nil`. 
     *                          This is used literally.
     *
     * - throws: `ClientError.argument` if `contentType` is `nil`
     */
    public func setContentType(contentType: String?) throws {
        if contentType == nil {
            throw ClientError.argument(NSLocalizedString(
                "Response Message content type must not be nil", comment: ""))
        }
        headers["Content-Type"] = Array(arrayLiteral: contentType!)
    }
    
    /**
     * Set message URL. The URL is not checked for validity.
     *
     * - parameter url: URL of the HTTP request.
     */
    public func setUrl(url: String) {
        self.url = url
    }
    
    /**
     * Set message body.
     *
     * - parameter body: Body of the HTTP response message, should not be `nil`.
     */
    public func setBody(body: Foundation.Data) {
        self.body = body
    }
    
    /**
     * Set message body using a String. The encoding for the body is UTF-8.
     *
     * - parameter body: body of the HTTP message, should not be `nil`
     */
    public func setBody(body: String) {
        self.body = body.data(using: .utf8)
    }
    
    /**
     * Set http response status code.
     *
     * - parameter statusCode: Status code for HTTP response.
     */
    public func setStatusCode(statusCode: StatusCode) {
        self.statusCode = statusCode
    }
    
    /**
     * Sets the Id of *RequestMessage* for that this response was created.
     *
     * - parameter requestId: Http request message id.
     */
    public func setRequestId(requestId: String) {
        self.requestId = requestId
    }
    
    /**
     * Initialize this builder from a JSON object
     *
     * - parameter jsonObject: The JSON object to read from.
     *
     * - returns: extracted item as JSON structure
     *
     * - throws: `ClientError.state` if `payload` or `requestId` value is `nil`
     */
    override public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String,Any> {
        
        let jsonDict: Dictionary = try super.fromJSON(data: data)
        var statusCode : StatusCode? = nil
        var url: String? = nil
        var body: Foundation.Data? = nil
        var requestId: String? = nil
        var headers: Dictionary<String,Any>? = nil
        for (key, value) in jsonDict {
            if key == "payload" {
                let payload: Dictionary? = (value as! Dictionary<String, Any>)
                if payload == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Response Message payload must not be nil",
                        comment: ""))
                }
                for (key1, value1) in payload! {
                    if key1 == "statusCode" {
                        statusCode = StatusCode.valueOf(code: value1 as! Int)
                    }
                    if key1 == "url" {
                        url = (value1 as! String)
                    }
                    if key1 == "requestId" {
                        requestId = (value1 as! String)
                        if requestId == nil {
                            throw ClientError.state(NSLocalizedString(
                                "Response Message request ID must not be nil",
                                comment: ""))
                        }
                    }
                    if key1 == "body" {
                        let bodyStr: String =  value1 as! String
                        body = Foundation.Data(base64Encoded: bodyStr,
                            options: Foundation.Data.Base64DecodingOptions(rawValue: 0))!
                    }
                    if key1 == "headers" {
                        headers = (value1 as! Dictionary<String,Any>)
                    }
                }
            }
        }
        setStatusCode(statusCode: statusCode!)
        setUrl(url: url!)
        setBody(body: body!)
        setRequestId(requestId: requestId!)
        
        if let headers = headers {
            for (key, value) in headers {
                let vals = value as! Array<String>
                var vaList: [String?]? = []
                for val in vals {
                    vaList!.append(val)
                }
                try addHeader(name: key, values: vaList)
            }
        }
        
        return jsonDict
    }
    
    /**
     * Creates new instance of a *ResponseMessage* using values from this 
     * *ResponseMessageBuilder*.
     *
     * - returns: Instance of *ResponseMessage*
     *
     */
    override public func build() throws -> ResponseMessage {
        // TODO: What is thrown ?
        return try ResponseMessage(builder: self)
    }
}

/**
 * Implementation of protocol `Equatable`
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: ResponseMessage, rhs: ResponseMessage) -> Bool {
    if lhs === rhs {
        return true
    }
    if !((lhs as Message) == (rhs as Message)) {
        return false
    }
    if !(lhs.body! == rhs.body!) {
        return false
    }
    
    var he: [String : [String]] = [ : ]
    for key in lhs.headers!.keys {
        var hev: [String] = []
        for v in lhs.headers![key]!! {
            hev.append(v!)
        }
        he[key] = hev
    }
    var he2: [String : [String]] = [ : ]
    for key in rhs.headers!.keys {
        var hev: [String] = []
        for v in rhs.headers![key]!! {
            hev.append(v!)
        }
        he2[key] = hev
    }
    // TODO: Need a way to compare Dictionary's
    if !(NSDictionary(dictionary: he).isEqual(to: he2)) {
        return false
    }
    
    if !(lhs.statusCode == rhs.statusCode) {
        return false
    }
    if !(lhs.url == rhs.url) {
        return false
    }
    return (lhs.requestId == rhs.requestId)
}

/**
 * ResponseMessage extends `Message` class. It stores HTTP headers, status code
 * and URL and body. The body is in byte array. Default encoding for string body
 * is UTF-8.
 * This class is immutable.
 */
public class ResponseMessage: Message {
    
    /** HTTP response message header. Header is a name and list of values pair. 
     */
    fileprivate var headers: [String : [String?]?]?
    /** Status code of the HTTP response message */
    fileprivate var statusCode: StatusCode?
    /** URL of the HTTP response message */
    fileprivate var url: String?
    /** Body of the HTTP response message */
    fileprivate var body: Foundation.Data?
    /** Id of the request message. Can be used for pairing request-response */
    fileprivate var requestId: String?
    
    /**
     * `ResponseMessage` constructor takes a `ResponseMessageBuilder` and sets
     * values to each field. If the value is `nil`, it sets a default value.
     *
     * - parameter builder: generic type of message builder
     *
     * - throws: `ClientError.state` if `builder.source` is `nil` or is the
     *                                empty string.
     */
    fileprivate init(builder: ResponseMessageBuilder) throws {
        try super.init(builder: builder)

        self.requestId = builder.requestId ?? nil
        self.headers = builder.headers
        self.url = builder.url ?? ""
        self.body = builder.body ??
            Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0)
        self.statusCode = builder.statusCode ?? .OK
    }
    
    /**
     * Get Http response message headers.
     *
     * - returns a dictionary representing headers, never `nil`
     */
    public func getHeaders() -> [String : [String?]?] {
        return self.headers!
    }
    
    /**
     * Get Http response message header values for given header name.
     *
     * - parameter name: header name, should not be `nil`
     *
     * - returns: an array of header values, never `nil`
     */
    public func getHeaderValues(name: String?) -> [String?]? {
        if (name != nil) {
            return headers![(name?.lowercased())!]!
        }
        return []
    }
    
    /**
     * Get response message header value at given index.
     *
     * - parameter name: header name, should not be `nil`.
     *
     * - parameter index: index in array of header values.
     *
     * - returns: header value, may return `nil` if header name does not exist 
     *            or index is out of range.
     */
    public func getHeaderValue(name: String?, index: Int) -> String? {
        if (name == nil) {
            return nil
        }
        
        var values: [String?]? = headers![(name?.lowercased())!]!
        
        if (values != nil) && (values!.count > index) {
            return values![index]
        } else {
            return nil
        }
    }
    
    /**
     * Get first value from response message header.
     *
     * - parameter name: header name, should not be `nil`
     *
     * - returns: first header value, may return `nil` if the header name does 
     *            not exist or no values were set for given header name.
     */
    public func getHeaderValue(name: String?) -> String? {
        if (name == nil) {
            return nil
        }
        
        var values: [String?]? = headers![(name?.lowercased())!]!
        
        if (values != nil) && (values!.count > 0) {
            return values![0]
        } else {
            return nil
        }
    }
    
    /**
     * Get request message URL.
     *
     * - returns: URL, never `nil`
     */
    public func getURL() -> String {
        return self.url!
    }
    
    /**
     * Get HTTP response message body. If message body contains string value, 
     * default encoding is UTF-8.
     *
     * - returns: body, never `nil`
     */
    public func getBody() -> Foundation.Data {
        return self.body!
    }
    
    /**
     * Get HTTP response message body in String. Returns the inner 
     * representation of message body to String.
     *
     * - returns: Http message body as String, never `nil`
     */
    public func getBodyString()-> String {
        return String(data: self.body!, encoding: .utf8)!
    }
    
    /**
     * Get Http response message status code.
     *
     * - returns: statusCode, never `nil`
     */
    func getStatusCode() -> StatusCode {
        return self.statusCode!
    }
    
    /**
     * Get ID of the `RequestMessage`. `RequestMessage` Id can be used for
     * pairing request-response.
     *
     * - returns: Request id, never `nil`
     */
    public func getRequestId() -> String {
        return self.requestId!
    }
    
    /**
     * Get message type.
     *
     * - returns: type, never `nil`.
     */
    override public func getType() -> MsgType {
        return .RESPONSE
    }
    
    /**
     * Exports data from `RequestMessage` to `String` using JSON interpretation
     * of the message.
     *
     * - returns: JSON interpretation of the message as String.
     */
    override public func toString() throws -> String {
        return toJsonString()
    }
    
    /**
     * Exports response message as a JSON string.
     *
     * - returns: response message as JSON string
     */
    override public func toJsonString() -> String {
        var result: String
        var addings: String = ""
        
        addings = addings + "\"payload\":{"
        addings = addings + "\"statusCode\":"
        addings = addings + String(statusCode!.getCode())
        addings = addings + ","         // <----
        addings = addings + "\"url\":\"" + url! + "\","
        addings = addings + "\"requestId\":\"" + requestId! + "\","
        if headers != nil {
            addings = addings + "\"headers\":{"
            let headerKeys: [String] = Array(self.headers!.keys)
            var z: Int = 0
            for k in headerKeys {
                addings = addings + "\"" + k + "\":["
                var zz: Int = 0
                if headers![k] != nil {
                    for s in (headers![k])!! {
                        addings = addings + "\"" + s! + "\""
                        zz += 1
                        if zz < headers![k]!!.count {
                            addings = addings + ","
                        }
                    }
                    addings = addings + "]"
                }
                z += 1
                if z < headers!.count {
                    addings = addings + ","
                }
            }
        }
        addings = addings + "},"
        addings = addings + "\"body\":\"" + body!.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0)) + "\""
        addings = addings + "}"
        
        result = toJsonString(expand: false, additionalItems: addings)
        
        return result
    }
    
    /**
     * Implementation of protocol `Hashable`
     *
     * - returns: the hash value
     */
    override public var hashValue : Int {
        var result: Int = super.hashValue
        var he: [String : [String]] = [ : ]
        for key in headers!.keys {
            var hev: [String] = []
            for v in headers![key]!! {
                hev.append(v!)
            }
            he[key] = hev
        }
        // TODO: Find a way to create a Dictionary from a Dictionary
        result = 31 &* result &+ NSDictionary(dictionary: he).hashValue
        result = 31 &* result &+ (statusCode?.hashValue)!
        result = 31 &* result &+ (url?.hashValue)!
        result = 31 &* result &+ (requestId?.hashValue)!
        // DJB hash function for arrays
        let bodyArray: [UInt8] = [UInt8](repeating: 0, count: body!.count)
        body!.copyBytes(to: UnsafeMutablePointer(mutating: bodyArray), count: body!.count)
        result = 31 &* result &+ bodyArray.reduce(5381) { ($0 << 5) &+ $0 &+ Int($1) }
        return result;
    }
}
