/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package com.oracle.iot.client.impl.device

import Foundation

// TODO: When extending StorageObjectBase

fileprivate var delegateMap: Dictionary<AdvancedStorageObject, StorageObjectImpl> =
    Dictionary<AdvancedStorageObject, StorageObjectImpl>()

fileprivate var callbackQueue = DispatchQueue.global(qos: .utility)

fileprivate class SOIProgressCallback: ProgressCallback {
    
    override init() {
        super.init()
    }
    
    override public func progress(progress:Progress) {
        let delegate: AdvancedStorageObject = progress.getStorageObject()
        // TODO: Need to synchronize
        guard let storageObject:StorageObjectImpl =
            delegateMap[delegate] else {
                return
        }
        
        let oldStatus:SyncStatus = storageObject.getSyncStatus()
        let state:ProgressState = progress.getState()
        
        switch state {
        case .completed:
            storageObject.setSyncStatus(syncStatus: SyncStatus.inSync)
            delegateMap.removeValue(forKey: delegate)
            break
        case .canceled, .failed:
            storageObject.setSyncStatus(syncStatus: SyncStatus.syncFailed)
            delegateMap.removeValue(forKey: delegate)
            break
        case .inProgress, .initiated, .queued, .unknown:
            break
        }
        if oldStatus != storageObject.getSyncStatus() {
            if storageObject.getInputPath() != nil {
                storageObject.handleStateChange()
            }
            if let cb:SyncCallback = storageObject.getSyncCallback() {
                // TODO: define callbackQueue in StorageObjectBase
                // as "dispatcher"
                callbackQueue.async(execute: {
                    var syncEvents:[SyncEvent] = storageObject.getSyncEvents()
                    while !syncEvents.isEmpty {
                        let syncEvent:SyncEvent = (syncEvents.removeFirst())
                        cb.onSync(event:syncEvent)
                    }
                })
            }
        }
    }
}

fileprivate var progressCallback = SOIProgressCallback()

/**
 * StorageObjectImpl
 */
public class StorageObjectImpl : StorageObjectBase {

    // The name of an attribute or a field
    private var nameForSyncEvent:String?
    private var deviceForSyncEvent:VirtualDeviceImpl?
    private var storageDispatcher:StorageDispatcherImpl!

    public init(dcd: AdvancedDirectlyConnectedDevice,
                delegate: AdvancedStorageObject) {
        storageDispatcher = StorageDispatcherImpl.getStorageDispatcher(dcd:dcd) as! StorageDispatcherImpl
        super.init(delegate:delegate)
    }
    
    /**
     * Notify the library to sync content with the storage cloud.
     * The direction of the sync, upload or download, depends on whether
     * the `setInputPath` input path or the
     * `setOutputPath` output path has been set.
     * This method does not start any uploads or downloads if `getSyncStatus`
     * is other than `SyncStatus.notInSync.
     *
     * This is a non-blocking call. The sync is performed on a separate thread.
     * The status of the sync can be monitored by setting a `SyncCallback`
     * on this `StorageObject`.
     *
     * If the input path cannot be read, or the output path cannot be
     * written, an `ClientError.file` is thrown. Any I/O exceptions
     * during the background sync are reported through the
     * `AbstractVirtualDevice.ErrorCallback` error callback
     * of the virtual device.
     *
     * - throws:
     *       - `ClientError.file` if the the input path cannot be read,
     *                          or the output path cannot be written
     *       - `ClientError.state` if both input path and output path are `nil`
     */
    public override func sync() throws {
        if getSyncStatus() == SyncStatus.notInSync {
            try super.sync()
            if !storageDispatcher.isClosed() {
                let delegate:AdvancedStorageObject = getDelegate()
                delegateMap[delegate] = self
                storageDispatcher.setProgressCallback(callback: progressCallback)
                try storageDispatcher.queue(storageObject: delegate)
            }
        } else {
            try super.sync()
        }
    }

    override func handleStateChange() {
        guard let dfse = deviceForSyncEvent else {
            return
        }
        // TODO: Determine throws chain. Anything that involves getting
        // a MessageDispatcher can throw, unlike SE
        // Just print an error here for now
        do {
            try dfse.handleStorageObjectStateChange(storageObject:self)
        } catch {
            print(error)
        }
    }
    
    override func createSyncEvent() -> SyncEvent {
        return SyncEventImpl(source:self, virtualDevice:deviceForSyncEvent!,
                             name:nameForSyncEvent!)
    }
    
    func setSyncEventInfo(virtualDevice:VirtualDeviceImpl,
            nameForSyncEvent:String ) {
        self.deviceForSyncEvent = virtualDevice
        self.nameForSyncEvent = nameForSyncEvent
    }
}
