/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

/**
 * VirtualDevice for a device-client adds methods to handle write-only and
 * executable actions.
 *
 */
public protocol VirtualDevice: AbstractVirtualDevice {
    
    /**
     * Set callback for handling an action
     *
     * - parameter actionName: The action name
     *
     * - parameter callback: The function to invoke
     *
     * - throws: `ClientError.argument` if `actionName` is not an action on 
     *           the model.
     */
    func setOnCall(actionName: String,
                callback: @escaping (VirtualDevice, AnyObject?) -> Void) throws
    
    /**
     * Create a custom data object for this VirtualDevice. The custom data
     * object will be created with the given format URN.
     *
     * - parameter format: the custom data message format URN.
     *
     * - returns: a `Data` object
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `format` or if the format type is not
     *           `DeviceModelFormatType.DATA`
     */
    func createData(format: String) throws -> Data
    
    /**
     * Create an `Alert` for this VirtualDevice. The alert will be created with
     * the given format URN.
     *
     * - parameter format: the alert format URN.
     *
     * - returns: an `Alert` object
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *            obtained from `alertUrn` or if the format is not
     *            `DeviceModelFormatType.ALERT`
     */
    func createAlert(format: String) throws -> Alert
    /**
     * Offer to set the value of an attribute. The attribute
     * value is set depending upon any policy that may have been
     * configured for the attribute. If there is no policy for the
     * given attribute, offer behaves as if the
     * `AbstractVirtualDevice.set(attributeName:attributeValue)` method were called.
     * The value is validated according to the constraints in the device model.
     * If the value is not valid, a `ClientError.argument` exception is raised.
     *
     * - parameter attributeName: the name of an attribute from the device type model
     *
     * - parameter attributeValue: the value to set
     *
     * - returns: this VirtualDevice instance
     *
     * - throws: `ClientError.argument` if the value is not valid,
     *                                  the attribute is not in the model,
     *                                  `attributeName` is `nil`, or if the 
     *                                  attribute is read-only
     */
    func offer(attributeName:String, attributeValue:AnyObject) throws -> VirtualDevice
}

class VirtualDeviceImpl: DevicePolicyManagerChangeListener, VirtualDevice, DeviceAnalog {
    
    var base: VirtualDeviceBase?
    
    // used from AlertImpl
    var msgDirectlyConnectedDevice: AdvancedDirectlyConnectedDevice?
    var attributeMap: [String : VirtualDeviceAttribute] = [ : ]
    var devicePolicyManager:DevicePolicyManager?
    
    private let pipelineDataCacheQueue:DispatchQueue =
        DispatchQueue(label:"pipelineDataCache queue", qos: .utility)
    // Data pertaining to this virtual device and its attributes for computing policies.
    // The key is attribute name (or nil for the device model policies), value is a
    // list. Each element in the list corresponds to a function in the pipeline for
    // the attribute, and the map is used by the function to hold data between calls.
    // Note that there is a 1:1 correspondence between the configuration
    // data for the attribute, and the pipeline data for the attribute.
    private var pipelineDataCache:Dictionary<String,[Dictionary<String,AnyObject>]> =
            Dictionary<String,[Dictionary<String,AnyObject>]>()
    private let offerQueue:DispatchQueue = DispatchQueue(label:"offer queue", qos:.utility)
    // TODO: Why was this global ?
    //fileprivate var scheduledPoliciesQueue:DispatchQueue =
    //    DispatchQueue.global(qos:.userInitiated)
    fileprivate var scheduledPoliciesQueue:DispatchQueue =
        DispatchQueue(label:"scheduledPoliciesQueue", qos:.userInitiated)
    /* no slide
    // [window : ScheduledPolicyData]
    var scheduledPolicies:[Int64:ScheduledPolicyData] = [Int64:ScheduledPolicyData]()
    */
    // Window and slide are the key.
    // { {window,slide} : ScheduledPolicyData }
    var scheduledPolicies:[ScheduledPolicyData.Key: ScheduledPolicyData] =
        [ScheduledPolicyData.Key:ScheduledPolicyData]()
    
    var timedPolicyThread:TimedPolicyThread!
    private var ERROR_CALLBACK_BRIDGE: ErrorCallbackBridge = ErrorCallbackBridge()
    private var DELIVERY_CALLBACK: OnErrorDeliverCallback = OnErrorDeliverCallback()
    
    var messsageDispatcher: MessageDispatcher?
    var requestDispatcher: RequestDispatcher?
    
    var adapter: VDIAdapter!
    var requestHandler: VDIRequestHandler?
    
    var endpointId: String?

    /**
     * - throws: `ClientError.network` if a `MessageDispatcher` cannot be
     *           obtained.
     */
    init(msgDirectlyConnectedDevice: AdvancedDirectlyConnectedDevice,
                endpointId: String, deviceModel: DeviceModel) throws {
        super.init()
        self.endpointId = endpointId
        self.msgDirectlyConnectedDevice = msgDirectlyConnectedDevice
        self.adapter = VDIAdapter(impl: self)
        self.base = VirtualDeviceBase(adapter: self.adapter,
                                      endpointId: endpointId,
                                      deviceModel: deviceModel)
        self.attributeMap = VirtualDeviceImpl.createAttributeMap(
            virtualDevice: self, deviceModel: deviceModel)
        
        timedPolicyThread = TimedPolicyThread(virtualDevice:self)
        
        self.devicePolicyManager =
            DevicePolicyManager.getDevicePolicyManager(dcd:self.msgDirectlyConnectedDevice!)
    
        // TODO: deal with async getPolicy
        // set up device model policies
        /*
        if let policy:DevicePolicy =
            try self.devicePolicyManager?.getPolicy(deviceModelURN:deviceModel.getURN(),
                                                    deviceId:getEndpointId()) {
            self.policyAdded(devicePolicy:policy)
        }
        */
        /* TODO when getPolicy is async
        try self.devicePolicyManager!.getPolicy(deviceModelURN:deviceModel.getURN(),
                                                deviceId:getEndpointId(),
                                                callback:{ (policy,error) in
            if error != nil {
                log(type:OSLogType.fault,
                    message: "VirtualDeviceImpl %@ policy error for model %@: %@",
                    args: self.getEndpointId(), deviceModel.getURN(), "\(error!)")
                return
            }
            if let policy = policy {
                self.policyAdded(policy:policy)
            }
            // Include following attribute processing
        })
        */

        self.devicePolicyManager?.addChangeListener(changeListener:self)
        self.messsageDispatcher = MessageDispatcher.getMessageDispatcher(
                        msgDirectlyConnectedDevice: msgDirectlyConnectedDevice)
        self.requestDispatcher = RequestDispatcher.getInstance()
        self.requestHandler = VDIRequestHandler(impl: self)
        // handler for all attribute and action requests
        try requestDispatcher!.registerRequestHandler(endpointId: endpointId,
            path: "deviceModels/" + deviceModel.getURN(), handler: requestHandler)
        
        try ERROR_CALLBACK_BRIDGE.add(virtualDevice: self)
        self.messsageDispatcher!.setOnError(callback: ERROR_CALLBACK_BRIDGE)
        self.messsageDispatcher!.setOnDelivery(callback: DELIVERY_CALLBACK)
    }
    
    private static func createAttributeMap(virtualDevice: VirtualDeviceImpl,
            deviceModel: DeviceModel) -> [String : VirtualDeviceAttribute] {
        
        var map: [String : VirtualDeviceAttribute] = [ : ]
        
        for attribute in (deviceModel as! DeviceModelImpl).getDeviceModelAttributes().values {
            let vda: VirtualDeviceAttribute = VirtualDeviceAttribute(
                                virtualDevice: virtualDevice, model: attribute)
            map[attribute.getName()!] = vda
            if let alias: String = attribute.getAlias(), !alias.isEmpty {
                map[alias] = vda
            }
        }
        return map;
    }
    
    /**
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `format` or if the format type is not
     *           `DeviceModelFormatType.DATA`
     */
    func createData(format: String) throws -> Data {
        return try DataImpl(virtualDevice: self, formatUrn: format)
    }
    
    func getDeviceModel() -> DeviceModel {
        return base!.getDeviceModel()
    }
    
    func getEndpointId() -> String {
        return base!.getEndpointId()
    }
    
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func get(attributeName: String) throws -> AnyObject? {
        let attribute: VirtualDeviceAttribute =
                                try adapter.getAttribute(name: attributeName)
        let val:AnyObject? = attribute.get()
        return val
    }
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func getLastKnown(attributeName: String) throws -> AnyObject? {
        let attribute: VirtualDeviceAttribute =
                                try adapter.getAttribute(name: attributeName)
        let val: AnyObject? = attribute.getLastKnown()
        return val
    }
    /**
     * - throws: `ClientError.argument` if attempting to set a read-only
     *           attribute, or setting an invalid value
     */
    @discardableResult
    func set(attributeName: String,
             attributeValue: AnyObject) throws -> VirtualDevice {
        try base!.set(attributeName: attributeName, value: attributeValue)
        return self
    }
    
    func update() -> VirtualDevice {
        base!.update()
        return self
    }
    /*
     * - throws: `ClientError.argument` if an attributes is nil or is the empty
     *            string or if an attribute type is string, if `newValue` is 
     *            nil or the empty string.
     */
    func finish() throws {
        try base!.finish()
    }

    func offer(attributeName:String, attributeValue:AnyObject) throws -> VirtualDevice {
        
        let attribute:VirtualDeviceAttribute = try adapter.getAttribute(name:attributeName)
        
        if !attribute.isSettable() {
            throw ClientError.argument(String(format:NSLocalizedString(
                "attempt to modify read-only attribute '%@'", comment: ""), attributeName))
        }
        
        let _offer:(DevicePolicy?)->() = { policy in
            do {
                if policy == nil {
                    try self.set(attributeName:attributeName,
                                 attributeValue:attributeValue)
                    return
                }
                
                let pipeline:[DevicePolicy.Function]? = policy!.getPipeline(attributeName:attributeName)
                if pipeline == nil || pipeline!.isEmpty {
                    try self.set(attributeName:attributeName,
                                 attributeValue:attributeValue)
                    return
                }

                var pipelineData:[Dictionary<String,AnyObject>] =
                    try self.getPipelineData(attribute:attributeName)
                
                let policyValue:AnyObject? =
                    self.offer0(attribute:attribute.getDeviceModelAttribute()!,
                                attributeValue:attributeValue, pipeline:pipeline!,
                                pipelineData:&pipelineData)
                // add back pipelineData
                self.putPipelineData(attribute: attributeName,
                                     pipelineData: pipelineData)
                if let policyValue = policyValue {
                    if self.base!.isUpdateMode() {
                        try self.set(attributeName:attributeName, attributeValue:policyValue)
                    } else {
                        //
                        // Handle calling offer outside of an update when there
                        // are computed metrics involved.
                        //
                        // Call updateFields ito ensure the computed metrics get run,
                        // and will put this attribute and computed attributes into
                        // one data message.
                        //
                        var updatedAttributes:[Pair<VirtualDeviceAttribute, AnyObject>] =
                            [Pair<VirtualDeviceAttribute, AnyObject>]()
                        
                        updatedAttributes.append(Pair<VirtualDeviceAttribute, AnyObject>(
                            key:attribute, value:policyValue))
                        
                        try self.adapter.updateFields(updatedAttributes:&updatedAttributes)
                    }
                }
            } catch {
                log(type:OSLogType.fault,
                    message:"VirtualDeviceImpl.offer %@ error for %@: %@",
                    args: self.getEndpointId(), self.getDeviceModel().getURN(),
                    "\(error)")
            }
        }
        
        let devicePolicy:DevicePolicy? = try devicePolicyManager?.getPolicy(
                                        deviceModelURN:getDeviceModel().getURN(),
                                        deviceId:getEndpointId()
            /*, callback: { (policy,error) in
                if error != nil {
                    log(type:OSLogType.fault,
                        message:"VirtualDeviceImpl.offer %@ policy error for %@: %@",
                        args: self.getEndpointId(), self.getDeviceModel().getURN(),
                        "\(error!)")
                    return
                }
                _offer(policy)
        }*/)
        _offer(devicePolicy)
        return self
    }
    
    // This is the main logic for handling a policy pipeline.
    // Returns true if attribute is set.
    private func offer0(attribute:DeviceModelAttribute?,
                        attributeValue:AnyObject?,
                        pipeline:[DevicePolicy.Function]?,
                        pipelineData: inout [Dictionary<String,AnyObject>]) -> AnyObject? {
        
        // TODO: Looks like a bug in SE called from one of the processExpiredFunction
        // functions attribute is literally passed as nil
        let attributeName:String = attribute?.getName() ?? ""
        var policyValue:AnyObject? = attributeValue
        
        guard pipeline != nil && !pipeline!.isEmpty else {
            return policyValue
        }
        
        return offerQueue.sync(execute: {
            
            // InProcessValue helps resolve $(<attribute-name>) in formulas
            // and conditions
            DeviceFunction.putInProcessValue(endpointId:self.getEndpointId(),
                                             deviceModelURN:self.getDeviceModel().getURN(),
                                             attribute:attributeName,
                                             value:policyValue)
            
            for index in pipeline!.startIndex..<pipeline!.endIndex {
                
                let function:DevicePolicy.Function = pipeline![index]
                var functionData:Dictionary<String,AnyObject>? = nil
                if index < pipelineData.count {
                    functionData = pipelineData[index]
                } else {
                    functionData = Dictionary<String,AnyObject>();
                    pipelineData.append(functionData!)
                }
                
                let key:String = function.getId()
                let parameters:Dictionary<String,Any> =
                    function.getParameters()
                
                let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId:key)
                
                if deviceFunction == nil {
                    continue
                }

                if deviceFunction!.apply(deviceAnalog:self,
                                        attribute:attributeName,
                                        configuration:parameters,
                                        data:&functionData!,
                                        value:policyValue) {
                    // Add back
                    pipelineData[index] = functionData!
                    if let valueFromPolicy:Any = deviceFunction!.get(
                            deviceAnalog:self, attribute:attributeName,
                            configuration:parameters,
                            data:&functionData!) {
                        
                        // functionData is modified add it back
                        pipelineData[index] = functionData!
                        
                        policyValue = cast(attributeType:attribute!.getType(),
                                                 newValue:valueFromPolicy) as AnyObject
                        DeviceFunction.putInProcessValue(endpointId:getEndpointId(),
                            deviceModelURN:getDeviceModel().getURN(),
                            attribute:attributeName, value:policyValue!)
                    } else {
                        // FINEST
                        log(type:OSLogType.fault,
                            message:"%@ got nil value from policy %@",
                            args: attributeName,
                                  deviceFunction!.getDetails(configuration: parameters))
                        return nil
                    }
                } else {
                    // functionData is modified add it back
                    pipelineData[index] = functionData!
                    
                    // FINE
                    if deviceFunction!.getId().hasPrefix("filter") {
                        log(type:OSLogType.debug,
                            message:"VirtualDevice: %@ : offer '%@' = 'rejected by policy ' %@",
                            args:getEndpointId(), attributeName, policyValue as! CVarArg,
                            deviceFunction!.getDetails(configuration: parameters))
                    }
                    return nil
                }
            }
            return policyValue
        })
    }
    
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func setOnChange(attributeName: String,
                     callback: @escaping (ChangeEvent)->Void) throws {
        let attribute: VirtualDeviceAttribute =
                                try adapter.getAttribute(name: attributeName)
        attribute.setOnChange(callback: callback)
    }
    
    func setOnChange(callback: @escaping (ChangeEvent)->Void) {
        base!.setOnChange(callback: callback)
    }
    
    func setOnError(callback: @escaping (ErrorEvent)->Void) {
        base!.setOnError(callback: callback)
    }
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func setOnError(attributeName: String,
                    callback: @escaping (ErrorEvent)->Void) throws {
        let attribute: VirtualDeviceAttribute =
                                try adapter.getAttribute(name: attributeName)
        attribute.setOnError(callback: callback)
    }
    
    var actionMap: [String : ((VirtualDevice, AnyObject?) -> Void)?]?
    let actionMapLock:DispatchQueue = DispatchQueue(label:"VirtualDevice.actionMapQueue",
                                                    qos:.utility)
    
    static func getDeviceModelAction(deviceModel: DeviceModel?,
                                     actionName: String) -> DeviceModelAction? {
        if (deviceModel != nil) {
            var actions: [String : DeviceModelAction] =
                    (deviceModel as! DeviceModelImpl).getDeviceModelActions()
            if actions.keys.count > 0 {
                let act: DeviceModelAction? = actions[actionName]!
                if act != nil {
                    return act!
                }
                for action in actions.values {
                    if actionName == action.getAlias() {
                        return action
                    }
                }
            }
        }
        return nil
    }
    
    /**
     * Set callback for handling an action
     *
     * - parameter actionName: The action name
     *
     * - parameter callback: The function to invoke
     *
     * - throws: `ClientError.argument` if `actionName` is not an action on the
     *            model.
     */
    func setOnCall(actionName: String,
            callback: @escaping (VirtualDevice, AnyObject?) -> Void) throws {

        try self.actionMapLock.sync() {
            if actionMap == nil {
                actionMap = [ : ]
            }
            
            var actions: [String : DeviceModelAction] =
                (base!.getDeviceModel() as! DeviceModelImpl).getDeviceModelActions()
            var deviceModelAction: DeviceModelAction? = actions[actionName]
            if (deviceModelAction == nil) {
                for action in actions.values {
                    if actionName == action.getAlias() {
                        deviceModelAction = action
                        break
                    }
                }
            }
            if deviceModelAction == nil {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "Action %@ not found in model", comment: ""), actionName))
            }
            
            actionMap![deviceModelAction!.getName()] = callback
            if deviceModelAction!.getAlias() != nil {
                actionMap![deviceModelAction!.getAlias()!] = callback
            }
        }
    }
    
    /**
     * - throws: `ClientError.argument` if a device model format cannot be
     *           obtained from `format` or if the format is not
     *           `DeviceModelFormatType.ALERT`
     */
    func createAlert(format: String) throws -> Alert {
        return try AlertImpl(virtualDevice: self, alertUrn: format)
    }
    
    // called from VirtualDeviceAttribute
    /**
     * Call the `onChange` callback for `virtualDeviceAttribute`
     *
     * - parameter virtualDeviceAttribute: the changed attribute
     *
     * - parameter newValue: the new attribute value
     *
     * - throws: 
     *    - `ClientError.network` if a `MessageDispatcher` cannot be
     *      obtained.
     *    - `ClientError.argument` if `DataMessageBuilder.builder.source`
     *      or `DataMessageBuilder.builder.format` is `nil` or is the
     *      empty string.
     */
    func processOnChange(virtualDeviceAttribute: VirtualDeviceAttribute,
                         newValue: AnyObject) throws {
        
        let builder: DataMessageBuilder = DataMessageBuilder()
        var storageObject:StorageObjectImpl? = nil
        do {
            try processOnChange(builder: builder,
                                attribute: virtualDeviceAttribute,
                                newValue: newValue,
                                storageObject: &storageObject)
        } catch {
            return
        }

        let dataMessage:DataMessage = try builder.build()
        do {
            try queueMessage(message:dataMessage, storageObject:storageObject)
        } catch {
            // MessageDispatcher queue is full.
            var attributes:Set<VirtualDeviceAttribute> = Set<VirtualDeviceAttribute>()
            attributes.update(with:virtualDeviceAttribute)
            notifyException(attributes:attributes, error:error as! ClientError)
        }
    }
    
    /**
     * - throws: 
     *     - `ClientError.network` if a `MessageDispatcher` cannot be
     *       obtained.
     *     - `ClientError.argument` if `DataMessageBuilder.builder.source`
     *        or `DataMessageBuilder.builder.format` is `nil` or is the
     *        empty string.
     */
    func processOnChange(updatedAttributes: [VirtualDeviceAttribute : AnyObject]) throws {
        if updatedAttributes.count == 0 {
            return
        }
        
        let builder: DataMessageBuilder = DataMessageBuilder()
        var storageObject:StorageObjectImpl? = nil
        for (key, value) in updatedAttributes {
            do {
                try processOnChange(builder: builder, attribute: key,
                                    newValue: value,
                                    storageObject: &storageObject)
            } catch {
                return
            }
        }
        builder.setReliability(reliability: Message.Reliability.guaranteed_delivery)
        let dataMessage:DataMessage = try builder.build()
        let messageDispatcher:MessageDispatcherImpl? =
            MessageDispatcher.getMessageDispatcher(
                msgDirectlyConnectedDevice: msgDirectlyConnectedDevice!) as? MessageDispatcherImpl
        if let so = storageObject  {
            messageDispatcher?.addStorageObjectDependency(storageObject: so,
                                                         clientMsgId: dataMessage.getClientId())
        }
        try messageDispatcher?.queue(message: dataMessage)
    }
    
    // called from updateFields
    func processOnChange(updatedAttributes:[Pair<VirtualDeviceAttribute, AnyObject>]) throws {
        
        if updatedAttributes.isEmpty {
            return
        }
        
        var keySet:Set<VirtualDeviceAttribute> = Set<VirtualDeviceAttribute>()
        
        let builder:DataMessageBuilder = DataMessageBuilder()
        var storageObject:StorageObjectImpl? = nil
        
        for entry in updatedAttributes {
            
            let attribute:VirtualDeviceAttribute = entry.getKey()
            keySet.update(with:attribute)
            let newValue:AnyObject = entry.getValue()
            
            do {
                try processOnChange(builder:builder, attribute:attribute,
                                    newValue:newValue,
                                    storageObject:&storageObject)
            } catch {
                print(error)
                return
            }
        }
        let dataMessage:DataMessage = try builder.build()
        
        do {
            try queueMessage(message:dataMessage, storageObject:storageObject)
        } catch {
            // MessageDispatcher queue is full.
            notifyException(attributes: keySet, error: error as! ClientError)
        }
    }
    
    private let eventDispatcher:DispatchQueue = DispatchQueue.global(qos:.utility)
    
    private func notifyException(attributes:Set<VirtualDeviceAttribute>, error:ClientError) {
        
        guard let errorCallback = base!.getOnErrorCallback() else {
            return
        }
        
        var head:NamedValueImpl? = nil
        var tail:NamedValueImpl? = nil
        for attribute in attributes {
            let name:String = attribute.getDeviceModelAttribute()!.getName()!
            let value:AnyObject = attribute.get()!
            let next:NamedValueImpl = NamedValueImpl(name:name, value:value)
            if head != nil {
                tail!.setNext(next:next)
                tail = next
            } else {
                head = next
                tail = head
            }
        }
        
        let msg:String = "\(error)"
        let errorEvent:ErrorEvent = ErrorEventImpl(virtualDevice:self,
                                                   value:head!, message:msg)
        
        eventDispatcher.async() {
            errorCallback(errorEvent)
        }
    }
    
    func putPipelineData(attribute:String, pipelineData:[Dictionary<String,AnyObject>]) {
        pipelineDataCacheQueue.sync() {
            self.pipelineDataCache[attribute] = pipelineData
        }
    }
    
    // TODO: Should this return the removed value ?
    func removePipelineData(attribute:String) {
        pipelineDataCacheQueue.sync() {
            _ = self.pipelineDataCache.removeValue(forKey:attribute)
        }
    }

    // Get the pipeline data for the attribute
    func getPipelineData(attribute:String) throws -> [Dictionary<String,AnyObject>] {
        
        // TODO: support async getPolicy
        guard let devicePolicy:DevicePolicy = try devicePolicyManager?.getPolicy(
            deviceModelURN:getDeviceModel().getURN(),
            deviceId:getEndpointId()) else {
                return [Dictionary<String,AnyObject>]()
        }
        
        guard let pipeline:[DevicePolicy.Function] =
            devicePolicy.getPipeline(attributeName:attribute), !pipeline.isEmpty else {
            return [Dictionary<String,AnyObject>]()
        }
        var pipelineData:[Dictionary<String,AnyObject>]? = nil
        // a nil attribute means a devicemodel policy
        let pipelineKey:String = attribute
        pipelineData = pipelineDataCacheQueue.sync(execute: {
            return self.pipelineDataCache[pipelineKey]
        })
        if pipelineData == nil {
            pipelineData = [Dictionary<String,AnyObject>]()
        }
        
        if pipelineData!.count < pipeline.count {
            for _ in pipelineData!.count..<pipeline.count {
                pipelineData!.append(Dictionary<String,AnyObject>())
            }
        }
        putPipelineData(attribute:pipelineKey, pipelineData:pipelineData!)
        return pipelineData!
    }
    
    // Also called from AlertImpl
    func queueMessage(message:Message, storageObject:StorageObjectImpl?) throws {
        
        let pair:Pair<Message,StorageObject?> =
            Pair<Message,StorageObject?>(key: message, value: storageObject)
        
        var pairs:[Pair<Message,StorageObject?>] =
            [Pair<Message,StorageObject?>]()
        pairs.append(pair)
        
        let deviceModelURN:String = getDeviceModel().getURN()
        
        // TODO: Fix when getPolicy is asynchronous
        //
        // Handling of policies applying to all attributes here...
        //
        if let devicePolicy:DevicePolicy = try devicePolicyManager?.getPolicy(
                    deviceModelURN:deviceModelURN, deviceId:getEndpointId()),
                let pipeline:[DevicePolicy.Function] =
                    devicePolicy.getPipeline(attributeName: DevicePolicy.ALL_ATTRIBUTES()) {

            // Some policies are negated by an alert of a given severity
            // (batching policies, for example)
            var alertSeverity:AlertMessageSeverity? = nil
            if message is AlertMessage {
                let alertMessage:AlertMessage = message as! AlertMessage
                alertSeverity = alertMessage.getSeverity()
            }

            if pipeline.count > 1 {
                // TODO: handle more than one function in a pipeline for all-attributes
                // Print out a warning message about too many function for all-attributes pipeline
                for index in 0..<pipeline.count {
                    
                    let function:DevicePolicy.Function = pipeline[index]
                    let id:String = function.getId()
                    let parameters:[String:Any] = function.getParameters()
                    let deviceFunction:DeviceFunction? =
                        DeviceFunction.getDeviceFunction(functionId: id)
                    
                    if index == 0 {
                        // WARNING
                        log(type:.fault,
                            message: "Only one function allowed for all-attribute pipeline. " +
                                     "Applying: %@",
                            args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                    } else {
                        // WARNING
                        log(type:.fault,
                            message:"Ignoring: %@",
                            args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                    }
                }
            }
            var pipelineData:[Dictionary<String,AnyObject>] =
                try getPipelineData(attribute: DevicePolicy.ALL_ATTRIBUTES())
            
            // TODO: handle more than one function in a pipeline for all-attributes
            // Note: A warning is printed out, several lines above this, if there
            // is more than one function for an all-attributes policy.
            // Leaving original line here, commented-out, for future reference.
            for index in 0..<1 /*pipeline.indices*/ {
                
                let function:DevicePolicy.Function = pipeline[index]
                let id:String = function.getId()
                let parameters:Dictionary<String,Any> = function.getParameters()
                let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId:id)
                if deviceFunction == nil {
                    continue
                }
                
                //if "batchByTime" == deviceFunction!.getId() {
                //continue
                //}
                
                var alertOverridesPolicy:Bool = false
                if alertSeverity != nil {
                    var configuredSeverity:AlertMessageSeverity = AlertMessageSeverity.critical
                    if let criterion:String = parameters["alertSeverity"] as? String {
                        configuredSeverity = AlertMessageSeverity.valueOf(severity:criterion)
                    }
                    alertOverridesPolicy =
                        configuredSeverity.rawValue <= (alertSeverity == nil ?
                            AlertMessageSeverity.critical.rawValue : alertSeverity!.rawValue)
                }
                
                var functionData:Dictionary<String, AnyObject>? = nil
                if index < pipelineData.endIndex {
                    functionData = pipelineData[index]
                } else {
                    functionData = Dictionary<String,AnyObject>()
                }

                if deviceFunction!.apply(deviceAnalog:self,
                                         attribute:nil,
                                         configuration:parameters,
                                         data:&functionData!, value:pair) ||
                                                        alertOverridesPolicy {
                    // Add back functionData
                    pipelineData[index] = functionData!
                    // Add back the pipelineData
                    putPipelineData(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                    pipelineData: pipelineData)
                    
                    // If the policy was scheduled
                    let window:Int64 = DeviceFunction.getWindow(configuration:parameters)
                    if window > 0 {
                        let slide:Int64 = DeviceFunction.getSlide(configuration: parameters,
                                                                  window:window)
                        let key:ScheduledPolicyData.Key = ScheduledPolicyData.Key(window:window,
                                                                                  slide:slide)
                        if let scheduledPolicyData:ScheduledPolicyData =
                                // JCL has "slide" not "key"
                                scheduledPolicies[key] {
                            let timeZero:Int64 = getCurrentTimeMillis()
                            var updatedAttributes:[Pair<VirtualDeviceAttribute, AnyObject>] =
                                [Pair<VirtualDeviceAttribute, AnyObject>]()
                            try scheduledPolicyData.processExpiredFunction(
                                            virtualDevice:self,
                                            updatedAttributes: &updatedAttributes,
                                            timeZero:timeZero/*, no slide
                                            devicePolicy: devicePolicy*/)
                            if !updatedAttributes.isEmpty {
                                try self.adapter.updateFields(updatedAttributes:&updatedAttributes)
                            }
                            return
                        }
                    }
                    
                    let value:[Pair<Message,StorageObject?>] =
                                deviceFunction!.get(deviceAnalog:self,
                                attribute:nil,
                                configuration:parameters,
                                data:&functionData!) as! [Pair<Message,StorageObject?>]
                    // overwites original pairs, copy return from get
                    pairs = [Pair<Message,StorageObject?>](value)
                    // INFO
                    log(type:OSLogType.info,
                        message:"VirtualDevice: %@ dispatching %d messages per policy '%@'",
                        args:getEndpointId(), pairs.count,
                            deviceFunction!.getDetails(configuration: parameters))
                    
                    // Add back functionData
                    pipelineData[index] = functionData!
                    // Add back the pipelineData
                    putPipelineData(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                    pipelineData: pipelineData)
                } else {
                    // Add back functionData
                    pipelineData[index] = functionData!
                    // Add back the pipelineData
                    putPipelineData(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                    pipelineData: pipelineData)
                    return
                }
            }
        }
        
        do {
            var messages:[Message] = [Message]()
            // TODO: why doesn't this use member variable ?
            let messageDispatcher:MessageDispatcherImpl? =
                MessageDispatcher.getMessageDispatcher(
                    msgDirectlyConnectedDevice:msgDirectlyConnectedDevice!) as? MessageDispatcherImpl
            
            for n in 0..<pairs.count {
                messages.append(pairs[n].getKey())
                if let so:StorageObject = pairs[n].getValue() {
                    messageDispatcher?.addStorageObjectDependency(storageObject:so,
                                            clientMsgId:message.getClientId())
                    try so.sync()
                }
            }
            
            try messageDispatcher?.queue(messages:messages)
            
        } catch {
            // SEVERE
            log(type:OSLogType.error,
                message: "\(error)")
            throw error
        }
    }
    
    func toString() -> String {
        return base!.toString()
    }
    
    func getValue(attributeType: DeviceModelAttributeType,
                  attributeName: String? = nil,
                  payload: Foundation.Data,
                  callback: @escaping (AnyObject?, ClientError?)->()) {
        
        var payloadStructure: Dictionary<String, AnyObject>?
        var clientError:ClientError? = nil
        do {
            payloadStructure = try JSONSerialization.jsonObject(
                with: payload,
                options: JSONSerialization.ReadingOptions()) as? Dictionary<String, AnyObject>
        } catch {
            clientError = ClientError.argument(NSLocalizedString(
                "Illegal value", comment: ""))
            callback(nil, clientError)
            return
        }
        let value: AnyObject? = payloadStructure!["value"]
        if value == nil {
            clientError = ClientError.argument(NSLocalizedString(
                "Bad payload", comment: ""))
            callback(nil, clientError)
            return
        }
        self.getValue(attributeType: attributeType,
                      attributeValue: value!,
                      attributeName: attributeName,
                      callback: callback)
    }
    
    func getValue(attributeType: DeviceModelAttributeType,
                  attributeValue: AnyObject,
                  attributeName: String? = nil,
                  callback: @escaping (AnyObject?, ClientError?)->()) {
        
        // TODO: value should be "unusable" if an error is returned
        var value: AnyObject = attributeValue
        var clientError:ClientError? = nil
        switch (attributeType) {
        case .BOOLEAN:
            if !(value is Bool) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break
            
        case .INTEGER:
            if !(value is Int) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break
            
        case .NUMBER:
            if !(value is NSNumber) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break
            
        case .DATETIME:
            if !(value is NSNumber || value is Date) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break
            
        case .STRING:
            if !(value is String) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break
        case .URI:
            if !(value is String) {
                clientError = ClientError.argument(NSLocalizedString(
                    "URI value is not STRING", comment: ""))
                break;
            }
            let uri = value as! String
            if StorageConnectionBase.isStorageCloudURI(uri: uri) {
                do {
                    try msgDirectlyConnectedDevice!.createStorageObject(uri: uri,
                            callback: {(delegate, error) in
                                
                        var clientError:ClientError? = nil
                        if let error = error {
                            clientError = ClientError.network(String(format: NSLocalizedString(
                                "Cannot get value for attribute %@. %@", comment:""),
                                         attributeName!, "\(error)"))
                            value = ExternalObject(uri: uri)
                        } else
                        // TODO: Can error and delegate be nil ?
                        if let delegate = delegate {
                            let storageObjectImpl:StorageObjectImpl =
                                StorageObjectImpl(dcd: self.msgDirectlyConnectedDevice!,
                                                  delegate:delegate)
                            storageObjectImpl.setSyncEventInfo(virtualDevice: self,
                                                    nameForSyncEvent: attributeName!)
                            value = storageObjectImpl
                        } else {
                            clientError = ClientError.network(String(format: NSLocalizedString(
                                "Cannot get value for attribute %@", comment:""),
                                         attributeName!))
                            value = ExternalObject(uri: uri)
                        }
                        callback(value, clientError)
                    })
                    // Return don't break, the callback is invoked by
                    // createStorageObject
                    return
                } catch {
                    clientError = ClientError.network("\(error)")
                }
            } else {
                value = ExternalObject(uri: uri)
            }
            break
        }
        callback(value, clientError)
    }
    
    /**
     * - throws: `ClientError.argument` if `attributes` is nil or is the empty
     *           string or if the attribute type is string, if `newValue` is 
     *           nil or the empty string.
     */
    private func processOnChange(builder: DataMessageBuilder,
            attribute: VirtualDeviceAttribute, newValue: AnyObject,
            storageObject: inout StorageObjectImpl?) throws {
        let deviceModelAttribute: DeviceModelAttribute =
                                        attribute.getDeviceModelAttribute()!
        
        let attributeName: String = deviceModelAttribute.getName()!
        builder.setFormat(format: base!.getDeviceModel().getURN() + ":attributes")
        builder.setSource(source: base!.getEndpointId())
        
        switch (deviceModelAttribute.getType()) {
        case .INTEGER:
            if newValue is String {
                try builder.addDataItem(key: attributeName,
                                        value: Double(newValue as! String)!)
            } else {
                try builder.addDataItem(key: attributeName,
                                        value: Double(truncating: newValue as! NSNumber))
            }
            break
        case .NUMBER:
            if newValue is String {
                try builder.addDataItem(key: attributeName,
                                        value: Double(newValue as! String)!)
            } else {
                try builder.addDataItem(key: attributeName,
                                        value: Double(truncating: newValue as! NSNumber))
            }
            break
        case .STRING:
            try builder.addDataItem(key: attributeName,
                                    value: newValue as! String)
            break
        case .BOOLEAN:
            try builder.addDataItem(key: attributeName,
                                    value: newValue.boolValue)
            break
        case .DATETIME:
            let value:Double!
            if newValue is Date {
                value = (newValue as! Date).timeIntervalSince1970
            } else if newValue is NSNumber {
                value = Double(truncating:(newValue as! NSNumber))
            } else {
                throw ClientError.argument(String(format:NSLocalizedString(
                    "@: attribute %@ value must be Date or Number", comment:""),
                "processOnChange()", attributeName))
            }
            try builder.addDataItem(key: attributeName, value: value)
            break
        case .URI:
            
            if newValue is StorageObjectImpl {
                let so = newValue as! StorageObjectImpl
                if so.getSyncStatus() == SyncStatus.notInSync ||
                        so.getSyncStatus() == SyncStatus.syncPending {
                    storageObject = so
                }
                so.setSyncEventInfo(virtualDevice: self, nameForSyncEvent: attributeName)
                try so.sync()
            }
            try builder.addDataItem(key: attributeName, value: (newValue as! ExternalObject).getURI())
            
            break
        }
    }

    func handleStorageObjectStateChange(storageObject: StorageObjectImpl) throws {
        let messageDispatcher:MessageDispatcherImpl? =
            MessageDispatcher.getMessageDispatcher(
                msgDirectlyConnectedDevice: msgDirectlyConnectedDevice!) as? MessageDispatcherImpl
        messageDispatcher?.removeStorageObjectDependency(storageObject:storageObject)
    }
    //
    // Routine for handling invocation of a policy function when the window's
    // slide expires. This routine gets the value of the function, and then
    // processes the remaining functions in the pipeline (if any).
    //
    func processExpiredFunction(
        updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>],
        attributeName:String,
        pipeline:[DevicePolicy.Function]?,
        pipelineData: inout [Dictionary<String,AnyObject>]) {
        
        guard let pipeline = pipeline, !pipeline.isEmpty else {
            return
        }
        
        do {
            let attribute:VirtualDeviceAttribute =
                try adapter.getAttribute(name:attributeName)
            let deviceModelAttribute:DeviceModelAttribute =
                attribute.getDeviceModelAttribute()!
            
            let function:DevicePolicy.Function = pipeline[0]
            let functionId:String = function.getId()
            let config:Dictionary<String,Any> = function.getParameters()
            var data:Dictionary<String,AnyObject> = pipelineData[0]
            guard let deviceFunction:DeviceFunction =
                DeviceFunction.getDeviceFunction(functionId: functionId) else {
                    // SEVERE
                    log(type:OSLogType.error,
                        message:"device function %@ not found",
                        args:functionId)
                return
            }
            
            if let value:Any = deviceFunction.get(deviceAnalog:self,
                                    attribute:deviceModelAttribute.getName(),
                                    configuration:config,
                                    data:&data) {
                // Add back pipeline data
                pipelineData[0] = data
                
                var newValue:Any? = value
                
                if pipeline.count > 1 {
                    // Need Array to convert ArraySlice
                    var pdataslice:[Dictionary<String,AnyObject>] = Array(pipelineData[1..<pipelineData.count])
                    // process remaining policies in the pipeline
                    // TODO: Make sure that if offer0 modifieds pdataslice that 
                    // the results are reflected in pipelineData dictionary entries
                    newValue = offer0(attribute:deviceModelAttribute,
                        attributeValue:newValue! as AnyObject,
                        pipeline: Array(pipeline[1..<pipeline.count]),
                        pipelineData:&pdataslice)
                    
                    // Add back to pipelineData
                    pipelineData.replaceSubrange(1..<pipelineData.count,
                                                 with:pdataslice)
                }
                
                if let policyValue:Any =
                    cast(attributeType:(attribute.getDeviceModelAttribute()?.getType())!,
                         newValue:newValue) {
                    // FINE
                    /*
                    log(type:.debug,
                        message: " %@ : Set : '%@' =  %@",
                        args:getEndpointId(), attributeName, "\(policyValue)")
                    */
                    updatedAttributes.append(Pair(key: attribute,
                                                  value: policyValue as AnyObject))
                }
            }
        } catch {
            // IllegalStateException why not throw ?
            // WARNING
            print(error)
        }
    }
    
    //
    // Routine for handling invocation of a device model policy when the window's
    // slide expires. This routine gets the value of the function, and then
    // processes the remaining functions in the pipeline (if any).
    //
    func processExpiredFunction(pipeline:[DevicePolicy.Function]?,
                                pipelineData: inout [Dictionary<String,AnyObject>]) {
        
        guard let pipeline = pipeline, !pipeline.isEmpty else {
            return
        }
        
        if pipeline.count > 1 {
            // TODO: handle more than one function in a pipeline for all-attributes
            // Print out a warning message about too many function for all-attributes pipeline
            for index in 0..<pipeline.count {
                
                let function:DevicePolicy.Function = pipeline[index]
                let id:String = function.getId()
                let parameters:[String:Any] = function.getParameters()
                let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId: id)
                
                if index == 0 {
                    // WARNING
                    log(type:.fault,
                        message: "Only one function allowed for all-attribute pipeline. " +
                        "Applying: %@",
                        args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                } else {
                    // WARNING
                    log(type:.fault,
                        message:"Ignoring: %@",
                        args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                }
            }
        }
        
        do {
            let function:DevicePolicy.Function = pipeline[0]
            let functionId:String = function.getId()
            let config:Dictionary<String,Any> = function.getParameters()
            var data:Dictionary<String,AnyObject> = pipelineData[0]
            guard let deviceFunction:DeviceFunction =
                    DeviceFunction.getDeviceFunction(functionId:functionId) else {
                // SEVERE
                log(type:OSLogType.error,
                    message:"device function %@ not found",
                    args:functionId)
                return
            }
            
            let value:Any? = deviceFunction.get(deviceAnalog: self,
                                                attribute:nil,
                                                configuration:config,
                                                data:&data)
            // Add back to pipelineData
            pipelineData[0] = data
            
            // TODO: handle more than one function in a pipeline for all-attributes
            // Leave code here for future reference.
            /*
            if value != nil && pipeline.count > 1 {
            
                // TODO: Need to revisit this to make sure that
                // pipelineData is actually modified in the offer0 call
                // via pdataslice
                var pdataslice:[Dictionary<String,AnyObject>] =
                    Array(pipelineData[1..<pipelineData.count])
                // process remaining policies in the pipeline
                value = offer0(attribute: nil,
                               attributeValue:value! as AnyObject,
                               pipeline: Array(pipeline[1..<pipeline.count]),
                               pipelineData: &pdataslice)
                // Add back to pipelineData
                pipelineData.replaceSubrange(1..<pipelineData.count,
                                             with:pdataslice)
            }
            */
            if value != nil {
                
                let pairs:Array<Pair<Message, StorageObject?>> =
                    value as! Array<Pair<Message, StorageObject?>>
                
                if pairs.isEmpty {
                    return
                }
                var messages:[Message] = [Message]()
                // TODO: why not use local member ?
                let messageDispatcher:MessageDispatcherImpl? =
                    MessageDispatcher.getMessageDispatcher(
                        msgDirectlyConnectedDevice:msgDirectlyConnectedDevice!)
                            as? MessageDispatcherImpl
                
                for n in 0..<pairs.count {
                    let messagePair:Pair<Message, StorageObject?> = pairs[n]
                    messages.append(messagePair.getKey())
                    let so:StorageObject? = messagePair.getValue()
                    if let so = so {
                        messageDispatcher?.addStorageObjectDependency(storageObject:so, clientMsgId:messages[n].getClientId())
                        try so.sync()
                    }
                }
                try messageDispatcher?.queue(messages: messages)
            }
        } catch {
            // SEVERE
            print(error)
        }
    }
    
    let computedMetricQueue:DispatchQueue = DispatchQueue.global(qos:.userInitiated)
    // The key is the set of attributes that are referred to in the computedMetric formula.
    // The value is the attribute that is computed.
    private var computedMetricTriggerMap:[Pair<Set<String>,String>] =
        [Pair<Set<String>,String>]()
    
    func checkComputedMetrics(updatedAttributes:Set<String>?) throws -> Set<String> {
        
        guard let updatedAttributes = updatedAttributes, !updatedAttributes.isEmpty,
            !computedMetricTriggerMap.isEmpty else {
                return Set<String>()
        }
        
        // This will be the set of attributes that have computed metrics
        // that are triggered by the set of updated attributes.
        var computedAttributes:Set<String> = Set<String>()
        
        computedMetricQueue.sync() {
            for entry in computedMetricTriggerMap {
                
                // This is the set of attributes that the formula refers to.
                let key:Set<String> = entry.getKey()
                
                // If the set of attributes that the formula refers to
                // is a subset of the updated attributes, then compute
                // the value of the attribute.
                if updatedAttributes.isStrictSuperset(of:key) {
                    computedAttributes.update(with:entry.getValue())
                }
            }
        }
        
        if !computedAttributes.isEmpty {
            for attributeName in computedAttributes {
                let attribute:VirtualDeviceAttribute = try adapter.getAttribute(name:attributeName)
                if !attribute.isSettable() {
                    // WARNING
                    log(type:OSLogType.fault,
                        message:"attempt to modify read-only attribute '%@'",
                        args: attributeName)
                    computedAttributes.remove(attributeName)
                    continue
                }
                
                // TODO: support async getPolicy
                let devicePolicy:DevicePolicy? = try devicePolicyManager?.getPolicy(
                    deviceModelURN: getDeviceModel().getURN(),
                    deviceId:getEndpointId())
                if devicePolicy == nil {
                    continue
                }
                guard let pipeline:[DevicePolicy.Function] =
                    devicePolicy!.getPipeline(attributeName:attributeName), !pipeline.isEmpty else {
                    continue
                }
                
                var pipelineData:[Dictionary<String,AnyObject>] =
                    try getPipelineData(attribute:attributeName)
                
                // offer0 returns true if the attribute was set. If the attribute was not set,
                // then remove the attribute name from the list of attributesToCompute
                let policyValue:AnyObject? =
                    offer0(attribute:attribute.getDeviceModelAttribute(),
                           attributeValue:attribute.get()!,
                           pipeline:pipeline,
                           pipelineData:&pipelineData)

                putPipelineData(attribute: attributeName,
                                pipelineData: pipelineData)
                if let policyValue = policyValue {
                    let _ = try attribute.update(value:policyValue);
                } else {
                    computedAttributes.remove(attributeName)
                }
            }
        }
        // TODO: Remove only debug
        //print("computedAttributes : \(computedAttributes)")
        
        return computedAttributes
    }
    
    private func cast(attributeType:DeviceModelAttributeType,
                      newValue:Any?) -> Any? {
        guard let newValue = newValue, !(newValue is NSNull) else {
            return nil
        }
        var castValue:Any? = nil
        switch (attributeType) {
        case .INTEGER:
            //let number:NSNumber = NSNumber(newValue)
            //let roundedValue:Int64 = Math.round(number.doubleValue())
            castValue = Int(truncating: newValue as! NSNumber)/*roundedValue*/ as AnyObject
            break
        case .NUMBER:
            //let number:NSNumber = NSNumber(newValue)
            castValue = Double(truncating: newValue as! NSNumber) as AnyObject
            break
        case .STRING:
            castValue = (newValue as! String) as AnyObject
            break
        case .BOOLEAN:
            castValue = (newValue as! Bool) as AnyObject
            break
        case .DATETIME:
            if newValue is Date {
                castValue = (newValue as! Date).timeIntervalSince1970 as AnyObject
            } else {
                //let number:Number = Number(newValue)
                castValue = Int64(truncating: newValue as! NSNumber) as AnyObject
            }
            break
        case .URI:
            castValue = (newValue as! String) as AnyObject
            break
        }
        return castValue
    }

    //
    // Window based policy support (as in "window", not Windows OS).
    // Have one scheduled task for each policy "slide" value. The slide value is
    // how much to move the window, so we want to run the policy when the slide expires.
    // When the slide expires, the runnable will call back each VirtualDeviceAttribute
    // that has a policy for that slide value.
    //

    func addScheduledPolicy(window:Int64, slide:Int64, timeZero:Int64,
                            attributeName:String, pipelineIndex:Int) {
        scheduledPoliciesQueue.sync() {
            let key:ScheduledPolicyData.Key = ScheduledPolicyData.Key(window:window,
                                                                      slide:slide)
            var scheduledPolicyData:ScheduledPolicyData! = scheduledPolicies[key]
            if scheduledPolicyData == nil {
                scheduledPolicyData = ScheduledPolicyData(window:window,
                                                          slide:slide,
                                                          timeZero:timeZero)
                scheduledPolicies[key] = scheduledPolicyData
                timedPolicyThread.addTimedPolicyData(data: scheduledPolicyData!)
                // TODO: (moved for slide)
                // Didn't I move this from here before for some reason ?
                if !timedPolicyThread.isAlive && !timedPolicyThread.isCancelled {
                    timedPolicyThread.start()
                }
            }
            scheduledPolicyData.addAttribute(attributeName:attributeName,
                                             pipelineIndex:pipelineIndex)
        }
    }
    
    func removeScheduledPolicy(slide:Int64, attributeName:String,
                               pipelineIndex:Int, window:Int64) {
        scheduledPoliciesQueue.sync() {
            let key:ScheduledPolicyData.Key = ScheduledPolicyData.Key(window:window,
                                                                      slide:slide)
            if let scheduledPolicyData:ScheduledPolicyData = scheduledPolicies[key] {
                scheduledPolicyData.removeAttribute(attributeName:attributeName,
                                                    pipelineIndex:pipelineIndex)
                if scheduledPolicyData.isEmpty {
                    scheduledPolicies.removeValue(forKey:key)
                    timedPolicyThread.removeTimedPolicyData(data: scheduledPolicyData)
                }
            }
        }
    }
    
    /**
     * A callback for notification that a policy has been added.
     *
     * - parameter devicePolicies: the policies that have been added.
     */
    override
    func policyAssigned(devicePolicy:DevicePolicy, assignedDevices:Set<String>) {
        guard !assignedDevices.isEmpty &&
                assignedDevices.contains(getEndpointId()) else {
            return
        }
        let timeZero:Int64 = TimeManager.currentTimeMillis()
        for (attributeName,functions) in devicePolicy.getPipelines() {
            policyAssigned(attributeName:attributeName, newPipeline:functions,
                           timeZero:timeZero)
        }
    }
    
    private func policyAssigned(attributeName:String,
                                newPipeline:[DevicePolicy.Function]?,
                                timeZero:Int64) {
        
        guard let newPipeline = newPipeline, !newPipeline.isEmpty else {
            return
        }
        
        for index in newPipeline.indices {
            
            let function:DevicePolicy.Function = newPipeline[index]
            let id:String = function.getId()
            let parameters:Dictionary<String,Any> = function.getParameters()
            let window:Int64 = DeviceFunction.getWindow(configuration: parameters)
            
            if window > -1 && ("eliminateDuplicates" != id  &&
                    "detectDuplicates" != id) {
                let slide:Int64 = DeviceFunction.getSlide(configuration:parameters,
                                                          window:window)
                addScheduledPolicy(window:window, slide:slide, timeZero:timeZero,
                                   attributeName:attributeName,
                                   pipelineIndex:index)
            }
            
            // If the first policy in the chain is a computed metric,
            // see if it refers to other attributes.
            if index == newPipeline.startIndex && "computedMetric" == id {
                // TODO: Duplicate code in MessagingPolicyImple
                // TODO: Can this be nil ?
                if var formula:String = parameters["formula"] as? String {
                    // If there is no formula or embedded variables
                    // then there aren't any triggerAttributes
                    var triggerAttributes:Set<String>? = nil
                    // Extract attributes where the syntax is $(attr)
                    // Range .upperBound points to after the "("
                    var attrvar:String? = nil
                    // If ther are no attributes, do nothing
                    var attrBegin = formula.range(of: "$(")
                    // Don't create triggerAttributes unles there is at least
                    // one embedded attribute
                    if attrBegin != nil {
                        triggerAttributes = Set<String>()
                    }
                    while attrBegin != nil {
                        // point formula to after "$("
                        attrvar = String(formula[attrBegin!.upperBound..<formula.endIndex])
                        // Find the end of attribute, nil if no ")"
                        if attrvar != nil {
                            if let attrEnd = attrvar!.range(of:")") {
                                // the rest of formula after the attribute
                                formula = String(attrvar![attrEnd.upperBound..<attrvar!.endIndex])
                                attrvar = String(attrvar![attrvar!.startIndex..<attrEnd.lowerBound])
                                if attrvar != nil && attrvar != attributeName {
                                    triggerAttributes!.update(with:attrvar!)
                                }
                            }
                        }
                        attrBegin = formula.range(of: "$(")
                    }
                    // If triggerAttributes is not nil, there it will not be
                    // empty unless the embedded attribute equals attributeName
                    if triggerAttributes != nil && !triggerAttributes!.isEmpty {
                        computedMetricQueue.sync() {
                            computedMetricTriggerMap.append(
                                Pair<Set<String>,String>(key:triggerAttributes!,
                                                         value:attributeName))
                        }
                    }
                }
            }
        }
    }
    
    /**
     * A callback for notification that a policy has been removed.
     *
     * - parameter devicePolicies: the policies that have been removed.
     */
    override
    func policyUnassigned(devicePolicy:DevicePolicy,
                          unassignedDevices:Set<String>) throws {
        guard !unassignedDevices.isEmpty &&
                unassignedDevices.contains(getEndpointId()) else {
            return
        }
        // debug
        log(type:.debug,
            message:"%@: endpointId:%@ : policy unassigned : %@",
            args: "VirtualDeviceImpl.policyUnssigned()",
                  getEndpointId(), devicePolicy.getId())
        
        var updatedAttributes:[Pair<VirtualDeviceAttribute,AnyObject>] =
            [Pair<VirtualDeviceAttribute,AnyObject>]()
        for (attributeName,functions) in devicePolicy.getPipelines() {
            try policyUnassigned(updatedAttributes:&updatedAttributes,
                                 attributeName:attributeName,
                                 oldPipeline:functions)
        }
        if !updatedAttributes.isEmpty {
            //
            // Call updateFields to ensure the computed metrics get run,
            // and will put all attributes into one data message.
            //
            try self.adapter.updateFields(updatedAttributes:&updatedAttributes)
        }
    }

    private func policyUnassigned(updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>],
                               attributeName:String,
                               oldPipeline:[DevicePolicy.Function]?) throws {
    
        guard let oldPipeline = oldPipeline, !oldPipeline.isEmpty else {
            return
        }
        //
        // The order in which the oldPipeline is finalized is important.
        // First, remove any scheduled policies so they don't get executed.
        // Any pending data will be committed in the next step.
        // Second, commit any "in process" values. This may cause a
        // computedMetric to be triggered.
        // Third, remove any computed metric triggers.
        // Lastly, remove any data for this pipeline from the policy data
        // cache
        //
        for index in oldPipeline.indices {
            let function:DevicePolicy.Function = oldPipeline[index]
            let id:String = function.getId()
            let parameters:Dictionary<String,Any> = function.getParameters()
            let window:Int64 = DeviceFunction.getWindow(configuration: parameters)
            if window > -1 && ("eliminateDuplicates" != id &&
                    "detectDuplicates" != id) {
                let slide:Int64 = DeviceFunction.getSlide(configuration: parameters,
                                                          window:window)
                removeScheduledPolicy(slide:slide, attributeName:attributeName,
                                      pipelineIndex:index, window:window)
            }
        }
        
        // commit values from old pipeline
        var pipelineData:[Dictionary<String,AnyObject>] =
            try getPipelineData(attribute:attributeName)
        if !pipelineData.isEmpty {
            if attributeName == DevicePolicy.ALL_ATTRIBUTES() {
                processExpiredFunction(pipeline:oldPipeline,
                                       pipelineData:&pipelineData)
            } else {
                processExpiredFunction(updatedAttributes:&updatedAttributes,
                                       attributeName:attributeName,
                                       pipeline:oldPipeline,
                                       pipelineData:&pipelineData)
            }
            // add back pipelineData, probably not needed since it is about
            // to be removed
            putPipelineData(attribute:attributeName, pipelineData:pipelineData)
            
            if attributeName != DevicePolicy.ALL_ATTRIBUTES() {
                // remove this attribute from the computedMetricTriggerMap
                if computedMetricTriggerMap.count > 0 {
                    for index in (0..<(computedMetricTriggerMap.count - 1)).reversed() {
                        let pair:Pair<Set<String>, String> = computedMetricTriggerMap[index]
                        if attributeName == pair.getValue() {
                            computedMetricTriggerMap.remove(at: index)
                        }
                    }
                }
            }
            // remove data from cache
            removePipelineData(attribute:attributeName)
        }
    }
    //
    // DeviceAnalog implementation
    //
    public func setAttributeValue(attribute:String, value:AnyObject) throws {
        try self.set(attributeName:attribute, attributeValue:value)
    }

    public func getAttributeValue(attribute:String) throws -> AnyObject? {
        return try self.get(attributeName:attribute)
    }
    
    public func call(actionName:String, args:[Any?]) {
        
        guard let callable:(VirtualDevice,AnyObject?)->() =
            self.actionMapLock.sync(execute: {
                return actionMap?[actionName] as? (VirtualDevice, AnyObject?) -> ()}) else {
                    // WARNING
                    log(type:.fault,
                        message:"%@ does not contain action '%@'",
                        args:getDeviceModel().getURN(), actionName)
                    return
        }
        
        guard let deviceModelAction:DeviceModelAction = VirtualDeviceImpl
            .getDeviceModelAction(deviceModel:getDeviceModel(),
                                  actionName:actionName),
                let argType:DeviceModelAttributeType =
                    deviceModelAction.getArgType() else {
            // argType is nil. Action takes no args.
            callable(self, nil)
            return
        }

        // TODO: currently, call only supports one arg
        //let arg:Any = args.first as Any
        guard !args.isEmpty, let arg = args[0] else {
            // WARNING
            log(type:.fault,
                message:"%@  action '%@' requires an argument",
                args: getDeviceModel().getURN(), actionName);
            return
        }
        
        var goodArg:Bool = false
        switch (argType) {
        case .INTEGER,.NUMBER:
            goodArg = arg is NSNumber
            if goodArg {
                // Assumption here is that lowerBound <= upperBound
                let val:NSNumber = arg as! NSNumber
                if let upper:NSNumber = deviceModelAction.getUpperBound() {
                    if Double(truncating: val) > Double(truncating: upper) {
                        // WARNING
                        log(type:.fault,
                            message:"%@ action '%@' arg out of range: %f > %f",
                            args: getDeviceModel().getURN(),
                            actionName, val, upper)
                        // even though the arg is out of range, pass it to the action
                        // TODO is this the right thing to do?
                    }
                }
                if let lower = deviceModelAction.getLowerBound() {
                    if Double(truncating: val) < Double(truncating: lower) {
                        // WARNING
                        log(type:.fault,
                            message:"%@ action '%@' arg out of range: %f < %f",
                            args: getDeviceModel().getURN(),
                            actionName, val, lower)
                        // even though the arg is out of range, pass it to the action
                        // TODO is this the right thing to do?
                    }
                }
            }
            break
        case .DATETIME:
            goodArg = arg is Date || arg is NSNumber
            break
        case .BOOLEAN:
            goodArg = arg is Bool
            break
        case .STRING,.URI:
            goodArg = arg is String
            break

        }
        
        if !goodArg {
            // WARNING
            log(type:.fault,
                message: "%@ action '%@' : wrong argument type. " +
                "Expected %@ found %@",
                args: getDeviceModel().getURN(), actionName,
                "\(argType)", "\(arg)")
            return
        }
        callable(self, arg as AnyObject)
    }

    public func queueMessage(message:Message) throws {
        try self.queueMessage(message:message, storageObject: nil)
    }
    
    //
    // Hooks for handling Alert and Data onError callbacks.
    //
    // When AlertImpl#raise() or DataImpl#submit() are called, the message clientId
    // and the callback are added to onErrorCallbacks map via addOnErrorCallback.
    // The message's clientId is the key, and a weak reference to the callback is
    // the value. If the message from calling Alert#raise() or Data#submit() cannot
    // be delivered, then ErrorCallbackBridge#failed(List<Message>, Exception) method
    // ends up getting called (from MessageDispatcherImpl.Transmitter send method).
    // The ErrorCallbackBridge#invokeErrorCallback does a removeOnErrorCallback,
    // if the message is an alert format or a data format, and invokes that callback
    // instead of the one for the virtual device.
    // If the message is delivered successfully, the entry is removed by way of
    // MessageDispatcher calling DELIVERY_CALLBACK
    //
    class OnErrorDeliverCallback:MessageDispatcher.DeliveryCallback {
        override func delivered(messages:[Message]) {
            for message in messages {
                if (message.getType() == .ALERT ||
                        message.getType() == .DATA) {
                    _ = removeOnErrorCallback(messageId: message.getClientId())
                }
            }
        }
    }
    
    // Hack to overcome swift limitiation that can't compare a closure.
    class CallbackWrapper {
        var message:AnyHashable
        var callback:(ErrorEvent)->()
        
        init(message:AnyHashable, callback: @escaping (ErrorEvent)->()) {
            self.callback = callback
            self.message = message
        }
    }

    fileprivate static var onErrorCallbacks:[String:CallbackWrapper] =
        [String:CallbackWrapper]()
    
    fileprivate static let onErrorCallbackQueue:DispatchQueue =
        DispatchQueue(label:"onErrorCallbackQueue", qos: .utility)
    
    static func addOnErrorCallback(message:AnyHashable, messageId:String,
                            callback: @escaping (ErrorEvent)->()) {
        onErrorCallbackQueue.sync() {
            let wrapper = CallbackWrapper(message:message, callback: callback)
            onErrorCallbacks[messageId] = wrapper
        }
        
    }
    
    static func removeOnErrorCallback(messageId:String) -> ((ErrorEvent)->())? {
        return onErrorCallbackQueue.sync(execute: {
            // If there is no entry for messageId, then there is no callback
            guard let wrapper:CallbackWrapper =
                onErrorCallbacks.removeValue(forKey:messageId) else {
                    return nil
            }
            return wrapper.callback
        })
    }
    
    // Replace existing callbeck with callback.
    // If callback is nil remove any existing entries.
    static func replaceOnErrorCallback(message:AnyHashable,
                                       callback: ((ErrorEvent)->())?) {

        onErrorCallbackQueue.sync() {
            var id:String? = nil
            var cw:CallbackWrapper? = nil
            for (messageId, wrapper) in onErrorCallbacks {
                if message != wrapper.message {
                    continue
                }
                id = messageId
                cw = wrapper
                break
            }
            
            if let id = id, let cw = cw  {
                // If callback is nil remove the entry
                if let callback = callback {
                    // replace the callback
                    cw.callback = callback
                } else {
                    // Remove the callback if callback is nil
                    onErrorCallbacks.removeValue(forKey: id)
                }
            }
        }
    }
}
