/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Implementation of protocol Equatable
 */
func ==(lhs: VirtualDeviceAttribute, rhs: VirtualDeviceAttribute) -> Bool {
    if lhs === rhs {
        return true
    }
    if !(lhs.virtualDevice === rhs.virtualDevice) {
        return false
    }
    if !(lhs.model == rhs.model) {
        return false
    }
    if (lhs.value === rhs.value &&
        lhs.lastKnownValue === rhs.lastKnownValue) ||
       (lhs.value == nil && rhs.value == nil &&
            lhs.lastKnownValue == nil && rhs.lastKnownValue == nil) {
        return true
    }
    
    var result:Bool = false
    switch lhs.model.getType(){
    case .STRING:
        result = lhs.value as? String == rhs.value as? String &&
                 lhs.lastKnownValue as? String == rhs.lastKnownValue as? String
        break
    case .INTEGER:
        result = lhs.value as? Int == rhs.value as? Int &&
                 lhs.lastKnownValue as? Int == rhs.lastKnownValue as? Int
        break
    case .NUMBER:
        result = lhs.value as? NSNumber == rhs.value as? NSNumber &&
                 lhs.lastKnownValue as? NSNumber == rhs.lastKnownValue as? NSNumber
        break
    case .BOOLEAN:
        result = lhs.value as? Bool == rhs.value as? Bool &&
                 lhs.lastKnownValue as? Bool == rhs.lastKnownValue as? Bool
        break
    case .URI:
        result = lhs.value as? String == rhs.value as? String &&
                 lhs.lastKnownValue as? String == rhs.lastKnownValue as? String
        break
    case .DATETIME:
        result = (lhs.value as? Date == rhs.value as? Date &&
                  lhs.lastKnownValue as? Date == rhs.lastKnownValue as? Date) ||
                 (lhs.value as? NSNumber == rhs.value as? NSNumber &&
                  lhs.lastKnownValue as? NSNumber == rhs.lastKnownValue as? NSNumber)
        break
    }
    /* TODO: FIXME
     if !( lhs.changeCallback === rhs.changeCallback) {
     return false
     }
     if !( lhs.errorCallback === rhs.errorCallback) {
     return false
     }
     */
    return result
}

/**
 * VirtualDeviceAttribute is an attribute in a VirtualDevice model. The
 * `AbstractVirtualDevice.get(String)` and `AbstractVirtualDevice.set(String,
 * AnyObject)` methods
 * are wrappers around the `VirtualDeviceAttribute` instances of a
 * `VirtualDevice`.
 */
class VirtualDeviceAttribute: VirtualDeviceAttributeBase, Hashable {

    var virtualDevice: VirtualDeviceImpl
    
    /**
     * Implementation of protocol Hashable
     */
    var hashValue : Int {
        var result: Int = model.hashValue
        if value != nil {
            switch model.getType(){
            case .STRING:
                result = 31 &* result &+ (value as! String).hashValue
                break
            case .INTEGER:
                result = 31 &* result &+ (value!.int32Value.hashValue).hashValue
                break
            case .NUMBER:
                result = 31 &* result &+ (value!.doubleValue.hashValue).hashValue
                break
            case .BOOLEAN:
                result = 31 &* result &+ (value as! Bool).hashValue
                break
            case .DATETIME:
                if value is Date {
                    result = 31 &* result &+ (value as! Date).hashValue
                } else if value is NSNumber {
                    result = 31 &* result &+ (value as! NSNumber).hashValue
                }
                break
            case .URI:
                result = 31 &* result &+ (value as! ExternalObject).hashValue
                break
            }
        } else {
            result = 31 &* result
        }
        return result
    }

    init(virtualDevice: VirtualDeviceImpl, model: DeviceModelAttribute) {
        self.virtualDevice = virtualDevice
        super.init(model: model)
    }
    
    override func set(value: AnyObject) throws {
        // validate throws ClientError.argument if value is not valid
        do {
            try validate(attribute: getDeviceModelAttribute()!, value: value)
        } catch {
            throw ClientError.argument(NSLocalizedString(
                "illegal value", comment: ""))
        }
        
        setLastKnown(lastKnown: value)
        self.value = value
        try (getVirtualDevice()).processOnChange(virtualDeviceAttribute: self,
                                                 newValue: get()!);
    }

    override func isSettable() -> Bool {
        // An attribute is always settable on the device-client side
        return true
    }
    
    func getVirtualDevice() -> VirtualDeviceImpl {
        return virtualDevice
    }
    
    func setLastKnown(lastKnown: AnyObject) {
        self.lastKnownValue = lastKnown
    }
    
    /**
     * Same as set, but doesn't generate an event. Returns true if the value
     * was updated.
     *
     * - parameter value: the new value with which the attribute will be updated
     *
     * - returns: true if the value was updated
     *
     * - throws: `ClientError.argument` if `value` is not the correct type or is
     *           is an invalid value for `attribute`.
     */
    func update(value: AnyObject) throws -> Bool {
        
        // validate throws ClientError.argument if value is not valid
        do {
            try validate(attribute: getDeviceModelAttribute()!, value: value)
        } catch {
            throw ClientError.argument(String(format:NSLocalizedString(
                "illegal value: %@", comment: ""), "\(error)"))
        }
        
        setLastKnown(lastKnown: value)
        self.value = value
        
        return true
    }

    func toString() -> String {
        let modelName = model.getName()
        var result: String = "name : " + modelName! + "\n"
        result = result + "description : "
        result = result + model.getDescription()! + "\n"
        result = result + "alias : " + model.getAlias()! + "\n"
        result = result + "model : " + model.getModel()! + "\n"
        result = result + "access : " + model.getAccess().rawValue + "\n"
        result = result + "type : " + model.getType().rawValue + "\n"
        
        let modelType: DeviceModelAttributeType = model.getType()
        if (modelType == .NUMBER) || (modelType == .INTEGER) {
            result = result + "lowerBound : " + 
                model.getLowerBound()!.description + "\n"
            result = result + "upperBound : " + 
                model.getUpperBound()!.description + "\n"
        }
        // can only be processed if the optional value value and 
        // lastKnowValue are set
        if(value != nil && lastKnownValue != nil) {
            switch modelType {
            case .STRING:
                result = result + "value : " + (value as! String) + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! String) + "\n"
                break
            case .INTEGER:
                result = result + "value : " + 
                    (value as! Int).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! Int).description + "\n"
                break
            case .NUMBER:
                result = result + "value : " + 
                    (value as! NSNumber).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! NSNumber).description + "\n"
                break
            case .BOOLEAN:
                result = result + "value : " + 
                    (value as! Bool).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! Bool).description + "\n"
                break
            case .URI:
                result = result + "value : " +
                    (value as! ExternalObject).getURI() + "\n"
                break
            case .DATETIME:
                var dtString:String!
                if value is Date {
                    dtString = (value as! Date).description
                } else { // Must be NSNumber
                    dtString = (value as! NSNumber).description
                }
                result = result + "value : \(dtString)\n"
                if lastKnownValue is Date {
                    dtString = (lastKnownValue as! Date).description
                } else { // Must be NSNumber
                    dtString = (lastKnownValue as! NSNumber).description
                }
                result = result + "lastKnownValue : \(dtString)\n"
                break
            }
        }
        return result
    }
}
