/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

class ApplicationEnumerator : PageableEnumerator<Application> {
    
    private var secureConnection: HttpSecureConnection
    private var filter: Filter?
    
    /**
     *    Instantiate a new iterator.
     *
     *     - Parameter offset: The offset of the first item to fetch
     *
     *     - Parameter limit: The maximum number of items requested per REST call
     *
     *     - Parameter filter: A filter to constrain the response. If *nil*, 
     *       all applications are enumerated.
     *
     *     - Parameter secureConnection: The server connection
     */
    init(offset: Int, limit: Int, filter: Filter?, secureConnection: HttpSecureConnection) {
        /*
         - Throws: IOException If an I/O error occurs when retrieving the total count of elements matching the filter
         
         - Throws: GeneralSecurityException When key or signature algorithm class
         cannot be loaded, or the key is not in
         the trusted assets store, or the
         private key is invalid
         */
        self.filter = filter
        self.secureConnection = secureConnection
        
        super.init(offs: offset, lim: limit)
    }
    
    /**
     * Iterate through the pages.
     *
     * - parameter callback: Callback containing the `Pageable<T>` If
     *                       an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    func iterate(callback: @escaping (Pageable<Application>?, ClientError?) -> Void) {
        do {
            try self.getCount(countResponse: { countResponse, error in
                if error != nil {
                   callback(nil, ClientError.network(NSLocalizedString("Cannot retrieve count of applications", comment: "")))
                } else {
                    if let countObject = countResponse {
                        let data = countObject.getData()
                        
                        if data.count == 0 {
                            self.setSize(size: -1)
                        } else {
                            var jsonObject:JSON!
                            do {
                                jsonObject = try JSON(data: data as Foundation.Data)
                            } catch {
                                callback(nil, ClientError.network(String(format:NSLocalizedString(
                                    "%@: JSON threw exception: %@",
                                    comment:""), "ApplicationEnumerator.iterate()",
                                                 "\(error)")))
                                return
                            }
                            if let size = jsonObject.dictionaryValue["count"]?.intValue {
                                self.setSize(size: size)
                            } else {
                                self.setSize(size: -1)
                            }
                        }
                        callback(self, nil)
                    } else {
                        callback(nil, ClientError.argument(NSLocalizedString("ApplicationEnumerator iterate() countResponse should not be nil", comment: "")))
                    }
                }
            })
        } catch {
            callback(nil, ClientError.network(NSLocalizedString("Exception while retrieving count of applications", comment: "")))
        }
    }
    
    /**
     * Performs a new REST call to get the count of applications matching the
     * filter
     *
     * - Returns: The count of devices matching the request, or *-1* if unknown.
     */
    private func getCount(countResponse: @escaping (HttpResponse?, ClientError?) -> Void) throws -> Void {
        let req = ApplicationEnumerationRequest(applicationId: nil, path: "count", filter: filter, offset: self.voffset, limit: self.vlimit)
        try self.secureConnection.get(restApi: req.request, headers: req.headers, response: { getResponse, error in
            countResponse(getResponse, error)
        })
    }
    
    /**
     * Performs a new REST call to refill internal cache.
     *
     * - Parameter offset: The offset of first element to fetch.
     *
     * - Parameter limit: The maximum number of elements to fetch.
     *
     * - Parameter callback: Callback containing the *Pageable<T>* If
     *                       an error occurs *page* is set to nil and
     *                       *error* indicates the error.
     */
    override func load(offset: Int, limit: Int, callback: @escaping (PagedResponse<Application>?, ClientError?) -> Void) throws -> Void {
        let req = ApplicationEnumerationRequest(applicationId: nil, path: nil, filter: nil, offset: offset, limit: limit)
        try self.secureConnection.get(restApi: req.request, headers: req.headers, response: { countResponse, error in
            if (error != nil) {
                callback(nil, ClientError.network(NSLocalizedString("Cannot retrieve page of applications for offset " + offset.description, comment: "")))
            } else {
                if let countObject = countResponse {
                    let data = countObject.getData()
                    do {
                        let jsonObject = try JSON(data: data)
                        let page = try ApplicationEnumerationResponse.fromJson(jsonObject: jsonObject)
                        callback(page, nil)
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString("Exception while retrieving page of applications for offset " + offset.description, comment: "")))
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("ApplicationEnumerator load() countResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    class ApplicationEnumerationRequest {
        
        fileprivate let headers = ["Content-Type" : "application/json", "Accept" : "application/json"]
        
        fileprivate let request: String
        
        /**
         *     Creates an *ApplicationEnumerationRequest* for 
         *     *"/iot/api/v2/apps/{applicationId}/{path}?offset={offset}
         *     &limit={limit}"*
         *
         *       - Parameter applicationID: The application identifier to extend 
         *         to the path. If *nil*, not added to the path.
         *
         *       - Parameter path: The path to extend the request. 
         *         If *nil*, not added to the request
         *
         *       - Parameter filter: A filter to constrain the response. 
         *         If *nil*, all applications are enumerated.
         *
         *       - Parameter offset: The offset of the first element to fetch
         *
         *       - Parameter limit: The maximum number of elements to fetch
         */
        init(applicationId: String?, path: String?, filter: Filter?, offset: Int, limit: Int) {
            
            var req = RestApi.V2.getReqRoot() + "/apps"
        
            
            if (applicationId != nil) {
                req += "/" + applicationId!
            }
            
            if (path != nil) {
                req += "/" + path!
            }
            
            var sep = "?"
            
            req += sep + "totalResults=true"
            sep = "&"
            
            if limit > 0 {
                req += sep + "limit=" + (limit as NSNumber).stringValue
                sep = "&"
            }
            
            if offset > 0 {
                req += sep + "offset=" + (offset as NSNumber).stringValue
                sep = "&"
            }
            
            if (filter != nil) {
                // URL build
                let json = filter!.toJson()
                
                let a = json[0].stringValue
                let b = json[1].dictionaryObject as! [String : String]

                
                let op = b.keys.first
                let app = b.values.first
                
                let query = "q={\"\(a)\":{\"" + op! + "\":\"" + app! + "\"}}"
         
                req += sep + query
            }
            
            self.request = req
        }
        
        func getRequest() -> String {
            return self.request
        }
        
        func getHeaders() -> [String : String] {
            return self.headers
        }
        
    }
    
    fileprivate class ApplicationEnumerationResponse : PagedResponse<Application> {
        
        private override init(elements: [Application], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
            super.init(elements: elements, offset: offset, total: total, hasMore: hasMore, links: links)
        }
        
        fileprivate static func fromJson(jsonObject: JSON) throws -> ApplicationEnumerationResponse {
            let items = jsonObject.dictionaryValue["items"]?.arrayValue
            let links = jsonObject.dictionaryValue["links"]?.arrayValue
            var hasMore = jsonObject.dictionaryValue["hasMore"]?.boolValue
            if hasMore == nil {
                hasMore = false
            }
            
            var offset = jsonObject.dictionaryValue["offset"]?.intValue
            if offset == nil {
                offset = -1
            }
            
            var total = jsonObject.dictionaryValue["totalResults"]?.intValue
            if total == nil {
                total = -1
            }
            
            var list: [Application] = [Application]()
            if items != nil && items?.count != 0 {
                for item: JSON in items! {
                    list.append(Application.fromJson(object: item))
                }
            }
            
            var map: [String : String] = [String : String]()
            if links != nil && links?.count != 0 {
                for value: JSON in links! {
                    let rel = value.dictionaryValue["rel"]?.stringValue
                    let href = value.dictionaryValue["href"]?.stringValue
                    if rel?.isEmpty == false && href?.isEmpty == false {
                        map[rel!] = href
                    }
                }
            }
            
            return ApplicationEnumerationResponse(elements: list, offset: offset!, total: total!, hasMore: hasMore!, links: map)
        }
    }
    
}
