/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * A *DeviceModelIterator* can be used to enumerate device models matching a set of criteria. When iterating, it may generates REST calls to get the next chunk of deviceModels.
 */
public final class DeviceModelIterator : PageableEnumerator<String> {
    
    /**
     * The *HttpSecureConnection* for server connection
     */
    private let secureConnection: HttpSecureConnection
    
    /**
     * The application identifier in which we should look for devices.
     */
    private let appId: String
    
    /**
     * Instantiate a new iterator.
     *
     * - parameter appId: The application identifier in which we should look for devices.
     *
     * - parameter offset: The offset of the first element to fetch.
     *
     * - parameter limit: The maximum number of items requested per REST call.
     *
     * - parameter secureConnection: A connection to the server.
     */
    init(appId: String, offset: Int, limit: Int, secureConnection: HttpSecureConnection) {
        
        self.appId = appId
        self.secureConnection = secureConnection
        
        super.init(offs: offset, lim: limit)
    }
    
    /**
     * Iterate through the pages.
     *
     * - parameter callback: Callback containing the `Pageable<T>` If
     *                       an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    func iterate(callback: @escaping (Pageable<String>?, ClientError?) -> Void) {
        do {
            try self.getCount(countResponse: { countResponse, error in
                if error != nil {
                    callback(nil, error)
                } else {
                    if let countObject = countResponse {
                        let data = countObject.getData()
                        
                        if data.count == 0 {
                            self.setSize(size: -1)
                        } else {
                            var jsonObject:JSON!
                            do {
                                jsonObject = try JSON(data: data as Foundation.Data)
                            } catch {
                                callback(nil, ClientError.argument(String(format:NSLocalizedString(
                                    "%@: JSON threw exception: %@",
                                        comment: ""),"DeviceModelIterator.iterate()", "\(error)")))
                                return
                            }
                            if let size = jsonObject.dictionaryValue["count"]?.intValue {
                                self.setSize(size: size)
                            } else {
                                self.setSize(size: -1)
                            }
                        }
                        callback(self, nil)
                    } else {
                        callback(nil, ClientError.argument(NSLocalizedString("DeviceModelIterator iterate() countResponse should not be nil", comment: "")))
                    }
                }
            })
        } catch {
            callback(nil, ClientError.network(NSLocalizedString("Exception while retrieving count of device models for application identifier " + self.appId, comment: "")))
        }
    }

    /**
     * Performs a new REST call to get the count of device models
     *
     * - Returns: The count of devices models, or *-1* if unknown.
     */
    private func getCount(countResponse: @escaping (HttpResponse?, ClientError?) -> Void) throws -> Void {
        let req = DeviceModelRequest(appId: self.appId, path: "count", offset: self.voffset, limit: self.vlimit)
        try self.secureConnection.get(restApi: req.request(), headers: req.headers(), response: { getResponse, error in
            countResponse(getResponse, error)
        })
    }
    
    /**
     * Performs a new REST call to refill internal cache.
     *
     * - Parameter offset: The offset of first element to fetch.
     *
     * - Parameter limit: The maximum number of elements to fetch.
     *
     * - Parameter callback: Callback containing the *Pageable<T>* If
     *                       an error occurs *page* is set to nil and
     *                       *error* indicates the error.
     */
    override func load(offset: Int, limit: Int, callback: @escaping (PagedResponse<String>?, ClientError?) -> Void) throws -> Void {
        let req = DeviceModelRequest(appId: self.appId, path: "", offset: offset, limit: limit)
        try self.secureConnection.get(restApi: req.request(), headers: req.headers(), response: { getResponse, error in
            if (error != nil) {
                callback(nil, error)
            } else {
                if let getObject = getResponse {
                    let data = getObject.getData()
                    do {
                        let jsonObject = try JSON(data: data)
                        let page = try DeviceModelResponse.fromJson(jsonObject: jsonObject)
                        callback(page, nil)
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString("Exception while retrieving page of device models for offset " + offset.description, comment: "")))
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("DeviceModelIterator load() getResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    
    /**
     * REST API request to get the list of device models.
     */
    class DeviceModelRequest {
        
        /// List of headers required for the REST API
        
        private let vheaders = ["Content-Type" : "application/json", "Accept" : "application/json"]
        
        /// The REST request including parameters
        
        private let vrequest: String
        
        /**
         * Creates a *DeviceModelRequest* matching the specified criteria.
         *
         * - Parameter appId: The application identifier in which we should look for devices.
         *
         * - Parameter limit: The maximum number of result to return.
         *
         * - Parameter offset: The offset of the results in the entire list of device models matching the specified criteria.
         */
        init(appId: String, path: String, offset: Int, limit: Int) {
            var req: String?
            var query: String = "totalResults=true"
            
            if appId.isEmpty == true {
                req = RestApi.V2.getReqRoot()
            } else {
                req = RestApi.V2.getReqRoot()+"/apps/" + appId
            }
            
            req = req! + "/deviceModels"
            
            if path.isEmpty == false {
                req = req! + "/" + path
            }
            
            if limit > 0 {
                query = query + "&limit=" + String(limit)
            }
            
            if offset > 0 {
                query = query + "&offset=" + String(offset)
            }
            
            self.vrequest = req! + "?" + query
        }
        
        /**
         * Get the REST call (including parameters) representing this request
         *
         * - Returns: The REST call (including parameters)
         */
        func request() -> String {
            return self.vrequest
        }
        
        /**
         * Get the list of HTTP headers required for this request
         *
         * - Returns: HTTP headers required to perform this request
         */
        func headers() -> [String : String] {
            return vheaders
        }
    }
    
    
    fileprivate class DeviceModelResponse : PagedResponse<String> {

        private override init(elements: [String], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
            super.init(elements: elements, offset: offset, total: total, hasMore: hasMore, links: links)
        }
        
        fileprivate static func fromJson(jsonObject: JSON) throws -> DeviceModelResponse {
            let items = jsonObject.dictionaryValue["items"]?.arrayValue
            let links = jsonObject.dictionaryValue["links"]?.arrayValue
            var hasMore = jsonObject.dictionaryValue["hasMore"]?.boolValue
            if hasMore == nil {
                hasMore = false
            }
        
            var offset = jsonObject.dictionaryValue["offset"]?.intValue
            if offset == nil {
                offset = -1
            }
        
            var total = jsonObject.dictionaryValue["totalResults"]?.intValue
            if total == nil {
                total = -1
            }
            
            var list: [String] = [String]()
            if items != nil && items?.count != 0 {
                // for each item, create an endpoint from json stream
                for item: JSON in items! {
                    let urn = item.dictionaryValue["urn"]?.stringValue
                    if urn != nil  {
                        list.append(urn!)
                    }
                }
            }
            
            var map: [String : String] = [String : String]()
            if links != nil && links?.count != 0 {
                for value: JSON in links! {
                    let rel = value.dictionaryValue["rel"]?.stringValue
                    let href = value.dictionaryValue["href"]?.stringValue
                    if rel?.isEmpty == false && href?.isEmpty == false {
                        map[rel!] = href
                    }
                }
            }
            
            return DeviceModelResponse(elements: list, offset: offset!, total: total!, hasMore: hasMore!, links: map)
        }
    }
}
