/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 *  A Filter that can be used for queries.
 *
 * Example:
 *
 * Devices activated in the last hour that implement a `com:acme:device`
 * device model.
 *
 * The following code:
 *
 * ```swift
 *  f = Filter.and(
 *        Filter.gte(Device.Field.CREATED.alias(),
 *          (System.currentTimeMillis() - 3600000L)),
 *        Filter.eq(Device.Field.STATE.alias(), "ACTIVATED"),
 *        Filter.like(Device.Field.DEVICE_MODELS.alias() + ".urn",
 *          "urn:com:acme:device:%"));
 * ```
 *
 * will create the following Json object:
 *
 * ```swift
 * {"$and":[
 *    {"created":{"$gte":1457137772894}},
 *    {"state":{"$eq":"ACTIVATED"}},
 *   {"deviceModels.urn":{"$like":"urn:com:acme:device:%"}}]}
 * ```
 */
public class Filter {
    
    class SimpleFilter : Filter {
        
        let value: JSON
        
        init(op: String, name: String, value: String) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: [String]) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: Bool) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: [Bool]) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: Double) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: [Double]) {
            self.value = JSON(arrayLiteral: name, [op : value])
        }
        
        init(op: String, name: String, value: CLongLong) {
            self.value = JSON(arrayLiteral: name, [op : String(value)])
        }
        
        init(op: String, name: String, value: [CLongLong]) {
            self.value = JSON(arrayLiteral: name, [op : String(describing: value)])
        }
        
        override func toJson() -> JSON {
            return self.value
        }
        
        override func toJsonURI() -> String {
            return self.value.rawString()!
        }
        
        override func isEmpty() -> Bool {
            return value.isEmpty
        }
    }

    class CompositeFilter : Filter {
        
        private var operands: [Filter] = []
        private let operat0r: String?
        
        init (operat0r: String, operands: [Filter]) {
            self.operands = operands
            self.operat0r = operat0r
        }
        
        // TODO: Make this more efficient. Use JSON to construct the
        // JSON structure.
        override func toJson() -> JSON {
            return JSON(toJsonURI())
        }
        
        // TODO: No need for a method to produce a String, just the 
        // toJson should be sufficient.
        override func toJsonURI() -> String {
            var sb = String()
            sb.append("{\"" + self.operat0r! + "\":")
            if (self.operands.count > 1) {
                sb.append("[")
            }
            
            for f in self.operands {
                let fil = f as! SimpleFilter
                
                let opn1 = fil.toJson().arrayValue[0].stringValue
                let opr = fil.toJson().arrayValue[1].dictionaryValue.first!.0
                let opn2 = fil.toJson().arrayValue[1].dictionaryValue.first!.1.stringValue
                
                sb.append("{\"" + opn1 + "\":")
                sb.append("{\"" + opr + "\":\"" + opn2 + "\"}},")
            }
            
            sb = String(sb[sb.startIndex..<sb.index(before: sb.endIndex)])
            if (self.operands.count > 1) {
                sb.append("]")
            }
            sb.append("}")
            
            return sb
        }
        
    }
    
    /**
     * Creates a query filter that performs the logical AND between sub filters 
     * provided as argument: `(filter1 and filter2 and ...)`.
     *
     * - parameter filters: Sub filters to combine.
     *
     * - returns: A new filter `{ "$and": [ filter1, filter2, .... ] }`.
     */
    public static func and(filters: [Filter]) -> Filter {
        return CompositeFilter(operat0r: "$and", operands: filters)
    }
    
    /**
     * Creates a query filter that performs the logical OR between sub filters 
     * provided as argument: `(filter1 or filter2 or ...)`.
     *
     * - parameter filters: Sub filters to combine.
     *
     * - returns: A new filter `{ "$or": [ filter1, filter2, .... ] }`.
     */
    public static func or(filters: [Filter]) -> Filter {
        return CompositeFilter(operat0r: "$or", operands: filters)
    }
    
    /**
     * Creates a query filter that performs the logical NOT on the filter 
     * provided as argument: `(not filter)`.
     *
     *  - parameter filter: filter on which to apply the NOT.
     *
     *  - returns: A new filter `{ "$not": filter }`.
     */
    public static func not(filter: Filter) -> Filter {
        return CompositeFilter(operat0r: "$not", operands: [filter])
    }
    
    /**
     *  Creates a query filter that performs `(op1 == op2)`.
     *
     *   - parameter op1: First operand.
     *
     *   - parameter op2: Second operand.
     *
     *   - returns: A new filter `{ op1: { "$eq": op2 } }`.
     */
    public static func eq(op1: String, op2: String) -> Filter {
        return SimpleFilter(op: "$eq", name: op1, value: op2)
    }
    
    /**
     * Creates a query filter that performs `(op1 == op2)`.
     *
     *  - parameter op1: First operand.
     *
     *  - parameter op2: Second operand.
     *
     *  - returns: A new filter `{ op1: { "$eq": op2 } }`.
     */
    public static func eq(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$eq", name: op1, value: op2)
    }
    
    /**
     * Creates a query filter that performs `(op1 == op2)`.
     *
     *  - parameter op1: First operand.
     *
     *  - parameter op2: Second operand.
     *
     *  - returns: A new filter `{ op1: { "$eq": op2 } }`.
     */
    public static func eq(op1: String, op2: Bool) -> Filter {
        return SimpleFilter(op: "$eq", name: op1, value: op2)
    }
    
    /**
     * Creates a query filter that performs `(op1 == op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$eq": op2 } }`.
     */
    public static func eq(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$eq", name: op1, value: op2)
    }
    
    /**
     * Creates a query filter that checks if `op1` exists.
     *
     * - parameter op1: First operand.
     *
     * - parameter e: Either `true` to check that exists, or `false` to check
     *                that not exists.
     *
     * - returns: A new filter performing `{ op1: { "$exists" : true/false }}`.
     */
    public static func exists(op1: String, e: Bool) -> Filter {
        return  SimpleFilter(op: "$exists", name: op1, value: e)
    }
    
    /**
     * Creates a query filter that checks if the value matches the pattern.
     *
     * - parameter value: The value to check.
     *
     * - parameter pattern: The pattern to match.
     *
     * - returns: A new filter performing `{ op1: { "$like" : pattern }}`.
     */
    public static func like(value: String, pattern: String) -> Filter {
        return SimpleFilter(op: "$like", name: value, value: pattern)
    }
    
    /**
     * Creates a query filter that performs `(key IN [value1, value2, ...])`.
     *
     * - parameter key: The key to check.
     *
     * - parameter values: The list of values to compare with.
     *
     * - returns: A new filter `{ key: { "$in" : [value1, value2, ...] }}`.
     */
    public static func _in(key: String, values: [String]) -> Filter {
        return SimpleFilter(op: "$in", name: key, value: values)
    }
    
    /**
     * Creates a query filter that performs `(key IN [value1, value2, ...])`.
     *
     * - parameter key: The key to check.
     *
     * - parameter values: The list of values to compare with.
     *
     * - returns: A new filter `{ key: { "$in" : [value1, value2, ...] }}`.
     */
    public static func _in(key: String, values: [CLongLong]) -> Filter {
        return SimpleFilter(op: "$in", name: key, value: values)
    }
    
    /**
     * Creates a query filter that performs `(key IN [value1, value2, ...])`
     *
     *  - parameter key: The key to check.
     *
     *  - parameter values: The list of values to compare with.
     *
     *  - returns: A new filter `{ key: { "$in" : [value1, value2, ...] }}`.
     */
    public static func _in(key: String, values: [Double]) -> Filter {
        return SimpleFilter(op: "$in", name: key, value: values)
    }
    
    /**
     * Creates a query filter that performs `(key IN [value1, value2, ...])`.
     *
     * - parameter key: The key to check.
     *
     * - parameter values: The list of values to compare with.
     *
     * - returns: A new filter `{ key: { "$in" : [value1, value2, ...] }}`.
     */
    public static func _in(key: String, values: [Bool]) -> Filter {
        return SimpleFilter(op: "$in", name: key, value: values)
    }
    
    /**
     * Creates a query filter that performs `(op1 != op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$ne": op2 } }`.
     */
    public static func ne(op1: String, op2: String) -> Filter {
        return SimpleFilter(op: "$ne", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 != op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$ne": op2 } }`.
     */
    public static func ne(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$ne", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 != op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$ne": op2 } }`.
     */
    public static func ne(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$ne", name: op1, value: op2);
    }
    /**
     * Creates a query filter that performs `(op1 != op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$ne": op2 } }`.
     */
    public static func ne(op1: String, op2: Bool) -> Filter {
        return SimpleFilter(op: "$ne", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &lt;= op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$lte": op2 } }`.
     */
    public static func lte(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$lte", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &lt;= op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$lte": op2 } }`.
     */
    public static func lte(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$lte", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &gt;= op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$gte": op2 } }`.
     */
    public static func gte(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$gte", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &gt;= op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$gte": op2 } }`.
     */
    public static func gte(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$gte", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &lt; op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$lt": op2 } }`.
     */
    public static func lt(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$lt", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &lt; op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$lt": op2 } }`.
     */
    public static func lt(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$lt", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &gt; op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$gt": op2 } }`.
     */
    public static func gt(op1: String, op2: CLongLong) -> Filter {
        return SimpleFilter(op: "$gt", name: op1, value: op2);
    }
    
    /**
     * Creates a query filter that performs `(op1 &gt; op2)`.
     *
     * - parameter op1: First operand.
     *
     * - parameter op2: Second operand.
     *
     * - returns: A new filter `{ op1: { "$gt": op2 } }`.
     */
    public static func gt(op1: String, op2: Double) -> Filter {
        return SimpleFilter(op: "$gt", name: op1, value: op2);
    }
    
    // TODO: Only need toJson no need for a method that returns a String.
    /**
     * Return the JSON object for this query filter.
     * 
     * - returns: a JSON representation of the filter.
     */
    func toJson() -> JSON {
        fatalError("Subclass need to implement the 'toJson()' method.")
    }
    
    func toJsonURI() -> String {
        fatalError("Subclass need to implement the 'toJsonURI()' method.")
    }
    
    func isEmpty() -> Bool {
        fatalError("Subclass need to implement the 'isEmpty()' method.")
    }

}
