/*
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package com.oracle.iot.client

import Foundation

/**
 * StorageObject provides information about content in cloud storage.
 */
public class AdvancedStorageObject: AdvancedExternalObject {

    private var name:String!
    private var contentType:String!
    private var encoding:String?
    private var date:String?
    private var length:Int = -1
    private var inputStream:InputStream?
    private var outputStream:OutputStream?
    private var timeToLive:Int = Config.STORAGE_TTL
    private var mutable:Bool = true
    private var metadata:[String:String] = [String:String]()

    /**
     * Create an `AdvancedStorageObject`.
     *
     * - parameter uri: the full URL of the object in the Storage Cloud
     *
     * - parameter name: name of the object used in the Storage Cloud
     *
     * - parameter contentType: the content type of the object, or `nil`
     *
     * - parameter encoding: the encoding of the object, or `nil`
     *
     * - parameter date: the last modified date of the object
     *
     * - parameter length: length of the object
     *
     * - parameter mutable: mutable `true` means mutable, `false` means immutable
     *        If `mutable` == `false`, then the AdvancedStorageObject will be made
     *        immutable by setting it’s *Write-Once-Read-Many* (WORM) policy when
     *        uploading it to the container to prevent the users
     *        from deleting it for a 7 days or modifying it.
     *        Number of days is 7 by default and can be overridden by setting property
     *        `Config.STORAGE_WORM_DAYS`. If the property is "0" then no WORM policy
     *        will be applied.
     */
    init(uri:String, name:String, contentType:String?, encoding:String?,
         date:String?, length:Int, mutable:Bool=true) {
        super.init(uri:uri)
        self.name = name
        self.contentType = contentType ?? StorageObject.OCTET_STREAM
        self.encoding = encoding
        self.date = date
        self.length = length
        self.mutable = mutable
    }

    /**
     * Get the the name of this object in the storage cloud. This is name and
     * path of the file that was uploaded to the storage cloud.
     * - returns: the name of this object in the storage cloud
     */
    public func getName() -> String {
        return name
    }

    /**
     * Get the mime-type of the content. See [ANA Media Types](
     * http://www.iana.org/assignments/media-types/media-types.xhtml)
     *
     * - returns: The mime-type of the content
     */
    public func getContentType() -> String {
        return contentType
    }

    /**
     * Get the compression scheme of the content.
     * - returns: the compression scheme of the content, or `nil` if the
     *            content is not compressed
     */
    public func getEncoding() -> String? {
        return encoding
    }

    /**
     * Get the length of the content in bytes. This is the number of bytes 
     * required to upload or download the content.
     *
     * - returns: The length of the content, or -1 if unknown
     */
    public func getLength() -> Int {
        return length
    }

    /**
     * Get the date and time the content was created or last modified in cloud 
     * storage. This may be `nil` if the content has not been uploaded.
     * The date and time stamp format is `ISO 8601`.
     *
     * - returns: The date the content was last modified in cloud storage, or
     *            `nil` if the content has not been uploaded.
     */
    public func getDate() -> String? {
        return date
    }

    /**
     * Called by the library to obtain the input stream when uploading content
     *
     * - returns: The input stream for the content being uploaded
     */
    public func getInputStream() -> InputStream? {
        return inputStream
    }

    /**
     * Set an input stream for content to be uploaded. The implementation
     * allows for either the input stream to be set, or the output stream to be 
     * set, but not both. If `inputStream` is not `nil`,
     * the output stream will be set to `nil.
     *
     * - parameter inputStream: A stream from which the content will be read
     *
     * - throws: `ClientError.state` if a transfer is currently in progress
     */
    public func setInputStream(inputStream:InputStream?) {
        self.inputStream = inputStream
        if inputStream != nil {
            self.outputStream = nil
        }
    }

    /**
     * Called by the library to obtain the output stream when downloading
     * content
     *
     * - returns: The output stream for the content being downloaded
     */
    public func getOutputStream() -> OutputStream? {
        return outputStream
    }

    /**
     * Set an output stream for content to be downloaded. The implementation
     * allows for either the input stream to be set, or the output stream to be 
     * set, but not both. If `outputStream` is not `nil`, the input stream will 
     * be set to `nil`.
     *
     * - parameter outputStream A stream to which the content will be written
     *
     * - throws: `ClientError.state` if a transfer is currently in progress
     */
    public func setOutputStream(outputStream:OutputStream?) {
        self.outputStream = outputStream
        if outputStream != nil {
            self.inputStream = nil
        }
    }
    
    /**
     * Used internally by the client library to update fields after upload.
     * Not intended for general use.
     * - parameter date: the last-modified date for the object in the Storage
     *                    Cloud
     * - parameter length: length the length for the object in the Storage 
     *                      Cloud
     */
    func setAttributes(date:String, length:Int) {
        self.date = date
        self.length = length
    }
    
    /**
     * Synchronize content with the Storage Cloud Service.
     *
     * - parameter callback: `callback` is called when the network activity is
     *                       complete.
     *
     * - throws: `ClientError.network` if there is an raised by the runtime,
     *                                 or a failure reported by the storage cloud
     *      `ClientError.security` if there is an exception establishing a 
     *                             secure connection to the storage cloud
     *      `ClientError.argument` if the `StorageObject` does not return an 
     *                             `InputStream` or `OutputStream`
     */
    public func sync(callback: @escaping (ClientError?) -> ()) throws {
        fatalError("Subclass need to implement the 'sync()' method.")
    }
    
    /**
     * Returns the mutable state of the StorageObject.
     *
     * - returns: `true` if the `StorageObject` is mutable.
     */
    public func isMutable() -> Bool {
        return self.mutable
    }

    /**
     * Returns the time to live of the StorageObject when creating it in the
     * Storage Cloud Service.
     *
     * - returns: the time to live in seconds
     */
    public func getTimeToLive() -> Int {
        return self.timeToLive
    }
    
    /**
     * Set the time to live for Storage Object in seconds
     * Time to live will be sent to the Storage Object when uploadind
     * a file.
     *
     * - parameter timeToLive: the time to live in seconds
     *
     * - throws: `ClientError.argument` if `timeToLive` is not greater than zero
     */
    public func setTimeToLive(timeToLive:Int) throws {
        guard timeToLive > 0 else {
            throw ClientError.argument(NSLocalizedString(
                "storage object time to live must be greater than zero", comment:""))
        }
        self.timeToLive = timeToLive
    }
    
    /**
     * Sets the metadata for the `AdvancedStorageObject`.
     * All metdata will be added to Storage Cloud Service as custom metadata
     * with the *X-Object-Meta-KeyName* header.
     *
     * - parameter key: The metadata key
     *
     * - parameter value: The metadata value
     *
     * - throws: `ClientError.argument` if the key or value is empty
     */
    public func setCustomMetadata(key:String, value:String) throws {
        guard !key.isEmpty && !value.isEmpty else {
            throw ClientError.argument(String(format:NSLocalizedString(
                "The metadata key '%@' and value '%@' must not be empty", comment:""),
                                              key, value))
        }
        metadata[key] = value
    }
    
    /**
     * This method return unmodifiable copy of metadata.
     *
     * @return Map of metadata. Map may be empty.
     */
    public func getCustomMetadata() -> [String:String] {
        return self.metadata
    }
}
