/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * DeviceModelAction is the model of an action in a *DeviceModel*.
 */
public class DeviceModelAction {
    
    internal var name: String?
    internal let description: String?
    internal let argType: DeviceModelAttributeType?
    internal var lowerBound: NSNumber?
    internal var upperBound: NSNumber?
    internal let alias: String?
    
    public init(name: String, description: String,
                type: DeviceModelAttributeType?,
                lowerBound: NSNumber?, upperBound: NSNumber?,     
                alias: String) {
        self.name = name
        self.description = description
        self.argType = type
        if type == nil {
            self.lowerBound = nil
            self.upperBound = nil
        } else {
            switch self.argType! {
            case .INTEGER:
                self.lowerBound = lowerBound
                self.upperBound = upperBound
                break
            case .NUMBER:
                self.lowerBound = lowerBound
                self.upperBound = upperBound
                break
            default:
                self.lowerBound = nil
                self.upperBound = nil
            }
        }
        self.alias = alias
    }
    
    public convenience init(name: String, description: String, alias: String) {
        self.init(name: name, description: description, type: nil,
                  lowerBound: nil, upperBound: nil, alias: alias)
    }
    
    /**
     * Get the action name.
     *
     * - returns: the action name from the device model
     */
    public func getName() -> String {
        return self.name!
    }
    
    /**
     * A human friendly description of the action. If the model does not
     * contain a description, this method will return an empty String.
     *
     * - returns: the action description, or an empty string
     */
    public func getDescription() -> String {
        return self.description!
    }
    
    /**
     * The data type of the argument to the action. If the action does not
     * take an argument, then this method will return nil.
     *
     * - returns: the action argument's data type, or nil
     */
    public func getArgType() -> DeviceModelAttributeType? {
        return self.argType
    }
    
    /**
     * For *DeviceModelAttributeType.NUMBER* and
     * *DeviceModelAttributeType.INTEGER only, give the lower bound of the
     * acceptable range of values for the action's argument.
     * nil is always returned for actions other than *.NUMBER*
     * or *.INTEGER* type.
     *
     * - returns: a Number, or nil if no lower bound has been set
     */
    public func getLowerBound() -> NSNumber? {
        return self.lowerBound
    }
    
    /**
     * For *DeviceModelAttributeType.NUMBER* and
     * *DeviceModelAttributeType.INTEGER only, give the upper bound of the
     * acceptable range of values for the action's argument.
     * nil is always returned for actions other than *.NUMBER*
     * or *.INTEGER* type.
     *
     * - returns: a Number, or nil if no upper bound has been set
     */
    public func getUpperBound() -> NSNumber? {
        return self.upperBound
    }
    
    /**
     * An alternative name for the action. This optional action provides a
     * more descriptive name for the action. For example, "temperature" might
     * be an alias for the action name "5700". The alias can be used where ever
     * the action name would be used in the *AbstractVirtualDevice* API.
     *
     * -returns: The action alias.
     */
    public func getAlias() -> String? {
        return self.alias!
    }
    
    public func toString() -> String {
        var str: String = "name = " + name!
        str = str + ", description = " + description!
        if(argType != nil) { str = str + ", type = " + (argType?.rawValue)! }
        if(lowerBound != nil) { str = str + ", lowerBound = " + (lowerBound?.description)! }
        if(upperBound != nil) {str = str + ", upperBound = " + (upperBound?.description)! }
        str = str + ", alias = " + alias!
        return str
    }
}
