/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

fileprivate var DOCUMENT_DIRECTORY_URL:URL? = nil

/**
 * DeviceModel Factory.
 */
class DeviceModelFactory {

    static let DEVICEMODEL_FILENAME_FORMAT:String = "dm-%@.json"
    static let ALLOW_DRAFT_MODELS:Bool = Config.ALLOW_DRAFT_DEVICE_MODELS
    
    private init() {}
    
    /**
     * Get a device model from the IoT server.
     *
     * - parameter secureConnection: a connection to the server.
     *
     * - parameter urn: the URN of desired device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be `nil` and an
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the
     *                       reason can be taken from error
     */
    static func getDeviceModel(secureConnection: HttpSecureConnection,
                               urn: String,
                               callback: @escaping (DeviceModel?, ClientError?) -> Void) throws {
        // TODO: FIXME Does not really throw
        try getDeviceModel(aid: nil, secureConnection: secureConnection,
                           urn: urn, callback: callback)
    }
    
    private static func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data,
                                        options: []) as? [String : AnyObject]
        }
        return nil
    }
    
    /**
     * Get a device model from the IoT server.
     *
     * - parameter aid: the IoT application identifier
     *
     * - parameter secureConnection: a connection to the server.
     *
     * - parameter urn: the URN of desired device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be `nil` and an 
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the
     *                       reason can be taken from error
     *
     * On error `callback` receives a `ClientError.network` error if:
     *    the object is not found (StatusCode.NOT_FOUND),
     *    the request fails (!StatusCode.OK), or
     *    the returned data is nil or JSON parsing fails
     */
    static func getDeviceModel(aid: String?, secureConnection: HttpSecureConnection,
            urn: String, callback: @escaping (DeviceModel?,ClientError?)->()) throws {
        
        // TODO: Do we really need to encode the filename ?
        // let filename:String = urn.addingPercentEncoding(withAllowedCharacters: .urlPathAllowed)
        let modelFilename:String = String(format:DEVICEMODEL_FILENAME_FORMAT, urn)
        if let modelJson:String = getFileFromLocalStore(filename:modelFilename) {
            do {
                let dm:DeviceModel =
                    try DeviceModelParser.fromJson(root: JSON(parseJSON:modelJson))
                callback(dm, nil)
                return
            } catch {
                print (error)
                callback(nil, ClientError.file(NSLocalizedString("\(error)", comment: "")))
            }
        }
        
        let deviceModelApiFormat:String =
            "%@/apps/%@/deviceModels?q={\"urn\":{\"$eq\":\"%@\"}}"
        
        // Download device model from server and cache it before returning to caller
        var uri: String
        if aid != nil {
            uri = String(format:deviceModelApiFormat,
                         RestApi.V2.getReqRoot(), aid!, urn)
        } else {
            /*
             * TODO: The server does not allow devices to list device
             *       models and use a filter like enterprise clients can,
             *       so we filter in the method below, when the server
             *       allows it, use the the code below instead.
             *
             * Filter f = Filter.and(Filter.eq("urn", urn),
             *            Filter.eq("draft", false));
             * uri = RestApi.V2.getReqRoot() +  "/deviceModels"
             * uri = uri + "&q=" + f.toJson().toString()
             */
            uri = RestApi.V2.getReqRoot() + "/deviceModels/" + urn
        }

        // TODO: FIXME exceptions are thrown from with the secureConnection 
        // completion handler
        // TODO: FIXME http errors are checked on return from 
        // callback not before.
        let accessToken = secureConnection.accessToken
        let headers: [String : String] = [
            "Content-Type" : "application/json",
            "Accept" : "application/json",
            "X-EndpointId" : secureConnection.trustedAssetsManager.getClientId(),
            "Authorization" : accessToken.getTokenType() + " " +
                accessToken.getToken()
        ]
        
        try secureConnection.get(restApi: uri, headers: headers,
                                 response: { httpResponse, err in
            if err != nil {
                callback(nil, err)
                return
            }
            guard let httpObject = httpResponse else {
                callback(nil, ClientError.argument(NSLocalizedString(
                    String(format:"DeviceModelFactory.getDeviceModel: " +
                        "cannot get device model %@, GET returned nil httpObject",
                        urn), comment: "")))
                return
            }
            
            do {
                switch(httpObject.getStatus()) {
                case StatusCode.OK.getCode():
                    let data:Foundation.Data = httpObject.getData()
                    if data.isEmpty {
                        callback(nil, ClientError.state(NSLocalizedString(
                            String(format:"DeviceModelFactory.getDeviceModel: " +
                            "cannot get device model %@ data is empty", urn),
                            comment: "")))
                        break
                    }
                    var dmj = try JSON(data: data)
                    if let items = dmj.dictionaryValue["items"]?.array {
                        if items.count > 0 {
                            dmj = items[0]
                        }
                    }
                    let dm:DeviceModel = try DeviceModelParser.fromJson(root: dmj)
                    if let dmString = dmj.rawString() {
                        // Don't fail if the device model cannot be cached
                        if !saveFileToLocalStore(filename:modelFilename,
                                                 content:dmString) {
                            log(type:.fault,
                                message:"DeviceModelFactory.getDeviceModel: " +
                                    "cannot cache deviceModel %@",
                                args:urn)
                        }
                    }
                    callback(dm, nil)
                break
                case StatusCode.NOT_FOUND.getCode():
                    callback(nil, ClientError.network(NSLocalizedString(
                        String(format:"DeviceModelFactory.getDeviceModel: " +
                            "device model %@ not found.", urn), comment: "")))
                break
                default:
                    callback(nil, ClientError.network(NSLocalizedString(
                        String(format:"DeviceModelFactory.getDeviceModel: " +
                            "request failed for device model %@ with status code %d.",
                            urn, httpObject.getStatus()), comment: "")))
                    break
                }
            } catch {
                callback(nil, ClientError.network(NSLocalizedString(
                    String(format:"DeviceModelFactory.getDeviceModel: " +
                        "cannot get device model %@: %@", urn, "\(error)"),
                    comment: "")))
            }
        })
    }
}
