/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Alamofire
import SwiftyJSON

public class SecureConnection : Hashable {

    public var hashValue:Int {
        var hash:Int = 37
        hash = 37 &* hash &+ self.trustedAssetsManager.getServerScheme().hashValue
        hash = 37 &* hash &+ self.trustedAssetsManager.getServerHost().hashValue
        hash = 37 &* hash &+ self.trustedAssetsManager.getServerPort().hashValue
        return hash
    }
    
    public static func ==(lhs:SecureConnection, rhs:SecureConnection) -> Bool {
        return lhs === rhs || (
            lhs.trustedAssetsManager.getServerScheme() ==
                rhs.trustedAssetsManager.getServerScheme() &&
            lhs.trustedAssetsManager.getServerHost() ==
                rhs.trustedAssetsManager.getServerHost() &&
            lhs.trustedAssetsManager.getServerPort() ==
                rhs.trustedAssetsManager.getServerPort())
     }
    
    // TODO: Why is "unowned" needed ? Causes trustedAssetsManager reference to
    // get lost in the EnterpriseLib
    private var _trustedAssetsManager:TrustedAssetsManager!
    /*unowned*/ var trustedAssetsManager:TrustedAssetsManager {
        get {
            return _trustedAssetsManager
        }
        set {
            _trustedAssetsManager = newValue
        }
    }

    private var _useOnlySharedSecret:Bool = false
    var useOnlySharedSecret: Bool {
        get {
            return _useOnlySharedSecret
        }
        set {
            _useOnlySharedSecret = newValue
        }
    }
    // TODO: From IOT-37050
    private var accessTokenQueue: DispatchQueue =
        DispatchQueue(label:"Access Token Queue",
                      qos:.utility)
    private var _accessToken:AccessToken
    var accessToken: AccessToken {
        get {
            return accessTokenQueue.sync(execute: {
                return _accessToken.clone()
            })
        }
        set {
            accessTokenQueue.sync() {
                _accessToken = newValue
            }
        }
    }
    
    init(trustedAssetsManager:TrustedAssetsManager, useOnlySharedSecret:Bool) {
        self._accessToken = AccessToken(expires: 0, tokenType: "", token: "")
        self.trustedAssetsManager = trustedAssetsManager
        self._useOnlySharedSecret = useOnlySharedSecret
        self._accessToken = AccessToken(expires: 0, tokenType: "", token: "")
    }
    
    // TODO: Convert uses of get with headers to this method.
    // headers code is duplicated everywhere
    func get(restApi: String,
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        fatalError("Subclass must implement 'get'")
    }
    
    func get(restApi: String, headers: [String : String],
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        fatalError("Subclass must implement 'get'")
    }
    
    func post(restApi: String, payload: Foundation.Data?, headers: [String : String],
              response: @escaping ((HttpResponse?, ClientError?) -> ())) throws{
        fatalError("Subclass must implement 'post'")
    }
    
    func put(restApi: String, payload: Foundation.Data, headers: [String : String],
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        fatalError("Subclass must implement 'put'")
    }
    
    func delete(restApi: String, headers: [String : String],
                response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws {
        fatalError("Subclass must implement 'delete'")
    }
    
    func patch(restApi: String, payload: Foundation.Data, headers: inout [String : String],
               response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        fatalError("Subclass must implement 'patch'")
    }
    
    func isClosed() -> Bool {
        fatalError("Subclass must implement 'isClosed'")
    }

    public func getTrustedAssetsManager() -> TrustedAssetsManager {
        return trustedAssetsManager
    }
    
    public func getEndpointId() -> String? {
        return trustedAssetsManager.getEndpointId()
    }
    
    // Closeable "override"
    public func close() throws {
        disconnect()
        self.trustedAssetsManager.close()
    }
    
    public func disconnect() {
    }
    
    func usesOnlySharedSecret() -> Bool {
        return useOnlySharedSecret
    }
    
}

class HttpSecureConnection : SecureConnection {

    var defaultHttpManager:Alamofire.SessionManager?
    // Create a DispatchQueue for executing the Alamofire request
    // completion handler. It must not be on the main thread. Note that
    // since it is not on the main thread it cannot update UI components
    // All requests must set the queue parameter to this queue.
    // See Transport.swift
    var responseQueue:DispatchQueue {
        get {
            return DispatchQueue(label:"HttpSecureConnection response queue",
                          qos: .utility)
        }
    }
    
    var closed:Bool = true
    
    let syncQueue:DispatchQueue = DispatchQueue.global(qos: .utility)
    
    let renewAccessTokenGroup:DispatchGroup = DispatchGroup()
    let renewAccessTokenQueue:DispatchQueue = DispatchQueue(label:"Renew Access Token Queue",
                                                            qos:.utility)
    private func waitForAccessToken() {
        //print(stackTrace())
        renewAccessTokenGroup.wait()
    }
    private func getAccessToken() {
        //print(stackTrace())
        //print("\n#### Requesting access token ######\n")
        renewAccessTokenGroup.enter()
    }
    
    override
    init(trustedAssetsManager: TrustedAssetsManager, useOnlySharedSecret: Bool) {
        super.init(trustedAssetsManager:trustedAssetsManager, useOnlySharedSecret:useOnlySharedSecret)
        let configuration = URLSessionConfiguration.default
        if (!Config.DISABLE_LONG_POLLING) {
            configuration.timeoutIntervalForRequest = Config.LONG_POLLING_TIMEOUT
            configuration.timeoutIntervalForResource = Config.LONG_POLLING_TIMEOUT
        }
        // else {
        //  keep default Alamofire's timeout (60 seconds)
        // }
        
        if (self.trustedAssetsManager.getServerTrustPolicy() != nil) {
            let serverTrustPolicies: [String: ServerTrustPolicy] =
                [self.trustedAssetsManager.getServerHost(): self.trustedAssetsManager.getServerTrustPolicy()!]
            
            configuration.httpAdditionalHeaders = Alamofire.SessionManager.defaultHTTPHeaders
            
            self.defaultHttpManager = Alamofire.SessionManager(
                configuration: configuration,
                serverTrustPolicyManager: ServerTrustPolicyManager(policies: serverTrustPolicies)
            )
        } else {
            self.defaultHttpManager = Alamofire.SessionManager(configuration: configuration)
        }
        closed = false
    }
    
    deinit {
        self.defaultHttpManager = nil
    }
    override
    func get(restApi: String,
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        let headers: [String : String] = [
            "Content-Type" : "application/json" ,
            "Accept" : "application/json",
        ]
        try invoke(method: "GET", restApi: restApi,
                   payload: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0),
                   headers: headers, response: { (invokeGet, error) in
            response(invokeGet, error)
        })
    }
    override
    func get(restApi: String, headers: [String : String],
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        try invoke(method: "GET", restApi: restApi,
                   payload: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0),
                   headers: headers, response: { (invokeGet, error) in
            response(invokeGet, error)
        })
    }
    override
    func post(restApi: String, payload: Foundation.Data?, headers: [String : String],
              response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        try invoke(method: "POST", restApi: restApi, payload: payload ,
                   headers: headers, response: { (invokePost, error) in
            response(invokePost, error)
        })
    }
    override
    func put(restApi: String, payload: Foundation.Data, headers: [String : String],
             response: @escaping ((HttpResponse?, ClientError?) -> ())) throws  {
        try invoke(method: "PUT", restApi: restApi, payload: payload, headers: headers,
                   response: { (invokePut, error) in
            response(invokePut, error)
        })
    }
    override
    func delete(restApi: String, headers: [String : String],
                response: @escaping ((HttpResponse?, ClientError?) -> ())) throws {
        try invoke(method: "DELETE", restApi: restApi,
                   payload: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0),
                   headers: headers, response: { (invokeDelete, error) in
            response(invokeDelete, error)
        })
    }
    override
    func patch(restApi: String, payload: Foundation.Data, headers: inout [String : String],
               response: @escaping ((HttpResponse?, ClientError?) -> ())) throws  {
        headers["X-HTTP-Method-Override"] = "PATCH"
        try invoke(method: "POST", restApi: restApi, payload: payload, headers: headers,
                   response: { (invokePatch, error) -> () in
            response(invokePatch, error)
        })
    }
    
    var tokenRequest: Int = 0

    // Should be private but can't be because of tests
    func invoke(method: String, restApi: String, payload: Foundation.Data?,
                headers: [String : String],
                response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws  {
        
        // Check for valid access token
        // First get a copy of the token, it blocks if it is being updated.
        let token = accessToken
        if !token.getToken().isEmpty && !token.hasExpired() {
            self.callRestMethod(method: method, restApi: restApi, payload: payload,
                                headers: headers, callback: response)
        } else {
            tokenRequest += 1
            let tr: Int = tokenRequest
            renewAccessTokenQueue.async(execute: {
                do {
                    // block if there are any pending renew access token
                    // requests
                    //self.renewAccessTokenGroup.wait()
                    self.waitForAccessToken()
                    // When unblocked check again for an access token
                    // If there is one, run the rest api call.
                    // If not call renew accessToken
                    let token = self.accessToken
                    if !token.getToken().isEmpty && !token.hasExpired() {
                        self.callRestMethod(method: method, restApi: restApi, payload: payload,
                                            headers: headers, callback: response)
                    } else {
                        try self.renewAccessToken(tr: tr, callback: { (error) in
                            if error != nil {
                                response(nil, error)
                                return
                            }
                            self.callRestMethod(method: method, restApi: restApi,
                                                payload: payload, headers: headers,
                                                callback: response)
                        })
                    }
                } catch let error as ClientError {
                    response(nil, error)
                } catch {
                    response(nil, ClientError.network("\(error)"))
                }
            })
        }
    }
    
    var expTokenReq: Int = 100
    
    private func callRestMethod(method: String, restApi: String, payload: Foundation.Data?,
                                headers: [String : String],
                                callback: @escaping ((HttpResponse?, ClientError?) -> Void)) {
        
        let serverHost: String = getTrustedAssetsManager().getServerHost()
        let serverPort: String = getTrustedAssetsManager().getServerPort()
        var URL: String = "\(getTrustedAssetsManager().getServerScheme())://\(serverHost):\(serverPort)\(restApi)"
        
        var _headers = headers
        let token = accessToken
        _headers["Authorization"] = token.getTokenType() + " " + token.getToken()
        if RestApi.V2.isWebApi() == false {
            if !self.getTrustedAssetsManager().getEndpointId().isEmpty {
                _headers["X-EndpointId"] = self.getTrustedAssetsManager().getEndpointId()
            }
        }
        
        URL = URL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
        
        let httpClient = HttpClient(url: URL, httpManager: self.defaultHttpManager!)
        
        self.invoke(httpClient: httpClient, method: method, payload: payload,
                    headers: _headers, callback: { (firstResponse, error) -> Void in
            if error != nil {
                callback(nil, error)
                return
            }
            if firstResponse == nil {
                callback(nil, ClientError.argument(NSLocalizedString(
                    "HttpSecureConnection callRestMethod() firstResponse should not be nil",
                    comment: "")))
                return
            }
            
            // Check if credentials have expired (401) or client-secret used (403)
            if firstResponse!.getStatus() != StatusCode.UNAUTHORIZED.getCode() &&
                firstResponse!.getStatus() != StatusCode.FORBIDDEN.getCode() {
                callback(firstResponse, nil)
                return
            }
            // Need to get a new token
            // queue an access token request
            self.renewAccessTokenQueue.async(execute: {
                do {
                    self.expTokenReq += 1
                    let tr:Int = self.expTokenReq
                    
                    // block if there are any pending renew access token
                    // requests
                    //self.renewAccessTokenGroup.wait()
                    self.waitForAccessToken()
                    // Check to see if we have a different token
                    let newToken = self.accessToken
                    let tokenAuth = newToken.getTokenType() + " " + newToken.getToken()
                    
                    // Still have the same token, get a new one
                    if _headers["Authorization"] == tokenAuth {
                        try self.renewAccessToken(tr: tr, callback: { (error) in
                            if error != nil {
                                callback(nil, error)
                                return
                            }
                            // Update the header with the new token
                            let token = self.accessToken
                            _headers["Authorization"] = token.getTokenType() + " " +
                                token.getToken()
                            
                            self.invoke(httpClient: httpClient, method: method,
                                        payload: payload, headers: _headers,
                                        callback: { (response, error) -> Void in
                                            if (error != nil) {
                                                callback(nil, error)
                                                return
                                            } else {
                                                callback(response, nil)
                                                return
                                            }
                            })
                        })
                    } else {
                        // Already have a new token use it
                        _headers["Authorization"] = tokenAuth
                        self.invoke(httpClient: httpClient, method: method,
                                    payload: payload, headers: _headers,
                                    callback: { (response, error) -> Void in
                                        if (error != nil) {
                                            callback(nil, error)
                                            return
                                        } else {
                                            callback(response, nil)
                                            return
                                        }
                        })
                    }
                } catch let error as ClientError {
                    callback(nil, error)
                } catch {
                    callback(nil, ClientError.network(NSLocalizedString(
                        "\(error.localizedDescription)", comment: "")))
                }
            })
        })
    }
    
    private func invoke(httpClient: HttpClient, method: String, payload: Foundation.Data?, headers: [String : String], callback: @escaping ((HttpResponse?, ClientError?) -> Void)) -> Void {
        do {
            if method == "GET" {
                try httpClient.get(headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else if method == "POST" {
                var ipayload:Foundation.Data? = payload
                if(payload == nil) {
                    ipayload = String("[]").data(using: .utf8)!
                }
                try httpClient.post(data: ipayload!, headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else if method == "DELETE" {
                try httpClient.delete(headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else { // Must be "PUT"
                try httpClient.put(data: payload!, headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            }
        } catch {
            callback(nil, ClientError.state(NSLocalizedString("Error while executing HTTP method: " + method, comment: "")))
        }
        
    }

    private func renewAccessToken(tr: Int, callback: @escaping ((ClientError?) -> ())) throws {
        // Enter the group to block any incoming requests that need the access token
        //renewAccessTokenGroup.enter()
        getAccessToken()
        
        let trustedAssetsManager = self.getTrustedAssetsManager()
        let serverHost = trustedAssetsManager.getServerHost()
        let serverPort = trustedAssetsManager.getServerPort()
        
        RestApi.USE_WEBAPI = false
        let restAPI = RestApi(version: "v2")
        
        // build the URL String
        let oauthRequestURL:String = trustedAssetsManager.getServerScheme() + "://" +
            serverHost + ":" + serverPort + restAPI.getReqRoot() + trustedAssetsManager.getOAuthRequestPath()
        
        let headers: [String:String] = ["Content-Type" : "application/x-www-form-urlencoded", "Accept" : "application/json"]
        // create the client Assertion
        let client_asertion:String?
        var useSharedSecret:Bool = true
        
        if (trustedAssetsManager.isActivated() && !trustedAssetsManager.getClientId().isEmpty) {
            useSharedSecret = false
        }
        
        // sha
        client_asertion = try self.buildClientAssertion(tam: trustedAssetsManager, secureHttpConnection: self,
                                                        useOnlySharedSecret: useSharedSecret)
        
        // generate the OAuth2 parameters; parameters must come from the TAM
        let oauthParameters: [String:String] = ["grant_type" : "client_credentials",
                                                "client_assertion_type" : "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
                                                "client_assertion" : client_asertion!,
                                                "scope" : trustedAssetsManager.isActivated() ? "":"oracle/iot/activation"]
        
        Config.postCounter += 1
        var getToken = {}
        getToken = {
            // Always check to make sure an access token request has not just
            // been completed. This block will be on a serial queue for access
            // token requests. When the access token is available, there should
            // not be much delay between the serialized requests since each
            // execution will return immediately. And new "invoke" requests
            // will see the access token and not be placed in the queue.
            let token = self.accessToken
            if !token.getToken().isEmpty && !token.hasExpired() {
                self.renewAccessTokenGroup.leave()
                callback(nil)
                return
            }
            self.defaultHttpManager!.request(oauthRequestURL,
                                             method: .post,
                                             parameters: oauthParameters,
                                             encoding: URLEncoding.default,
                                             headers: headers)
                .validate(statusCode: 200..<404)
                .responseJSON(queue: self.responseQueue, completionHandler: { (response) -> Void in
                    var clientError: ClientError? = nil
                    switch response.result {
                    case .success(let value):
                        if response.response?.statusCode == StatusCode.BAD_REQUEST.getCode() ||
                            response.response?.statusCode == StatusCode.FORBIDDEN.getCode() ||
                            response.response?.statusCode == StatusCode.UNAUTHORIZED.getCode() {
                            self.nextTry += 1
                            let responseTime = HttpSecureConnection.parseResponseTime(responseData: response.data!)
                            TimeManager.setCurrentTimeMillis(serverTime: responseTime)
                            if self.nextTry < 2 {
                                getToken()
                                return
                            } else {
                                clientError = ClientError.network(NSLocalizedString("Failed to renew access token after 2 attempts",
                                                                                    comment: ""))
                            }
                        } else {
                            let json = JSON(value)
                            if ((json["error"] != JSON.null) || (json["error_description"] != JSON.null)) {
                                clientError = ClientError.network(NSLocalizedString(
                                    "Renew token error: " + json["error_description"].string!, comment: ""))
                            } else if let myAccessToken = json["access_token"].string,
                                let myTokenType = json["token_type"].string,
                                let myExpiresIn = json["expires_in"].int64 {
                                
                                let token: AccessToken = AccessToken(expires: 0, tokenType: "", token: "")
                                token.token = myAccessToken
                                token.tokenType = myTokenType
                                token.expires = myExpiresIn
                                token.expirationTime += token.expires
                                self.accessToken = token
                                
                                self.nextTry = 0
                                
                            } else {
                                self.nextTry = 0
                                clientError = ClientError.network(NSLocalizedString("Renew token: retrieved JSON error", comment: ""))
                            }
                        }
                    case .failure(let error):
                        print(error)
                        self.nextTry = 0
                        clientError = ClientError.network(NSLocalizedString("Renew token: JSON response validation failed with error " + error.localizedDescription, comment: ""))
                    }
                    self.renewAccessTokenGroup.leave()
                    callback(clientError)
                })
        }
        getToken()
    }

    override
    func close() { 
        // TODO: In case we need to wait for tasks to finish then use
        // finishTasksAndInvalidate() but be aware method returns immediately
        // => handle callback from SessionDelegate
        syncQueue.sync() {
            // TODO: See if this is causing intermittent "invalidated sessions"
            //self.defaultHttpManager?.session.invalidateAndCancel()
            self.trustedAssetsManager.close()
            disconnect()
            closed = true
        }
    }
    
    override
    func isClosed() -> Bool {
        return syncQueue.sync() {
            return closed
        }
    }
    
    // TODO: Needs synchronization
    override
    func disconnect() {
        self.accessToken = AccessToken(expires: 0, tokenType: "", token: "")
    }
    
    private static func parseResponseTime(responseData: Foundation.Data) -> Int64 {
        var jsonObject:JSON!
        do {
            jsonObject = try JSON(data: responseData)
        } catch {
            print(String(format:NSLocalizedString(
                "%@:JSON threw exception: %@", comment:""),
                "HttpSecureConnection.parseResponseTime()", "\(error)"))
            return TimeManager.currentTimeMillis()
        }

        if jsonObject["currentTime"] == JSON.null {
            return TimeManager.currentTimeMillis()
        } else {
            return jsonObject["currentTime"].int64Value
        }
    }
    
    private var nextTry = 0
    
    // JWT expiration claim adjustment value, in milliseconds. This is added
    // to the current millisecond time since January 1, 1970 in order to get
    // some time "not to far" in the future for the JWT "exp" claim.
    private let EXP_CLAIM_DELTA:Int64 = 15 * 1000 * 60 // 15 minutes
    func buildClientAssertion(tam:TrustedAssetsManager, secureHttpConnection:
        HttpSecureConnection, useOnlySharedSecret:Bool) throws -> String {
        
        var id:String
        let expTime = (TimeManager.currentTimeMillis() + EXP_CLAIM_DELTA) / 1000
        
        if (!useOnlySharedSecret && tam.isActivated()) {
            id = tam.getClientId()
        } else {
            id = tam.getEndpointId()
            if id.isEmpty {
                id = tam.getClientId()
            }
        }
        
        let sharedSecret = useOnlySharedSecret || !tam.isActivated()
        var alg:String  = sharedSecret ? "HS256" : "RS256"
        // TODO: Note was not in Enterprise version
        // sha
        if(tam.isActivated()) {
            alg = "RS256"
        }
        // TODO: swift 3 Use Dictionary not NSDictionary
        // set up the Dictionary for the JSon objects
        let header:NSMutableDictionary = NSMutableDictionary()
        header.setValue("JWT", forKey: "typ")
        header.setValue(alg, forKey: "alg")
        
        let headerStr = self.generateJSONString(para: header) // the header JSON string
        
        var claimStr = "{\"iss\":\"";
        claimStr += id
        claimStr += "\",\"sub\":\""
        claimStr += id
        claimStr += "\",\"aud\":\"oracle/iot/oauth2/token\",\"exp\":"
        claimStr += String(expTime)
        claimStr += "}"
        
        let inputToSign = self.encodeStringBase64String(str: headerStr) + "." + self.encodeStringBase64String(str: claimStr)
        
        var signed:String = ""
        // TODO: The if clause in the Enterprise version included "|| tam.getEndpointId().isEmpty == false)"
        // Is it needed here ?
        // SHA
        if(tam.isActivated() == false /*|| tam.getEndpointId().isEmpty == false*/) {
            
            signed = try tam.signWithSharedSecretString(data: inputToSign, algorithm: "HmacSHA256", activationId: nil)
            
            var data = self.dataFromHexadecimalString(str: signed)
            let count = data!.count / MemoryLayout<UInt8>.size
            
            // create an array of Uint8
            var array = [UInt8](repeating: 0, count: count)
            
            // copy bytes into array
            (data! as Foundation.Data).copyBytes(to: &array, count:count * MemoryLayout<UInt8>.size)
            //Convert [UInt8] to Data
            data = Foundation.Data(bytes: UnsafePointer<UInt8>(array), count: array.count)
            
            //Encode to base64
            let base64Data = data!.base64EncodedData(options: .lineLength64Characters)
            
            if let tempStr = String(data: base64Data, encoding: .utf8) {
                signed = tempStr
            } else {
                print("not a valid UTF-8 sequence")
            }
            
        }
        else {
            signed = try self.trustedAssetsManager.signWithPrivateKeyString(str: inputToSign, algorithm: "PKCS1256")!
        }
        
        var result = inputToSign + "." + signed
        result = result.replacingOccurrences(of: "+", with: "-").replacingOccurrences(of: "/", with: "_")
        
        let nextToLast:String.Index = result.index(before: result.endIndex)
        if result[nextToLast] == "=" {
            result = String(result[result.startIndex..<nextToLast])
        }
        
        return result
    }
    
    func generateJSONString(para:NSMutableDictionary) -> String {
        
        JSONSerialization.isValidJSONObject(para)
        let jsonData = try! JSONSerialization.data(withJSONObject: para, options: JSONSerialization.WritingOptions())
        
        let jsonString = String(data: jsonData, encoding: .utf8)!
        return jsonString
    }
    
    /// - returns: String representing encoded String a BASE64 String
    func encodeStringBase64String(str: String) -> String {
        let utf8str = self.encodeStringBase64(str: str)
        let base64Encoded = utf8str.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
        return base64Encoded
    }
    
    /// - returns: Data representing the UTF8 byte array converted input String
    func encodeStringBase64(str: String) -> Foundation.Data {
        return str.data(using: .utf8)!
    }
    
    /// - returns: Data represented by this hexadecimal string. Returns nil if string contains characters outside the 0-9 and a-f range.
    
    func dataFromHexadecimalString(str: String) -> Foundation.Data? {
        let trimmedString = str.trimmingCharacters(in: CharacterSet(charactersIn: "<> ")).replacingOccurrences(of: " ", with: "")
        
        // make sure the cleaned up string consists solely of hex digits, and that we have even number of them
        
        let regex = try! NSRegularExpression(pattern: "^[0-9a-f]*$", options: .caseInsensitive)
        
        let found = regex.firstMatch(in: trimmedString, options: [], range: NSMakeRange(0, trimmedString.count))
        if found == nil || found?.range.location == NSNotFound || trimmedString.count % 2 != 0 {
            return nil
        }
        
        // everything ok, so now let's build Data
        
        // TODO: swift 3 Don't use "NSMutableData"
        let data = NSMutableData(capacity: trimmedString.count / 2)
        
        var index = trimmedString.startIndex
        while index < trimmedString.endIndex {
            let range = index..<trimmedString.index(index, offsetBy: 2)
            let byteString = trimmedString[range]
            
            let num = UInt8(byteString.withCString { strtoul($0, nil, 16) })
            data?.append([num] as [UInt8], length: 1)
            
            //index = trimmedString.index(after: trimmedString.index(index, offsetBy: 2))
            index = trimmedString.index(index, offsetBy: 2)
        }
        
        return data as Foundation.Data?
    }
    
    static func createHttpSecureConnection(tam: TrustedAssetsManager, useOnlySharedSecret: Bool) -> HttpSecureConnection {
        return HttpSecureConnection(trustedAssetsManager: tam, useOnlySharedSecret: useOnlySharedSecret)
    }
}
