/*
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

import Foundation
import SwiftyJSON
import Alamofire
import CommonCrypto

//package com.oracle.iot.client.impl

/**
 * The `StorageConnection` transfers content to the Storage Cloud Service.
 * There is one StorageConnection instance per client.
 */
public class StorageConnectionBase: StorageConnection {

    // TODO: Need better way to handle and setup defaults
    private static let STORAGE_CLOUD_HOST:String =
        Config.STORAGE_CONNECTION_HOST_NAME
    
    // If true add the "<endpointId>/" prefix to the storage object name
    private static let virtualStorageDirectories:Bool =
        !Config.DISABLE_STORAGE_OBJECT_PREFIX
    
    /**
     * Write-Once-Read-Many (WORM) policy period
     * This policy prevents the users from deleting it for a period
     * (7, by default) days and modifying it.
     */
    
    private static let WORM_EXPIRATION_MAX_DAYS:Int =
        Config.WORM_EXPIRATION_MAX_DAYS
    // only used if Config.MAX_STORAGE_DAYS is invalid
    private static let WORM_EXPIRATION_DEFAULT_DAYS:String = "7"
    private static let WORM_HEADER:String = "X-Worm-Expiration-Days"
    private static let WORM_EXPIRATION_DAYS:String? = {
        let configWED:Int = Config.WORM_EXPIRATION_DAYS
        if configWED == 0 {
            // Don't use WORM Policy
            return nil
        } else if configWED > 0 {
            if configWED > WORM_EXPIRATION_MAX_DAYS {
                return (WORM_EXPIRATION_MAX_DAYS as NSNumber).stringValue
            } else {
                return (configWED as NSNumber).stringValue
            }
        } else {
            // If property contains wrong value set 7 days
            return WORM_EXPIRATION_DEFAULT_DAYS
        }
    }()

    /**
     * REST resource for storage authentication
     */
    private static let REST_STORAGE_AUTHENTICATION:String =
        RestApi.V2.getReqRoot() + "/provisioner/storage"
    private static let AUTH_TOKEN_HEADER:String = "X-Auth-Token"
    private static let RETRY_LIMIT:Int = Config.STORAGE_CONNECTION_RETRY_LIMIT
    private var secureConnection:SecureConnection!
    private var authToken:String? = nil
    
    private var closed:Bool = true
    
    private var defaultHttpManager:Alamofire.SessionManager?
    
    // Serial queue for authentication requests. Whenever a request for an
    // access token is made, a workitem is placed on this queue. Since it is
    // serial, the first work item in gets executed first and will set the
    // authentication token, any queued requests will see that access token
    private var authQueue: DispatchQueue!
    
    // Create a DispatchQueue for executing the Alamofire request
    // completion handler. It must not be on the main thread. Note that
    // since it is not on the main thread it cannot update UI components
    // All requests must set the queue parameter to this queue.
    // See Transport.swift
    private var responseQueue:DispatchQueue {
        get {
            return DispatchQueue(label: "StorageConnectionImpl response queue",
                          qos: .utility)
        }
    }

    private var _storageAuthenticationResponse: StorageAuthenticationResponse?
    private var storageAuthenticationResponse: StorageAuthenticationResponse? {
        get {
            // This should not be possible. Constructor should throw
            // if authenticate fails.
            if _storageAuthenticationResponse == nil {
                do {
                    try authenticate()
                } catch {
                    print("cannot get storageAuthenticationResponse: \(error)")
                    return nil
                }
            }
            return authQueue.sync {
                return _storageAuthenticationResponse?.clone()
            }
        }
        set {
            authQueue.sync {
                _storageAuthenticationResponse = newValue
            }
        }
    }
    
    public init(secureConnection:SecureConnection) {
        self.secureConnection = secureConnection
        self.authQueue = DispatchQueue(
            label: "StorageConnectionImpl Authentication Queue",
            qos: .utility)
        self.closed = false

        super.init()
        self.initHttpManager()

        // TODO: What happens when this call fails ?
        do {
            try authenticate()
        } catch {
            print("authenticate threw exception: \(error)")
        }
    }
    
    private func initHttpManager() {
        let configuration = URLSessionConfiguration.default
        let tam:TrustedAssetsManager = secureConnection.getTrustedAssetsManager()
        if tam.getServerTrustPolicy() != nil {
            let serverTrustPolicies: [String: ServerTrustPolicy] =
                [tam.getServerHost(): tam.getServerTrustPolicy()!]
            
            configuration.httpAdditionalHeaders = Alamofire.SessionManager.defaultHTTPHeaders
            self.defaultHttpManager = Alamofire.SessionManager(
                configuration: configuration,
                serverTrustPolicyManager:
                ServerTrustPolicyManager(policies: serverTrustPolicies))
        } else {
            self.defaultHttpManager = Alamofire.SessionManager(configuration: configuration)
        }
    }

    // returns com.oracle.iot.client.impl.StorageObjectImpl
    public func createStorageObject(clientId:String, name:String,
                                    contentType:String?, mutable:Bool)
            throws -> AdvancedStorageObject {
        if let sar = storageAuthenticationResponse {
            
            var path:String = name
            if StorageConnectionBase.virtualStorageDirectories && !clientId.isEmpty {
                path = clientId + "/" + name;
            }
            
            let contentStorageLocation:String =
                sar.getStorageContainerUrl()! + "/" + path
            
            return createStorageObject(storageContainerUrl:contentStorageLocation,
                            name:name, contentType:contentType,
                            encoding: nil, date: nil, length: -1, mutable:mutable)
        } else {
            throw ClientError.state("cannot obtain storage authentication")
        }
    }
    
    public func createStorageObject(storageContainerUrl:String, name:String,
                             contentType:String?, encoding:String?, date:String?,
                             length:Int, mutable:Bool) -> AdvancedStorageObject {
        fatalError("Subclass must implement createStorageObject()")
    }

    public func createStorageObject(storageUrl:String,
            callback: @escaping (AdvancedStorageObject?, ClientError?)->()) throws {
            
        // Verify that storageUrl is not empty and is a valid url
        guard let _:URL = URL(string: storageUrl) else {
            throw ClientError.argument(NSLocalizedString(
                "Storage Cloud URL is invalid.", comment: ""))
        }
        
        var sar = storageAuthenticationResponse
        if sar == nil {
            throw ClientError.state(NSLocalizedString(
                "Cannot obtain storage authentication", comment: ""))
        }

        // TODO: Does this validation have to happen everytime ?
        // What happens if the URL changes after this process gets the
        // file url ?
        guard let containerUrl:String = sar?.getStorageContainerUrl() else {
            throw ClientError.argument(NSLocalizedString(
                "Storage Cloud Service URL is invalid.", comment: ""))
        }
        
        let fullContainerUrl:String = containerUrl + "/"
        if !storageUrl.hasPrefix(fullContainerUrl) {
            log(type:.fault,
                message:"%@: Storage url '%@' does not match the storage container url %@",
                args: "StorageConnectionBase.createStorageObject()",
                storageUrl, containerUrl)
            throw ClientError.argument(NSLocalizedString(
                "Storage Cloud URL is invalid.", comment: ""))
        }
        
        // storageContainerUrl = <fullContainerUrl>/<moreurl>
        
        // The url path is used for "name" which could have several levels.
        var name:String = String(storageUrl[fullContainerUrl.endIndex..<storageUrl.endIndex])
        if StorageConnectionBase.virtualStorageDirectories {
            let slashsubstring:Range = storageUrl.range(of:"/", options:.backwards)!
            name = String(storageUrl[storageUrl.index(after:slashsubstring.lowerBound)..<storageUrl.endIndex])
        }
        
        if name.isEmpty {
            log(type:.fault,
                message: "%@: Storage url '%@' does not include the file path",
                args: "StorageConnectionBase.createStorageObject()",
                storageUrl)
            throw ClientError.argument(NSLocalizedString(
                "Storage cloud url is invalid", comment: ""))
        }
    
        func getStorageObject(retry:Int) -> () {
            
            var storageObject: AdvancedStorageObject? = nil
            var clientError:ClientError? = nil

            if retry >= StorageConnectionImpl.RETRY_LIMIT {
                callback(nil, ClientError.network(String(format:NSLocalizedString(
                    "Failed to create storage object after %d attempts", comment: ""),
                    StorageConnectionImpl.RETRY_LIMIT)))
                return
            }

            if sar == nil {
                sar = self.storageAuthenticationResponse
            }
            if sar == nil {
                callback(nil, ClientError.state(NSLocalizedString(
                    "Cannot obtain storage authentication", comment:"")))
                return
            }
            
            let authToken = sar?.getAuthToken()
            let headHeaders: [String : String] = [
                "X-Auth-Token" : authToken!
            ]

            let headurl = storageUrl
            self.defaultHttpManager?.request(headurl, method: .head,
                headers: headHeaders).response(queue: self.responseQueue,
                completionHandler: { response in
                    let status = response.response?.statusCode
                    if status == 200 {
                        var date:String? = "\(Date())"
                        var contentType = StorageObject.OCTET_STREAM
                        var encoding = "utf-8"
                        var length = Int(0)
                        var worm_policy:String = ""
                        if let respHeaders = response.response?.allHeaderFields {
                            if respHeaders["Last-Modified"] != nil {
                                let lastmodified = respHeaders["Last-Modified"] as! String
                                date = lastmodified //ISO8601DateFormatter().date(from: lastmodified)
                            }
                            if respHeaders["Content-Type"] != nil {
                                contentType = respHeaders["Content-Type"] as! String
                            }
                            if respHeaders["Content-Length"] != nil {
                                length = Int(respHeaders["Content-Length"] as! String)!
                            }
                            if respHeaders["X-Worm-Expiration-At"] != nil {
                                worm_policy = respHeaders["X-Worm-Expiration-At"] as! String
                            }
                        }
                        if response.response?.textEncodingName != nil {
                            encoding = response.response!.textEncodingName!
                        }
                        storageObject = self.createStorageObject(storageContainerUrl:storageUrl,
                            name: name,
                            contentType:contentType,
                            encoding: encoding,
                            date:date,
                            length:length,
                            mutable:worm_policy.isEmpty)
                        
                        callback(storageObject, nil)
                    } else if status == 401 {
                        // token expired
                        // Token will be renewed
                        print("createStorageObject token expired")
                        // TODO: Need a better wat to saefely expire the token
                        self.storageAuthenticationResponse = nil
                        sar = self.storageAuthenticationResponse!
                        getStorageObject(retry:(retry + 1))
                    } else if retry < StorageConnectionImpl.RETRY_LIMIT {
                        getStorageObject(retry:(retry + 1))
                    } else {
                        clientError = ClientError.network(String(format: NSLocalizedString(
                            "HTTP %d", comment:""), status ?? "500"))
                        callback(nil, clientError)
                    }
                 })
        }
        getStorageObject(retry: 0)
    }

    @objc public class func isStorageCloudURI(uri:String) -> Bool {
        return URL(string:uri)?.host!.range(of:StorageConnectionImpl.STORAGE_CLOUD_HOST) != nil
    }

    @objc public func close() /*throws*/ {
        authQueue.sync(execute: {
            if !closed {
                closed = true
            }
        })
    }
    
    public override func sync(storageObject: AdvancedStorageObject,
                     callback: @escaping (ClientError?)->()) throws {
        
        // At this point storageObject is a StorageObjecteDelegate instance
        let delegate:StorageObjectDelegate = storageObject as! StorageObjectDelegate
        // TODO: If neither stream is set isn't that an error case ?
        if storageObject.getInputStream() != nil {
            try upload(storageObject:delegate, callback:callback)
        } else if storageObject.getOutputStream() != nil {
            try download(storageObject:delegate, callback:callback)
        } else {
            callback(ClientError.state(
                NSLocalizedString("inputStream and outputStream are not set",
                                  comment: "")))
        }
    }

    // TODO: Should the storage objects be part of the callback arguments ?
    private func upload(storageObject: StorageObjectDelegate,
                        callback: @escaping (ClientError?)->()) throws {
        try transfer(storageObject: storageObject,
                     callback: { (responseCode, error) in
                        var clientError: ClientError? = error
                        if clientError == nil {
                            if let responseCode = responseCode {
                                if responseCode != 201 {
                                    clientError = ClientError.network(String(format:NSLocalizedString(
                                        "Upload failed %d", comment: ""), responseCode))
                                }
                            }
                        }
                        callback(clientError)
        })
    }
    
    // TODO: Should the storage objects be part of the callback arguments ?
    private func download(storageObject: StorageObjectDelegate,
                          callback: @escaping (ClientError?)->()) throws {
        try transfer(storageObject: storageObject,
                     callback: { (responseCode, error) in
                        var clientError:ClientError? = error
                        if clientError == nil {
                            if let responseCode = responseCode {
                                if responseCode != 200 && responseCode != 206 {
                                    clientError = ClientError.network(String(format:NSLocalizedString(
                                        "Download failed %d", comment: ""), responseCode))
                                }
                            }
                        }
                        callback(clientError)
        })
    }

    private func transfer(storageObject: StorageObjectDelegate,
                          callback: @escaping (Int?, ClientError?)->()) throws {
        
        let _upload:Bool = storageObject.getInputStream() != nil
        let scPath:String = (storageObject.getURI().isEmpty // _upload
            ? (storageAuthenticationResponse?.getStorageContainerUrl())! + "/" +
                storageObject.getName()
            : storageObject.getURI())
        
        // TODO: Implement better retry mechanism, not recursion, 
        // use dispatch queue.
        var retries:Int = 0
        func doTransfer(scPath:String, storageObject: StorageObjectDelegate, retry:Int,
                        callback: @escaping (Int?, ClientError?) ->()) throws {
            if retries > StorageConnectionImpl.RETRY_LIMIT {
                let op:String = _upload ? "Upload" : "Download"
                callback(nil, ClientError.network(String(format:NSLocalizedString(
                    "%@ failed retries exceeded.", comment: ""), op)))
                return
            }

            var filepath = scPath
            try transferContent(scPath: scPath, storageObject:storageObject,
                callback: { (status, date, transferredBytes) in
                    defer {
                        retries += 1
                    }
                    // failed checksum, try again
                    if status == -1 {
                        do {
                            try doTransfer(scPath:filepath,
                                       storageObject:storageObject,
                                       retry:retries, callback:callback)
                        } catch {
                            print(error)
                            let op:String = _upload ? "Upload" : "Download"
                            callback(status, ClientError.network(String(format:NSLocalizedString(
                                "%@ failed with response code %d", comment: ""), op, status)))
                        }
                        return
                    }
                    // token expired
                    if status == 401 {
                        do {
                            try self.authenticate()
                            if _upload {
                                filepath = self.storageAuthenticationResponse!.getStorageContainerUrl()! +
                                    "/" + storageObject.getName()
                            }
                            try doTransfer(scPath:filepath,
                                           storageObject:storageObject,
                                           retry:retries, callback:callback)
                        } catch {
                            print(error)
                            callback(nil, ClientError.network("\(error)"))
                        }
                        return
                    }
                    if _upload {
                        if status == 201 {
                            // TODO: Should be StorageObjectDelegate
                            storageObject.setAttributes(date: date,
                                                      length: transferredBytes)
                        }
                        callback(status, nil)
                        return
                    } else if status == 200 || status == 206 {
                        callback(status, nil)
                        return
                    }

                    //don't retry other errors
                    let op:String = _upload ? "Upload" : "Download"
                    callback(status, ClientError.network(String(format:NSLocalizedString(
                    "%@ failed with response code %d", comment: ""), op, status)))
                    return
            })
        }
        try doTransfer(scPath: scPath, storageObject:storageObject, retry:retries,
            callback: callback)
    }

    private func transferContent(scPath:String,
                storageObject: StorageObjectDelegate,
                //callback: @escaping (Int, String, Int)->())
        callback: @escaping (Int, String, Int)->())
        throws {

        let _upload:Bool = storageObject.getInputStream() != nil
        if _upload {
            // TODO: Should pass inputStream
            try upload(storagePath: scPath, instream:storageObject.getInputStream()!,
                       storageObject: storageObject,
                       //fileToUpload: URL(string: storageObject.getURI())!,
                       callback: callback)
        } else {
            // TODO: Have to deal with destination as outputstream
            try download(storagePath: scPath, outstream: storageObject.getOutputStream()!,
                         storageObject: storageObject,
                         callback: callback)
        }
    }

    // Can this really throw
    private func upload(storagePath:String, instream:InputStream,
                        storageObject: StorageObjectDelegate,
                    callback: @escaping (Int, String, Int) -> ()) throws {
        
        //print("#### upload to " + storagePath + " ####")

        self.defaultHttpManager?.startRequestsImmediately = true

        var uploadHeaders: [String : String] = [
            "X-Auth-Token" : (storageAuthenticationResponse?.getAuthToken())!,
            "Content-Type" : storageObject.getContentType(),
        ]
        if storageObject.getEncoding() != nil {
            uploadHeaders["Content-Encoding"] = storageObject.getEncoding()!
        }
        for (k,v) in storageObject.getCustomMetadata() {
            uploadHeaders["X-Object-Meta-" + k] = v
        }
        if let wormdays = StorageConnectionBase.WORM_EXPIRATION_DAYS {
            if !storageObject.isMutable() {
                uploadHeaders[StorageConnectionBase.WORM_HEADER] = wormdays
            }
        }
        if storageObject.getTimeToLive() > 0 {
            uploadHeaders["X-Delete-After"] = (storageObject.getTimeToLive() as NSNumber).stringValue
        }
        
        var transferredBytes:Int = 0
        let uploadRequest:UploadRequest? =  self.defaultHttpManager?.upload(
            instream,
            to: URL(string:storagePath)!,
            method: .put,
            headers: uploadHeaders)
                .response(
                    queue: self.responseQueue,
                    completionHandler: {response in
                //print(response)
                // What happens if not open ?
                instream.close()
                let status = response.response?.statusCode ??
                                            StatusCode.BAD_REQUEST.getCode()
                // Successful transfer
                var date:String = "\(Date())"
                if status == 201 || status == 200 || status == 206 {
                    // TODO: is this necessary ? Is it already IOS806
                    if let sdate = response.response?.allHeaderFields["Last-Modified"] {
                        date = sdate as! String //ISO8601DateFormatter().date(from:sdate as! String) ?? Date()
                    } else {
                        date = "\(Date())"
                    }
                    // TODO: verify checksum
                    // checksum mismatch return status == -1
//                    let checksum = -1
//                    if response.response?.allHeaderFields["ETag"] != nil {
//                        checksum = Int(response.response?.allHeaderFields["ETag"] as! String)!
//                    }
                }
                callback(status, date, transferredBytes)
        })
        
        // Note the absense of () for the uploadProgress method call.
        // When they are present the compiler cannot disambiguate between
        // the method and the uploadProgress property.
        // You get this error
        // Cannot call value of non-function type 'Progress'
        uploadRequest?.uploadProgress { p in
            /*
            print ("totalUnitCount: \(p.totalUnitCount)\n" +
                "completedUnitCount: \(p.completedUnitCount)\n" +
                "localizedDescription: \(p.localizedDescription)\n" +
                "localizedAdditionalDescription: \(p.localizedAdditionalDescription)\n" +
                "isCancellable: \(p.isCancellable)\n" +
                "isPausable: \(p.isPausable)\n" +
                "fractionalCompleted: \(Float(p.fractionCompleted))\n")
            */
            if storageObject.isCanceled() {
                if p.isCancellable {
                    print("upload cancelled")
                    p.cancel()
                } else {
                    print("upload is not cancellable")
                }
            }
            transferredBytes = (p.completedUnitCount as NSNumber).intValue
            storageObject.setTransferredBytes(transferredBytes: transferredBytes)

        }
        return
    }
    
    // Can this really throw ?
    private func download(storagePath:String, outstream:OutputStream,
                          storageObject: StorageObjectDelegate,
                         callback: @escaping (Int, String, Int) -> ()) throws {
        
        //print("#### download " + storagePath + " ####")
        
        let downloadHeaders: [String : String] = [
            "X-Auth-Token" : (storageAuthenticationResponse?.getAuthToken()!)!
        ]

        var transferredBytes:Int = 0
        self.defaultHttpManager?.startRequestsImmediately = true
        let downloadRequest:DownloadRequest? =
            self.defaultHttpManager?.download(URL(string:storagePath)!,
                method: .get,
                headers: downloadHeaders)
                    .response(queue: self.responseQueue,
                              completionHandler: { response in
                    //print(response)
                    // TODO: This probably should be String
                    var date:String = "\(Date())"
                    let status = response.response?.statusCode ??
                                                StatusCode.BAD_REQUEST.getCode()
                    // succesful download
                    if status == 200 || status == 206 {
                        // TODO: is this necessary ? Is it already IOS806
                        if let sdate = response.response?.allHeaderFields["Last-Modified"] {
                            date = sdate as! String //ISO8601DateFormatter().date(from:sdate as! String) ?? Date()
                        }
                        if let responseheaders = response.response?.allHeaderFields {
                            for (k,v) in responseheaders {
                                if k is String && (k as! String).hasPrefix("X-Object-Meta-") {
                                    let header:String = k as! String
                                    var key:String = ""
                                    if let keyRange:Range = header.range(of:"X-Object-Meta-") {
                                        key = String(header[keyRange.upperBound..<header.endIndex])
                                    } else {
                                        continue
                                    }
                                    do {
                                        try storageObject.setCustomMetadata(key:key,
                                                                            value:v as! String)
                                    } catch {
                                        // TODO: What should happen here ?
                                        // TODO: Should the transfer fail ?
                                        // TODO: JCL ignores the exception
                                        //status = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                                        log(type:.fault,
                                            message:"%@: setCustomMetadata(%@,%@) threw exception %@",
                                            args:"StorageConnectionBase.download()",
                                            key, v as! String, "\(error)")
                                    }
                                }
                            }
                        }
                        // TODO: verify checksum
                        // checksum mismatch return status == -1
                        // let checksum = -1
                        // if response.response?.allHeaderFields["ETag"] != nil {
                        //     checksum = Int(response.response?.allHeaderFields["ETag"] as! String)!
                        // }
                        // Closes outstream
                        // This must be done in the response handler
                        self.copyFile(infile:response.temporaryURL!,
                             outstream:outstream)
                    }
                    callback(status, date, transferredBytes)
                })
        
        // Note the absense of () for the uploadProgress method call.
        // When they are present the compiler cannot disambiguate between
        // the method and the uploadProgress property.
        // You get this error
        // Cannot call value of non-function type 'Progress'
        downloadRequest?.downloadProgress { p in
            /*
            print ("totalUnitCount: \(p.totalUnitCount)\n" +
                "completedUnitCount: \(p.completedUnitCount)\n" +
                "localizedDescription: \(p.localizedDescription)\n" +
                "localizedAdditionalDescription: \(p.localizedAdditionalDescription)\n" +
                "isCancellable: \(p.isCancellable)\n" +
                "isPausable: \(p.isPausable)\n" +
                "fractionalCompleted: \(Float(p.fractionCompleted))\n")
            */
            if storageObject.isCanceled() {
                if p.isCancellable {
                    print("download cancelled")
                    p.cancel()
                } else {
                    print("download is not cancellable")
                }
            }
            transferredBytes = (p.completedUnitCount as NSNumber).intValue
            storageObject.setTransferredBytes(transferredBytes: transferredBytes)
        }
        return
    }
    
    private func copyFile(infile:URL, outstream:OutputStream) {
        let instream = InputStream(url: infile)
        var buff:[UInt8] = [UInt8](repeating: 0,
                                   count: Config.STORAGE_CONNECTION_CHUNK_SIZE)
        var readin:Int = 0
        instream?.open()
        outstream.open()
        repeat {
            readin = (instream?.read(&buff, maxLength: buff.count))!
            //print(readin)
            outstream.write(&buff, maxLength: readin)
        } while (instream?.hasBytesAvailable)!
        outstream.close()
        instream?.close()
    }
/*
    func MD5(_ string: String) -> String? {
        let algorithm = HMAC.getAlgorithm(algorithm: "hmacmd5")
        let length = algorithm.digestLength //Int(CC_MD5_DIGEST_LENGTH)
        var digest = [UInt8](repeating: 0, count: length)
        if let d = string.data(using: String.Encoding.utf8) {
            d.withUnsafeBytes { (body: UnsafePointer<UInt8>) in
                CC_MD5(body, CC_LONG(d.count), &digest)
            }
        }
        return (0..<length).reduce("") {
            $0 + String(format: "%02x", digest[$1])
        }
    }
 */
    
    private func authenticate() throws  {
        
        //print("#### Getting storage service access token ####")
        
        // TODO: Use DispatchGroup ?
        // A semaphore for authenticate to wait on until the network
        // request completes.
        let authSem = DispatchSemaphore(value:0)
        var clientError:ClientError? = nil
        // This queue must not be on the main thread and any calls resulting
        // from the executed code must not rely on the main thread otherwise
        // there will be deadlock if "authenticate" is called on the main
        // thread.
        try authQueue.sync(execute: {
            let headers: [String : String] = [
                "Content-Type" : "application/json" ,
                "Accept" : "application/json",
                ]
            try secureConnection.get(
                restApi: StorageConnectionImpl.REST_STORAGE_AUTHENTICATION,
                headers: headers,
                response: { response, err in
                    if err != nil {
                        clientError = err
                    } else {
                        let status = response?.getStatus()
                        if status == StatusCode.OK.getCode() {
                            if let authjson = response?.getData() {
                                do {
                                    let auth:JSON = try JSON(data: authjson)
                                    // TODO: Use fromJSON api
                                    self._storageAuthenticationResponse = StorageAuthenticationResponse(
                                        storageContainerUrl: auth.dictionaryValue["storageContainerUrl"]!.string!,
                                        authToken: auth.dictionaryValue["authToken"]!.string!)
                                } catch {
                                    clientError = ClientError.network(String(format:NSLocalizedString(
                                        "%@: JSON threw exception: %@", comment:""),
                                        "StorageConnectionBase:authenticate()", "\(error)"))
                                }
                            } else {
                                clientError = ClientError.network(NSLocalizedString(
                                    "StorageConnectionBase:authenticate() empty data from successful authentication request",
                                        comment:""))
                            }
                        } else {
                            var dataString = ""
                            if status! >= StatusCode.BAD_REQUEST.getCode() &&
                                response!.getData().count > 0 {
                                dataString = String(data: response!.getData(), encoding: .utf8)!
                            }
                            // TODO: getVerboseStatus
                            clientError = ClientError.network(String(format:NSLocalizedString(
                                "%@ %@ : HTTP %@ %@", comment: ""),
                                 "GET", StorageConnectionImpl.REST_STORAGE_AUTHENTICATION,
                                 StatusCode.valueOf(code: status!).getDescription(),
                                 dataString))
                        }
                    }
                    authSem.signal()
            })
        })
        _ = authSem.wait(timeout:DispatchTime.now() + .seconds(10))
        if let clientError = clientError {
            throw clientError
        }
    }
}
