/*
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package oracle.iot.client
import Foundation

/**
 * A callback interface for receiving an event when content referred to by
 * an attribute value has been successfully synchronized, or has failed
 * to be synchronized.
 * See `setOnSync`
 */
open class SyncCallback {

    public init() {
    }
    /**
     * Callback for receiving an event when content referred to by
     * an attribute value has been successfully synchronized, or has failed
     * to be synchronized.
     * This callback will never be invoked by the
     * client-library with a `nil` argument.
     *
     * - parameter event: The error event
     */
    open func onSync(event:SyncEvent) {
        fatalError("Subclass need to implement the 'onSync()' method.")
    }
}


/**
 * An event passed to the `onSync` callback when content referred to
 * by an attribute value has been successfully synchronized, or has failed
 * to be synchronized.
 */
public protocol SyncEvent {

    /** 
     * Get the virtual device that is the source of the event.
     * - returns: the virtual device, or `nil` if `sync` was not called
     *            independently
     */
    func getVirtualDevice() -> AbstractVirtualDevice?

    /**
     * Get the name of the attribute, action, or format that this event
     * is associated with.
     * - returns: the name, or `nil` if `sync` was called independently
     */
    func getName() -> String?

    /**
     * Get the StorageObject that is the source of this event.
     * - returns: the storageObject, never `nil`
     */
    func getSource() -> StorageObject
}

class SyncEventImpl : SyncEvent {
    
    private let name:String?
    private let source: StorageObject!
    private let virtualDevice:VirtualDevice?
    
    init(source:StorageObject, virtualDevice:VirtualDevice?,
                     name:String?) {
        self.source = source
        self.virtualDevice = virtualDevice
        self.name = name
    }
    
    /**
     * Get the virtual device that is the source of the event.
     *
     * - returns: the virtual device, never `nil`
     */
    public func getVirtualDevice() -> AbstractVirtualDevice? {
        return virtualDevice
    }
    
    /**
     * Get the name of the attribute, action, or format that this event
     * is associated with.
     *
     * - returns: the name, never `nil`
     */
    public func getName() -> String? {
        return name
    }
    
    /**
     * Get the `StorageObject` that is the source of this event.
     *
     * - returns: the storageObject, never `nil`
     */
    public func  getSource() -> StorageObject {
        return source
    }
}

/**
 * The status of whether or not the content is in sync with the storage cloud.
 */
public enum SyncStatus {
    /**
     * The content is in sync with the storage cloud. This means that
     * the upload or download is complete.
     */
    case inSync

    /**
     * The content is not in sync with the storage cloud because
     * the upload or download failed.
     */
    case syncFailed
    
    /**
     * The content is not in sync with the storage cloud because
     * it has not yet been uploaded or download.
     */
    case notInSync

    /**
     * The content is not in sync with the storage cloud, but a
     * sync is pending
     */
    case syncPending
}

/**
 * StorageObject provides information about content in cloud storage.
 *
 * ```
 *     // Upload example
 *     var lenna:StorageObject =
 *         directlyConnectedDevice.createStorageObject(name:"lenna.jpg",
 *                                              contentType: "image/jpeg")
 *     lenna.setInputPath(inputPath:"images/lenna.jpg")
 *
 *     // onSync is called when the content referenced by the storageObject
 *     // is in sync with the storage cloud, or the sync has failed.
 *     lenna.setOnSync(callback: { event in
 *         let storageObject:StorageObject = event.getSource()
 *         if storageObject.getSyncStatus() == SyncStatus.inSync {
 *             // image was uploaded and can be deleted
 *         } else if storageObject.getSyncStatus() == SyncStatus.syncFailed {
 *             // image was not uploaded, take action!
 *         }
 *     })
 *
 *     virtualDevice.set(attributeName:"image", attributeValue:lenna)
 *
 *     // Download example
 *     // onChange is called when the attribute value changes.
 *     virtualDevice.setOnChange(attributeName:"image", callback: { event in 
 *         var namedValue:NamedValue = event.getNamedValue()
 *         var storageObject:StorageObject = namedValue.getValue() as StorageObject
 *         if storageObject.getContentLength() < availableDiskSpace {
 *             // syncTo will kick off the async download of the content
 *             storageObject.setOnSync(callback: { event in
 *                 let storageObject:StorageObject = event.getSource()
 *                 if storageObject.getSyncStatus() == SyncStatus.inSync {
 *                     // image was downloaded and can now be used
 *                 } else if storageObject.getSyncStatus() == SyncStatus.syncFailed {
 *                     // image was not downloaded, take action!
 *                 }
 *             }
 *             storageObject.setOutputPath(outputPath:"downloads" +
 *                                                    storageObject.getName())
 *             storageObject.sync()
 *         }
 *     })
 * }
 * ```
 */
public class StorageObject : ExternalObject {
    
    public static let OCTET_STREAM = "application/octet-stream"

    private let delegate: AdvancedStorageObject
    //private var storageDispatcher:StorageDispatcher!
    private var /*volatile*/ syncStatus:SyncStatus!
    private var /*volatile*/ callback:SyncCallback?
    private var inputPath:String?
    private var outputPath:String?
    
    /**
     * Create a `StorageObject` instance.
     * - parameter delegate: A messaging API `StorageObject` to which
     *             this API will delegate.
     */
    init(delegate:AdvancedStorageObject) {
        self.delegate = delegate
        super.init(uri: delegate.getURI())
        self.syncStatus = SyncStatus.notInSync
    }
    
    // TODO: HACK to overcome inability to use nil value in Pair
    // TODO: Only used in DeviceFunctionTest
    // TODO: REMOVE WHEN IOT-55312 gets merged
    init() {
        self.delegate = AdvancedStorageObject(uri: "", name: "name",
                                              contentType: nil,
                                              encoding: nil,
                                              date: "",
                                              length: -1)
        super.init(uri:"")
    }

    /**
     * Get the the name of this object in the storage cloud.
     *
     * - returns: the name of this object in the storage cloud
     */
    public func getName() -> String {
        return delegate.getName()
    }

    /**
     * Get the mime-type of the content. See [IANA Media Types](
     * http://www.iana.org/assignments/media-types/media-types.xhtml)
     *
     * - returns: The mime-type of the content
     */
    public func getContentType() -> String {
        return delegate.getContentType()
    }

    /**
     * Get the compression scheme of the content.
     *
     * - returns: the compression scheme of the content, or `nil` if the
     *            content is not compressed
     */
    public func getEncoding() -> String? {
        return delegate.getEncoding()
    }

    /**
     * Get the length of the content in bytes. This is the number of bytes
     * required to upload or download the content.
     *
     * - returns: The length of the content, or -1 if unknown
     */
    public func getLength() -> Int {
        return delegate.getLength()
    }

    /**
     * Get the date and time the content was created or last modified in cloud 
     * storage. This may be `nil` if the content has not been uploaded.
     * The date and time stamp format is `ISO 8601`
     *
     * - returns: The date the content was last modified in cloud storage, or 
     *            `nil` if the content has not been uploaded.
     */
    public func getDate() -> String? {
        // TODO: Should be String
        return delegate.getDate()
    }

    /**
     * Get the status of whether or not the content is in sync with the 
     * storage cloud.
     *
     * - returns: the status of whether or not the content is in sync with the 
     *            storage cloud.
     */
    public func getSyncStatus() -> SyncStatus {
        return syncStatus
    }

    /**
     * Get the input path.
     *
     * - returns: the input path, or `nil` if not set
     */
    public func getInputPath() -> String? {
        return inputPath
    }

    /**
     * Set the input path for uploading content to the storage cloud. The 
     * implementation allows for either the input path to be set, or the 
     * output path to be set, but not both.
     *
     * If the `inputPath` parameter is not `nil`, the `outputPath` will be set 
     * to `nil`.
     * If the `inputPath` parameter is not `nil` and does not equal the 
     * current input path, the sync status will be reset to 
     * `SyncStatus.notInSync`.
     *
     * This method throws `ClientError.state` if `getSyncStatus()` returns
     * `SyncStatus.SYNC_PENDING`.
     *
     * - parameter inputPath: the path from which to read the content for
     *                         upload
     *
     * - throws: 
     *       - `ClientError.file` if `inputPath` does not identify an existing 
     *                          file
     *       - `ClientError.argument` if sync status is `SYNC_PENDING`
     */
    public func setInputPath(inputPath:String?) throws {

        if syncStatus == SyncStatus.syncPending {
            throw ClientError.state(NSLocalizedString("sync pending", 
                                                      comment: ""))
        }

        if self.inputPath == nil || self.inputPath! != inputPath {
            self.inputPath = inputPath
            if self.inputPath != nil {
                self.outputPath = nil
                self.syncStatus = SyncStatus.notInSync
                delegate.setOutputStream(outputStream: nil)
                delegate.setInputStream(inputStream: InputStream(fileAtPath:inputPath!)!)
            }
        }
    }

    /**
     * Get the output path.
     *
     * - returns: the output path, or `nil` if not set
     */
    public func getOutputPath() -> String? {
        return outputPath
    }

    /**
     * Set the output path for downloading content from the storage cloud. The
     * implementation allows for either the output path to be set, or the 
     * input path to be set, but not both.
     *
     * If the `outputPath` parameter is not `nil`, the input path will be set
     * to `nil`.
     * If the `outputPath` parameter is not `nil` and does not equal the
     * current output path, the sync status will be reset to 
     * `SyncStatus.notInSync`
     *
     * This method throws `ClientError.state` if
     * `getSyncStatus()` returns `SyncStatus.SYNC_PENDING`
     *
     * - parameter outputPath: the path to which the content will be downloaded
     *                          If the path does not exist, it will be created.
     *
     * - throws:
     *       - `ClientError.file` if the `outputPath` cannot be written.
     *       - `ClientError.state` if called when sync status is `SYNC_PENDING`
     */
    public func setOutputPath(outputPath:String?) throws {

        if syncStatus == SyncStatus.syncPending {
            throw ClientError.state(NSLocalizedString("sync pending",
                comment: ""))
        }

        if self.outputPath == nil || self.outputPath! != outputPath {
            self.outputPath = outputPath
            if self.outputPath != nil {
                self.inputPath = nil
                self.syncStatus = SyncStatus.notInSync
                delegate.setInputStream(inputStream: nil)
                let components:[String] = outputPath!.components(separatedBy: "/")
                var dirpath = ""
                for i in 0...(components.count - 2) {
                     dirpath += components[i] + "/"
                }
                
                do {
                    try FileManager.default.createDirectory(atPath: dirpath,
                            withIntermediateDirectories: true, attributes: nil)
                } catch {
                    throw ClientError.file("Unable to create directory \(dirpath) \(error)")
                }
                delegate.setOutputStream(outputStream: OutputStream(toFileAtPath:outputPath!,
                        append:false)!)
            }
        }
    }
    
    /**
     * Sets the metadata for the StorageObject. All metdata will be added to
     * Storage Cloud Service as custom metadata
     * with the *X-Object-Meta-KeyName* header.
     *
     * - parameter key: The metadata key
     *
     * - parameter value: The metadata value
     *
     * - throws: `ClientError.argument` if the key or value is empty
     *  `ClientError.state` if the key or value is `nil`
     */
    public func setCustomMetadata(key:String, value:String ) throws {
        try delegate.setCustomMetadata(key:key, value:value)
    }
    
    /**
     * This method return unmodifiable copy of metadata.
     *
     * - returns: Map of metadata. Map may be empty.
     */
    public func getCustomMetadata() -> [String:String] {
        return delegate.getCustomMetadata()
    }
    
    /**
     * Returns the mutable state of the StorageObject.
     *
     * - returns: `true` if the `StorageObject` is mutable.
     */
    public func isMutable() -> Bool {
        return delegate.isMutable()
    }
    
    /**
     * Returns the time to live of the StorageObject when creating it in the
     * Storage Cloud Service.
     *
     * - returns: Time to live in seconds, or *-1* if unknown
     */
    public func getTimeToLive() -> Int {
        return delegate.getTimeToLive()
    }
    
    /**
     * Set the time to live for Storage Object in seconds
     * Time to live will be set to the Storage Object while file uploading.
     *
     * - parameter timeToLive: in seconds
     *
     * - throws: `ClientError.argument` if `timeToLive` is less than or equal to 0
     *  `ClientError.state` if the storage object status is `SyncStatus.syncPending`
     */
    public func setTimeToLive(timeToLive:Int) throws {
        guard timeToLive > 0 else {
            throw ClientError.argument(NSLocalizedString("timeToLive must be greater than 0", comment:""))
        }
        guard syncStatus != SyncStatus.syncPending else {
            throw ClientError.state(NSLocalizedString("sync pending", comment:""))
        }
    
        try delegate.setTimeToLive(timeToLive:timeToLive)
    }
    
    /**
     * Notify the library to sync content with the storage cloud.
     * The direction of the sync, upload or download, depends on whether
     * the `setInputPath` input path or the
     * `setOutputPath` output path has been set.
     * This method does not start any uploads or downloads if `getSyncStatus`
     * is other than `SyncStatus.notInSync.
     * 
     * This is a non-blocking call. The sync is performed on a separate thread.
     * The status of the sync can be monitored by setting a `SyncCallback`
     * on this `StorageObject`.
     * 
     * If the input path cannot be read, or the output path cannot be
     * written, an `ClientError.file` is thrown. Any I/O exceptions
     * during the background sync are reported through the
     * `AbstractVirtualDevice.ErrorCallback` error callback
     * of the virtual device.
     *
     * - throws: 
     *       - `ClientError.file` if the the input path cannot be read,
     *                          or the output path cannot be written
     *       - `ClientError.state` if both input path and output path are `nil`
     */
    public func sync() throws {
        fatalError("Subclass must implement sync()")
    }

    /**
     * Set a callback that is invoked when the content referred to by this
     * `StorageObject` is synchronized.
     *
     * - parameter callback: a callback to invoke when there is an error 
     *                        setting a value, if `nil`, the existing callback
     *                        will be removed
     */
    public func setOnSync(callback: SyncCallback?) {
        self.callback = callback
    }

    /**
     * Used internally by the client library. Not intended for general use.
     * - parameter syncStatus: the new SyncStatus
     */
    func setSyncStatus(syncStatus: SyncStatus) {
        self.syncStatus = syncStatus
    }
    
    /**
     * Used internally by the client library. Not intended for general use.
     * - returns: the SyncCallback
     */
    func getSyncCallback() -> SyncCallback? {
        return callback
    }
}
