/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package com.oracle.iot.client.impl

import Foundation

fileprivate var dispatcher:DispatchQueue = DispatchQueue.global(qos: .utility)

/**
 * StorageObjectBase is the common implementation between the
 * implementations of DeviceLib and EnterpriseLib StorageObjectImpl.
 */
public class StorageObjectBase : StorageObject {

    private var delegate:AdvancedStorageObject!
    private var syncEvents:[SyncEvent]!
    
    public override init(delegate:AdvancedStorageObject) {
        super.init(delegate: delegate)
        self.delegate = delegate
        self.syncEvents = [SyncEvent]()
    }

    public override func sync() throws {
        if getSyncStatus() == SyncStatus.notInSync {
            if getOutputPath() == nil && getInputPath() == nil {
                throw ClientError.state(NSLocalizedString(
                    "input path or output path must be set", comment: ""))
            }
            setSyncStatus(syncStatus: SyncStatus.syncPending);
            addSyncEvent(syncEvent: createSyncEvent())
        } else {
            addSyncEvent(syncEvent: createSyncEvent())
        }
    }

    public override func setOnSync(callback: SyncCallback?) {
        super.setOnSync(callback: callback)
        syncEvents.removeAll()
    }

    func getDelegate() -> AdvancedStorageObject {
        return delegate
    }

    func getSyncEvents() -> [SyncEvent] {
        return syncEvents
    }

    func addSyncEvent(syncEvent: SyncEvent) {
        guard let cb = getSyncCallback() else {
            return
        }
        switch (getSyncStatus()) {
            case .syncPending, .notInSync:
                DispatchQueue.global(qos:.utility).async(execute: {
                    self.syncEvents.append(syncEvent)
                })
                break;
            case .syncFailed, .inSync:
                //Immediate Callback
                dispatcher.async(execute: {
                    cb.onSync(event: syncEvent)
                })
                break;
        }
    }

    func handleStateChange() {
        fatalError("Subclass must implement handleStateChange()")
    }

    func createSyncEvent() -> SyncEvent {
        fatalError("Subclass must implement createSyncEvent()")
    }

    class func getDispatcher() -> DispatchQueue {
        return dispatcher
    }
}
