/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Heimdall
import Security
import Alamofire
import CommonCrypto

extension TAMconstants {
    // Keychain attribute names for inserting, searching and deletion
    static let ACTIVE:String        = "ACTIVE"
    static let SHARED_SECRET:String = "SHARED_SECRET"
    static let ENDPOINT:String      = "ENDPOINT"
    static let DEVICEID:String      = "DEVICEID"
    static let HOST:String          = "HOST"
    static let PORT:String          = "PORT"
    static let INSTANCE:String      = "INSTANCE"
}

/**
 * The UnifiedTrustedAssetManager class implementation
 */
class UnifiedTrustedAssetsManager :TrustedAssetsManagerImpl {

    private var isProvisionedFromKeychain = false
    
    /**
     * The init method of the TAM has to be private
     */
    private override init() {
        super.init()
    }

    /**
     * - throws: `ClientError.security` if server policy cannot be set.<br/>
     *           `ClientError.file` if the provision file is empty or has the
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    @objc convenience init(context: Dictionary<String,Any>?) throws {
        // sequence first read the provisioning file then check if the TAM is 
        // already active and if yes load item from keychain
        self.init()
        // We always have to read the elements from the encrypted blob in the 
        // provisioning file to know if we have a file
        // for an already activated device or a another device
        if let context = context {
            if let path = context[TAMconstants.pathKey] as? String {
                // if we have a path given as information
                let pwd = context[TAMconstants.paswdKey] as? String
                // The information path and password  must be present
                let up = UnifiedProvisioner()
                self.isProvisioned = try up.loadFileFromPath(path: path,
                                                    password: pwd!, tam: self)
            } else {
                let fileName = context[TAMconstants.fileKey] as? String
                let fileEXT = context[TAMconstants.fileExtKey] as? String
                let sub = context[TAMconstants.subdKey] as? String
                let pwd = context[TAMconstants.paswdKey] as? String
                // The minimal file name and password input must be present
                let up = UnifiedProvisioner()
                self.isProvisioned = try up.loadFile(fileName: fileName!,
                                                     withExtension: fileEXT!,
                                                     subdirectory: sub,
                                                     password: pwd!, tam: self)
            }
            self.isProvisionedFromKeychain = false
        }
        // at this stage we know the deviceId of the provisioning file and can 
        // check if there is already data for an activated TAM available
        // we allow only allow one TAM instance for an activated device and one 
        // TAM instance for a non-activated device we ensure there is a unique
        // TAM object per given deviceId from a provisioning file
        self.isActive = isActivePersistent()
        if self.isActive {
            self.isProvisioned = self.loadFromKeychain()
            if (self.isProvisioned) {
                self.isProvisioned = setServerPolicyFromCertificates()
            }
            if (!self.isProvisioned) {
                throw ClientError.security(NSLocalizedString(
                    "Failed to set a Server policy from the TAM", comment: ""))
            }
        } else {
            // there are no data for an active TAM present in the keychain
            // but an instance was already created
            self.isProvisioned = setServerPolicyFromCertificates()
        }
    }

    
    override var description: String {
        var desc:String = String()
        desc += "\r\n"
        desc += "---------------------------\r\n"
        desc += "Activation state="
        desc += String(self.isActive)
        desc += "\r\n"
        desc += "deviceId="
        desc += String(self.deviceId)
        desc += "\r\n"
        desc += "Endpoint ID ="
        desc += String(self.endpointId)
        desc += "\r\n"
        desc += "Shared Secret="
        desc += String(self.sharedSecret)
        desc += "\r\n"
        desc += "---------------------------\r\n"
        return desc
    }

    /**
     * Retrieves the trust anchor or most-trusted Certification Authority (CA)
     * certificates to be used to validate the IoT CS server certificate chain.
     *
     * - returns: A `Foundation.Data` array of DER-encoded trust anchor
     *            certificates (byte arrays).
     */
    override func getTrustAnchorCertificates() -> [Foundation.Data]? {
        var certArray = [CFData]()
        var certData:CFData?
        if (self.rootCACert != nil) {
            certData = SecCertificateCopyData(self.rootCACert!)
            if (certData != nil ) {
                certArray.append(certData!)
            }
        }
        // TODO: Why is this implementation different than 
        // DefaultTrustedAssetsManager (TrustedAssetsManagerImpl) ?
        return certArray as [Foundation.Data]
    }
    
    /**
     * Sets the assigned endpoint ID and certificate as returned by the
     * activation procedure. Upon a call to this method, a compliant
     * implementation of the *TrustedAssetsManager* interface must ensure
     * the persistence of the provided endpoint credentials. This method can
     * only be called once; unless the *TrustedAssetsManager* has been
     * reset.
     *
     * If the client is a pre-activated enterprise application, the endpoint ID
     * has already been provisioned and calling this method MUST fail.
     *
     * - parameter endpointId: the assigned endpoint ID.
     *
     * - parameter certificate: the DER-encoded certificate issued by the server
     *   or an empty array if no certificate was provided by the server.
     *
     * - throws: `ClientError.state` if already activated, or private keys have
     *          not been generated, or `endpointId` is already assigned, or
     *          if certificate data has not been encoded.<br/>
     *          `ClientError.argument` if `endpointId` is an empty string.
     */
    override func setEndPointCredentials(endpointId: String,
                                         certificate: Foundation.Data) throws {
        // if the TAM is already active we throw an error
        if (self.isActive) {
            throw ClientError.state(NSLocalizedString("Already activated.",
                                                      comment: ""))
        }
        // the key pair must already be generated
        if (self.heimdall == nil) {
            throw ClientError.state(NSLocalizedString(
                "Private key not yet generated.", comment: ""))
        }
        // if for some reason already an endpointID is set
        if (!self.endpointId.isEmpty) {
            throw ClientError.state(NSLocalizedString(
                "EnpdpointId already assigned.", comment: ""))
        }
        // if the given endpointId is empty we throw an error
        if (endpointId.isEmpty) {
            throw ClientError.argument(NSLocalizedString(
                "Endpoint must not be an empty string.", comment: ""))
        }
        // if no certificate is given we set the endpoint id
        if (certificate.count == 0) {
            self.endpointId = endpointId
            let endid = KeychainEntry(Name: TAMconstants.ENDPOINT+self.deviceId,
                                      data: self.endpointId )
            try insertItemWithAttributes(attributes: endid.attributesToSave)
        }
        // if a certificate is given we set the certificate and endpointID but 
        // only if the data form a valid certificate
        if (certificate.count != 0) {
            // first we check if we can create a certificate object
            if let endpointCertificate = SecCertificateCreateWithData(nil,
                                                    certificate as CFData) {
                self.endpointCert = endpointCertificate
                self.endpointId = endpointId
                let endid = KeychainEntry(
                            Name: TAMconstants.ENDPOINT + self.deviceId,
                            data: self.endpointId )
                try insertItemWithAttributes(attributes: endid.attributesToSave)
            } else {
                // we throw an error if the certificate is not correctly 
                // formatted
                throw ClientError.state(NSLocalizedString(
                    "Certificate data are not DER encoded", comment: ""))
            }
        }
        // when we reach this point we can set the TAM to
        //activated and store the items of the TAM persistent in the keychain
        self.setToActivated()
        try self.storeInKeychain()
        return
    }

    /**
     * Sets the state of the TAM to Activated will only work if the provisioning 
     * and initialisation
     * worked correctly
     */
    override func setToActivated() -> Void {
        if (self.isProvisioned){
            self.isActive = true
            do {
               let active = KeychainEntry(
                            Name: TAMconstants.ACTIVE + self.deviceId,
                            data: TAMconstants.ACTIVE )
               try insertItemWithAttributes(attributes: active.attributesToSave)
            } catch let error {
                debugPrint("Error in setToActivated \(error)")
            }
        }
    }
    
    /**
     * Set the Server trust policy and validate it against the given host name
     *
     * - parameter hostName: the host to define and validate the policies for
     *
     * - parameter trust: the SecTrust object for this host
     *
     * - parameter certificateChain: the certificate chain for the host
     *
     * - returns: `true` if the server trust policy with the validated
     *            certificate chain and the validated host is valid
     */
    override func setServerTrustPolicy(hostName: String, trust: SecTrust,
                                certificateChain: [SecCertificate]) -> Bool {
        let host = hostName
        let serverTrust = trust
        let certificates = certificateChain
        let serverTrustPolicy = ServerTrustPolicy.pinCertificates(
            certificates: certificates,
            validateCertificateChain: true,
            // TODO: Why is this false when the "exact" impl in 
            // DefaultTrustedASsetsManager
            // (TrustedAseetsManagerImpl) is true ?
            validateHost: false
        )
        
        let serverTrustIsValid = serverTrustPolicy.evaluate(serverTrust,
                                                            forHost: host)
        if (serverTrustIsValid) {
            self.serverPolicy = serverTrustPolicy
        }
        return serverTrustIsValid
    }
    
    override func setCertificateinKeyChain(certificate: SecCertificate,
                                           tagName: String) throws -> Bool {
        let secCertData:Foundation.Data =
            SecCertificateCopyData(certificate) as Foundation.Data
        let cerSubject:String =
            String(describing: SecCertificateCopySubjectSummary(certificate))
        let queryFilter = [
            String(kSecClass)              : kSecClassCertificate,
            String(kSecAttrLabel)          : tagName,
            String(kSecAttrSubject)        : cerSubject,
            // that makes the difference to store to multiple certificate in
            // the keychain
            String(kSecAttrSerialNumber)   : tagName,
            String(kSecValueData)          : secCertData,
            ] as [String : Any]
        let result = SecItemAdd(queryFilter as CFDictionary, nil)
        if ( result == errSecDuplicateItem) {
            throw ClientError.security(NSLocalizedString("Certificate already in keychain \(tagName).", comment: ""))
        }
        if result != noErr {
            throw ClientError.security(NSLocalizedString("Error in setting a certificate into the keychain.", comment: ""))
        }
        return true
    }

    /**
     * Creates an RSA key-pair and stores them in the keychain
     *
     * - parameter keySize: key length in bits
     *
     * - parameter tagPrefix: for the keys to find them in the keychain
     *
     * - throws: `ClientError.argument` if `tagPrefix` is the empty string, or
     *           `keySize` is less than or equal to zero<br/>
     *           `ClientError.state` if already activated<br/>
     *           `ClientError.security` if key pair generation fails
     *
     */
    override func generateKeyPair(algorithm: String,
                                  keySize: Int) throws -> Void {
        if (algorithm.isEmpty) {
            throw ClientError.argument(NSLocalizedString(
                "Prefix to identify the keys in the keychain must be set",
                comment: ""))
        }
        if (keySize <= 0) {
            throw ClientError.argument(NSLocalizedString(
                "Key size cannot be negative or zero", comment: ""))
        }
        // TODO: Why is this check only in UnifiedTrustedAssetsManager ?
        if (!self.endpointId.isEmpty) {
            throw ClientError.state(NSLocalizedString(
                "Already activated: EndpointId already assigned", comment: ""))
        }

        // create the key without the prefix
        // creates the Heimdall instance with the key-pair, if the key pair 
        // generation fails the heimdall instance is not created
        
        // TODO: Why doesn't this code reassign "self.tagPublic" and
        // self.tagPrivate like DefaultTrustedAssetsManager did
        // (now in TrustedAssetsManagerImpl)

        let tagPublic = self.keyPublic + self.deviceId
        let tagPrivate = self.keyPrivate + self.deviceId
        self.heimdall = Heimdall(publicTag: tagPublic, privateTag: tagPrivate,
                                 keySize: keySize)
        if (self.heimdall == nil) {
            throw ClientError.security(NSLocalizedString(
                "KeyPair generation failed", comment: ""))
        }
    }
    
    /**
     * Resets the trust material back to its provisioning state; in particular,
     * the key pair is erased. The client will have to go, at least,through 
     * activation again; depending on the provisioning policy in place, the 
     * client may have to go through registration again.
     */
    override func reset() throws {
        // TODO: FIXME does not throw only prints to log
        // deletes the key pair and set the internal state of the TAM to false
        // the TAM has to be reprovisioned and activated
        if (heimdall != nil) {
            let _ = heimdall!.destroy()
            heimdall = nil
        }
        if (self.isActive) {
            // only if the TAM is active we have stored the data in the keychain 
            // and have to clean up now delete all entries in the keychain
            if (!self.serverHost.isEmpty) {
                let host = KeychainEntry(Name: TAMconstants.HOST+self.deviceId)
                do {
                    try removeItemWithAttributes(
                        attributes: host.attributesForFetch)
                } catch _ { }
                self.serverHost = String()
            }
            // ---
            if (!self.serverPort.isEmpty) {
                let port = KeychainEntry(Name: TAMconstants.PORT+self.deviceId)
                do {
                    try removeItemWithAttributes(
                        attributes: port.attributesForFetch)
                } catch _ { }
                self.serverPort = String()
            }
            // ---
            if (!self.sharedSecret.isEmpty){
                let secret = KeychainEntry(
                             Name: TAMconstants.SHARED_SECRET + self.deviceId)
                do {
                    try removeItemWithAttributes(
                        attributes: secret.attributesForFetch)
                } catch _ { }
                self.sharedSecret = String()
            }
            // ---
            if (!self.deviceId.isEmpty) {
               let devid = KeychainEntry(
                           Name: TAMconstants.DEVICEID + self.deviceId)
               do {
                    try removeItemWithAttributes(
                        attributes: devid.attributesForFetch)
               } catch _ { }
            }
            // ---
            // delete the Endpoint ID only if it is set
            if (!self.endpointId.isEmpty) {
                let endid = KeychainEntry(
                            Name: TAMconstants.ENDPOINT + self.deviceId)
                try removeItemWithAttributes(
                        attributes: endid.attributesForFetch)
                self.endpointId = String()
            }
            if (self.rootCACert != nil) {
                let res = removeCertificateFromKeyChain(
                    tagName: "rootCA"+self.deviceId)
                if res {
                    debugPrint("Successful removed the rootCA \(self.deviceId)")
                }
                self.serverPolicy = nil
                self.rootCACert = nil
            }
            // ---
            if (self.isActive) {
               let activ = KeychainEntry(
                           Name: TAMconstants.ACTIVE + self.deviceId)
                try removeItemWithAttributes(
                    attributes: activ.attributesForFetch)
                self.isActive = false
            }
        }
        self.deviceId = String()
        self.isProvisioned = false
        self.isProvisionedFromKeychain = false
        return
    }
    
    /**
     * Closes the TAM. Reset the Instance flag of the TAM object in the 
     * keychain.
     * The TAM object should be deinitialized after this call.
     */
    override func close() {
        serverPolicy = nil
        rootCACert = nil
        endpointCert = nil
        heimdall = nil
        isActive = false
        return
    }

    // Impl

    private func isActivePersistent() -> Bool {
        do {
            let active = KeychainEntry(Name: TAMconstants.ACTIVE+self.deviceId,
                                       data: TAMconstants.ACTIVE )
            var result = try self.fetchItemWithAttributes(
                                  attributes: active.attributesForFetch)
            result = active.dataFromAttributes(attributes: result!)
            let activeStr = result!["token"] as! String
            if (activeStr.isEmpty) {
                return false
            } else if activeStr == TAMconstants.ACTIVE {
                return true
            } else {
                
            }
        } catch _ {
            // if no result the TAM is not ACTIVE
            return false
        }
        return false
    }
    
 
    private func storeInKeychain() throws {
        let host = KeychainEntry(Name: TAMconstants.HOST + self.deviceId,
                                 data: self.serverHost )
        try insertItemWithAttributes(attributes: host.attributesToSave)
        let port = KeychainEntry(Name: TAMconstants.PORT + self.deviceId,
                                 data: self.serverPort )
        try insertItemWithAttributes(attributes: port.attributesToSave)
        let secret = KeychainEntry(
                     Name: TAMconstants.SHARED_SECRET + self.deviceId,
                     data: self.sharedSecret )
        try insertItemWithAttributes(attributes: secret.attributesToSave)
        let devid = KeychainEntry(
                    Name: TAMconstants.DEVICEID + self.deviceId,
                    data: self.deviceId )
        try insertItemWithAttributes(attributes: devid.attributesToSave)
        if (self.rootCACert != nil){
            let kr = try self.setCertificateinKeyChain(
                          certificate: self.rootCACert!,
                          tagName: "rootCA"+self.deviceId)
            if (kr) {
                debugPrint("certificate stored in keyChain " +
                    String(describing: self.rootCACert))
            }
        }
    }
    

    private func loadFromKeychain() -> Bool {
        do {  // loads the TAM items from the keychain
            if (self.heimdall == nil) {
                   let tagPublic = self.keyPublic + self.deviceId
                   let tagPrivate = self.keyPrivate + self.deviceId
                   self.heimdall = Heimdall(publicTag: tagPublic,
                                            privateTag: tagPrivate,
                                            keySize: self.RSAkeySize)
            }
            let host = KeychainEntry(Name: TAMconstants.HOST + self.deviceId)
            var result = try fetchItemWithAttributes(
                                attributes: host.attributesForFetch)
            result = host.dataFromAttributes(attributes: result!)
            self.serverHost = result!["token"] as! String
            // ---
            let port = KeychainEntry(Name: TAMconstants.PORT + self.deviceId)
            result = try fetchItemWithAttributes(
                                attributes: port.attributesForFetch)
            result = port.dataFromAttributes(attributes: result!)
            self.serverPort = result!["token"] as! String
            // ---
            let secret = KeychainEntry(
                         Name: TAMconstants.SHARED_SECRET + self.deviceId)
            result = try fetchItemWithAttributes(
                         attributes: secret.attributesForFetch)
            result = secret.dataFromAttributes(attributes: result!)
            self.sharedSecret = result!["token"] as! String
            // ---
            let devid = KeychainEntry(
                Name: TAMconstants.DEVICEID + self.deviceId)
            result = try fetchItemWithAttributes(
                         attributes: devid.attributesForFetch)
            result = devid.dataFromAttributes(attributes: result!)
            self.deviceId = result!["token"] as! String
            self.clientId = self.deviceId  // check this
            // ---
            let endid = KeychainEntry(
                        Name: TAMconstants.ENDPOINT + self.deviceId)
            result = try fetchItemWithAttributes(
                            attributes: endid.attributesForFetch)
            result = endid.dataFromAttributes(attributes: result!)
            self.endpointId = result!["token"] as! String
            self.rootCACert = self.getCertificateFromKeyChain(
                                    tagName: "rootCA"+self.deviceId)
            self.isProvisioned = true
            self.isProvisionedFromKeychain = true
            return true
        } catch {
            debugPrint("Error when loading item from the keychain \(error)")
            self.isProvisioned = false
            return false
        }
    }

    /**
     * Deinitializes the TAM instance. Will be called by the system or by the 
     * object referencing this TAM object.
     */
    deinit {
        // TODO: Should this do something, user super classes which 
        // calls reset ?
        // Any static state ?
    }
    
//------------------------------------ Provisioning methods -------------------
    
    fileprivate func setSharedSecret(SHARED_SECRET sharedS: String) {
        self.sharedSecret = sharedS
    }

    // The setting function for the provisioning from the URL
    fileprivate func setServerURI(SERVER_URI uri: String) -> Bool {
        guard let serverURL:URL = URL(string: uri) else { return false }
        guard let sh = serverURL.host else { return false }
        self.serverHost = sh
        self.serverScheme = serverURL.scheme!
        if self.serverScheme.lowercased() == "http" { return false }
        var sp = (serverURL as NSURL).port
        if (sp == nil && self.serverScheme == "https") { sp = 443 }
        self.serverPort = (sp?.stringValue)!
        return true
    }
    
    private func setServerPolicyFromCertificates() -> Bool {
        if (self.rootCACert == nil) {
            self.serverPolicy = nil
            return true
        }
        // if a rootCACert is provisioned we set a policy object and evaluate it
        let trust = trustWithCertificates(certificates: [self.rootCACert!])
        let anchorSet = setRootCertificateAsLoneAnchorCertificateForTrust(
                                        trust: trust!,root: self.rootCACert!)
        if (!anchorSet) {
            return false
        }
        // ---- set policies and check if they are valid
        let policies = [SecPolicyCreateBasicX509()]
        SecTrustSetPolicies(trust!, policies as CFTypeRef)
        let valid = trustIsValid(trust: trust!)
        if (!valid) {
            return false
        }
        var validPolicy = false
        if (valid) {
            validPolicy = self.setServerTrustPolicy(
                                hostName: self.getServerHost(), trust: trust!,
                                certificateChain: [self.rootCACert!])
            // if the evaluation of the policy fails we return false
            if (!validPolicy) {
                return false
            }
        }
        return true
    }

    fileprivate func setClientID(CLIENT_ID clientID: String) {
        self.clientId = clientID
        self.deviceId = clientID
    }
    
    fileprivate func setEndpointIDProvisioned(ENDPOINT_ID endpoint: String) {
        // check do we have to check if the tam is active or deny setting of 
        // the endpoint if it is set already
        self.endpointId = endpoint
    }
    
    fileprivate func setTrustAnchor(TRUST_ANCHOR cert: Foundation.Data ) -> Bool {
        let certificate = SecCertificateCreateWithData(nil,cert as CFData)
        if ((certificate != nil)) {
            self.rootCACert = certificate
            return true
        }
        return false
    }
}
 
class UnifiedProvisioner {
    
    private static let FORMAT_VERSION:UInt8 = 33
    private static let MAX_FORMAT_VERSION:UInt8 = 126
    
    private static let SERVER_URI_TAG:UInt8 = 1
    private static let ACTIVATION_ID_TAG:UInt8 = 2
    private static let SHARED_SECRET_TAG:UInt8 = 3
    private static let ENDPOINT_ID_TAG:UInt8 = 4
    private static let TRUST_ANCHOR_TAG:UInt8 = 5
    private static let PRIVATE_KEY_TAG:UInt8 = 6
    private static let PUBLIC_KEY_TAG:UInt8 = 7
    private static let CONNECTED_DEVICE_TAG:UInt8 = 8
    
    private static let AES_BLOCK_SIZE:Int = 16
    private static let AES_KEY_SIZE:Int = 128
    private static let PBKDF2_ITERATIONS:Int = 10000
    
    init() {}
    
    /**
     * - throws: `ClientError.file` if the provision file is empty or has the
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    fileprivate func loadFileFromPath(path: String, password: String,
            tam: UnifiedTrustedAssetsManager? = nil) throws -> Bool {
        let fileURL = URL(fileURLWithPath: path)
        guard let pData = try? Foundation.Data(contentsOf: fileURL) else {
            return false
        }
        return try parseFileContent(pData: pData, password: password, tam: tam)
    }
    
    // Load the unified provisioner file
    // -checks the version information
    // -cuts if the comments at the end of the file if present
    // -decrypt the file
    // -parse the decrypted TLV structure
    // -set the data in the tam object
    /**
     * - throws: `ClientError.file` if the resourece url cannot be formed or
     *           if the provision file is empty or has the
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    func loadFile(fileName: String, withExtension: String,
                  subdirectory: String? = nil , password: String,
                  tam: UnifiedTrustedAssetsManager? = nil) throws -> Bool {
        
        guard let url:URL = TrustedAssetsManagerImpl.URLForResource(
                fileName: fileName, withExtension: withExtension,
                subdirectory: subdirectory) else {
                    
            if (subdirectory != nil) {
                throw ClientError.file(String(format: NSLocalizedString(
                    "Wrong file information filename = %@" +
                    " file-extension = %@ subdirectory = %@", comment: ""),
                    fileName, withExtension, subdirectory!))
            } else {
                throw ClientError.file(String(format: NSLocalizedString(
                    "Wrong file information filename = %@" +
                    " file-extension = %@", comment: ""),
                    fileName, withExtension))
            }
        }
        let pData = try Foundation.Data(contentsOf: url)
        return try parseFileContent(pData: pData, password: password, tam: tam)
    }
    
    /**
     * - throws: `ClientError.file` if the provision file is empty or has the 
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    private func parseFileContent(pData: Foundation.Data, password: String,
                    tam: UnifiedTrustedAssetsManager? = nil ) throws -> Bool {
        // decode BASE64
        // make an array of UInt8 = Byte        
        var pArray = pData.withUnsafeBytes({(databytes: UnsafePointer<UInt8>) in
            return Array(UnsafeBufferPointer(
                start: databytes, count: pData.count))
        })
        if (pArray.count == 0) {
            throw ClientError.file(NSLocalizedString("No Data in File",
                                                     comment: ""))
        }
        if (pArray.first != UnifiedProvisioner.FORMAT_VERSION) {
            throw ClientError.file(NSLocalizedString(
                "Wrong Version of the provisioning file", comment: ""))
        }
        
        let spec:UInt8 = 0x23
        var mark:Int=1
        for i in 1 ..< pArray.count {
            if (pArray[i] == spec){
                mark = i
                break
            }
        }
        // if mark is still 1 then there is no comment at the end of the file
        if (mark == 1) {
            mark = pArray.count
        }
        let cipherTextSlice = pArray[1...(mark-1)]
        let cipherText64 = Array(cipherTextSlice)
        // in cipher Text we have the BASE64 encoded bytes that have to be 
        // decoded
        let data = Foundation.Data(bytes: UnsafePointer<UInt8>(
                                    cipherText64), count: cipherText64.count)
        let cipherText = Foundation.Data(base64Encoded: data,
            options: Foundation.Data.Base64DecodingOptions.ignoreUnknownCharacters)!
        // the cipher Text as decoded Data in clear text is the BASE64 encoded 
        // clear text at the end of the file
        // Split the cipherText in the Data into the salt and the real text to 
        // be enciphered allocate a new Array to hold the cipher text
        var ciTA = [UInt8](repeating: 0, count: cipherText.count)
        cipherText.copyBytes(to: &ciTA, count: ciTA.count)
        let saltSlice = ciTA[0...(UnifiedProvisioner.AES_BLOCK_SIZE-1)]
        let salt = Array(saltSlice)
        // cipherTextRaw is the encrypted part of the file
        // we have the salt and the password as String now we can run the PBKDF2
        let passwordA: [UInt8] = Array(password.utf8)
        do {
            let value = try PBKDF2(password: passwordA, salt: salt,
                                   iterations: 10000, keyLength: 16,
                                   algorithm: .sha1).calculate()
            let keyD = Foundation.Data(bytes: UnsafePointer<UInt8>(value),
                                       count: value.count)
            // key and a cipher text are present
            let algorithm = CCAlgorithm(kCCAlgorithmAES128)
            if let clearText = CryptoHelper.decryptUP(data: cipherText,key: keyD,
                                                      algorithm: algorithm) {
                pArray = (clearText.withUnsafeBytes({(databytes: UnsafePointer<UInt8>) in
                    return Array(UnsafeBufferPointer(start: databytes,
                                                     count: clearText.count))
                }))
            } else {
                throw ClientError.file(NSLocalizedString(
                    "Can't decrypt the input file", comment: ""))
            }
        } catch _ {
             throw ClientError.file(NSLocalizedString(
                "Can't decrypt the input file", comment: ""))
        }
        
        // only after deciphering and parsing we can read the items for the 
        // TAM object
        var upServerURI = String()
        var upSharedSecret = String()
        var upEndpointId = String()
        var upClientID = String()
        var upCertData: Foundation.Data?
        var tlv:TLV
        var offsetNext = 0
        var offset = 0
        let length:Int = pArray.count
        repeat {
            tlv = try TLV.init(data: pArray,offset: offset)
            switch tlv.tag {
                case UnifiedProvisioner.SERVER_URI_TAG:
                   upServerURI = String.init(bytes: tlv.value, encoding: .utf8)!
                case UnifiedProvisioner.ACTIVATION_ID_TAG:
                    upClientID = String.init(bytes: tlv.value, encoding: .utf8)!
                case UnifiedProvisioner.SHARED_SECRET_TAG:
                    upSharedSecret = String.init(bytes: tlv.value,
                                                 encoding: .utf8)!
                case UnifiedProvisioner.ENDPOINT_ID_TAG:
                    upEndpointId = String.init(bytes: tlv.value,
                                               encoding: .utf8)!
                case UnifiedProvisioner.PRIVATE_KEY_TAG:
                    debugPrint("Private Key \(tlv)")
                case UnifiedProvisioner.PUBLIC_KEY_TAG:
                    debugPrint("Public Key \(tlv)")
                case UnifiedProvisioner.TRUST_ANCHOR_TAG:
                    upCertData = Foundation.Data(
                        bytes: UnsafePointer<UInt8>(tlv.value),
                        count: tlv.value.count)
                case UnifiedProvisioner.CONNECTED_DEVICE_TAG:
                    let id:TLV = try TLV.init(data: tlv.value, offset: 0)
                    let ss:TLV = try TLV.init(data: tlv.value,
                                              offset: id.offsetToNext)
                    tam?.addShareSecret(hardwareId:
                            String.init(bytes: id.value, encoding: .utf8)!,
                                        sharedSecret:
                            String.init(bytes: ss.value, encoding: .utf8)!)
            default:
                debugPrint("Value unknown \(tlv.tag)")
            }
            offsetNext = tlv.offsetToNext
            offset = offsetNext
        } while offset < length
                
        // only set the TAM item if the TAM object is not nil
        if (tam != nil) {
            let psr = tam!.setServerURI(SERVER_URI: upServerURI)
            if (!psr) {
                throw ClientError.file(String(format: NSLocalizedString(
                    "Not a valid Server URL %@ in provisioning file",
                    comment: ""), upServerURI))
            }
            tam!.setClientID(CLIENT_ID: upClientID)
            tam!.setSharedSecret(SHARED_SECRET: upSharedSecret)
            tam!.setEndpointIDProvisioned(ENDPOINT_ID: upEndpointId)
            if upCertData != nil {
                let trustanchor = tam!.setTrustAnchor(TRUST_ANCHOR: upCertData!)
                if (!trustanchor) {
                    return false
                }
            }
        }
        return true
    }
}

fileprivate struct TLV: CustomStringConvertible {
    let tag:UInt8
    let length:UInt16
    let value:[UInt8]
    var offsetToNext:Int
    
    /**
     * - Throws `ClientError.file` if the provisioning file is not formatted
     *                             correctly
     */
    fileprivate init(data: Array<UInt8>, offset: Int) throws {
        offsetToNext = offset
        self.tag = data[offsetToNext]
        offsetToNext += 1
        self.length = TLV.twoBytes(value1: data[offsetToNext],
                                   value2: data[offsetToNext+1])
        offsetToNext += 2
        // If length is 0 do not decrement
        let offl:Int = Int(offsetToNext) +
            (Int(self.length) == 0 ? Int(self.length) : Int(self.length) - 1)
        let isIndexValid = data.indices.contains(offl)
        if (!isIndexValid) {
            throw ClientError.file(NSLocalizedString(
                "Error in provisioning file: TLV is not correctly " +
                "formatted, check the creation of the provisioning file " +
                "and that the correct password is used", comment: ""))
        }
        let valueSl = data[Int(offsetToNext)...offl]
        offsetToNext = offsetToNext + Int(self.length)
        self.value = Array(valueSl)
    }
    
    var description: String {
        var desc:String = String()
        desc += "TAG="
        desc += String(self.tag)
        desc += "\r\n"
        desc += "Length="
        desc += String(self.length)
        desc += "\r\n"
        desc += "Value ="
        desc += String(describing: self.value)
        desc += "\r\n"
        desc += "Next Offset="
        desc += String(self.offsetToNext)
        return desc
    }
    
    // converts two UInt8 value to a UInt16 value (bigEndian)
    private static func twoBytes(value1: UInt8, value2: UInt8) -> UInt16{
        let array : [UInt8] = [value1, value2]
        var value : UInt16 = UInt16(array[1]) * 256 + UInt16(array[0])
        value = UInt16(bigEndian: value)
        return value
    }
}
