/*
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */
import Foundation
import os
import SwiftyJSON

fileprivate var DOCUMENT_DIRECTORY_URL:URL? = nil

// TODO: make configurable
func haveLocalStore() -> Bool {
    return Config.USE_LOCAL_STORE
}

/**
 * Look for file in the documents directory and return a URL to the file
 */
func getDocumentURL(filename:String) -> URL? {
    if DOCUMENT_DIRECTORY_URL == nil {
        do {
            let ddu = try FileManager.default.url(for:.documentDirectory,
                                                  in:.userDomainMask,
                                                  appropriateFor: nil, create: true)
            DOCUMENT_DIRECTORY_URL = ddu
        } catch {
            log(type:.fault,
                message:"Cannot get documentDirector: %@",
                args: "\(error)")
            return nil
        }
    }
    return DOCUMENT_DIRECTORY_URL!.appendingPathComponent(filename)
}

/**
 * Try to locate the file in the ".documentDirectory" or "Bundle.main.path"
 * if searchBundle is true
 */
func getFileContent(filename:String, searchBundle:Bool = false) -> String? {
    
    var content:String? = nil
    if let documentFileURL:URL = getDocumentURL(filename:filename) {
        do {
            content = try String(contentsOf: documentFileURL, encoding: .utf8)
        } catch {
            // File does not exist
            let nserror = error as NSError
            if nserror.domain == "NSCocoaErrorDomain" && nserror.code == 260 {
                return nil
            } else {
                log(type:.debug,
                    message:"getFileContent: cannot get content of file %@: %@",
                    args:documentFileURL.path, "\(error)")
            }
            
        }
        return content
    }
    
    if !searchBundle {
        return nil
    }
    
    do {
        if let bundleFileStr = Bundle.main.path(forResource: filename,
                                                ofType: nil) {
            let bundleFileURL = URL(fileURLWithPath: bundleFileStr)
            content = try String(contentsOf: bundleFileURL, encoding: .utf8)
        }
    } catch {
        log(type:.fault,
            message:"getFileContent: cannot get content of file %@: %@",
            args: filename, "\(error)")
        
    }
    return content
}

/**
 * Return the content of filename, searching first in the document directory
 * and if the file is not there, search the bundle directory.
 */
func getFileFromLocalStore(filename:String) -> String? {
    return getFileContent(filename:filename, searchBundle:true)
}

/**
 * Saves content to the documents directory to a file named filename
 */
func saveFileToLocalStore(filename:String, content:String) -> Bool {
    if let fileURL:URL = getDocumentURL(filename:filename) {
        do {
            try content.write(to: fileURL, atomically: true, encoding: .utf8)
            return true
        } catch {
            log(type:.fault,
                message:"saveFileToLocalStore: cannot save content to file %@: %@",
                args:filename, "\(error)")
        }
    }
    return false
}

func deleteFileFromLocalStore(filename:String) {
    guard let url:URL = getDocumentURL(filename: filename) else {
        return
    }
    do {
        try FileManager.default.removeItem(atPath: url.path)
    } catch {
        log(type:.fault,
            message:"Cannot delete file %@ from local store: %@",
            args: url.path, "\(error)")
    }
}

func getClassName(instance:Any?) -> String {
    return String(describing: type(of: instance))
}


func getCurrentTimeMillis() -> Int64 {
    return Int64(Date().timeIntervalSince1970 * 1000.0)
}

fileprivate let dateFormatter = DateFormatter()
func formatDate(date:Date) -> String {
    dateFormatter.timeStyle = .medium
    dateFormatter.dateStyle = .medium
    return dateFormatter.string(from: date)
}

struct ConcurrentMap<K:Hashable,V> {
    var queue:DispatchQueue! = nil
    var map:Dictionary<K,V>! = nil
    init(label:String) {
        self.queue = DispatchQueue(label:label, qos:.utility)
        self.map = Dictionary<K,V>()
    }
    func get(key:K) -> V? {
        return queue.sync(execute: {
            return map[key]
        })
    }
    mutating func put(key:K, value:V?) {
        queue.sync() {
            map[key] = value
        }
    }
    func sync(execute: ()->V?) -> V? {
        return queue.sync(execute: execute)
    }
    func sync(execute: ()->()) {
        queue.sync(execute: execute)
    }
}
